# Building externals modules:

INCLUDES = -I. -I$(TOP)/kernel $(GC_INCLUDES) $(DLD_INCLUDES)
  LIBS = $(EXTRA_LIBRARIES) -L$(TOP)/kernel -loskernel -lm $(GC_LIB) $(DLD_LIB)
CFLAGS = $(OPTIMIZE) $(INCLUDES) $(EXTRA_INCLUDES) $(DEFINES) $(EXTRA_DEFINES)

# Rules for building the module:
#    - statically (both a library, and an executable): "static:" and "exec:"
#    - dynamically (a loadable module): "dynamic:"

static: libos$(MODULE).a

libos$(MODULE).a: $(OBJECTS)
	rm -rf libos$(MODULE).a
	$(AR) $(ARFLAGS) libos$(MODULE).a $(OBJECTS) $(EXTRA_OBJECTS)
	$(RANLIB) libos$(MODULE).a

exec: os$(MODULE)

os$(MODULE): static $(MAIN_OBJECTS)
	$(CC) -o os$(MODULE) $(MAIN_OBJECTS) libos$(MODULE).a $(LIBS)

.SUFFIXES:	.so .c

.c.so:
	$(CC) $(SHCFLAGS) $(CFLAGS) -o $*.so -c $*.c

dynamic: $(MODULE).so

$(MODULE).so: $(DYNAMIC_OBJECTS)
	$(SHLD) $(SHLDFLAGS) -o $(MODULE).so $(DYNAMIC_OBJECTS) $(EXTRA_LIBRARIES)

# Cleaning:

clean:
	rm -rf os$(MODULE) $(MODULE).so libos$(MODULE).a 
	rm -rf $(OBJECTS) $(DYNAMIC_OBJECTS) $(MAIN_OBJECTS)
	rm -rf core *~ $(EXTRA_CLEAN)

# Installation:

install: exec.inst module.inst lib.inst script.inst doc.inst 

bininst: exec.inst module.inst

exec.inst:
	-@if test -f "os$(MODULE)"; then { \
		if test ! -d $(OS_BINDIR); then \
			$(TOP)/utils/mkdirhier $(OS_BINDIR); \
		fi; \
		if test -n "$(STANDALONE_NAME)"; then \
			$(INSTALL) os$(MODULE) $(OS_BINDIR)/$(STANDALONE_NAME); \
		else \
			$(INSTALL) os$(MODULE) $(OS_BINDIR); \
		fi; \
	}; fi

module.inst:
	-@if test -f "$(MODULE).so"; then { \
		if test ! -d "$(OS_MODULEDIR)"; then \
			$(TOP)/utils/mkdirhier $(OS_MODULEDIR); \
		fi; \
		$(INSTALL) $(MODULE).so $(OS_MODULEDIR); \
	}; fi

lib.inst:
	-@if test -n "$(SCM_LIBDIR)"; then { \
		if test ! -d "$(OS_LIBDIR)/$(MODULE)"; then \
			$(TOP)/utils/mkdirhier $(OS_LIBDIR)/$(MODULE) ;\
		fi; \
		for i in `cat $(SCM_LIBDIR)/FILES`; do \
			echo "Installing: library file $$i for $(MODULE)"; \
			$(INSTALL) "$(SCM_LIBDIR)/$$i" $(OS_LIBDIR)/$(MODULE);\
		done; \
	}; fi

script.inst:
	-@if test -n "$(SCM_SCRIPTSDIR)" -a -d "$(SCM_SCRIPTSDIR)"; then { \
		if test ! -d "$(OS_SCRIPTSDIR)"; then \
			$(TOP)/utils/mkdirhier $(OS_SCRIPTSDIR); \
		fi; \
		for i in `cat $(SCM_SCRIPTSDIR)/FILES`; do \
			echo "Installing: script $$i for $(MODULE)"; \
			$(INSTALL) "$(SCM_SCRIPTSDIR)/$$i" $(OS_SCRIPTSDIR); \
		done; \
	}; fi

doc.inst:
	-@if test ! -d "$(OS_HTMLDIR)"; then \
		$(TOP)/utils/mkdirhier $(OS_HTMLDIR); \
	fi
	-@if test ! -d "$(OS_LIBDIR)/$(MODULE)"; then \
		$(TOP)/utils/mkdirhier "$(OS_LIBDIR)/$(MODULE)"; \
	fi
	$(TOP)/kernel/oskernel -f $(TOP)/kernel/scripts/autodir \
		-path $(TOP)/kernel/lib \
		-d . \
		`if test -n "$(SCM_LIBDIR)"; then echo "-d $(SCM_LIBDIR)"; fi` \
		`if test -n "$(SCM_SCRIPTSDIR)"; then echo "-d $(SCM_SCRIPTSDIR)"; fi` \
		-oscm $(OS_LIBDIR)/$(MODULE) \
		-ohtml $(OS_HTMLDIR) \
		-ohtmlhelp "$(MODULE)-help.html" \
		-ohtmlindex "$(MODULE)-idx.html" \
		-ohtmlmodule "$(MODULE)-m.html"
	echo "$(MODULE) $(MODULE)-m.html" >> $(OS_HTMLDIR)/INSTALLED


