/***************************************************************************
 * arit.c
 * arit.c,v 1.2 1995/03/17 14:16:26 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * OScheme implementation of arithmetic. OScheme arithmetic is limited to
 * the underlying hardware 'int' implementation.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>
#include <math.h>

#include <os_config.h>
#include <os_interp.h>
#include <os_obj.h>
#include <os_string.h>
#include <os_atom.h>
#include <os_gc.h>
#include <os_vector.h>
#include <os_error.h>
#include <os_arit.h>
#include <os_cons.h>
#include <os_cblock.h>
#include <os_subr.h>
#include <os_port.h>
#include <os_regexp.h>

OsClass OsIntegerClass = (OsClass) 0 ;
OsClass OsFloatClass   = (OsClass) 0 ;

static OsError ose_div_zero = (OsError) 0 ;

static OsInteger os_integers[OS_INTEGER_PREDEFINED_RANGE] ;

/*
 * I use this macro heavily inside this file:
 */

#define NOT_A_NUMBER(o) \
    OsErrorTrigger (1 \
                    , OsE_BadType, (o) \
                    , OsList2Make ((OsObject) OsIntegerClass \
                                   , (OsObject) OsFloatClass)) ;

/***************************************************************************
 * OsInteger:
 * ----------
 * Integer stuff
 ***************************************************************************/

/*
 * Constructors:
 */

INLINE OsInteger OsIntegerMake (ival)
    int ival ;
{
    OsInteger i ;
    if ( (OS_LOWER_INTEGER_PREDEFINED <= ival)
        && (ival < OS_HIGHEST_INTEGER_PREDEFINED) ) {
        return (os_integers[ival-(OS_LOWER_INTEGER_PREDEFINED)]) ;
    }
    i = (OsInteger) OsGCMallocAtomic (sizeof (OsIntegerRec)) ;
    i->os_class = OsIntegerClass ;
    i->ival     = ival ;
    return (i) ;
}

/*
 * Subrs:
 */

/* AUTODOC
 * (number? OBJECT)
 * Is OBJECT a number ?
 */

static OsObject OsSubrNumberP (OsObject object) {
    if ( OsIsInteger (object) || OsIsFloat (object) )
        return (OsA_True) ;
    return (OsA_False) ;
}

static OsObject OsSubrAlwaysTrue (OsObject object) {
    return (OsA_True) ;
}

static OsObject OsSubrRationalP (OsObject object) {
    return (OsIsInteger (object) ? OsA_True : OsA_False) ;
}

/* AUTODOC
 * (integer? OBJECT)
 * Is OBJECT an integer ?
 * Returns #t if object is an integer, #f otherwise.
 */

static OsObject OsSubrIntegerP (OsInteger i) {
    if ( OsIsInteger (i) ) {
        return (OsA_True) ;
    } else if ( OsIsFloat (i) ) {
        return ((rint (OsFloatValue(i)) == OsFloatValue (i))
                ? OsA_True
                : OsA_False) ;
    } else {
        return (OsA_False) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (real? OBJECT)
 * Is given object a real ?
 */

static OsObject OsSubrRealP (OsObject object) {
    return ((OsIsInteger (object) || OsIsFloat (object))
             ? OsA_True
             : OsA_False) ;
}

/* AUTODOC
 * (exact? OBJECT)
 * Is given OBJECT an exact number ?
 */

static OsObject OsSubrExactP (OsObject object) {
    return (OsIsInteger (object) ? OsA_True : OsA_False) ;
}

/* AUTODOC
 * (inexact? OBJECT)
 * Is given OBJECT an inexact number ?
 */

static OsObject OsSubrInexactP (OsObject object) {
    return (OsIsFloat (object) ? OsA_True : OsA_False) ;
}

/* AUTODOC
 * (zero? NUMBER)
 * Is given NUMBER 0 ?
 */

static OsObject OsSubrZeroP (OsObject object) {
    if ( OsIsInteger (object) ) {
        return (( ! OsIntegerValue (object)) ? OsA_True : OsA_False) ;
    } else if ( OsIsFloat (object) ) {
        return (OsFloatIsZero (object) ? OsA_True : OsA_False) ;
    } else {
        NOT_A_NUMBER (object) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (negative? NUMBER)
 * Is given NUMBER negative ?
 */

static OsObject OsSubrNegativeP (OsObject object) {
    if ( OsIsInteger (object) ) {
        return ((OsIntegerValue (object) < 0) ? OsA_True : OsA_False) ;
    } else if ( OsIsFloat (object) ) {
        return ((OsFloatValue (object) < ((double) 0)) ? OsA_True : OsA_False);
    } else {
        NOT_A_NUMBER (object) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (positive? NUMBER)
 * Is given NUMBER positive ?
 */

static OsObject OsSubrPositiveP (OsObject object) {
    if ( OsIsInteger (object) ) {
        return ((OsIntegerValue (object) > 0) ? OsA_True : OsA_False) ;
    } else if ( OsIsFloat (object) ) {
        return ((OsFloatValue(object) > ((double) 0)) ? OsA_True : OsA_False) ;
    } else {
        NOT_A_NUMBER (object) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (+ [NUMBERS*])
 * Add the given NUMBERS, and returns the result.
 * Given any number of NUMBER, this primitive returns their sum.
 */

static OsObject OsSubrPlus (unsigned int argc, OsObject *argv) {
    int    iret = 0 ;
    double fret = (double) 0 ;
    while ( argc-- ) {
        if ( OsIsInteger (*argv) ) {
            iret += OsIntegerValue (*argv) ;
        } else if ( OsIsFloat (*argv) ) {
            goto float_cont ;
        } else {
            NOT_A_NUMBER (*argv) ;
        }
        argv++ ;
    }
    return ((OsObject) OsIntegerMake (iret)) ;
    
  float_cont:
    fret = OsFloatValue (*argv) + ((double) iret) ; argv++ ;
    while ( argc-- ) {
        if ( OsIsInteger (*argv) ) {
            fret += ((double) OsIntegerValue(*argv)) ;
        } else if ( OsIsFloat (*argv) ) {
            fret += OsFloatValue (*argv) ;
        } else {
            NOT_A_NUMBER (*argv) ;
        }
        argv++ ;
    } 
    return ((OsObject) OsFloatMake (fret)) ;
}

/* AUTODOC
 * (- NUM [NUMBERS*])
 * Substract to NUM the given NUMBERS, and returns the result.
 * Returns NUM minus the sum of the remaining NUMBERS.
 */

static OsObject OsSubrMinus (unsigned int argc, OsInteger *argv) {
    int    iret ;
    double fret ;

    if ( argc < 1 ) {
        return ((OsObject) OsIntegerMake (0)) ;
    } else if ( argc == 1 ) {
        if ( OsIsInteger (*argv) ) {
            return ((OsObject) OsIntegerMake (- OsIntegerValue(*argv))) ;
        } else if ( OsIsFloat (*argv) ) {
            return ((OsObject) OsFloatMake (- OsFloatValue (*argv))) ;
        } else {
            NOT_A_NUMBER (*argv) ;
        }
    } 
    /* argc > 1 */
    if ( OsIsInteger (*argv) ) {
        iret = OsIntegerValue(*argv) ; argv++ ; argc-- ;
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                iret -= OsIntegerValue (*argv) ;
            } else if ( OsIsFloat (*argv) ) {
                fret = ((double) iret) - OsFloatValue (*argv) ; argv++ ;
                goto float_cont ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return ((OsObject) OsIntegerMake (iret)) ;
    } else if ( OsIsFloat (*argv) ) {
        fret = OsFloatValue (*argv) ; argv++ ; argc-- ;
      float_cont:
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                fret -= ((double) OsIntegerValue(*argv)) ;
            } else if ( OsIsFloat (*argv) ) {
                fret -= OsFloatValue (*argv) ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return ((OsObject) OsFloatMake (fret)) ;
    } else {
        NOT_A_NUMBER (*argv) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (* [NUMBERS*])
 * Multiply the given NUMBERS, and returns the result.
 * Given any number of NUMBER, this primitive returns their product. With no
 * argument, it returns 1.
 */


static OsObject OsSubrMult (unsigned int argc, OsInteger *argv) {
    int    iret = 1 ;
    double fret = (double) 1 ;
    while ( argc-- ) {
        if ( OsIsInteger (*argv) ) {
            iret *= OsIntegerValue (*argv) ;
        } else if ( OsIsFloat (*argv) ) {
            goto float_cont ;
        } else {
            NOT_A_NUMBER (*argv) ;
        }
        argv++ ;
    }
    return ((OsObject) OsIntegerMake (iret)) ;
    
  float_cont:
    fret = OsFloatValue (*argv) * ((double) iret) ; argv++ ;
    while ( argc-- ) {
        if ( OsIsInteger (*argv) ) {
            fret *= ((double) OsIntegerValue (*argv)) ;
        } else if ( OsIsFloat (*argv) ) {
            fret *= OsFloatValue (*argv) ;
        } else {
            NOT_A_NUMBER (*argv) ;
        }
        argv++ ;
    } 
    return ((OsObject) OsFloatMake (fret)) ;
}

/* AUTODOC
 * (/ NUM [NUMBERS*])
 * Divide NUM by the given NUMBERS, and returns the result.
 * Divide NUM by all given NUMBERS. If no NUMBERS are provided, returns the 
 * inverse of NUM.
 */

static OsObject OsSubrDiv (unsigned int argc, OsInteger *argv) {
    int    iret ;
    double fret ;

    if ( argc < 1 ) {
        return ((OsObject) OsIntegerMake (1)) ;
    } else if ( argc == 1 ) {
        if ( OsIsInteger (*argv) ) {
            if ( ! OsIntegerValue (*argv) )
                (void) OsErrorTrigger (1, ose_div_zero, *argv) ;
            return ((OsObject) OsIntegerMake (1 / OsIntegerValue (*argv))) ;
        } else if ( OsIsFloat (*argv) ) {
            if ( OsFloatIsZero (*argv) ) 
                (void) OsErrorTrigger (1, ose_div_zero, *argv) ;
            return ((OsObject) OsFloatMake (1 / OsFloatValue (*argv))) ;
        } else {
            NOT_A_NUMBER (*argv) ;
        }
    } 
    /* argc > 1 */
    if ( OsIsInteger (*argv) ) {
        iret = OsIntegerValue (*argv) ; argv++ ; argc-- ;
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( ! OsIntegerValue (*argv) )
                    (void) OsErrorTrigger (1, ose_div_zero, *argv) ;
                iret /= OsIntegerValue (*argv) ;
            } else if ( OsIsFloat (*argv) ) {
                if ( OsFloatIsZero (*argv) ) 
                    (void) OsErrorTrigger (1, ose_div_zero, *argv) ;
                fret = ((double) iret) / OsFloatValue (*argv) ; argv++ ;
                goto float_cont ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return ((OsObject) OsIntegerMake (iret)) ;
    } else if ( OsIsFloat (*argv) ) {
        fret = OsFloatValue (*argv) ; argv++ ; argc-- ;
      float_cont:
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( ! OsIntegerValue (*argv) )
                    (void) OsErrorTrigger (1, ose_div_zero, *argv) ;
                fret /= ((double) OsIntegerValue (*argv)) ;
            } else if ( OsIsFloat (*argv) ) {
                if ( OsFloatIsZero (*argv) ) 
                    (void) OsErrorTrigger (1, ose_div_zero, *argv) ;
                fret /= OsFloatValue (*argv) ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return ((OsObject) OsFloatMake (fret)) ;
    } else {
        NOT_A_NUMBER (*argv) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (= [NUMBERS]*)
 * Test the equality of NUMBERS.
 * Compare each NUMBERS, return #t if they are all equals, #f otherwise.
 */

static OsObject OsSubrEgal (unsigned int argc, OsInteger *argv) {
    int    ival ;
    double fval ;

    if ( argc <= 1 ) {
        return (OsA_True) ;
    }
    if ( OsIsInteger (*argv) ) {
        ival = OsIntegerValue (*argv) ; argv++ ; argc-- ;
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( OsIntegerValue (*argv) != ival )
                    return (OsA_False) ;
            } else if ( OsIsFloat (*argv) ) {
                if ( ((int) OsFloatValue (*argv)) != ival )
                    return (OsA_False) ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return (OsA_True) ;
    } else if ( OsIsFloat (*argv) ) {
        fval = OsFloatValue (*argv) ; argv++ ; argc-- ;
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( ((double) OsIntegerValue (*argv)) != fval )
                    return (OsA_False) ;
            } else if ( OsIsFloat (*argv) ) {
                if (OsFloatValue (*argv) != fval)
                    return (OsA_False) ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return (OsA_True) ;
    } else {
        NOT_A_NUMBER (*argv) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (< [NUMBERS]*)
 * Test that are in strict decreasing order.
 * Returns #t if NUMBERS are in strict decreasing order, #f otherwise.
 */

static OsObject OsSubrLT (unsigned int argc, OsInteger *argv) {
    int    ival ;
    double fval ;

    if ( argc <= 1 ) {
        return (OsA_True) ;
    }
    if ( OsIsInteger (*argv) ) {
        ival = OsIntegerValue (*argv) ; argv++ ; argc-- ;
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( OsIntegerValue (*argv) <= ival )
                    return (OsA_False) ;
                ival = OsIntegerValue (*argv) ;
            } else if ( OsIsFloat (*argv) ) {
                if ( OsFloatValue (*argv) <= ((double) ival) )
                    return (OsA_False) ;
                fval = OsFloatValue (*argv) ; argv++ ;
                goto float_cont ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return (OsA_True) ;
    } else if ( OsIsFloat (*argv) ) {
        fval = OsFloatValue (*argv) ; argv++ ; argc-- ;
      float_cont:
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( ((double) OsIntegerValue (*argv)) <= fval )
                    return (OsA_False) ;
                fval = ((double) OsIntegerValue (*argv)) ;
            } else if ( OsIsFloat (*argv) ) {
                if (OsFloatValue (*argv) <= fval)
                    return (OsA_False) ;
                fval = OsFloatValue (*argv) ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return (OsA_True) ;
    } else {
        NOT_A_NUMBER (*argv) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (> [NUMBERS]*)
 * Test that are in strict increasing order.
 * Returns #t if NUMBERS are in strict increasing order, #f otherwise.
 */

static OsObject OsSubrGT (unsigned int argc, OsInteger *argv) {
    int    ival ;
    double fval ;

    if ( argc <= 1 ) {
        return (OsA_True) ;
    }
    if ( OsIsInteger (*argv) ) {
        ival = OsIntegerValue (*argv) ; argv++ ; argc-- ;
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( OsIntegerValue (*argv) >= ival )
                    return (OsA_False) ;
                ival = OsIntegerValue (*argv) ;
            } else if ( OsIsFloat (*argv) ) {
                if ( OsFloatValue (*argv) >= ((double) ival) )
                    return (OsA_False) ;
                fval = OsFloatValue (*argv) ; argv++ ;
                goto float_cont ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return (OsA_True) ;
    } else if ( OsIsFloat (*argv) ) {
        fval = OsFloatValue (*argv) ; argv++ ; argc-- ;
      float_cont:
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( ((double) OsIntegerValue (*argv)) >= fval )
                    return (OsA_False) ;
                fval = ((double) OsIntegerValue (*argv)) ;
            } else if ( OsIsFloat (*argv) ) {
                if (OsFloatValue (*argv) >= fval)
                    return (OsA_False) ;
                fval = OsFloatValue (*argv) ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return (OsA_True) ;
    } else {
        NOT_A_NUMBER (*argv) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (<= [NUMBERS]*)
 * Test that are in decreasing order.
 * Returns #t if NUMBERS are in decreasing order, #f otherwise.
 */

static OsObject OsSubrLTE (unsigned int argc, OsInteger *argv) {
    int    ival ;
    double fval ;

    if ( argc <= 1 ) {
        return (OsA_True) ;
    }
    if ( OsIsInteger (*argv) ) {
        ival = OsIntegerValue (*argv) ; argv++ ; argc-- ;
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( OsIntegerValue (*argv) < ival )
                    return (OsA_False) ;
                ival = OsIntegerValue (*argv) ;
            } else if ( OsIsFloat (*argv) ) {
                if ( OsFloatValue (*argv) < ((double) ival) )
                    return (OsA_False) ;
                fval = OsFloatValue (*argv) ; argv++ ;
                goto float_cont ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return (OsA_True) ;
    } else if ( OsIsFloat (*argv) ) {
        fval = OsFloatValue (*argv) ; argv++ ; argc-- ;
      float_cont:
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( ((double) OsIntegerValue (*argv)) < fval )
                    return (OsA_False) ;
                fval = ((double) OsIntegerValue (*argv)) ;
            } else if ( OsIsFloat (*argv) ) {
                if (OsFloatValue (*argv) < fval)
                    return (OsA_False) ;
                fval = OsFloatValue (*argv) ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return (OsA_True) ;
    } else {
        NOT_A_NUMBER (*argv) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (<= [NUMBERS]*)
 * Test that are in increasing order.
 * Returns #t if NUMBERS are in increasing order, #f otherwise.
 */

static OsObject OsSubrGTE (unsigned int argc, OsInteger *argv) {
    int    ival ;
    double fval ;

    if ( argc <= 1 ) {
        return (OsA_True) ;
    }
    if ( OsIsInteger (*argv) ) {
        ival = OsIntegerValue (*argv) ; argv++ ; argc-- ;
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( OsIntegerValue (*argv) > ival )
                    return (OsA_False) ;
            } else if ( OsIsFloat (*argv) ) {
                if ( OsFloatValue (*argv) > ((double) ival) )
                    return (OsA_False) ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return (OsA_True) ;
    } else if ( OsIsFloat (*argv) ) {
        fval = OsFloatValue (*argv) ; argv++ ; argc-- ;
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( ((double) OsIntegerValue (*argv)) > fval )
                    return (OsA_False) ;
            } else if ( OsIsFloat (*argv) ) {
                if (OsFloatValue (*argv) > fval)
                    return (OsA_False) ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return (OsA_True) ;
    } else {
        NOT_A_NUMBER (*argv) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (min [NUMBERS+])
 * Returns the smallest integer in the given list of NUMBERS.
 * Returns the smallest integer in the given list of NUMBERS.
 */

static OsObject OsSubrMin (unsigned int argc, OsInteger *argv) {
    OsInteger imin ;
    OsFloat   fmin ;
    if ( argc < 1 ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (1)
                               , OsIntegerMake (0)) ;
    }
    if ( OsIsInteger (*argv) ) {
        imin = (OsInteger) *argv ; argv++ ; argc-- ;
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( OsIntegerValue (*argv) < OsIntegerValue (imin) )
                    imin = (OsInteger) *argv ;
            } else if ( OsIsFloat (*argv) ) {
                if ( OsFloatValue (*argv) < ((double) OsIntegerValue(imin)) )
                    fmin = (OsFloat) *argv ;
                else
                    fmin = OsFloatMake ((double) OsIntegerValue (imin)) ;
                argv++ ;
                goto float_cont ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return ((OsObject) imin) ;
    } else if ( OsIsFloat (*argv) ) {
        fmin = (OsFloat) *argv ; argv++ ; argc-- ;
      float_cont:
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( ((double) OsIntegerValue (*argv)) < OsFloatValue (fmin) )
                    fmin = OsFloatMake ((double) OsIntegerValue (*argv)) ;
            } else if ( OsIsFloat (*argv) ) {
                if ( OsFloatValue (*argv) < OsFloatValue (fmin) )
                    fmin = (OsFloat) *argv ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return ((OsObject) fmin) ;
    } else {
        NOT_A_NUMBER (*argv) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (max [NUMBERS]*)
 * Returns the biggest integer in the given list of NUMBERS.
 * Returns the biggest integer in the given list of NUMBERS.
 */

static OsObject OsSubrMax (unsigned int argc, OsInteger *argv) {
    OsInteger imax ;
    OsFloat   fmax ;
    if ( argc < 1 ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (1)
                               , OsIntegerMake (0)) ;
    }
    if ( OsIsInteger (*argv) ) {
        imax = (OsInteger) *argv ; argv++ ; argc-- ;
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( OsIntegerValue (*argv) > OsIntegerValue (imax) )
                    imax = (OsInteger) *argv ;
            } else if ( OsIsFloat (*argv) ) {
                if ( OsFloatValue (*argv) > ((double) OsIntegerValue(imax)) )
                    fmax = (OsFloat) *argv ;
                else
                    fmax = OsFloatMake ((double) OsIntegerValue (imax)) ;
                argv++ ;
                goto float_cont ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return ((OsObject) imax) ;
    } else if ( OsIsFloat (*argv) ) {
        fmax = (OsFloat) *argv ; argv++ ; argc-- ;
      float_cont:
        while ( argc-- ) {
            if ( OsIsInteger (*argv) ) {
                if ( ((double) OsIntegerValue (*argv)) > OsFloatValue (fmax) )
                    fmax = OsFloatMake ((double) OsIntegerValue (*argv)) ;
            } else if ( OsIsFloat (*argv) ) {
                if ( OsFloatValue (*argv) > OsFloatValue (fmax) )
                    fmax = (OsFloat) *argv ;
            } else {
                NOT_A_NUMBER (*argv) ;
            }
            argv++ ;
        }
        return ((OsObject) fmax) ;
    } else {
        NOT_A_NUMBER (*argv) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (abs NUMBER)
 * Returns the magnitude of NUMBER.
 */

static OsObject OsSubrAbs (OsObject number) {
    if ( OsIsInteger (number) ) {
        return ((OsIntegerValue (number) < 0)
                ? (OsObject) OsIntegerMake (- OsIntegerValue (number))
                : (OsObject) number) ;
    } else if ( OsIsFloat (number) ) {
        return ((OsFloatValue (number) < ((double) 0))
                ? (OsObject) OsFloatMake (- OsFloatValue (number))
                : (OsObject) number) ;
    } else {
        NOT_A_NUMBER (number) ;
    }
    /* NOT REACHED */
}

                
/* AUTODOC
 * (quotient N1 N2)
 * Return the quotient of the division of N1 by N2.
 * Return the quotient of the division of N1 by N2. 
 * This function might trigger the 'error:div-by-zero error.
 */

static OsObject OsSubrQuotient (OsInteger n1, OsInteger n2) {
    OsShouldBeInteger (n1) ;
    OsShouldBeInteger (n2) ;
  retry:
    if ( n2->ival ) {
        return ((OsObject) OsIntegerMake (n1->ival / n2->ival)) ;
    } else {
        n2 = (OsInteger) OsErrorTrigger (1, ose_div_zero, n2) ;
        goto retry ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (remainder N1 N2)
 * Returns the remainder of the division of N1 by N2.
 * Returns the remainder of the division of N1 by N2.
 * This function might trigger the 'error:div-by-zero error.
 */

static OsObject OsSubrRemainder (OsInteger n1, OsInteger n2) {
    OsShouldBeInteger (n1) ;
    OsShouldBeInteger (n2) ;
  retry:
    if ( n2->ival ) {
        int ret = n1->ival - (n2->ival * (n1->ival / n2->ival)) ;
        return ((OsObject) OsIntegerMake (ret)) ;
    } else {
        n2 = (OsInteger) OsErrorTrigger (1, ose_div_zero, n2) ;
        goto retry ;
    }
    /* NOT REACHED */
}
    
/* AUTODOC
 * (modulo N1 N2)
 * Returns the N1 modulo N2.
 * Returns the N1 modulo N2.
 * This function might trigger the 'error:div-by-zero error.
 */

static OsObject OsSubrModulo (OsInteger n1, OsInteger n2) {
    OsShouldBeInteger (n1) ;
    OsShouldBeInteger (n2) ;
  retry:
    if ( n2->ival ) {
        int mod = OsIntegerValue (n1) % OsIntegerValue (n2) ;
        return ((OsObject) OsIntegerMake (((OsIntegerValue(n2) < 0)
                                           ? (mod > 0)
                                           : (mod < 0))
                                          ? mod + OsIntegerValue(n2)
                                          : mod)) ;
    } else {
        n2 = (OsInteger) OsErrorTrigger (1, ose_div_zero, n2) ;
        goto retry ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (odd? NUMBER)
 * Is number odd ?
 * Returns #t if NUMBER is odd, #f otherwise.
 */

static OsObject OsSubrOddP (OsInteger n) {
    OsShouldBeInteger (n) ;
    return ((n->ival & 1) ? OsA_True : OsA_False) ;
}

/* AUTODOC
 * (even? NUMBER)
 * Is number even ?
 * Returns #t if NUMBER is even, #f otherwise.
 */

static OsObject OsSubrEvenP (OsInteger n) {
    OsShouldBeInteger (n) ;
    return ((n->ival & 1) ? OsA_False : OsA_True) ;
}

/* AUTODOC
 * (integer->string NUMBER)
 * Returns a string represnting the given NUMBER.
 * The returned string will be 'read as the given NUMBER.
 */

static OsObject OsSubrIntegerToString (n)
    OsInteger n ;
{
    char tmp[OS_DISPLAY_BUFSIZ] ;
    n = OsShouldBeInteger (n) ;
    sprintf (tmp, "%d", n->ival) ;
    return ((OsObject) OsStringMake (tmp)) ;
}

/* AUTODOC
 * (string->integer NUMBER)
 * Parse the string as an integer, and return the result.
 * Parse the string as an integer, and return the result.
 */

static OsObject OsSubrStringToInteger (s)
    OsString s ;
{
    OsShouldBeString (s) ;
    return ((OsObject) OsIntegerMake (atoi(s->string))) ;
}

/* AUTODOC
 * (logor [NUMBERS*])
 * Returns the logical OR of the given list of numbers.
 * Returns the logical OR of the given list of numbers.
 */

static OsObject OsSubrLogor (argc, argv)
    unsigned int argc ;
    OsInteger    *argv ;
{
    int ret = 0 ;
    while ( argc-- ) {
        OsShouldBeInteger (*argv) ;
        ret |= (*argv++)->ival ;
    }
    return ((OsObject) OsIntegerMake (ret)) ;
}

/* AUTODOC
 * (logand [NUMBERS*])
 * Returns the logical AND of the given list of numbers.
 * Returns the logical AND of the given list of numbers.
 */

static OsObject OsSubrLogand (argc, argv)
    unsigned int argc ;
    OsInteger    *argv ;
{
    int ret = ~0 ;
    while ( argc-- ) {
        OsShouldBeInteger (*argv) ;
        ret &= (*argv++)->ival ;
    }
    return ((OsObject) OsIntegerMake (ret)) ;
}

/* AUTODOC
 * (logxor N1 N2)
 * Returns the logical XOR of N1 by N2.
 * Returns the logical XOR of N1 by N2.
 */

static OsObject OsSubrLogxor (n1, n2)
    OsInteger n1, n2 ;
{
    OsShouldBeInteger (n1) ;
    OsShouldBeInteger (n2) ;
    return ((OsObject) OsIntegerMake (n1->ival ^ n2->ival)) ;
}

/* AUTODOC
 * (not N1)
 * Returns the logical NOT of N1.
 * Returns the logical NOT of N1.
 */

static OsObject OsSubrLognot (n1)
    OsInteger n1 ;
{
    OsShouldBeInteger (n1) ;
    return ((OsObject) OsIntegerMake (~ n1->ival)) ;
}

/* AUTODOC
 * (floor NUMBER)
 * Return the largest integer not larger than NUMBER.
 */

static OsObject OsSubrFloor (OsObject number) {
    if ( OsIsInteger (number) ) {
        return ((OsObject) OsFloatMake (((double) OsIntegerValue (number)))) ;
    } else if ( OsIsFloat (number) ) {
        return ((OsObject) OsFloatMake (floor (OsFloatValue (number)))) ;
    } else {
        NOT_A_NUMBER (number) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (ceiling NUMER)
 * Returns the smallest integer not smaller than NUMBER.
 */

static OsObject OsSubrCeiling (OsObject number) {
    if ( OsIsInteger (number) ) {
        return ((OsObject) OsFloatMake ((double) OsIntegerValue (number))) ;
    } else if ( OsIsFloat (number) ) {
        return ((OsObject) OsFloatMake (ceil (OsFloatValue (number)))) ;
    } else {
        NOT_A_NUMBER (number) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (truncate NUMBER)
 * Returns the integer closest to NUMBER.
 */

static OsObject OsSubrTruncate (OsObject number) {
    if ( OsIsInteger (number) ) {
        return ((OsObject) number) ;
    } else if ( OsIsFloat (number) ) {
        if ( OsFloatValue (number) > ((double) 0) ) 
            return ((OsObject) OsFloatMake (floor (OsFloatValue (number)))) ;
        else
            return ((OsObject) OsFloatMake (ceil (OsFloatValue (number)))) ;
    } else {
        NOT_A_NUMBER (number) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (round NUMER)
 * Returns the closest integer to NUMBER, rounding to even if needed.
 * This primitive rounds NUMBER in such a way that if NUMBER is half-way
 * between two integers, than the even one is returned.
 */

static OsObject OsSubrRound (OsObject number) {
    if ( OsIsInteger (number) ) {
        return (number) ;
    } else if ( OsIsFloat (number) ) {
        return ((OsObject) OsFloatMake (rint (OsFloatValue (number)))) ;
    } else {
        NOT_A_NUMBER (number) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (string->number STRING [RADIX])
 * Parse STRING as a number, and return it.
 * This procedure takes a STRING and optionally a RADIX. It creates a number
 * of maxiaml precision for the given representation STRING. If the STRING can
 * not be parsed as a number, then the procedure returns #f.
 * WARNING: this procedure is not strictly conformant to R4RS: Although RADIX
 * can be any number, if the STRING contains characters other then digits, then
 * the STRING will be rejected as a number (practically, this means that only
 * RADIX less than 10 will be usable).
 */

static OsObject OsSubrStringToNumber (unsigned int argc, OsObject *argv) {
    static os_regexp *ire = (os_regexp *) 0 ;
    static os_regexp *fre = (os_regexp *) 0 ;
    int base = 10 ;
    OsString string ;

    switch (argc) {
      case 2:
        OsMustBeInteger (argv[1]) ;
        base = OsIntegerValue (argv[1]) ;
        /* NO BREAK INTENTIONAL ! */
      case 1:
        string = (OsString) argv[0] ;
        OsShouldBeString (string) ;
        break ;
      default:
        OsErrorTrigger (1, OsE_ArityShouldBe
                        , OsIntegerMake (1)
                        , OsIntegerMake (0)) ;
        /* NOT REACHED */
    }
    if ( ! ire ) 
        ire = os_regcomp ("^[+-]?[0-9]+$") ;
    if ( ! fre )
        fre = os_regcomp (
"^[+-]?(([0-9]*\\.[0-9]+([eE][0-9]+)?)|([0-9]+(\\.[0-9]*)?([eE][0-9]+)?))$"
) ;
    if ( os_regexec (ire, OsStringBytes (string)) ) {
        long number = strtol (OsStringBytes (string), (char **) 0, base) ;
        return ((OsObject) OsIntegerMake (number)) ;
    } else if ( os_regexec (fre, OsStringBytes (string)) ) {
        return ((OsObject) OsFloatMakeFromCString (OsStringBytes (string))) ;
    } else {
        return (OsA_False) ;
    }
    /* NOT REACHED */
}
                      
/* AUTODOC
 * (number->string NUMBER)
 * Convert given NUMBER to its string representation.
 * This procedure converts the given NUMBER into the string that represents it.
 * WARNING: This procedure is not strictly compliant with R4RS because it has
 * no RADIX optional argument.
 */

static OsObject OsSubrNumberToString (OsObject number) {
    if ( OsIsInteger (number) || OsIsFloat (number) ) {
        OsString     string = OsStringMakeFromBytes (0, "") ;
        OsOutputPort port   = OsOutputStringMake (string) ;
        OsSendDisplay (number, port) ;
        OsOutputPortClose (port) ;
        return ((OsObject) string) ;
    }
    NOT_A_NUMBER (number) ;
    /* NOT REACHED */
}

/* AUTODOC
 * (exact->inexact NUMBER)
 * Convert exact NUMBER to its closest inexact representation.
 */

static OsObject OsSubrExactToInexact (OsObject number) {
    if ( OsIsInteger (number) ) {
        return ((OsObject) OsFloatMake ((double) OsIntegerValue (number))) ;
    } else if ( OsIsFloat (number) ) {
        return (number) ;
    } else {
        NOT_A_NUMBER (number) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (exact->inexact NUMBER)
 * Convert inexact NUMBER to its closest exact rerpesentation.
 */

static OsObject OsSubrInexactToExact (OsObject number) {
    if ( OsIsInteger (number) ) {
        return (number) ;
    } else if ( OsIsFloat (number) ) {
        return ((OsObject) OsIntegerMake (rint (OsFloatValue (number)))) ;
    } else {
        NOT_A_NUMBER (number) ;
    }
    /* NOT REACHED */
}

/*
 * Method impl for integers:
 */

static OsObject os_integer_display (OsInteger self
                                    , int sel
                                    , OsOutputPort port) {
    char tmp[OS_DISPLAY_BUFSIZ] ;
    sprintf (tmp, "%d", self->ival) ;
    OsOutputPortWriteCString (port, tmp) ;
    return ((OsObject) self) ;
}

static OsObject os_integer_equal (OsInteger self, int sel, OsInteger other) {
    if ( OsIsInteger (other) )
        return ((other->ival == self->ival) ? OsA_True : OsA_False) ;
    return (OsA_False) ;
}

void OsIntegerInitialize (OsInterp interp, int *pargc, char **argv) {
    int i ;
    OsInteger *pi ;
    OsIntegerClass = OsClassDeclare ("Integer", sizeof (OsIntegerRec)) ;
    OsDefineMsg (OsIntegerClass, OS_SEL_DISPLAY, os_integer_display) ;
    OsDefineMsg (OsIntegerClass, OS_SEL_WRITE, os_integer_display) ;
    OsDefineMsg (OsIntegerClass, OS_SEL_EQUAL, os_integer_equal) ;
    for (i = OS_LOWER_INTEGER_PREDEFINED, pi = os_integers ;
         i < OS_HIGHEST_INTEGER_PREDEFINED
         ; i++, pi++ ) {
        *pi = (OsInteger) OsGCMallocAtomic (sizeof (OsIntegerRec)) ;
        (*pi)->os_class = OsIntegerClass ;
        (*pi)->ival     = i ;
    }
    return ;
}

void OsIntegerSubrInitialize (OsInterp interp) {
    ose_div_zero = OsErrorMake (interp
                                , "error:div-by-zero"
                                , "Division by 0"
                                , 0, OSE_CATCHABLE) ;
    OsSubrMake (interp, "number?", OsSubrNumberP, 1) ;
    OsSubrMake (interp, "complex?", OsSubrAlwaysTrue, 1) ;
    OsSubrMake (interp, "rational?", OsSubrRationalP, 1) ;
    OsSubrMake (interp, "integer?", OsSubrIntegerP, 1) ;
    OsSubrMake (interp, "real?", OsSubrRealP, 1) ;
    OsSubrMake (interp, "exact?", OsSubrExactP, 1) ;
    OsSubrMake (interp, "inexact?", OsSubrInexactP, 1) ;
    OsSubrMake (interp, "=", OsSubrEgal, OS_NARY) ;
    OsSubrMake (interp, "<", OsSubrLT, OS_NARY) ;
    OsSubrMake (interp, ">", OsSubrGT, OS_NARY) ;
    OsSubrMake (interp, "<=", OsSubrLTE, OS_NARY) ;
    OsSubrMake (interp, ">=", OsSubrGTE, OS_NARY) ;
    OsSubrMake (interp, "min", OsSubrMin, OS_NARY) ;
    OsSubrMake (interp, "max", OsSubrMax, OS_NARY) ;
    OsSubrMake (interp, "+", OsSubrPlus, OS_NARY) ;
    OsSubrMake (interp, "-", OsSubrMinus, OS_NARY) ;
    OsSubrMake (interp, "*", OsSubrMult, OS_NARY) ;
    OsSubrMake (interp, "/", OsSubrDiv, OS_NARY) ;

    OsSubrMake (interp, "quotient", OsSubrQuotient, 2) ;
    OsSubrMake (interp, "remainder", OsSubrRemainder, 2) ;
    OsSubrMake (interp, "modulo", OsSubrModulo, 2) ;
    OsSubrMake (interp, "odd?", OsSubrOddP, 1) ;
    OsSubrMake (interp, "even?", OsSubrEvenP, 1) ;

    OsSubrMake (interp, "integer->string", OsSubrIntegerToString, 1) ;
    OsSubrMake (interp, "string->integer", OsSubrStringToInteger, 1) ;
    
    OsSubrMake (interp, "logor", OsSubrLogor, OS_NARY) ;
    OsSubrMake (interp, "logand", OsSubrLogand, OS_NARY) ;
    OsSubrMake (interp, "logxor", OsSubrLogxor, 2) ;
    OsSubrMake (interp, "lognot", OsSubrLognot, 1) ;

    OsSubrMake (interp, "zero?", OsSubrZeroP, 1) ;
    OsSubrMake (interp, "negative?", OsSubrNegativeP, 1) ;
    OsSubrMake (interp, "positive?", OsSubrPositiveP, 1) ;
    OsSubrMake (interp, "abs", OsSubrAbs, 1) ;

    OsSubrMake (interp, "floor", OsSubrFloor, 1) ;
    OsSubrMake (interp, "ceiling", OsSubrCeiling, 1) ;
    OsSubrMake (interp, "truncate", OsSubrTruncate, 1) ;
    OsSubrMake (interp, "round", OsSubrRound, 1) ;

    OsSubrMake (interp, "string->number", OsSubrStringToNumber, OS_NARY) ;
    OsSubrMake (interp, "number->string", OsSubrNumberToString, 1) ;

    OsSubrMake (interp, "inexact->exact", OsSubrInexactToExact, 1) ;
    OsSubrMake (interp, "exact->inexact", OsSubrExactToInexact, 1) ;
    return ;
}

/***************************************************************************
 * OsFloat:
 * --------
 * Float stuff
 ***************************************************************************/

OsFloat OsFloatMake (double value) {
    OsFloat flt = (OsFloat) OsObjectMake (OsFloatClass, 0) ;
    flt->fval = value ;
    return (flt) ;
}

OsFloat OsFloatMakeFromCString (char *string) {
    double fval ;
    sscanf (string, "%lf", &fval) ;
    return (OsFloatMake (fval)) ;
}

/*
 * Float class and methods:
 */

static OsObject os_float_display (OsFloat self, int sel, OsOutputPort port) {
    char tmp[OS_DISPLAY_BUFSIZ] ;
    sprintf (tmp, "%g", OsFloatValue (self)) ;
    OsOutputPortWriteCString (port, tmp) ;
    return ((OsObject) self) ;
}

static OsObject os_float_equal (OsFloat self, int sel, OsFloat other) {
    if ( OsIsFloat (other) )
        return ((OsFloatValue(self) == OsFloatValue(other))
                ? OsA_True
                : OsA_False) ;
    return (OsA_False) ;
}

void OsFloatInitialize (OsInterp interp, int *pargc, char **argv) {
    OsFloatClass = OsClassDeclare ("Float", sizeof (OsFloatRec)) ;
    OsDefineMsg (OsFloatClass, OS_SEL_DISPLAY, os_float_display) ;
    OsDefineMsg (OsFloatClass, OS_SEL_WRITE, os_float_display) ;
    OsDefineMsg (OsFloatClass, OS_SEL_EQUAL, os_float_equal) ;
    return ;
}

void OsFloatSubrInitialize (OsInterp interp) {
    return ;
}
