/***************************************************************************
 * bcomp.c
 * bcomp.c,v 1.2 1995/03/17 14:17:47 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * OScheme byte compiler, and the various controls it provides.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <oscheme.h>

#include "os_bcode.h"

static OsAtom   osa_trace_compiler  = (OsAtom) 0 ;
static OsAtom   osa_opt_inline_subr = (OsAtom) 0 ;
static OsAtom   osa_opt_debug       = (OsAtom) 0 ;

static OsObject osa_super           = (OsObject) 0 ;
static OsObject osa_send_msg        = (OsObject) 0 ;
static OsObject osa_lambda          = (OsObject) 0 ;

static OsError ose_compiler_error   = (OsError) 0 ;

/***************************************************************************
 * Compile time utilities
 ***************************************************************************/

typedef struct _OsCompEnv {
    unsigned long   options ;
    unsigned long   mode ;
    unsigned char   arity ;
    unsigned char   tail_depth ;
    unsigned char   tail_top ;
    OsVector	    env ;
    OsVector        data ;
    OsHashtable     datalib ;
    unsigned int    codelen ;
    unsigned char   *code ;
    unsigned char   *pc ;
} OsCompEnvRec ;

/*
 * Options management:
 */

#define OS_OPT_INLINE_SUBR	(1<<0)
#define OS_OPT_TAIL_CALL	(1<<1)
#define OS_OPT_DEBUG		(1<<2)

#define OS_INLINE_SUBRP(c)	((c)->options & OS_OPT_INLINE_SUBR)
#define OS_TAIL_CALLP(c)	((c)->options & OS_OPT_TAIL_CALL)
#define OS_DEBUGP(c)		((c)->options & OS_OPT_DEBUG)

#define OS_COMP_OPT(c,o)	((c)->options |= (o))

/*
 * Mode management macros:
 * -----------------------
 */

#define OS_MODE_PUSH		(1<<0)
#define OS_MODE_TAIL		(1<<1)
#define OS_MODE_FUNC		(1<<2)
#define OS_MODE_INLINED		(1<<3)
#define OS_MODE_TAILED		(1<<4)
#define OS_MODE_FUNCTAIL	(1<<5)	/* preserved */
#define OS_MODE_FUNCINLINED	(1<<6)

#define OS_PUSHP(c)		((c)->mode & OS_MODE_PUSH)
#define OS_TAILP(c)		((c)->mode & OS_MODE_TAIL)
#define OS_FUNCP(c)		((c)->mode & OS_MODE_FUNC)
#define OS_INLINEDP(c)		((c)->mode & OS_MODE_INLINED)
#define OS_TAILEDP(c)		((c)->mode & OS_MODE_TAILED)
#define OS_FUNCTAILP(c)		((c)->mode & OS_MODE_FUNCTAIL)
#define OS_FUNCINLINEDP(c)	((c)->mode & OS_MODE_FUNCINLINED)

#define OS_COMP_ARITY(c,a)	((c)->arity = (a))

static INLINE unsigned long OS_COMP_MODE (OsCompEnv cenv, unsigned long mode) {
    unsigned long oldmode = cenv->mode ;
    cenv->mode = (mode | (oldmode & OS_MODE_FUNCTAIL)) ;
    return (oldmode) ;
}

static INLINE unsigned long OS_COMP_ORMODE(OsCompEnv cenv, unsigned long mode){
    unsigned long oldmode = cenv->mode ;
    cenv->mode |= mode ;
    return (oldmode) ;
}

static void os_compenv_mdisplay (OsCompEnv cenv) {
    printf ("%cush %cail %cunc %cnlined %cailed %cunctail %cuncinline.\n"
            , (cenv->mode & OS_MODE_PUSH) ? 'P' : 'p'
            , (cenv->mode & OS_MODE_TAIL) ? 'T' : 't'
            , (cenv->mode & OS_MODE_FUNC) ? 'F' : 'f'
            , (cenv->mode & OS_MODE_INLINED) ? 'I' : 'i'
            , (cenv->mode & OS_MODE_TAILED) ? 'T' : 't'
            , (cenv->mode & OS_MODE_FUNCTAIL) ? 'F' : 'f'
            , (cenv->mode & OS_MODE_FUNCINLINED) ? 'F' : 'f') ;
    return ;
}

static void os_compenv_cdisplay (OsCompEnv cenv) {
    unsigned char *pc = cenv->code ;
    while ( pc < cenv->pc )
        os_decode_instr (&pc, (OsObject *) 0) ;
    return ;
}
    
/*
 * Environment management:
 * -----------------------
 * A compile time frame looks like:
 * #(<previous frame> <funcp> {<var>}+)
 * <funcP. is a boolean indicating whether the frame should be unlink'ed
 * through UNLINK instr (instead of UNLINKEXT, if flag is false).
 * The index of var minus *one* in the vector is its position in the runtime
 * stack, because of the saved env register.
 */

static void os_env_pushframe (OsCompEnv cenv, OsCons lframe) {
    OsCons   local      = lframe ;
    unsigned int nlocal = OsConsLength (local), i = 2 ;
    OsVector vframe     = OsVectorNMakeWithDefault (2+nlocal, OsA_Unbound) ;
    /* Assign to each variable a location on the stack. Allocate now the vector
       to be returned throuigh os_env_lookup, even if the frame index can has
       to be recomputed each time... */
    vframe->cells[0] = (OsObject) cenv->env ;
    vframe->cells[1] = (OS_FUNCP(cenv)) ? OsA_True : OsA_False ;
    while ( i < nlocal + 2) {
        vframe->cells[i++]   = OS_CONS_CAR (local) ;
        local = (OsCons) OS_CONS_CDR (local) ;
    }
    cenv->env = vframe ;
    return ;
}

static void os_env_popframe (OsCompEnv cenv) {
    if ( cenv->env == (OsVector) OsA_Nil ) 	/* sanity check */
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("os_env_popframe")
                               , OsStringMake ("no env to pop !")) ;
    cenv->env = (OsVector) cenv->env->cells[0] ;
    return ;
}

static int os_env_lookup (cenv, atom, pfloc, pvloc) 
    OsCompEnv cenv ;
    OsAtom    atom ;
    unsigned char *pfloc, *pvloc ;
{
    OsVector vframe = cenv->env ;
    unsigned char floc = (unsigned char) 0, vloc ;
    while ( vframe != (OsVector) OsA_Nil ) {
        vloc = (unsigned char) 2 ;
        while ( vloc < (unsigned char) vframe->csize ) {
            if ( ((OsObject) atom) == vframe->cells[vloc] ) {
                *pfloc = floc ;
                *pvloc = vframe->csize - vloc ;
                return (1) ;
            }
            vloc++ ;
        }
        floc++ ;
        vframe = (OsVector) vframe->cells[0] ;
    }
    return (0) ;
}

static unsigned char os_env_size (OsCompEnv cenv, unsigned char depth) {
    /* Returns the size of env at given depth */
    OsVector vframe = cenv->env ;
    while ( depth-- ) {
        if ( vframe == (OsVector) OsA_Nil )
            (void) OsErrorTrigger (1, OsE_Implementation
                                   , OsStringMake ("os_env_size")
                                   , OsStringMake ("null env !")) ;
        vframe = (OsVector) vframe->cells[0] ;
    }
    return ((unsigned char) (vframe->csize - 2)) ;
}

static INLINE unsigned char os_env_framesize (OsVector vframe) {
    return ((unsigned char) vframe->csize - 2) ;
}

static unsigned char os_env_depth (OsCompEnv cenv) {
    OsVector vframe = cenv->env ;
    unsigned char depth = 0 ;
    while ( vframe != (OsVector) OsA_Nil ) {
        depth++ ;
        vframe = (OsVector) vframe->cells[0] ;
    }
    return (depth) ;
}
                                                    
static int os_env_globalp (OsCompEnv cenv) {
    /* Is the current comp env the global one (needed to compile define) */
    return (cenv->env == (OsVector) OsA_Nil) ;
}

static int os_env_funcp (OsVector vframe) {
    return (vframe->cells[1] == OsA_True) ;
}

/*
 * Constant management:
 * --------------------
 * We could share more data types here...this just depends on the hash/cmp
 * function we use for the datalib stuff (currently defaults to Pointer)
 */
    
unsigned short os_data_register (OsCompEnv cenv, OsObject object) {
    unsigned long didx ;
    if ( (didx = (unsigned long) OsHashtableGet (cenv->datalib, object)) ) {
        return ((unsigned short) didx) ;
    } else {
        unsigned long didx = (unsigned long) OsVectorFastAppend (cenv->data
                                                                 , object) ;

        if ( OS_CHECK_SHORT (didx) )
            (void) OsErrorTrigger (1, OsE_Implementation
                                   , OsStringMake("os_data_register")
                                   , OsStringMake("max data vector exceeded"));
        OsHashtablePut (cenv->datalib, object, (void *) didx) ;
        return ((unsigned short) didx) ;
    }
    /* NOT REACHED */
}

/*
 * Special form support:
 * ---------------------
 * A special gets a special compiler function. The CAR of the special form
 * is used to dispatch to the appropriate compiler function.
 * NOTE: This utility uses the fact that (void *) = (void (*)()), which is
 * not portable (on PC for example).
 */

typedef OsObject (*OsSpecialFormCompiler)() ;
    
static OsHashtable os_special_hashtable = (OsHashtable) 0 ;

void os_special_form (OsAtom atom, OsSpecialFormCompiler compiler) {
    OsHashtablePut (os_special_hashtable, atom, compiler) ;
    return ;
}

OsSpecialFormCompiler os_special_form_lookup (OsAtom atom) {
    return ((OsSpecialFormCompiler) OsHashtableGet (os_special_hashtable
                                                    , atom)) ;
}

/*
 * Code emitting:
 * --------------
 * I have tried to avoid using the os_emit_byte like functions to keep
 * a proper interface between the exact code to emit and the function it
 * performs. 
 * I use inline functions here, also I could use macros.
 */

static INLINE unsigned char os_code_realloc (OsCompEnv cenv
                                             , unsigned char byte) {
    unsigned long ipc = cenv->pc - cenv->code ;
    cenv->codelen += OS_COMPILER_CODE_INCREMENT ;
    cenv->code = (unsigned char *) OsGCRealloc (cenv->code, cenv->codelen) ;
    cenv->pc = cenv->code + ipc ;
    return (*(cenv->pc++) = byte) ;
}

#define OS_EMIT_BYTE(c,b) \
    ((((c)->pc - (c)->code) < (c)->codelen) \
     ? *((c)->pc++) = (b) \
     : os_code_realloc ((c),(b)))

static INLINE void os_bump_code (OsCompEnv cenv
                                 , unsigned long start
                                 , unsigned long offset)
{
    unsigned char *src = cenv->code + start + offset ;
    unsigned char *dst = cenv->code + start ;
    unsigned long len  = (cenv->pc - cenv->code) - (start + offset) ;
    while ( len-- )
        *dst++ = *src++ ;
    cenv->pc = dst ;
    return ;
}

static INLINE void os_emit_byte (OsCompEnv cenv, unsigned char instr) {
    OS_EMIT_BYTE (cenv, instr) ;
}

static INLINE void os_emit_noop (OsCompEnv cenv, unsigned int count) {
    while ( count-- ) 
        os_emit_byte (cenv, OS_B_NOOP) ;
    return ;
}
    
static INLINE void os_emit_2byte (OsCompEnv cenv
                                  , unsigned char b1
                                  , unsigned char b2) {
    OS_EMIT_BYTE (cenv, b1) ; OS_EMIT_BYTE (cenv, b2) ;
}

/* !!! Must be consistent with OS_PUT_SHORT */
static INLINE void os_emit_short (OsCompEnv cenv, short s) {
    OS_EMIT_BYTE (cenv, (unsigned char) 0) ;
    OS_EMIT_BYTE (cenv, (unsigned char) 0) ;
    cenv->pc -= 2 ;
    OS_PUT_SHORT (cenv->pc, s) ;
    return ;
}

static INLINE void os_emit_cst (OsCompEnv cenv, OsObject object) {
    unsigned short didx = os_data_register (cenv, object) ;
    os_emit_byte (cenv, (OS_PUSHP(cenv) ? OS_B_CSTPUSH : OS_B_CST)) ;
    os_emit_short (cenv, didx) ;
    return ;
}

static INLINE void os_emit_localref (OsCompEnv cenv
                                     , unsigned char floc
                                     , unsigned char vloc) {
    switch (floc) {
      case 0:
        OS_EMIT_BYTE (cenv, (OS_PUSHP (cenv) ? OS_B_LREF0PUSH : OS_B_LREF0)) ;
        OS_EMIT_BYTE (cenv, vloc) ;
        return ;
      case 1:
        OS_EMIT_BYTE (cenv, (OS_PUSHP (cenv) ? OS_B_LREF1PUSH : OS_B_LREF1)) ;
        OS_EMIT_BYTE (cenv, vloc) ;
        return ;
      case 2:
        OS_EMIT_BYTE (cenv, (OS_PUSHP (cenv) ? OS_B_LREF2PUSH : OS_B_LREF2)) ;
        OS_EMIT_BYTE (cenv, vloc) ;
        return ;
      case 3:
        OS_EMIT_BYTE (cenv, (OS_PUSHP (cenv) ? OS_B_LREF3PUSH : OS_B_LREF3)) ;
        OS_EMIT_BYTE (cenv, vloc) ;
        return ;
      default:
        OS_EMIT_BYTE (cenv, (OS_PUSHP (cenv) ? OS_B_LREFNPUSH : OS_B_LREFN)) ;
        OS_EMIT_BYTE (cenv, floc) ;
        OS_EMIT_BYTE (cenv, vloc) ;
        return ;
    }
    /* NOT REACHED */
}

static INLINE void os_emit_localset (OsCompEnv cenv
                                     , unsigned char floc
                                     , unsigned char vloc) {
    switch (floc) {
      case 0:
        OS_EMIT_BYTE (cenv, OS_B_LSET0) ;
        OS_EMIT_BYTE (cenv, vloc) ;
        break ;
      case 1:
        OS_EMIT_BYTE (cenv, OS_B_LSET1) ;
        OS_EMIT_BYTE (cenv, vloc) ;
        break ;
      case 2:
        OS_EMIT_BYTE (cenv, OS_B_LSET2) ;
        OS_EMIT_BYTE (cenv, vloc) ;
        break ;
      case 3:
        OS_EMIT_BYTE (cenv, OS_B_LSET3) ;
        OS_EMIT_BYTE (cenv, vloc) ;
        break ;
      default:
        OS_EMIT_BYTE (cenv, OS_B_LSETN) ;
        OS_EMIT_BYTE (cenv, floc) ;
        OS_EMIT_BYTE (cenv, vloc) ;
        break ;
    }
    if ( OS_PUSHP (cenv) )
        os_emit_byte (cenv, OS_B_PUSH) ;
    return ;
}

static INLINE void os_emit_globalref (OsCompEnv cenv, OsAtom atom) {
    unsigned short didx = os_data_register (cenv, (OsObject) atom) ;
    OS_EMIT_BYTE (cenv, (OS_PUSHP (cenv) ? OS_B_GREFPUSH : OS_B_GREF)) ;
    os_emit_short (cenv, didx) ;
    return ;
}

static INLINE void os_emit_globalset (OsCompEnv cenv, OsAtom atom) {
    unsigned short didx ;
    didx = os_data_register (cenv, (OsObject) atom) ;
    OS_EMIT_BYTE (cenv, OS_B_GSET);
    os_emit_short (cenv, didx) ;
    if ( OS_PUSHP (cenv) )
        os_emit_byte (cenv, OS_B_PUSH) ;
    return ;
}

static INLINE void os_emit_globaldef (OsCompEnv cenv, OsAtom atom) {
    unsigned short didx ;
    didx = os_data_register (cenv, (OsObject) atom) ;
    OS_EMIT_BYTE (cenv, OS_B_GDEF);
    os_emit_short (cenv, didx) ;
    return ;
}

static INLINE void os_emit_link (OsCompEnv cenv, OsVector frame) {
    unsigned short didx ;
    didx = os_data_register (cenv, (OsObject) OsVectorTail(frame,2)) ;
    if ( OS_FUNCP (cenv) )
        OS_EMIT_BYTE (cenv, OS_B_LINK) ;
    else
        OS_EMIT_BYTE (cenv, OS_B_LINKEXT) ;
    os_emit_short (cenv, didx) ;
    return ;
}

static INLINE void os_emit_narylink (OsCompEnv cenv, OsVector frame) {
    unsigned short didx ;
    didx = os_data_register (cenv, (OsObject) OsVectorTail(frame,2)) ;
    if ( OS_FUNCP (cenv) )
        OS_EMIT_BYTE (cenv, OS_B_LINKN) ;
    else
        OS_EMIT_BYTE (cenv, OS_B_LINKNEXT) ;
    os_emit_short (cenv, didx) ;
    return ;
}

static INLINE unsigned long os_emit_unlink (OsCompEnv cenv
                                            , unsigned char howmany) {
    if ( OS_FUNCP (cenv) ) 
        OS_EMIT_BYTE (cenv, OS_B_UNLINK) ;
    else
        OS_EMIT_BYTE (cenv, OS_B_UNLINKEXT) ;
    OS_EMIT_BYTE (cenv, howmany) ;
    return ((unsigned long) (cenv->pc - cenv->code) - 2) ;
}


static INLINE unsigned long os_emit_tailunlink (OsCompEnv cenv) {
    /* Emit all unlinks needed to go back to tail_depth. If the last one
     * is a unlinkext, than we also emit a return, because in this case
     * ut is supposed there is no continuation */
    unsigned long iloc   = cenv->pc - cenv->code ;
    unsigned char depth  = cenv->tail_depth ;
    OsVector      vframe = cenv->env ;
    unsigned int  funcp  = 0 ;

    if ( depth ) {
        while ( depth-- ) {
            if ( (funcp = os_env_funcp (vframe)) )
                OS_EMIT_BYTE (cenv, OS_B_UNLINK) ;
            else
                OS_EMIT_BYTE (cenv, OS_B_UNLINKEXT) ;
            OS_EMIT_BYTE (cenv, os_env_framesize (vframe)) ;
            vframe = (OsVector) vframe->cells[0] ;
        }
        if ( ! funcp )
            OS_EMIT_BYTE (cenv, OS_B_RETURN) ;
    }
    return (iloc) ;
}

static void os_emit_sendmsg (OsCompEnv cenv, OsEObjSel selector) {
    unsigned short didx = os_data_register (cenv, (OsObject) selector) ;
    switch (selector->arity) {
      case 0:  os_emit_byte (cenv, OS_B_SENDMSG0) ; break ;
      case 1:  os_emit_byte (cenv, OS_B_SENDMSG1) ; break ;
      case 2:  os_emit_byte (cenv, OS_B_SENDMSG2) ; break ;
      case 3:  os_emit_byte (cenv, OS_B_SENDMSG3) ; break ;
      case 4:  os_emit_byte (cenv, OS_B_SENDMSG4) ; break ;
      case 5:  os_emit_byte (cenv, OS_B_SENDMSG5) ; break ;
      case 6:  os_emit_byte (cenv, OS_B_SENDMSG6) ; break ;
      case 7:  os_emit_byte (cenv, OS_B_SENDMSG7) ; break ;
      default: os_emit_byte (cenv, OS_B_SENDNMSG) ; break ;
    }
    os_emit_short (cenv, didx) ;
    return ;
}
    
static void os_emit_sendsupermsg (OsCompEnv cenv, OsEObjSel selector) {
    unsigned short didx = os_data_register (cenv, (OsObject) selector) ;
    switch (selector->arity) {
      case 0:  os_emit_byte (cenv, OS_B_SENDSUPERMSG0) ; break ;
      case 1:  os_emit_byte (cenv, OS_B_SENDSUPERMSG1) ; break ;
      case 2:  os_emit_byte (cenv, OS_B_SENDSUPERMSG2) ; break ;
      case 3:  os_emit_byte (cenv, OS_B_SENDSUPERMSG3) ; break ;
      case 4:  os_emit_byte (cenv, OS_B_SENDSUPERMSG4) ; break ;
      case 5:  os_emit_byte (cenv, OS_B_SENDSUPERMSG5) ; break ;
      case 6:  os_emit_byte (cenv, OS_B_SENDSUPERMSG6) ; break ;
      case 7:  os_emit_byte (cenv, OS_B_SENDSUPERMSG7) ; break ;
      default: os_emit_byte (cenv, OS_B_SENDSUPERNMSG) ; break ;
    }
    os_emit_short (cenv, didx) ;
    return ;
}
    
/*
 * Patchable instr are theone that have rel-to-pc args
 * We use C stack to keep track of locations to patch (seems better than
 * manually handling them, ala xscheme).
 */

static INLINE unsigned long os_emit_patchable (OsCompEnv cenv
                                               , unsigned char instr
                                               , unsigned long skip) {
    unsigned long iloc ;
    OS_EMIT_BYTE (cenv, instr) ;
    iloc = (unsigned long) (cenv->pc - cenv->code) ;
    cenv->pc += skip ;
    return (iloc) ;
}

static INLINE void os_emit_patch (OsCompEnv cenv, unsigned long iloc) {
    unsigned char *topatch = cenv->code + iloc ;
    unsigned long radr = ((unsigned long) (cenv->pc - cenv->code)) - iloc ;
    if ( radr > 0xFFFF )
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("os_emit_patch")
                               , OsStringMake ("mex relative jump exceeded")) ;
    OS_PUT_SHORT (topatch, (unsigned short) radr) ;
    return ;
}

static INLINE void os_emit_patch_adj (OsCompEnv cenv
                                      , unsigned long iloc
                                      , int adjust) {
    unsigned char  *topatch = cenv->code + iloc, *old = topatch ;
    unsigned short prev = OS_GET_SHORT (old) ;
    OS_PUT_SHORT (topatch, ((unsigned short) (((int) prev) + adjust))) ;
    return ;
}

static INLINE void os_noopize (OsCompEnv cenv
                               , unsigned long iloc
                               , unsigned char howmany) {
    unsigned char *topatch = cenv->code + iloc ;
    while ( howmany-- )
        *topatch++ = OS_B_NOOP ;
    return ;
}

#define os_emit_save(c)      os_emit_patchable((c),OS_B_SAVE,sizeof(short))
#define os_emit_briffalse(c) os_emit_patchable((c),OS_B_BRIFFALSE,sizeof(short))
#define os_emit_briftrue(c)  os_emit_patchable((c),OS_B_BRIFTRUE,sizeof(short))
#define os_emit_goto(c)	     os_emit_patchable ((c),OS_B_RGOTO,sizeof(short))

#define os_patch_save            os_emit_patch
#define os_patch_briffalse       os_emit_patch
#define os_patch_briftrue        os_emit_patch
#define os_patch_goto            os_emit_patch
#define os_noopize_save(c,l)     os_noopize((c), l-1, 3)

static INLINE unsigned long os_emit_debugpush (cenv, expr)
    OsCompEnv cenv ;
    OsObject  expr ;
{
    unsigned short didx = os_data_register (cenv, expr) ;
    unsigned long  iloc = (cenv->pc - cenv->code) ;
    os_emit_byte (cenv, OS_DBG_PUSH) ;
    os_emit_short (cenv, didx) ;
    return (iloc) ;
}

static INLINE void os_safize_dbgpush (cenv, iloc)
    OsCompEnv     cenv ;
    unsigned long iloc ;
{
    cenv->code[iloc] = OS_DBG_SAFEPUSH ;
    return ;
}

static INLINE void os_emit_debugpop (cenv, expr)
    OsCompEnv cenv ;
    OsObject  expr ;
{
    os_emit_byte (cenv, OS_DBG_POP) ;
    return ;
}


/***************************************************************************
 * Compiler entry point
 ***************************************************************************/

static OsCodeBlock os_build_codeblock (OsCompEnv cenv) {
    unsigned char *pc = cenv->code, *jmp ;
    bcode_descr   d ;
    
    /* This post-processing performs indirect jump by-passing
     * We *do not* add or remove instructions, just patch indirect jumps */
    while ( pc < cenv->pc ) {
        d = byte_codes + (*pc) ;
        if (*pc++ == OS_B_RGOTO) {
            jmp = pc + OS_GET_SHORT(pc) - 2 ;
            switch (*jmp++) {
              case OS_B_RGOTO:
              case OS_B_BRIFFALSE:
              case OS_B_BRIFTRUE:
              case OS_B_SAVE:
                  if ( *jmp++ == OS_B_RGOTO ) {
                      do {
                          jmp += (OS_GET_SHORT(jmp) - 2) ;
                      } while ( *jmp++ == OS_B_RGOTO ) ;
                      OS_PUT_SHORT (pc, ((unsigned long) (jmp - pc)) & 0xFFFF);
                  }
            }
            pc += d->skip ;
        }
    }
    return (OsCodeBlockMake (cenv->code
                             , (unsigned long) (cenv->pc - cenv->code)
                             , cenv->data)) ;
    
}

#define OPT_P(a)	((a)->vcell == OsA_True)

#define OS_COMP_ENV_NEW(_venv) \
    OsCompEnvRec _env ; \
    _env.options    = ((OPT_P(osa_opt_inline_subr) ? OS_OPT_INLINE_SUBR : 0) |\
                       /* compiler borken if not */  OS_OPT_TAIL_CALL        |\
                       (OPT_P(osa_opt_debug)       ? OS_OPT_DEBUG       : 0));\
    _env.mode       = OS_MODE_FUNCTAIL | OS_MODE_TAIL ; \
    _env.tail_depth = 0 ; \
    _env.tail_top   = 0 ; \
    _env.env        = (OsVector) OsA_Nil ; \
    _env.data       = OsVectorMake (OS_COMPILER_DEFAULT_DATALEN) ; \
    _env.datalib    = OsHashtableMake (0, OsHashPointer, OsCmpPointer) ; \
    _env.codelen    = OS_COMPILER_DEFAULT_CODELEN ; \
    _env.code       = (unsigned char *) OsGCMallocAtomic (_env.codelen) ; \
    _env.pc         = _env.code ; \
    _venv = &(_env) ;

OsCodeBlock OsCompileExpr (OsObject expr) {
    OsCompEnv cenv ;
    OS_COMP_ENV_NEW (cenv) ;
#ifdef OS_COMPILER_TRACEABLE
    if ( osa_trace_compiler->vcell == OsA_True ) {
        printf ("***** New compilation *****\n") ;
        OsSendDisplay (expr, OsCurrentOutputPort) ;
        putchar ('\n') ;
    }
#endif /* OS_COMPILER_TRACEABLE */
    OsSendCompile (expr, cenv) ;
    os_emit_byte (cenv, OS_B_EXIT) ;
    return (os_build_codeblock (cenv)) ;

}


static OsObject os_compile_expr (OsObject expr, OsCompEnv cenv) {
    static int level = 0 ;
    int        local_level ;
    OsObject   retval ;
#ifdef OS_COMPILER_TRACEABLE
    if ( osa_trace_compiler->vcell == OsA_True ) {
        local_level = level++ ;
        printf ("----- \n") ;
        OsSendDisplay (expr, OsCurrentOutputPort) ;
        printf ("\nENTER(%3d):", local_level) ;
        os_compenv_mdisplay (cenv) ;
    }
#endif /* OS_COMPILER_TRACEABLE */
    retval = OsSendCompile (expr, cenv) ;
#ifdef OS_COMPILER_TRACEABLE
    if ( osa_trace_compiler->vcell == OsA_True ) {
        printf ("LEAVE(%3d):", local_level) ;
        os_compenv_mdisplay (cenv) ;
    }
#endif /* OS_COMPILER_TRACEABLE */

    return (retval) ;
}

/*
 * Compile a symbol access
 */

static OsObject os_comp_symbolref (atom, sel, cenv)
    OsAtom    atom ;
    int sel ;
    OsCompEnv cenv ;
{
    unsigned char floc, vloc ;
    OsObject value = atom->vcell ;

    /* Test for a localref */
    if ( os_env_lookup (cenv, atom, &floc, &vloc) )  {
        os_emit_localref (cenv, floc, vloc) ;
        return ((OsObject) atom) ;
    }
    /* Test for not inlinable globalref */
    if ( ( ! OS_FUNCP (cenv) )
        || (! OS_INLINE_SUBRP (cenv))
        || ( ! OsIsCodeBlock (value)) ) {
        os_emit_globalref (cenv, atom) ;
        return ((OsObject) atom) ;
    } else { /* Try to inline, if this is a subr */
        OsCodeBlock block     = (OsCodeBlock) value ;
        unsigned char *code   = block->code->string ;

        switch (*code) {
          case OS_B_SUBR0:
          case OS_B_ASUBR0:
            if ( cenv->arity != 0 ) {
                os_emit_2byte (cenv, OS_B_LOAD, cenv->arity) ;
                os_emit_byte (cenv, OS_B_ASUBR0) ;
            } else {
                os_emit_byte (cenv, OS_B_SUBR0) ;
            }
            break ;
          case OS_B_SUBR1:
          case OS_B_ASUBR1:
            if ( cenv->arity != 1 ) {
                os_emit_2byte (cenv, OS_B_LOAD, cenv->arity) ;
                os_emit_byte (cenv, OS_B_ASUBR1) ;
            } else {
                os_emit_byte (cenv, OS_B_SUBR1) ;
            }
            break ;
          case OS_B_SUBR2:
          case OS_B_ASUBR2:
            if ( cenv->arity != 2 ) {
                os_emit_2byte (cenv, OS_B_LOAD, cenv->arity) ;
                os_emit_byte (cenv, OS_B_ASUBR2) ;
            } else {
                os_emit_byte (cenv, OS_B_SUBR2) ;
            }
            break ;
          case OS_B_SUBR3:
          case OS_B_ASUBR3:
            if ( cenv->arity != 3 ) {
                os_emit_2byte (cenv, OS_B_LOAD, cenv->arity) ;
                os_emit_byte (cenv, OS_B_ASUBR3) ;
            } else {
                os_emit_byte (cenv, OS_B_SUBR3) ;
            }
            break ;
          case OS_B_SUBR4:
          case OS_B_ASUBR4:
            if ( cenv->arity != 4 ) {
                os_emit_2byte (cenv, OS_B_LOAD, cenv->arity) ;
                os_emit_byte (cenv, OS_B_ASUBR4) ;
            } else {
                os_emit_byte (cenv, OS_B_SUBR4) ;
            }
            break ;
          case OS_B_SUBR5:
          case OS_B_ASUBR5:
            if ( cenv->arity != 5 ) {
                os_emit_2byte (cenv, OS_B_LOAD, cenv->arity) ;
                os_emit_byte (cenv, OS_B_ASUBR5) ;
            } else {
                os_emit_byte (cenv, OS_B_SUBR5) ;
            }
            break ;
          case OS_B_SUBR6:
          case OS_B_ASUBR6:
            if ( cenv->arity != 6 ) {
                os_emit_2byte (cenv, OS_B_LOAD, cenv->arity) ;
                os_emit_byte (cenv, OS_B_ASUBR6) ;
            } else {
                os_emit_byte (cenv, OS_B_SUBR6) ;
            }
            break ;
          case OS_B_SUBR7:
          case OS_B_ASUBR7:
            if ( cenv->arity != 7 ) {
                os_emit_2byte (cenv, OS_B_LOAD, cenv->arity) ;
                os_emit_byte (cenv, OS_B_ASUBR7) ;
            } else {
                os_emit_byte (cenv, OS_B_SUBR7) ;
            }
            break ;
          case OS_B_ASUBRN:
            os_emit_2byte (cenv, OS_B_LOAD, cenv->arity) ;
            os_emit_byte (cenv, OS_B_ASUBRN) ;
            break ;
          default:
            os_emit_globalref (cenv, atom) ;
            return ((OsObject) atom) ;
        }
        /* put cfunc in data, and advert the inlining */
        os_emit_short (cenv,os_data_register(cenv, block->data->cells[0]));
        OS_COMP_ORMODE (cenv, OS_MODE_INLINED) ;
        return ((OsObject) atom) ;
    }
    /* NOT REACHED */
}

/*
 * Compile a constant:
 * This one is accessed by os_object.c, for default OS_SEL_COMPILE selector
 * handling.
 */

OsObject os_comp_constant (cst, sel, cenv)
    OsObject  cst ;
    int sel ;
    OsCompEnv cenv ;
{
    os_emit_cst (cenv, cst) ;
    return (cst) ;
}

/*
 * Compile an if expression:
 * The difficult part here is that if the 'if' occurs in tail position, and
 * one of the two branches of the if is TAILEDP, than the other one needs
 * to be patch to provide at least the unlink part of unlinkjump.
 * Anyway, at runtime, when reaching the end of the if, both branches should
 * be either TAILEDP, or not TAILEDP.
 */

static OsObject os_comp_if (icons, cenv)
    OsCons    icons ;
    OsCompEnv cenv ;
{
    OsObject expr = OS_CONS_CDR (icons) ;
    unsigned long oldmode ;
    unsigned long pfalse, pout ;
    unsigned long ptailed = 0 ;
    
    OsMustBeCons (expr) ;
    /* Compile comdition into acc */
    oldmode = OS_COMP_MODE (cenv, 0) ;
    os_compile_expr (OS_CONS_CAR (expr), cenv) ;
    pfalse = os_emit_briffalse (cenv) ;
    /* Compile the iftrue branch with the oldmode */
    OS_COMP_MODE (cenv, oldmode) ;
    expr = OS_CONS_CDR (expr) ;
    OsMustBeCons (expr) ;
    os_compile_expr (OS_CONS_CAR(expr), cenv) ;
    /* If the call was tailed, than code following it is unreachable (due
     * to the emited unlink[ext]jump instr. No need to emit a goto in this
     * case. Otherwise, the other branch might be tailed, so we prepare
     * this branch (potentially bumping the tailunlink - see below) */
    if ( ! OS_TAILEDP (cenv) ) {
        ptailed = os_emit_tailunlink (cenv) ;
        pout    = os_emit_goto (cenv) ;
    }
    /* Compile the iffalse branch with the orig oldmode */
    expr = OS_CONS_CDR (expr) ;
    if ( ! OsIsCons (expr) ) {
        if ( expr == OsA_Nil )
            os_emit_cst (cenv, OsA_False) ;
        else
            (void) OsErrorTrigger (1, ose_compiler_error
                                   , OsStringMake ("invalid iffalse branch")) ;
    }
    OS_COMP_MODE (cenv, oldmode) ;
    os_patch_briffalse (cenv, pfalse) ;
    os_compile_expr (OS_CONS_CAR (expr), cenv) ;
    if ( ( ! ptailed) && ( ! OS_TAILEDP (cenv) ) ) {
        /* iftrue was tailed, iffalse is not. tailify iffalse
         * No need to patch 'goto': it was not even emitted ! */
        OS_COMP_ORMODE (cenv, oldmode | OS_MODE_TAILED) ;
        os_emit_tailunlink (cenv) ;
    } else if ( ptailed && ( ! OS_TAILEDP (cenv)) ) {
        /* iftrue wasn't tailed nor iffalse. bump tailification of iftrue */
        os_patch_goto (cenv, pout) ;
        os_bump_code (cenv, ptailed, pout - 1 - ptailed) ;
        os_emit_patch_adj (cenv, pfalse, - ((int) pout-1-ptailed)) ;
    } else if ( ptailed ) {
        /* They are two remaining cases:
         * iftrue is tailed, iffalse also: no problems, we do not need to
         *    patch the (unemited) goto instr.
         * iftrue is not tailed, iffalse is tailed: leave the emited unlink */
        os_patch_goto (cenv, pout) ;
    }
    return ((OsObject) icons) ;
}

/*
 * Compile a cons...this is, of course the hard part !
 */

static OsObject os_comp_cons (icons, sel, cenv)
    OsCons    icons ;
    int sel ;
    OsCompEnv cenv ;
{
    OsObject fun  = OS_CONS_CAR (icons) ;
    OsCons   args = (OsCons) OS_CONS_CDR (icons) ;
    OsSpecialFormCompiler sform = os_special_form_lookup ((OsAtom) fun) ;
    unsigned long psave, pdbg ;
    unsigned long oldmode ;
    unsigned char arity = (unsigned char) 0 ;

    if ( sform )
        return ((sform)(icons, cenv)) ;

    if ( OS_DEBUGP (cenv) )
        pdbg = os_emit_debugpush (cenv, icons) ;
    else
        pdbg = 0 ;
    
    /* For this std function call, compile args in push-mode */
    psave = os_emit_save (cenv) ;
    oldmode = OS_COMP_MODE (cenv, OS_MODE_PUSH) ;
    while ( OsIsCons (args) ) {
        os_compile_expr (OS_CONS_CAR (args), cenv) ;
        args = ((OsCons) OS_CONS_CDR (args)) ;
        arity++ ;
    }
    if ( args != ((OsCons) OsA_Nil) ) {
        (void) OsErrorTrigger (1, ose_compiler_error
                               , OsStringMake ("invalid dot in application")) ;
    }
    /* Now, args have been pushed, compile fun in acc register, signaling
     * that we are compiling the object in functionnal position 
     * Note that the trailing oldmode reset, also clean the INLINE and
     * TAILED flags
     * I used to have (as of 2 Feb 95):
     * OS_COMP_MODE (cenv, OS_MODE_FUNC | (oldmode & OS_MODE_TAIL)) ;
     * But this was optimizing too muchm think of
     * (define (f x) ((foo x) x))
     * The call to (foo x) was inlined, making the rest unreachable.
     * The following new version seems more reasonable:
     */
    if ( OsIsCons (fun) && (OS_CONS_CAR (fun) == osa_lambda) )
        OS_COMP_MODE (cenv, OS_MODE_FUNC | (oldmode & OS_MODE_TAIL)) ;
    else
        OS_COMP_MODE (cenv, OS_MODE_FUNC) ;
    OS_COMP_ARITY (cenv, arity) ;
    os_compile_expr (fun, cenv) ;
    if ( ! OS_INLINEDP (cenv) ) {
        /* DEBUG */
        if (OS_COMP_OPT (cenv, OS_OPT_TAIL_CALL)
            && (oldmode & OS_MODE_TAIL)
            && ( ! OS_FUNCINLINEDP (cenv)) 
            && (os_env_depth (cenv) != cenv->tail_depth + cenv->tail_top)) {
            printf ("TAIL-CALL failed %d %d %d\n"
                    , os_env_depth (cenv), cenv->tail_depth, cenv->tail_top) ;
        }
        if (OS_COMP_OPT (cenv, OS_OPT_TAIL_CALL)
            && (oldmode & OS_MODE_TAIL)
            && ( ! OS_FUNCINLINEDP (cenv))
            && cenv->tail_depth
            && (os_env_depth (cenv) == cenv->tail_depth + cenv->tail_top)) {
            /* tail_depth start at 0, get incremented *before* the compilation
             * of the tail call, this is why we "- 1" it */
            if ( oldmode & OS_MODE_FUNCTAIL )
                os_emit_byte (cenv, OS_B_UNLINKJUMP) ;
            else
                os_emit_byte (cenv, OS_B_UNLINKEXTJUMP) ;
            os_emit_byte (cenv, (unsigned char) cenv->tail_depth - 1) ;
            os_emit_byte (cenv, arity) ;
            os_emit_byte (cenv, os_env_size (cenv, cenv->tail_depth - 1)) ;
            os_bump_code (cenv, psave - 1, 3) ;
            if ( pdbg )
                os_safize_dbgpush (cenv, pdbg) ;
            oldmode |= OS_MODE_TAILED ;
        } else if ( OS_FUNCINLINEDP (cenv)
                   /* && OS_FUNCP (cenv) */
                   /* && OS_TAILEDP (cenv) */ ) {
            /* This uncomprehensible test makes:
             * (((lambda (z) (lambda (x) (cons x z))) 2) 1)
             * ((lambda (v f) (f v v)) 1 cons)
             * BOTH compile properly. It was found by diffing comp mode
             * at this point. */
            if ( /*BUG*/ ( OS_TAILEDP (cenv)) && OS_FUNCTAILP (cenv) ) {
                os_patch_save (cenv, psave) ;
            } else {
                os_bump_code (cenv, psave - 1, 3) ;
            }
        } else {
            os_emit_2byte (cenv, OS_B_JUMP, arity) ;
            os_patch_save (cenv, psave) ;
        }
    } else {
        os_bump_code (cenv, psave - 1, 3) ;
    }
    /* Propagate back if we were in TAIL */
    if ( OS_TAILEDP (cenv) )
        oldmode |= OS_MODE_TAILED ;
    OS_COMP_MODE (cenv, oldmode) ;
    /* Finally, is this was to be compiled in push-mode, push the result */
    if ( OS_PUSHP (cenv) )
        os_emit_byte (cenv, OS_B_PUSH) ;

    if ( OS_DEBUGP (cenv) )
        os_emit_debugpop (cenv) ;
    return ((OsObject) icons) ;
}

/*
 * Compile a lambda:
 * -----------------
 * Depending on whether the lambda is in functional position or not, we
 * do inline the lambda or not (without closing it in the last case)
 * Otherwise, this lambda is in arg place, so we are to compile it "from
 * scratch", but in the current env, build the block, etc...
 */

/* 
 * When the condition indicated here is met, the arity will not be loaded
 * to acc through a jump instr. We need to explictly load it.
 */

static void os_lambda_arity_check (OsCompEnv cenv) {
    if ( OS_FUNCP (cenv) && OS_FUNCINLINEDP (cenv) )
        os_emit_2byte (cenv, OS_B_LOAD, cenv->arity) ;
}

static unsigned char os_comp_lambda_formals (OsCons formals, OsCompEnv cenv) {
    OsCons        formal = formals ;
    unsigned char arity  = (unsigned char) 0 ;
    /* Check for a degenerated formal */
    if ( ! OsIsCons (formals) ) {
        if ( formals == (OsCons) OsA_Nil ) {
            os_env_pushframe (cenv, (OsCons) OsA_Nil) ;
            os_lambda_arity_check (cenv) ;
            os_emit_link (cenv, cenv->env) ;
            return ((unsigned char) 0) ;
        } else if ( OsIsAtom (formals) ) {
            os_env_pushframe (cenv, OsConsMake ((OsObject) formals, OsA_Nil)) ;
            os_lambda_arity_check (cenv) ;
            os_emit_narylink (cenv, cenv->env) ;
            return ((unsigned char) 1) ;
        } else {
            (void) OsErrorTrigger (1, ose_compiler_error
                                   , OsStringMake ("bad formals in lambda")) ;
        }
    }
    /* Check for an dot-ended list */
    do {
        if ( ! OsIsAtom (OS_CONS_CAR(formal)) ) {
            (void) OsErrorTrigger (1, ose_compiler_error
                                   , OsStringMake ("formal should be atom")) ;
        }
        arity++ ;
        formal = (OsCons) OS_CONS_CDR (formal) ;
    } while ( OsIsCons (formal) ) ;
    if ( formal == (OsCons) OsA_Nil ) {
        /* Great, this is really easy */
        os_env_pushframe (cenv, formals) ;
        os_lambda_arity_check (cenv) ;
        os_emit_link (cenv, cenv->env) ;
        return (arity) ;
    } else {
        OsObject start, *list = &start ;
        /* We need to make a real list before pushframe (formal is doted) */
        if ( ! OsIsAtom (formal) ) {
            (void) OsErrorTrigger (1, ose_compiler_error
                                   , OsStringMake ("invalid formal: dotted")) ;
        }
        do {
            *list   = (OsObject) OsConsMake (OS_CONS_CAR(formals), OsA_Nil) ;
            list    = &(OS_CONS_CDR(*list)) ;
            formals = (OsCons) OS_CONS_CDR (formals) ;
        } while ( OsIsCons (formals) ) ;
        *list = (OsObject) OsConsMake ((OsObject) formal, OsA_Nil) ;
        /* We are done */
        arity++ ;
        os_env_pushframe (cenv, (OsCons) start) ;
        os_lambda_arity_check (cenv) ;
        os_emit_narylink (cenv, cenv->env) ;
        return (arity) ;
    }
    /* NOT REACHED */
}

static void os_comp_lambda_body (OsCons body, OsCompEnv cenv) {
    unsigned long oldmode = OS_COMP_MODE (cenv, 0) ;
    /* Check for a degenrated body (more than R4RS ! */
    if ( ! OsIsCons (body) ) {
        if ( body != (OsCons) OsA_Nil ) {
            (void) OsErrorTrigger (1, ose_compiler_error
                                   , OsStringMake ("invalid lambda nody")) ;
        }
        os_comp_constant (body, cenv) ;
        return ;
    }
    /* Compile as implicit begin, enforcing tail-call mode for last */
    while ( OsIsCons (OS_CONS_CDR(body)) ) {
        os_compile_expr (OS_CONS_CAR(body), cenv) ;
        body = (OsCons) OS_CONS_CDR (body) ;
    }
    if ( OS_CONS_CDR (body) != OsA_Nil ) {
        (void) OsErrorTrigger (1, ose_compiler_error
                               , OsStringMake("lambda body with dotted pair"));
    }
    /* As compile_expr might emit the unlink instr, the comp env should
     * adertize whether the englobing expr is in FUNC position or not (needed
     * to know whether a unlink or unlinkext instr is needed).
     */
    OS_COMP_MODE (cenv, oldmode & OS_MODE_TAIL ) ;
    /* cenv->tail_depth++ ; */
    os_compile_expr (OS_CONS_CAR (body), cenv) ;
    /* cenv->tail_depth-- ; */
    if ( OS_TAILEDP (cenv) ) {
        OS_COMP_MODE (cenv, oldmode | OS_MODE_TAILED) ;
    } else {
        OS_COMP_MODE (cenv, oldmode) ;
    }
    return ;
}

static void os_lambda_compile (OsAtom name
                               , OsCons formals
                               , OsCons body
                               , OsCompEnv cenv) {
    unsigned char arity ;

    cenv->tail_depth++ ;
    
    if ( OS_FUNCP (cenv) ) {
        /* We need set the FUNCINLINED flags right now (instead of at the end)
         * because comp_lambda_formals arity check emiting depends on whether
         * we are inlining the lambda or not.
         * The proper way would be to have a specific flag to comp_formals */
        OS_COMP_ORMODE (cenv, OS_MODE_FUNCINLINED) ;
        arity = os_comp_lambda_formals (formals, cenv) ;
        os_comp_lambda_body (body, cenv) ;
        os_env_popframe (cenv) ;
        if ( ! OS_TAILEDP (cenv) ) {
            os_emit_unlink (cenv, arity) ;
        }
    } else {
        /* We need to start the lambda compilation in a fresh environment
         * because we will be used as a function ret value. Although we
         * start a new comp env, the (compiled) lambda env is the current
         * env, hence the affectation to nenv->env.
         * Although we generate linkEXT/unlinkEXT instructions, we have to
         * wait for runtime in order to build the close block, because we
         * do want to *share* the result of evaluating the lambda.
         */
        OsCompEnv nenv ;
        OsCodeBlock block ;
        unsigned long oldmode ;
        OS_COMP_ENV_NEW (nenv) ;
        nenv->mode      &= (~OS_MODE_FUNCTAIL) ;
        nenv->env        = cenv->env ;
        nenv->tail_depth = 1 ;
        nenv->tail_top   = cenv->tail_top + cenv->tail_depth - 1 ;
        arity = os_comp_lambda_formals (formals, nenv) ;
        os_comp_lambda_body (body, nenv) ;
        os_env_popframe (nenv) ;
        /* If lambda has been tailed, than code following it is unreachable
         * so it does not even needs a return (which has been emitted by
         * emit_tailunlink) */
        if ( ! OS_TAILEDP (nenv) ) {
            os_emit_unlink (nenv, arity) ;
            os_emit_byte (nenv, OS_B_RETURN) ;
        } 
        block = os_build_codeblock (nenv) ;
        oldmode = OS_COMP_MODE (cenv, 0) ;	/* just load cst to acc */
        os_emit_cst (cenv, (OsObject) block) ;
        os_emit_byte (cenv, OS_B_CLOSE) ;
        OS_COMP_MODE (cenv, oldmode) ;		/* check for a push */
        if ( OS_PUSHP (cenv) )
            os_emit_byte (cenv, OS_B_PUSH) ;
    }

    cenv->tail_depth-- ;
    
    return ;
}

static OsObject os_comp_lambda (icons, cenv)
    OsCons icons ;
    OsCompEnv cenv ;
{
    OsCons cons = (OsCons) OS_CONS_CDR (icons) ;
    OsCons formals ;
    
    if ( ! OsIsCons (cons) ) {
        (void) OsErrorTrigger (1, ose_compiler_error
                               , OsStringMake("invalid 'lambda: no formals"));
    }
    formals = (OsCons) OS_CONS_CAR (cons) ;
    cons    = (OsCons) OS_CONS_CDR (cons) ;
    if ( ! OsIsCons (cons) ) {
        (void) OsErrorTrigger (1, ose_compiler_error
                               , OsStringMake ("invalid 'lambda: no body)")) ;
     }
    os_lambda_compile ((OsAtom) 0, formals, cons, cenv) ;
    return ((OsObject) icons) ;
}
    
/*
 * Compiling begin:
 * ----------------
 * Propagates cenv mode only to the last form
 * AUTODOC
 * (begin EXPRESSION*)
 * Evaluate EXPORESSION in sequence.
 * The EXPRESSIONs are evaluated sequentially from left to right, and the value
 * of the last EXPRESSION is returned.  This EXPRESSION type is used to 
 * sequence side effects such as input and output.
 */

static OsObject os_comp_begin (icons, cenv)
    OsCons icons ;
    OsCompEnv cenv ;
{
    OsCons cons ;
    unsigned long oldmode = OS_COMP_MODE (cenv, 0) ;

    /* Check for degenerate begin form */
    if ( ! OsIsCons (icons) ) {
        if ( icons != (OsCons) OsA_Nil ) {
            (void) OsErrorTrigger (1, ose_compiler_error
                                   , OsStringMake ("empty 'begin")) ;
        }
        os_comp_constant (icons, cenv) ;
        return ((OsObject) icons) ;
    }
    cons = (OsCons) OS_CONS_CDR (icons) ;
    while ( OsIsCons (OS_CONS_CDR (cons)) ) {
        os_compile_expr (OS_CONS_CAR (cons), cenv) ;
        cons = (OsCons) OS_CONS_CDR (cons) ;
    }
    if ( OS_CONS_CDR (cons) != OsA_Nil ) {
        (void) OsErrorTrigger (1, ose_compiler_error
                               , OsStringMake ("dotted pair in 'begin form")) ;
    }
    OS_COMP_MODE (cenv, oldmode) ;
    os_compile_expr (OS_CONS_CAR (cons), cenv) ;
    return ((OsObject) icons) ;
}

/*
 * Compiling set!:
 *----------------
 * Nothing special...
 */

static OsObject os_comp_setbang (icons, cenv)
    OsCons icons ;
    OsCompEnv cenv ;
{
    unsigned long oldmode = OS_COMP_MODE (cenv, 0) ;
    OsCons cons = (OsCons) OS_CONS_CDR (icons) ;
    OsAtom atom ;
    unsigned char floc, vloc ;

    if ( ! OsIsCons (cons) ) {
        (void) OsErrorTrigger (1, ose_compiler_error
                               , OsStringMake ("invalid 'set!: no atom"));
    }
    atom = (OsAtom) OS_CONS_CAR (cons) ;
    if ( ! OsIsAtom (atom) ) {
        (void) OsErrorTrigger (1, ose_compiler_error
                               , OsStringMake ("invalid lvalue for set!")) ;
    }
    cons = (OsCons) OS_CONS_CDR (cons) ;
    if ( ! OsIsCons (cons) ) {
        (void) OsErrorTrigger (1, ose_compiler_error
                               , OsStringMake ("invalid 'set!: no value"));
    }    
    os_compile_expr (OS_CONS_CAR (cons), cenv) ;
    OS_COMP_MODE (cenv, oldmode) ;
    if ( os_env_lookup (cenv, atom, &floc, &vloc) ) 
        os_emit_localset (cenv, floc, vloc) ;
    else
        os_emit_globalset (cenv, atom) ;
    OS_COMP_MODE (cenv, oldmode) ;
    return ((OsObject) icons) ;
}

/*
 * Compiling define:
 */

static OsObject os_comp_define (icons, cenv)
    OsCons icons ;
    OsCompEnv cenv ;
{
    OsCons cons = (OsCons) OS_CONS_CDR (icons) ;
    OsAtom name ;
    unsigned long oldmode ;
    
    if ( ! OsIsCons (cons) ) {
        (void)OsErrorTrigger (1, ose_compiler_error
                              , OsStringMake ("invalid define form: no atom"));
    }
    oldmode = OS_COMP_MODE (cenv, cenv->mode & OS_MODE_FUNCTAIL) ;
    if ( OsIsCons (OS_CONS_CAR (cons)) ) {
        OsCons header  = (OsCons) OS_CONS_CAR (cons) ;
        OsCons formals = (OsCons) OS_CONS_CDR (header) ;
        name = (OsAtom) OS_CONS_CAR (header) ;
        if ( ! OsIsAtom (name) ) {
            (void) OsErrorTrigger (1, ose_compiler_error
                                   , OsStringMake ("invalid lambda name")) ;
        }
        os_lambda_compile (name, formals, (OsCons) OS_CONS_CDR (cons), cenv) ;
    } else {
        name = (OsAtom) OS_CONS_CAR (cons) ;
        if ( ! OsIsAtom (name) ) {
            (void) OsErrorTrigger (1, ose_compiler_error
                                   , OsStringMake ("invalid func parameter")) ;
         }
        cons = (OsCons) OS_CONS_CDR (cons) ;
        if ( ! OsIsCons (cons) ) {
            (void)OsErrorTrigger (1, ose_compiler_error
                                  , OsStringMake ("invalid define: no value"));
        }
        os_compile_expr (OS_CONS_CAR (cons), cenv) ;
    }
    /* Atom value is now in register, set atom */
    OS_COMP_MODE (cenv, oldmode) ;
    os_emit_globaldef (cenv, name) ;
/*    
    if ( os_env_globalp (cenv) ) {
        os_emit_globaldef (cenv, name) ;
    } else {
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("os_comp_define")
                               , OsStringMake ("define in lex scope")) ;
    }
 */
    return ((OsObject) icons) ;
}

/*
 * Compiling quote (easy ! but needed)
 */

static OsObject os_comp_quote (icons, cenv)
    OsCons icons ;
    OsCompEnv cenv ;
{
    OsCons cons = (OsCons) OsMustBeCons (icons) ;
    cons = (OsCons) OS_CONS_CDR (cons) ;
    OsMustBeCons (cons) ;
    os_emit_cst (cenv, OS_CONS_CAR (cons)) ;
    return ((OsObject) icons) ;
}

/*
 * Compiling or
 * AUTODOC
 * (or TEST*)
 * Returns the value of the first TEST that doesn't evaluate to #f.
 * The TEST expressions are evaluated from left to right, and the value of
 * the first expression that evaluates to a true value is returned.  Any
 * remaining expressions are not evaluated.  If all expressions evaluate to
 * false values, the value of the last expression is returned.  If there are
 * no expressions then is returned.
 * Example:
 * (or (= 2 2) (> 2 1)) 
 * = #t 
 */

static OsObject os_comp_or (icons, cenv)
    OsCons icons ;
    OsCompEnv cenv ;
{
    OsCons   cons   = (OsCons) OS_CONS_CDR (icons) ;
    OsVector vpatch = OsVectorMake (0) ;
    OsObject *cells, *ecells ;
    unsigned long oldmode ;
    
    if ( cons == (OsCons) OsA_Nil ) {
        os_emit_cst (cenv, OsA_False) ;
    } else {
        oldmode = OS_COMP_MODE (cenv, 0) ;
        OsMustBeCons (cons) ;
        while ( OsIsCons (OS_CONS_CDR(cons)) ) {
            OS_COMP_MODE (cenv, 0) ;
            os_compile_expr (OS_CONS_CAR (cons), cenv) ;
            OsVectorFastAppend (vpatch, (OsObject) os_emit_briftrue (cenv)) ;
            cons = (OsCons) OS_CONS_CDR (cons) ;
        }
        OS_COMP_MODE (cenv, oldmode) ;
        os_compile_expr (OS_CONS_CAR(cons), cenv) ;
        if ( OS_CONS_CDR(cons) != OsA_Nil ) {
            (void) OsErrorTrigger (1, ose_compiler_error
                                   , OsStringMake ("invalid expr: dotted")) ;
        }
        cells  = vpatch->cells ;
        ecells = vpatch->cells + vpatch->csize ;
        if ( (! OS_TAILEDP (cenv)) && (oldmode & OS_MODE_PUSH) ) {
            unsigned int pgoto = os_emit_goto (cenv) ;
            while ( cells < ecells ) {
                os_patch_briffalse (cenv, (unsigned long) *cells) ;
                cells++ ;
            }
            os_emit_byte (cenv, OS_B_PUSH) ;
            os_patch_goto (cenv, pgoto) ;
        } else {
            unsigned int pgoto ;
            if ( OS_TAILEDP (cenv) ) { /* tailed, but might have a cont */
                pgoto = os_emit_goto (cenv) ;
                while ( cells < ecells ) {
                    os_patch_briffalse (cenv, (unsigned long) *cells) ;
                    cells++ ;
                }
                os_emit_tailunlink (cenv) ;
                os_patch_goto (cenv, pgoto) ;
            } else {
                while ( cells < ecells ) {
                    os_patch_briffalse (cenv, (unsigned long) *cells) ;
                    cells++ ;
                }
            }
            if ( oldmode & OS_MODE_PUSH )
                os_emit_byte (cenv, OS_B_PUSH) ;
        }
    }
    return ((OsObject) icons) ;
}

/*
 * Compile and
 * AUTODOC
 * (and TEST*)
 * Returns the value of the first TEST that evaluates to #f.
 * The TEST expressions are evaluated from left to right, and the value of
 * the first expression that evaluates to a false value is returned.  Any
 * remaining expressions are not evaluated.  If all the expressions evaluate
 * to true values, the value of the last expression is returned.  If there
 * are no expressions then is returned. 
 */

static OsObject os_comp_and (icons, cenv)
    OsCons icons ;
    OsCompEnv cenv ;
{
    OsCons   cons   = (OsCons) OS_CONS_CDR (icons) ;
    OsVector vpatch = OsVectorMake (0) ;
    OsObject *cells, *ecells ;
    unsigned long oldmode ;
    if ( cons == (OsCons) OsA_Nil ) {
        os_emit_cst (cenv, OsA_True) ;
    } else {
        oldmode = OS_COMP_MODE (cenv, 0) ;
        OsMustBeCons (cons) ;
        while ( OsIsCons (OS_CONS_CDR (cons)) ) {
            OS_COMP_MODE (cenv, 0) ;
            os_compile_expr (OS_CONS_CAR (cons), cenv) ;
            OsVectorFastAppend (vpatch, (OsObject) os_emit_briffalse (cenv)) ;
            cons = (OsCons) OS_CONS_CDR (cons) ;
        }
        OS_COMP_MODE (cenv, oldmode) ;
        os_compile_expr (OS_CONS_CAR (cons), cenv) ;
        if ( OS_CONS_CDR (cons) != OsA_Nil ) {
            (void) OsErrorTrigger (1, ose_compiler_error
                                   , OsStringMake ("invalid 'and: dotted")) ;
        }
        cells  = vpatch->cells ;
        ecells = vpatch->cells + vpatch->csize ;
        if ( (! OS_TAILEDP (cenv)) && (oldmode & OS_MODE_PUSH) ) {
            unsigned int pgoto = os_emit_goto (cenv) ;
            while ( cells < ecells ) {
                os_patch_briffalse (cenv, (unsigned long) *cells) ;
                cells++ ;
            }
            os_emit_byte (cenv, OS_B_PUSH) ;
            os_patch_goto (cenv, pgoto) ;
        } else {
            unsigned int pgoto ;
            if ( OS_TAILEDP (cenv) ) { /* tailed, but might have a cont */
                pgoto = os_emit_goto (cenv) ;
                while ( cells < ecells ) {
                    os_patch_briffalse (cenv, (unsigned long) *cells) ;
                    cells++ ;
                }
                os_emit_tailunlink (cenv) ;
                os_patch_goto (cenv, pgoto) ;
            } else {
                while ( cells < ecells ) {
                    os_patch_briffalse (cenv, (unsigned long) *cells) ;
                    cells++ ;
                }
            }
            if ( oldmode & OS_MODE_PUSH )
                os_emit_byte (cenv, OS_B_PUSH) ;
        }
    }
    return ((OsObject) icons) ;
}

/*
 * Compile msg exprs: for [<obj> (<sel:> <arg>)* <sel>?], the compiler returns
 * a list: (sys:send-msg ...).
 */

static OsObject os_comp_msg (icons, cenv)
    OsObject  icons ;
    OsCompEnv cenv ;
{
    OsObject      cons = OS_CONS_CDR (icons) ;
    unsigned int  len  = OsConsLength ((OsCons) cons) ;
    OsObject      rcv ;
    OsAtom        *selv ;
    unsigned int  selc ;
    OsEObjSel     selector ;
    unsigned long oldmode ;
    unsigned long psave = (unsigned long) 0 ;
    
    if ( len < 2 ) {
        (void) OsErrorTrigger (1, ose_compiler_error
                               , OsStringMake ("invalid msg expr length")) ;
    }
    rcv  = OS_CONS_CAR (cons) ;
    cons = OS_CONS_CDR (cons) ;		/* len > 2, this one is a cons */
    psave = os_emit_save (cenv) ;
    os_emit_byte (cenv, OS_B_PUSH) ;	/* leave a stack slot for receiver */
    oldmode = OS_COMP_MODE (cenv, 0) ;
    if ( len == 2 ) {
        OsAtom sel = (OsAtom) OS_CONS_CAR (cons) ;
        OsShouldBeAtom (sel) ;
        selc    = 1 ;
        selv    = (OsAtom *) OsGCMalloc (sizeof (OsAtom)) ;
        selv[0] = sel ;
    } else {
        unsigned int  i ;
        OsAtom        *pselv ;
        
        selc = ((len-1) >> 1) + ((len-1) & 1) ;
        selv = (OsAtom *) OsGCMalloc (selc * sizeof (OsObject)) ;
        /* Compile parameters, in push mode */
        OS_COMP_MODE (cenv, OS_MODE_PUSH) ;
        for (i = 0, pselv = selv ; i < len - 1 ; i++ ) {
            if ( i & 1 ) {		/* arg part, compile & push */
                OS_COMP_MODE (cenv, OS_MODE_PUSH) ;
                os_compile_expr (OS_CONS_CAR (cons), cenv) ;
            } else {			/* sel part, save */
                OsAtom sel = (OsAtom) OS_CONS_CAR (cons) ;
                OsShouldBeAtom (sel) ;
                *pselv++ = sel ;
            }
            cons = OS_CONS_CDR (cons) ;
        }
    }
    /* Check arity */
    selector = OsEObjSelMake (selc, selv, ((len-1) >> 1)) ;
    if ( selector->arity != ((len-1) >> 1) ) {
        (void) OsErrorTrigger (1, ose_compiler_error
                               , OsStringMake ("invalid msg send arity")) ;
    }
    /*
     * (selc,selv) contains the selector, rcv is the receiver, args pushed
     * Check receiver against [rcv super] expr, in which case, we emit
     * a super-send bytecode instead of the normal send.
     */
    OS_COMP_MODE (cenv, 0) ;
    if ( OsIsCons (rcv)
        && (OsConsLength ((OsCons) rcv) == 3)
        && (OS_CONS_CAR (rcv) == osa_send_msg)
        && (OS_CONS_CAR (OS_CONS_CDR (OS_CONS_CDR (rcv))) == osa_super) ) {
        os_compile_expr (OS_CONS_CAR (OS_CONS_CDR (rcv)), cenv) ;
        os_emit_sendsupermsg (cenv, selector) ;
    } else {
        os_compile_expr (rcv, cenv) ;
        os_emit_sendmsg (cenv, selector) ;
    }
    os_patch_save (cenv, psave) ;
    OS_COMP_MODE (cenv, oldmode) ;
    if ( OS_PUSHP (cenv) )
        os_emit_byte (cenv, OS_B_PUSH) ;
    return ((OsObject) icons) ;
}

/*
 * Subrs:
 */

/* AUTODOC
 * (sys:compile EXPRESSION)
 * Compile the given EXPRESSION into a code block.
 * The EXPRESSION is byte compiled to produce a code-block. This function
 * should not be called directly (unless you know what you do). As the
 * EXPRESSION is not expanded before compilation, you should provide an
 * already expanded EXPRESSION.
 */

static OsObject OsSubrCompile (OsObject expr) {
    return ((OsObject) OsCompileExpr (expr)) ;
}

void OsCompInitialize (OsInterp i, int *pargc, char **argv) {
    OsDefineMsg (OsAtomClass, OS_SEL_COMPILE, os_comp_symbolref) ;
    OsDefineMsg (OsConsClass, OS_SEL_COMPILE, os_comp_cons) ;

    os_special_hashtable = OsHashtableMake (0, OsHashPointer, OsCmpPointer) ;
    os_special_form (OsAtomMake (i, "if", OsA_Unbound), os_comp_if) ;
    os_special_form (OsAtomMake (i, "lambda", OsA_Unbound), os_comp_lambda) ;
    os_special_form (OsAtomMake (i, "set!", OsA_Unbound), os_comp_setbang) ;
    os_special_form (OsAtomMake (i, "begin", OsA_Unbound), os_comp_begin) ;
    os_special_form (OsAtomMake (i, "define", OsA_Unbound), os_comp_define) ;
    os_special_form (OsAtomMake (i, "quote", OsA_Unbound), os_comp_quote) ;
    os_special_form (OsAtomMake (i, "or", OsA_Unbound), os_comp_or) ;
    os_special_form (OsAtomMake (i, "and", OsA_Unbound), os_comp_and) ;
    os_special_form (OsAtomMake (i, "sys:send-msg", OsA_Unbound), os_comp_msg);

/* AUTODOC
 * compiler:inline-subr
 * Should the compiler inline subrs ?
 * This boolean indicates wether primitive function can be inlined at compile
 * time. It defaults to #t (inlining). You should change this flags value only
 * when you really do redefine primitives, which is not recommended.
 */
    osa_opt_inline_subr = OsAtomMake (i, "compiler:inline-subr", OsA_True) ;
/* AUTODOC
 * compiler:debug
 * Should the compiler produce debuggable code ?
 * This flags indicates wether debugging infos should be emitted by the
 * compiler. It defaults to #f (no debug). Turn it on if you want to use the
 * debugger, and 'require "debugger".
 * WARNING: debugging is an alpha feature of an alpha release of oscheme.
 */
    osa_opt_debug       = OsAtomMake (i, "compiler:debug", OsA_False) ;
    
    osa_super    = (OsObject) OsAtomMake (i, "super", OsA_Unbound) ;
    osa_send_msg = (OsObject) OsAtomMake (i, "sys:send-msg", OsA_Unbound) ;
    osa_lambda   = (OsObject) OsAtomMake (i, "lambda", OsA_Unbound) ;
#ifdef OS_COMPILER_TRACEABLE
    osa_trace_compiler = OsAtomMake (i, "sys:trace-compiler", OsA_False) ;
#endif /* OS_COMPILER_TRACEABLE */

    return ;
}

void OsCompSubrInitialize (OsInterp interp) {
/* AUTODOC
 * error:compiler
 * Compiler error.
 * This is the compiler error context. When an error occurs at compilation,
 * the error is triggered. This error has one parameter, a human readable form
 * of the error diagnostic.
 */
    ose_compiler_error = OsErrorMake (interp
                                      , "error:compiler"
                                      , "compiler error %0"
                                      , 1, OSE_CATCHABLE) ;
    OsSecSubrMake (interp, "sys:compile", OsSubrCompile
                   , 1, OS_SECURITY_LEVEL_F) ;
    return ;
}



