/***************************************************************************
 * cblock.c
 * cblock.c,v 1.4 1995/03/17 14:19:38 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * The bytecode interpretor. See the file osbcode.h for infos
 * AUTODOC MODULE
 * Code block implementation. Code block are the internal representation of
 * closures and functions in oscheme.
 * Advanced users only !
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>
#include <string.h>
#include <os_config.h>

#ifdef OS_VM_TIMEABLE
#include <sys/types.h>
#include <sys/times.h>
#include <time.h>
#ifndef CLK_TCK
#    ifdef HZ
#        define CLK_TCK HZ
#    else
#        define CLK_TCK 60
#    endif
#endif

#define CLK_TO_MS(x)		((1000 * (x)) / CLK_TCK)
#define DELTA_STIME(s,e)	CLK_TO_MS((e)->tms_stime-(s)->tms_stime)
#define DELTA_UTIME(s,e)	CLK_TO_MS((e)->tms_utime-(s)->tms_utime)
#endif /* OS_VM_TIMEABLE */

#include <os_interp.h>
#include <os_obj.h>
#include <os_string.h>
#include <os_atom.h>
#include <os_vector.h>
#include <os_error.h>
#include <os_gc.h>
#include <os_cons.h>
#include <os_cblock.h>
#include <os_cont.h>
#include <os_eobj.h>
#include <os_port.h>
#include <os_subr.h>
#include <os_arit.h>
#include <os_sig.h>
#include <os_eobjP.h>

#define OS_BYTECODE_DESCR_HERE
#include "os_bcode.h"
#undef OS_BYTECODE_DESCR_HERE

#define OS_NULL_ENV	0xcafecafe
#define OS_DBG_FRAMESZ	3

OsClass OsCodeBlockClass   = (OsClass) 0 ;
OsClass OsClosedBlockClass = (OsClass) 0 ;
OsClass OsContBlockClass   = (OsClass) 0 ;
OsClass OsWindBlockClass   = (OsClass) 0 ;
OsClass OsBreakPointClass  = (OsClass) 0 ;

#ifdef OS_VM_TRACEABLE
static OsAtom osa_trace_machine = (OsAtom) 0 ;
#endif /* OS_VM_TRACEABLE */
#ifdef OS_VM_TIMEABLE
static OsAtom osa_times_machine = (OsAtom) 0 ;
#endif /* OS_VM_TIMEABLE */

static OsError  ose_break    = (OsError) 0 ;
static OsError  ose_bad_wind = (OsError) 0 ;

static OsCodeBlock os_exit_block = (OsCodeBlock) 0 ;

/*
 * Internal C API
 */

/*
 * Build a lisp environemnt given a stack based env pointer
 * The resulting environemnt is a list of frames, each frame being composed
 * with a var name followed by its value in the frame:
 * ((f0v0 value f0v1 value) ... (fnv0 ...))
 */

OsObject os_build_env (env)
    OsObject *env ;
{
    OsObject *p ;
    OsCons   lstart = (OsCons) OsA_Nil, *llist = &lstart ;
    OsCons   start, *list ;
    
    while ( env != (OsObject *) OS_NULL_ENV ) {
        OsVector vframe = (OsVector) *env ;
        start = (OsCons) OsA_Nil ;
        list  = &start ;
        for (p = env - vframe->csize ; p < env ; p++ ) {
            *list = OsConsMake (vframe->cells[vframe->csize-(env-p)], OsA_Nil);
            list  = (OsCons *) &(OS_CONS_CDR (*list)) ;
            *list = OsConsMake (*p, OsA_Nil) ;
            list  = (OsCons *) (&OS_CONS_CDR (*list)) ;
        }
        *llist = OsConsMake ((OsObject) start, OsA_Nil) ;
        llist  = (OsCons *) &(OS_CONS_CDR (*llist)) ;
        env = (OsObject *) env[1] ;
    }
    return ((OsObject) lstart) ;
}

/*
 * Build the full VM state at this time.
 * This is a vector of stacked frames. In stacked frame is a vector of two
 * entries:
 * - the current expr being evaluated
 * - its env (or nil if env has been dropped, due to tail-call)
 */

OsObject os_build_current_vmstate (OsInterp interp) {
    OsObject *sp = interp->dbg_sp;
    OsVector vstack = OsVectorMake (0) ;
    OsVector vframe ;
    while ( sp > interp->dbg_stack + OS_DBG_FRAMESZ ) {
        OsObject    *env  = (OsObject *) sp[-1] ;
        OsObject    expr  = sp[-2] ;

        if ( env )
            vframe = OsVector2Make (expr, os_build_env ((OsObject *) env)) ;
        else
            vframe = OsVector2Make (expr, OsA_Nil) ;
        (void) OsVectorFastAppend (vstack, (OsObject) vframe) ;
        sp -= OS_DBG_FRAMESZ ;
    }
    OsOutputPortFlush (OsCurrentOutputPort) ;
    return (vstack->csize ? ((OsObject) vstack) : ((OsObject) 0)) ;
}

/* 
 * Map the given block for potential places to break.
 * When such a code location is found, apply func:
 * OsObject (*func) (unsigned char *iloc, OsObject expr, OsObject closure)
 * - iloc is the instruction location
 * - expr is the debug info at this place
 * - closure is theprovided closure.
 * If function returns OsA_True, than return, else continue. os_break_map
 * returns OsA_True if interrupted, OsA_False else.
 */

OsObject os_break_map (block, func, closure)
    OsCodeBlock   block ;
    OsObject      (*func)() ;
    OsObject      closure ;
{
    unsigned char *pc, *epc ;
    bcode_descr   bd ;

    pc  = block->code->string ;
    epc = pc + block->code->len ;
    while ( pc < epc ) {
        bd = byte_codes + ((unsigned long) *pc) ;
        switch (*pc++) {
          case OS_DBG_PUSH:
          case OS_DBG_PUSHBRK:
          case OS_DBG_SAFEPUSH:
          case OS_DBG_SAFEPUSHBRK:
            {
                OsObject  expr    = block->data->cells[OS_GET_SHORT(pc)] ;
                
                if ( (*func)(pc, expr, closure) == OsA_True )
                    return (OsA_True) ;
            }
            continue ;
          default:
            pc += bd->skip ;
            continue ;
        }
    }
    return (OsA_False) ;
}

void os_decode_instr (port, io_pc, data)
    OsOutputPort  port ;
    unsigned char **io_pc ;
    OsObject      *data ;
{
    char tmp[OS_DISPLAY_BUFSIZ] ;
    unsigned char *pc = *io_pc ;
    unsigned int instr = ((unsigned int) *pc++) ;

    if ( instr >= OS_B_OPLAST ) {
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("os_decode_instr")
                               , OsStringMake ("unknown op code")) ;
    } else {
        bcode_descr d = byte_codes + instr ;
        char *fmt = d->fmt ;

        sprintf (tmp, "%#X %-15.15s"
                 , ((unsigned int) (pc-1)) & 0xFFFF , d->name) ;
        OsOutputPortWriteCString (port, tmp) ;
        while ( *fmt ) {
            switch (*fmt++) {
              case 'x':
                {
                    unsigned short didx = OS_GET_SHORT (pc) ;
                    if ( data )
                        sprintf (tmp, "\t%d\t;; (cfunc %p)"
                                 , didx, data[didx]) ;
                    else
                        sprintf (tmp, "\t%d\t;; (cfunc %d)"
                                 , didx, didx) ;
                }
                OsOutputPortWriteCString (port, tmp) ;
                break ;
              case 'd':
                {
                    unsigned short didx = OS_GET_SHORT (pc) ;
                    sprintf (tmp, "\t%d\t;; ", didx) ;
                    OsOutputPortWriteCString (port, tmp) ;
                    if ( data )
                        OsSendDisplay (data[didx], OsCurrentOutputPort) ;
                }
                break ;
              case 'n':
                sprintf (tmp, "\t%d", ((unsigned int) *pc++)) ;
                OsOutputPortWriteCString (port, tmp) ;
                break ;
              case 'g':
                {
                    unsigned long loc = (((unsigned long) pc) & 0xFFFF) ;
                    sprintf (tmp, "\t%#04lx", loc+OS_GET_SHORT(pc)) ;
                    OsOutputPortWriteCString (port, tmp) ;
                }
                break ;
              default:
                (void) OsErrorTrigger (1, OsE_Implementation
                                       , OsStringMake ("os_decode_instr")
                                       , OsStringMake ("invalid format")) ;
                /* NOT REACHED */
            }
        }
    }
    OsOutputPortWriteChar (port, '\n') ;
    *io_pc = pc ;
    return ;
}

static void  os_codeblock_decode (OsCodeBlock block, OsOutputPort port) {
    unsigned char *pc = block->code->string ;
    unsigned char *pe = block->code->string+block->code->len ;
    while ( pc <= pe ) {
        os_decode_instr (port, &pc, block->data->cells) ;
    }
    return ;
}

/*
 * C API:
 */

OsObject OsCodeBlockApplyWithArray (OsInterp interp
                                    , OsCodeBlock block
                                    , unsigned int argc, OsObject *argv
                                    , OsObject *perror) {
    REGISTER OsObject *sp = interp->sp ;
    interp->ridx = argc ;
    *++sp = (OsObject) 0 ;				/* fake data */
    *++sp = (OsObject) os_exit_block->code->string ;	/* exit from vm */
    while ( argc-- ) {
        *++sp = *argv++ ;
    }
    interp->sp = sp ;
    if ( perror ) 
        return (OsCodeBlockProtectExecute (interp, block, perror)) ;
    else
        return (OsCodeBlockExecute (interp, block)) ;
}

OsObject OsCodeBlockApplyWithList (OsInterp interp
                                   , OsCodeBlock block
                                   , OsCons list
                                   , OsObject *perror) {
    REGISTER OsObject *sp = interp->sp ;
    interp->ridx = 0 ;
    *++sp = (OsObject) 0 ;				/* fake data */
    *++sp = (OsObject) os_exit_block->code->string ;	/* exit from vm */
    while ( OsIsCons (list) ) {
        interp->ridx++ ;
        *++sp = OS_CONS_CAR (list) ;
        list  = (OsCons) OS_CONS_CDR (list) ;
    }
    interp->sp = sp ;
    if ( perror )
        return (OsCodeBlockProtectExecute (interp, block, perror)) ;
    else
        return (OsCodeBlockExecute (interp, block)) ;
}

typedef struct _os_protect {
    struct _os_protect *prev ;
    OsInterp           interp ;
    OsObject           error ;
    OsObject           *stackp, *dbg_stackp ;
    unsigned long      stamp ;
    OsCodeBlock        enter ;
    jmp_buf            protect ;
} os_protect_rec, *os_protect ;

static os_protect    protection_head  = (os_protect) 0 ;
static unsigned long protection_stamp = (unsigned long) 0 ;

#define NEW_PROTECTION(_i,_p) \
    (_p)->prev       = protection_head ; \
    (_p)->interp     = _i ; \
    (_p)->error      = (OsObject) 0 ; \
    (_p)->stackp     = (_i)->sp ; \
    (_p)->dbg_stackp = (_i)->dbg_sp ; \
    (_p)->stamp      = ++protection_stamp ; \
    (_p)->enter      = (OsCodeBlock) 0 ; \
    protection_head = (_p) ;

/*
 * os_restore_protection:
 * This function is called just *after* the context switch of call/cc. It has
 * to invoke all the dynamic-wind's before procedure, before entering the
 * continuation. This is done recursevly, to trigger thunk in the correct
 * order.
 * We use the top of the stack (the one after the siwtch) to evaluate those
 * thunk.
 */

static void os_invoke_entries (os_protect prot
                               , OsInterp interp
                               , OsObject *at, OsObject *dbg_at)
{
    if ( prot ) {
        os_invoke_entries (prot->prev, interp, at, dbg_at) ;
        if ( prot->enter ) {
            /* I think this is safe:
             * Before switch-interp, we grab the continuation, so executing
             * protection 'before' can safely take place at current top of
             * interpreter's stack, provided it is not the current one.
             */
            if ( prot->interp == interp ) {
                prot->interp->sp     = at ;
                prot->interp->dbg_sp = dbg_at ;
            }
            OsCodeBlockApplyWithArray (prot->interp
                                       , prot->enter
                                       , 0, (OsObject *) 0
                                       , (OsObject *) 0) ;
        }
    }
    return ;
}

static void os_restore_protection (os_protect prot) {
    os_protect pp = protection_head = prot ;
    OsObject   *expt_stackp     = prot->interp->sp + 1 ;
    OsObject   *expt_dbg_stackp = prot->interp->dbg_sp ;
    
    os_invoke_entries (pp, prot->interp, expt_stackp, expt_dbg_stackp) ;
    return ;
}

/*
 * os_uncatched_error:
 * When under a ProtectExecute (CProtectExecute, CodeBlockProtectExecute, ...)
 * one can call os_uncatched_error to make the corresponding exception goes
 * up the stack.
 */

void os_uncatched_error (error) 
    OsObject error ;
{
    if ( ! protection_head ) {
        OsString errmsg = OsErrorFormat ((OsVector) error) ;
        OsFatalError (errmsg->string) ;
    }
    protection_head->error = error ;
    longjmp (protection_head->protect, 2) ;
    /* NOT REACHED */
}

OsObject OsCodeBlockProtectExecute (OsInterp interp
                                    , OsCodeBlock block
                                    , OsObject *perred) {
    os_protect_rec protection ;
    OsObject       value ;
    
    *perred = (OsObject) 0 ;
    NEW_PROTECTION (interp, &protection) ;
    bzero (interp->sp + 1
           , sizeof(OsObject) * (interp->stacke - interp->sp - 1)) ;
    if ( ! setjmp (protection.protect) ) {
        value = OsCodeBlockExecute (interp, block) ;
        interp->sp      = protection.stackp ;
        interp->dbg_sp  = protection.dbg_stackp ;
        protection_head = protection.prev ;
        return (value) ;
    } else {
        *perred         = protection_head->error ;
        interp->sp      = protection.stackp ;
        interp->dbg_sp  = protection.dbg_stackp ;
        if ( protection.interp != OsCurrentInterp() )
            (void) OsInterpSwitch (protection.interp) ;
        protection_head = protection.prev ;
    }
    return ((OsObject) 0) ;
}

OsObject OsCProtectExecute (OsInterp interp
                            , OsObject (*fun)(), OsObject arg
                            , OsObject *perred) {
    os_protect_rec protection ;
    OsObject       value ;
    
    *perred = (OsObject) 0 ;
    NEW_PROTECTION (interp, &protection) ;
    if ( ! setjmp (protection.protect) ) {
        value           = (*fun)(arg) ;
        interp->sp      = protection.stackp ;
        interp->dbg_sp  = protection.dbg_stackp ;
        protection_head = protection.prev ;
        return (value) ;
    } else {
        *perred         = protection_head->error ;
        interp->sp      = protection.stackp ;
        interp->dbg_sp  = protection.dbg_stackp ;
        if ( protection.interp != OsCurrentInterp() )
            (void) OsInterpSwitch (protection.interp) ;
        protection_head = protection.prev ;
    }
    return ((OsObject) 0) ;
}

static void os_throw_wind (wind, value)
    OsWindBlock wind ;
    OsObject    value ;
{
    os_protect pp = protection_head ;
    while ( pp ) {
        if ( pp->stamp == wind->stamp ) {
            pp->error = value ;
            longjmp (pp->protect, 3) ;
        }
        pp = pp->prev ;
    }
    return ;
}

OsObject OsWindProtectExecute (OsInterp interp
                               , OsCodeBlock enter
                               , OsCodeBlock block
                               , OsObject *perror) {
    os_protect_rec protection ;
    OsObject       value, error ;
    OsWindBlock    wind = (OsWindBlock) OsObjectMake (OsWindBlockClass, 0) ;
    unsigned char  code[4], *p = code ;
    int            ret ;
    
    NEW_PROTECTION (interp, &protection) ;
    protection.enter = enter ;
    *p++ = OS_B_RETWIND ;
    OS_PUT_SHORT (p, (unsigned short) 0) ;
    wind->code       = OsStringMakeFromBytes (p - code, code) ;
    wind->data       = OsVector1Make ((OsObject) wind) ;
    wind->stamp      = protection_stamp ;
    if ( ! (ret = setjmp (protection.protect)) ) {
        value           = OsCodeBlockApplyWithArray (interp
                                                     , block
                                                     , 1, (OsObject *) &wind
                                                     , (OsObject *) 0) ;
        interp->sp      = protection.stackp ;
        interp->dbg_sp  = protection.dbg_stackp ;
        protection_head = protection.prev ;
        return (value) ;
    } else if ( ret == 3 ) {
        error           = protection_head->error ;
        interp->sp      = protection.stackp ;
        interp->dbg_sp  = protection.dbg_stackp ;
        if ( protection.interp != OsCurrentInterp() )
            (void) OsInterpSwitch (protection.interp) ;
        protection_head = protection.prev ;
        return (error) ;
    } else {
        /* skip our own protection, and throw error (continued) */
        *perror         = protection_head->error ;
        interp->sp      = protection.stackp ;
        interp->dbg_sp  = protection.dbg_stackp ;
        if ( protection.interp != OsCurrentInterp() )
            (void) OsInterpSwitch (protection.interp) ;
        protection_head = protection.prev ;
        return ((OsObject) 0) ;
    }
    /* NOT REACHED */
}
    
OsVector OsCurrentVMStateToError (OsInterp interp) {
    return ((OsVector) os_build_current_vmstate (interp)) ;
}

/*
 * Constructors:
 */

OsCodeBlock OsCodeBlockMake (unsigned char *code
                             , unsigned long codelen
                             , OsVector data) {
    OsCodeBlock code_block = (OsCodeBlock) OsObjectMake (OsCodeBlockClass, 0) ;
    code_block->code = OsStringMakeFromBytes (codelen, code) ;
    code_block->data = OsVectorFix (data) ;
    return (code_block) ;
}

OsClosedBlock OsCodeBlockClose (OsCodeBlock block) {
    OsClosedBlock cblock ;
    
    cblock = (OsClosedBlock) OsObjectMake (OsClosedBlockClass, 0) ;
    cblock->code = block->code ;
    cblock->data = block->data ;
    cblock->env  = (OsObject *) OS_NULL_ENV ;
    return (cblock) ;
}

static int OsContBlockMake (OsInterp interp
                            , OsObject acc, OsObject *env, int ridx
                            , OsObject *stack, OsObject *stackp
                            , OsObject *data, unsigned char *pc
                            , OsContBlock *pblock) {
    OsContBlock cblock = (OsContBlock) OsObjectMake (OsContBlockClass, 0) ;
    unsigned char code[4], *p = code ;
    unsigned int sz = sizeof (OsObject) * (stackp - stack + 1) ;
    
    *p++ = OS_B_RETCONT ;
    OS_PUT_SHORT (p, (short) 0) ;

    cblock->interp     = interp ;
    cblock->code       = OsStringMakeFromBytes (p - code, code) ;
    cblock->data       = OsVector1Make ((OsObject) cblock) ;
    cblock->acc        = acc ;
    cblock->env        = env ;
    cblock->ridx       = ridx ;
    cblock->stacklen   = stackp - stack ;
    cblock->stack      = (OsObject *) OsGCMalloc (sz) ;
    bcopy (stack, cblock->stack, sz) ;
    cblock->cells      = data ;
    cblock->pc         = pc ;
    cblock->protection = protection_head ;
    *pblock = cblock ;
    return (os_cstack_freeze (cblock)) ;
}


/*
 * Subrs:
 */

/* AUTODOC
 * (procedure? OBJECT)
 * Is OBJECT a procedure ?
 * Returns #t if the given OBJECT is a procedure, #f otherwise.
 */

static OsObject OsSubrProcedureP (OsObject proc) {
    return ((OsIsBlock(proc)) ? OsA_True : OsA_False) ;
}

/* AUTODOC
 * (cont->env CONTINUATION)
 * Reify the CONTINUATION environment, returns a list of bindings.
 * This function reify the CONTINUATION environment, and returns a list of
 * of frames. Each frames is a list whose length is even, containing a symbol
 * and its associated value in the frame
 */

static OsObject OsSubrContToEnv (OsContBlock cblock) {
    OsShouldBeContBlock (cblock) ;
    return ((OsObject) os_build_env (cblock->env)) ;
}

/* AUTODOC
 * (closure->env CLOSURE)
 * Reify the CLOSURE environment, returns a list of bindings.
 * This function reify the CLOSURE environment, and returns a list of
 * frames. Each frames is a list whose length is even, containing a symbol
 * and its associated value in the frame
 */

static OsObject OsSubrClosureToEnv (closure)
    OsClosedBlock closure ;
{
    OsShouldBeClosedBlock (closure) ;
    return ((OsObject) os_build_env (closure->env)) ;
}

/* AUTODOC
 * (codeblock:dump BLOCK)
 * Disassemble the given BLOCK of bytecode to (current-output-port)
 * This function disassemble a BLOCK of byte code, and dump it into human
 * readable form to the current output port.
 */

static OsObject OsSubrCodeBlockDump (codeblock)
    OsCodeBlock codeblock ;
{
    OsMustBeBlock (codeblock) ;
    os_codeblock_decode (codeblock, OsCurrentOutputPort) ;
    return (OsA_True) ;
}

/* AUTODOC
 * (codeblock:execute BLOCK)
 * Execute the given BLOCK of byte code.
 * This function sets the virtual machine to execute the given BLOCK of byte
 * code, and runs it.
 * Hackers only.
 */

static OsObject OsSubrCodeBlockExecute (argc, argv)
    unsigned int argc ;
    OsObject     *argv ;
{
    OsCodeBlock block ;
    OsObject    value ;
    
    if ( argc < 1 ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (1)
                               , OsIntegerMake (0)) ;
    }
    block = (OsCodeBlock) argv[0] ; argc-- ; argv++ ;
    OsMustBeBlock (block) ;
    value = OsCodeBlockApplyWithArray (OsCurrentInterp()
                                       , block
                                       , argc, argv
                                       , (OsObject *) 0) ;
    return (value) ;
}

/*
 * needs some global to list beak points:
 */

static OsCodeBlock bp_curblock = (OsCodeBlock) 0 ;

static void break_point_accumulate (iloc, expr, vector)
    unsigned char *iloc ;
    OsObject      expr ;
    OsVector      vector ;
{
    OsBreakPoint bp = (OsBreakPoint) OsObjectMake (OsBreakPointClass, 0) ;

    bp->block     = bp_curblock ;
    bp->expr      = expr ;
    bp->iloc      = iloc ;
    bp->condition = OsA_True ;
    (void) OsVectorFastAppend ((OsVector) vector, (OsObject) bp) ;
    return ;
}

/* AUTODOC
 * (codeblock:break-points BLOCK)
 * Returns a vector containing all possible break points in BLOCK
 * This function retreives all possible break points in the given BLOCK of
 * byte code. It returns a vector containing one break point object for each
 * possible break point.
 */
 
static OsObject OsSubrCodeBlockBreakPoints (block)
    OsCodeBlock block ;
{
    OsVector bpv = OsVectorMake (0) ;

    OsMustBeBlock (block) ;
    bp_curblock = block ;
    os_break_map (block, break_point_accumulate, (OsObject) bpv) ;
    bp_curblock = (OsCodeBlock) 0 ;
    return ((OsObject) bpv) ;
}

/* AUTODOC
 * (codeblock:break BREAKPOINT ON-OFF)
 * Activate/Deactivate the given BREAKPOINT.
 * This function activates the given BREAKPOINT. Activating a break-point will
 * make the virtual machine triggers the 'error:break-point error with the full
 * virtual machine state as context.
 */

static OsObject OsSubrCodeBlockBreak (bp, onoff)
    OsBreakPoint bp ;
    OsObject     onoff ;
{
    OsShouldBeBreakPoint (bp) ;
    switch (*(bp->iloc)) {
      case OS_DBG_PUSHBRK:
        if ( onoff == OsA_False )
            *(bp->iloc) = OS_DBG_PUSH ;
        return (OsA_False) ;
      case OS_DBG_SAFEPUSHBRK:
        if ( onoff == OsA_False )
            *(bp->iloc) = OS_DBG_SAFEPUSH ;
        return (OsA_False) ;
      case OS_DBG_PUSH:
        if ( onoff == OsA_True )
            *(bp->iloc) = OS_DBG_PUSHBRK ;
        return (OsA_True) ;
      case OS_DBG_SAFEPUSH:
        if ( onoff == OsA_True )
            *(bp->iloc) = OS_DBG_SAFEPUSHBRK ;
        return (OsA_True) ;
      default:
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("OsSubrCodeBlockBreak")
                               , OsStringMake ("invalid instr at location"));
    }
    /* NOT REACHED */
}

/*
 * Initialize:
 */

void OsCodeBlockInitialize (OsInterp interp, int *pargc, char **argv) {
    OsCodeBlockClass   =OsClassDeclare("CodeBlock",sizeof (OsCodeBlockRec)) ;
    OsClosedBlockClass =OsClassDeclare("ClosedBlock",sizeof(OsClosedBlockRec));
    OsContBlockClass   =OsClassDeclare("ContBlock",sizeof(OsContBlockRec)) ;
    OsWindBlockClass   =OsClassDeclare("WindBlock", sizeof (OsWindBlockRec)) ;
    OsBreakPointClass  =OsClassDeclare("BreakPoint",sizeof(OsBreakPointRec)) ;

#ifdef OS_VM_TRACEABLE
    osa_trace_machine = OsAtomMake (interp, "sys:trace-machine", OsA_False) ;
#endif /* OS_VM_TRACEABLE */
#ifdef OS_VM_TIMEABLE
    osa_times_machine = OsAtomMake (interp, "sys:times-machine", OsA_False) ;
#endif /* OS_VM_TIMEABLE */
    return ;
}

void OsCodeBlockSubrInitialize (OsInterp interp) {
    unsigned char exit_code[2] ;
    exit_code[0] = OS_B_EXIT ;
    exit_code[1] = OS_B_NOOP ;
    os_exit_block = OsCodeBlockMake (exit_code, 2, OsVectorMake(0)) ;
    ose_break    = OsErrorMake (interp
                                , "error:break-point"
                                , "User break point"
                                , 0, OSE_NORMAL) ;
    ose_bad_wind = OsErrorMake (interp
                                , "error:bad-wind"
                                , "Invalid wind for %0"
                                , 1, OSE_CATCHABLE) ;
    OsSubrMake (interp, "procedure?", OsSubrProcedureP, 1) ;
    OsSubrMake (interp, "codeblock:break-points",OsSubrCodeBlockBreakPoints,1);
    OsSubrMake (interp, "codeblock:break", OsSubrCodeBlockBreak, 2) ;
    OsSubrMake (interp, "codeblock:dump", OsSubrCodeBlockDump, 1) ;
    OsSecSubrMake (interp, "cont->env", OsSubrContToEnv
                   , 1, OS_SECURITY_LEVEL_E) ;
    OsSecSubrMake (interp, "closure->env", OsSubrClosureToEnv
                   , 1, OS_SECURITY_LEVEL_E) ;
    OsSecSubrMake (interp
                   , "codeblock:execute"
                   , OsSubrCodeBlockExecute
                   , OS_NARY, OS_SECURITY_LEVEL_F);
    return ;
}


/***************************************************************************
 * Virtual machine code
 ***************************************************************************/
 
OsObject OsCodeBlockExecute (OsInterp interp, OsCodeBlock code_block) {
    REGISTER unsigned char *pc ;
    REGISTER OsObject *stackp ;
    REGISTER OsObject acc ;
    REGISTER OsObject *env ;
    REGISTER OsObject *data ;
    REGISTER OsObject robj ;
    REGISTER int ridx, rjdx ;
    OsObject     *stacke ;
    OsCodeBlock  cont ;
    OsInterp     old_interp = (OsInterp) 0 ;
    
#ifdef OS_VM_TIMEABLE
    struct tms tstart, tend ;
#endif /* OS_VM_TIMEABLE */
    
/*
 * Initialize the stack, the stack pointer, etc...
 */
    pc     = code_block->code->string ;
    data   = code_block->data->cells ;
    env    = (OsIsClosedBlock (code_block)
              ? ((OsClosedBlock) code_block)->env
              : (OsObject *) OS_NULL_ENV) ;
    acc    = (OsObject) code_block ;
    ridx   = interp->ridx ;
    
    stackp = interp->sp ;
    stacke = interp->stacke ;

    if ( interp != OsCurrentInterp () )
        old_interp = OsInterpSwitch (interp) ;
    
/*
 * This macro checks that the given location is in the VM stack
 */
    
#define OS_STACKP(_l) \
   ((((OsObject*) _l) >= interp->stack) && (((OsObject*) _l) < stacke))

/*
 * This macro is called before leaving the function, to save local vars
 * in globals, and to check for signals: leaving the VM costs a function call
 * so adding the cost of a test should not be dreadfull.
 */

#ifdef OS_HAS_SIGNALS
#    define OS_VM_OUTGOING() \
         { interp->sp = stackp ; if (os_signals_pending) OsDoSignals() ;}
#else 
#    define OS_VM_OUTGOING() interp->sp = stackp
#endif /* OS_HAS_SIGNALS */

/*
 * Security checks:
 * Are performed on all subr call, with two checks:
 * - We are really using a subr object (prevent hacking the data part of byte
 *   code blocks),
 * - The security of interp is correct
 */

#ifdef OS_SECURE
#   define OS_VM_SUBR_CHECK(s) \
    (OsMustBeSubr(s), OsInterpCheckSecurity (interp, ((OsSubr) (s))->security))
#else /* OS_SECURE */
#   define OS_VM_SUBR_CHECK(s) (s)
#endif /* OS_SECURE */
        
/*
 * Outgoing C function call:
 */

#define OS_VM_CAPPLY0(f) \
    OS_VM_OUTGOING() ; \
    acc = ((OsObject (*)()) f)() ;
#define OS_VM_CAPPLY1(f) \
    OS_VM_OUTGOING() ; \
    stackp-- ; \
    acc = ((OsObject (*)()) f)(stackp[1]) ;
#define OS_VM_CAPPLY2(f) \
    OS_VM_OUTGOING() ; \
    stackp -= 2 ; \
    acc = ((OsObject (*)()) f)(stackp[1], stackp[2]) ;
#define OS_VM_CAPPLY3(f) \
    OS_VM_OUTGOING() ; \
    stackp -= 3 ; \
    acc = ((OsObject (*)()) f)(stackp[1], stackp[2], stackp[3]) ;
#define OS_VM_CAPPLY4(f) \
    OS_VM_OUTGOING() ; \
    stackp -= 4 ; \
    acc = ((OsObject (*)()) f)(stackp[1], stackp[2], stackp[3]  \
                               , stackp[4]) ; 
#define OS_VM_CAPPLY5(f) \
    OS_VM_OUTGOING() ; \
    stackp -= 5 ; \
    acc = ((OsObject (*)()) f)(stackp[1], stackp[2], stackp[3] \
                               , stackp[4], stackp[5]) ; 
#define OS_VM_CAPPLY6(f) \
    OS_VM_OUTGOING() ; \
    stackp -= 6 ; \
    acc = ((OsObject (*)()) f)(stackp[1], stackp[2], stackp[3] \
                               , stackp[4], stackp[5], stackp[6]) ;
#define OS_VM_CAPPLY7(f) \
    OS_VM_OUTGOING() ; \
    stackp -= 7 ; \
    acc = ((OsObject (*)()) f)(stackp[1], stackp[2], stackp[3] \
                               , stackp[4], stackp[5], stackp[6], stackp[7]) ;
#define OS_VM_CAPPLYN(f) \
    OS_VM_OUTGOING () ; \
    stackp -= ridx ; \
    acc = ((OsObject (*)()) f)(ridx, stackp+1) ;
    
/*
 * Outgoing C message implementations
 */
           
#define OS_VM_CSEND0(imp,sel) \
    OS_VM_OUTGOING() ; \
    acc = ((OsObject (*)()) SELIMP_CFUN(imp))(SELIMP_CLOSURE(imp), acc, sel) ;
#define OS_VM_CSEND1(imp,sel) \
    stackp-- ; \
    OS_VM_OUTGOING() ; \
    acc = ((OsObject (*)()) SELIMP_CFUN(imp))(SELIMP_CLOSURE(imp), acc, sel \
                                              , stackp[1]) ;
#define OS_VM_CSEND2(imp,sel) \
    stackp -= 2 ; \
    OS_VM_OUTGOING() ; \
    acc = ((OsObject (*)()) SELIMP_CFUN(imp))(SELIMP_CLOSURE(imp), acc, sel \
                                              , stackp[1], stackp[2]) ;
#define OS_VM_CSEND3(imp,sel) \
    stackp -= 3 ; \
    OS_VM_OUTGOING() ; \
    acc = ((OsObject (*)()) SELIMP_CFUN(imp))(SELIMP_CLOSURE(imp), acc, sel \
                                              , stackp[1], stackp[2] \
                                              , stackp[3]) ;
#define OS_VM_CSEND4(imp,sel) \
    stackp -= 4 ; \
    OS_VM_OUTGOING() ; \
    acc = ((OsObject (*)()) SELIMP_CFUN(imp))(SELIMP_CLOSURE(imp), acc, sel \
                                              , stackp[1], stackp[2] \
                                              , stackp[3], stackp[4]) ; 
#define OS_VM_CSEND5(imp,sel) \
    stackp -= 5 ; \
    OS_VM_OUTGOING() ; \
    acc = ((OsObject (*)()) SELIMP_CFUN(imp))(SELIMP_CLOSURE(imp), acc, sel \
                                              , stackp[1], stackp[2] \
                                              , stackp[3], stackp[4] \
                                              , stackp[5]) ; 
#define OS_VM_CSEND6(imp,sel) \
    stackp -= 6 ; \
    OS_VM_OUTGOING() ; \
    acc = ((OsObject (*)()) SELIMP_CFUN(imp))(SELIMP_CLOSURE(imp), acc, sel \
                                              , stackp[1], stackp[2] \
                                              , stackp[3], stackp[4] \
                                              , stackp[5], stackp[6]) ;
#define OS_VM_CSEND7(imp,sel) \
    stackp -= 7 ; \
    OS_VM_OUTGOING() ; \
    acc = ((OsObject (*)()) SELIMP_CFUN(imp))(SELIMP_CLOSURE(imp), acc, sel \
                                              , stackp[1], stackp[2] \
                                              , stackp[3], stackp[5] \
                                              , stackp[5], stackp[6] \
                                              , stackp[7]) ;
#define OS_VM_CSENDN(imp,sel) \
    stackp -= ridx ; \
    OS_VM_OUTGOING () ; \
    acc = ((OsObject (*)()) SELIMP_CFUN(imp))(SELIMP_CLOSURE(imp), acc, sel \
                                              , ridx, stackp) ;
    


     
#ifdef OS_VM_TIMEABLE
        if ( osa_times_machine->vcell != OsA_False ) {
            times (&tstart) ;
        }
#endif /* OS_VM_TIMEABLE */
        
    while ( 1 ) {

#ifdef OS_VM_TRACEABLE
        {
            unsigned char *cpc = pc ;
            if ( osa_trace_machine->vcell != OsA_False ) {
                printf ("[s=x%08x,e=x%08x,d=x%08x] ", stackp, env, data) ;
                os_decode_instr (OsCurrentOutputPort, &cpc, data) ;
            }
        }
#endif /* OS_VM_TRACEABLE */

#ifdef OS_VM_CHECKSTACK
        if ( ! OS_STACKP (stackp) ) {
            os_db_break ("*** Stack pointer out of stack.") ;
        }
#endif /* OS_VM_CHECKSTACK */
#ifdef OS_SECURE
        if ( stackp >= stacke ) {
            OsInterpStackSizeExceeded (interp) ;
        }
        OsInterpCheckClicks (interp, 1) ;
#endif /* OS_SECURE */
        switch ( *pc++ ) {
          case OS_B_NOOP:
            continue ;
            
          case OS_B_LREF0:
            acc = env[-((int) *pc++)] ;
            continue ;
          case OS_B_LREF1:
            acc = ((OsObject *) env[1])[-((int) *pc++)] ;
            continue ;
          case OS_B_LREF2:
            acc = ((OsObject *)
                   ((OsObject *)
                    ((OsObject *) env[1]))[1])[-((int) *pc++)] ;
            continue ;
          case OS_B_LREF3:
            acc = ((OsObject *)
                   ((OsObject *)
                    ((OsObject *)
                     ((OsObject *) env[1])[1])[1]))[-((int) *pc++)] ;
            continue ;
          case OS_B_LREFN:
            for (ridx = (int) *pc++, robj = (OsObject) env ; ridx-- ; )
                robj = ((OsObject *) robj)[1] ;
            acc = ((OsObject *) robj)[-((int) *pc++)] ;
            continue ;

          case OS_B_LREF0PUSH:
            *++stackp = acc = env[-((int) *pc++)] ;
            continue ;
          case OS_B_LREF1PUSH:
            *++stackp = acc = ((OsObject *) env[1])[-((int) *pc++)] ;
            continue ;
          case OS_B_LREF2PUSH:
            acc = ((OsObject *)
                   ((OsObject *)
                    ((OsObject *) env[1]))[1])[-((int) *pc++)] ;
            *++stackp = acc ;
            continue ;
          case OS_B_LREF3PUSH:
            acc = ((OsObject *)
                   ((OsObject *)
                    ((OsObject *)
                     ((OsObject *) env[1])[1])[1]))[-((int) *pc++)] ;
            *++stackp = acc ;
            continue ;
          case OS_B_LREFNPUSH:
            for (ridx = (int) *pc++, robj = (OsObject) env ; ridx-- ; )
                robj = ((OsObject *) robj)[1] ;
            acc = ((OsObject *) robj)[-((int) *pc++)] ;
            *++stackp = acc ;
            continue ;

          case OS_B_LSET0:
            env[-((int) *pc++)] = acc ;
            continue ;
          case OS_B_LSET1:
            ((OsObject *) env[1])[-((int) *pc++)] = acc ;
            continue ;
          case OS_B_LSET2:
            ((OsObject *)
             ((OsObject *)
              ((OsObject *) env[1]))[1])[-((int) *pc++)] = acc ;
            continue ;
          case OS_B_LSET3:
            ((OsObject *)
             ((OsObject *)
              ((OsObject *)
               ((OsObject *) env[1])[1])[1]))[-((int) *pc++)] = acc ;
            continue ;
          case OS_B_LSETN:
            for (ridx = (int) *pc++, robj = (OsObject) env ; ridx-- ; )
                robj = ((OsObject *) robj)[1] ;
            ((OsObject *) robj)[-((int) *pc++)] = acc ;
            continue ;

          case OS_B_GREF:
            acc = data[OS_GET_SHORT(pc)] ;
            if (((OsAtom) acc)->vcell == OsA_Unbound) {
                OS_VM_OUTGOING() ;
                acc = OsErrorTrigger (1, OsE_Unbound, acc) ;
            }
            acc = ((OsAtom) acc)->vcell ;
            continue ;
          case OS_B_GREFPUSH:
            acc = data[OS_GET_SHORT(pc)] ;
            if (((OsAtom) acc)->vcell == OsA_Unbound) {
                OS_VM_OUTGOING() ;
                acc = OsErrorTrigger (1, OsE_Unbound, acc) ;
            }
            acc = ((OsAtom) acc)->vcell ;
            *++stackp = acc ;
            continue ;
          case OS_B_GSET:
            /* robj used here for debug purposes */
            robj = data[OS_GET_SHORT(pc)] ;
            if ((((OsAtom) robj)->vcell) == OsA_Unbound) {
                OS_VM_OUTGOING() ;
                acc = OsErrorTrigger (1, OsE_Unbound, robj) ;
            }
            ((OsAtom) robj)->vcell = acc ;
            continue ;
          case OS_B_GDEF:
            ((OsAtom) data[OS_GET_SHORT(pc)])->vcell = acc ;
            continue ;
            
          case OS_B_CST:
            acc = data[OS_GET_SHORT(pc)] ;
            continue ;
          case OS_B_CSTPUSH:
            *++stackp = acc = data[OS_GET_SHORT(pc)] ;
            continue ;

          case OS_B_LINK:
            robj = data[OS_GET_SHORT(pc)] ; 
            if ( ((OsVector) robj)->csize != ridx ) {
                OS_VM_OUTGOING() ;
                (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                       , OsIntegerMake(((OsVector)robj)->csize)
                                       , OsIntegerMake(ridx)) ;
            }
            *++stackp = robj ; 
            *++stackp = (OsObject) env ;
            env = stackp - 1 ;
            continue ;
          case OS_B_LINKN:
            robj = data[OS_GET_SHORT(pc)] ;
            if ( ridx > (rjdx = ((OsVector) robj)->csize - 1) ) {
                OS_VM_OUTGOING() ;
                stackp   -= (ridx - rjdx) ;
                stackp[1] = (OsObject) OsListNMake (ridx - rjdx, stackp+1) ;
                stackp++ ;
            } else if ( rjdx == ridx ) {
                *++stackp = OsA_Nil ;
            } else {
                OS_VM_OUTGOING() ;
                (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                       , OsIntegerMake (rjdx)
                                       , OsIntegerMake (ridx)) ;
            }
            *++stackp = robj ;
            *++stackp = (OsObject) env ;
            env = stackp - 1 ;
            continue ;
          case OS_B_UNLINK:
            /*
             * At LINK time (see below, we have pushed the frame vector
             * description (1 word), and the previous env register (1 word)
             * this account for the added 2 to the prvided value.
             */
            env = (OsObject *) ((OsObject *) env[1]) ;
            stackp -= (((int) *pc) + 2) ;
            pc++ ;
            continue ;
          case OS_B_LINKEXT:
            /*
             * acc contains a closed code block, and it's time to set the env
             * register for its execution
             */
            robj = data[OS_GET_SHORT(pc)] ;
            if ( ((OsVector) robj)->csize != ridx ) {
                OS_VM_OUTGOING() ;
                (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                       , OsIntegerMake(((OsVector)robj)->csize)
                                       , OsIntegerMake(ridx)) ;
            }
            *++stackp = robj ;
            *++stackp = (OsObject) ((OsClosedBlock) acc)->env ;
            *++stackp = (OsObject) env ;
            env = stackp - 2 ;
            continue ;
          case OS_B_LINKNEXT:
            robj = data[OS_GET_SHORT(pc)] ;
            if ( ridx > (rjdx = ((OsVector) robj)->csize - 1) ) {
                OS_VM_OUTGOING() ;
                stackp   -= (ridx - rjdx) ;
                stackp[1] = (OsObject) OsListNMake (ridx - rjdx, stackp+1) ;
                stackp++ ;
            } else if ( rjdx == ridx ) {
                *++stackp = OsA_Nil ;
            } else {
                OS_VM_OUTGOING() ;
                (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                       , OsIntegerMake (rjdx)
                                       , OsIntegerMake (ridx)) ;
            }
            *++stackp = robj ;
            *++stackp = (OsObject) ((OsClosedBlock) acc)->env ;
            *++stackp = (OsObject) env ;
            env = stackp - 2 ;
            continue ;
          case OS_B_UNLINKEXT:
            env     = (OsObject *) *stackp-- ;
            stackp -= (((int) *pc++) + 2) ;
            continue ;

          case OS_B_SAVE:
            ridx = OS_GET_SHORT (pc) ;
            *++stackp = (OsObject) data ;
            *++stackp = (OsObject) (pc - 2 + ridx) ;
            continue ;
          case OS_B_JUMP:
            ridx = ((int) *pc++) ;
            if ( (! OsIsCodeBlock (acc))
                && ( ! OsIsClosedBlock (acc))
                && ( ! OsIsContBlock (acc))
                && ( ! OsIsWindBlock (acc)) ) {
                OS_VM_OUTGOING() ;
                acc = OsErrorTrigger (1, OsE_BadFunction, acc) ;
            }
            pc   = ((OsCodeBlock) acc)->code->string ;
            data = ((OsCodeBlock) acc)->data->cells ;
            continue ;
          case OS_B_UNLINKJUMP:
            /* Get rid of given number of env frames, which are NON externs */
            ridx    = (int) *pc++ ;
            while ( ridx-- )
                env = ((OsObject *) env[1]) ;
            data = stackp - (rjdx = ridx = ((int) *pc++)) ;
            /* Restore stack pointer, env is now set to a NON ext env
             * that's why we take env[1] to get the previous env. The added
             * (-1) is to pop env's value itself */
            robj = (OsObject) (OS_STACKP(env) ? env[1] : env[2]) ;
            if ( robj != ((OsObject) OS_NULL_ENV) ) {
                env    = (OsObject *) robj ;
                stackp = env - 1 - ((int) *pc++) ;
            } else {
                stackp  = env - ((int) *pc++) - 1 ;
                env     = (OsObject *) OS_NULL_ENV ;
            }
            /* bump the stack, and jump */
            while ( ridx-- )
                *++stackp = *++data ;
            ridx = rjdx ;
            if ( (! OsIsCodeBlock (acc))
                && ( ! OsIsClosedBlock (acc))
                && ( ! OsIsContBlock (acc))
                && ( ! OsIsWindBlock (acc)) ) {
                OS_VM_OUTGOING() ;
                acc = OsErrorTrigger (1, OsE_BadFunction, acc) ;
            }
            pc   = ((OsCodeBlock) acc)->code->string ;
            data = ((OsCodeBlock) acc)->data->cells ;
            continue ;
          case OS_B_UNLINKEXTJUMP:
            /* Instruction provides:
             * - Number of static frames to withdraw
             * - arity of the call
             * - size of deepest withdrawed env
             * We first withdraw the static frames
             */
            ridx    = (int) *pc++ ;
            while ( ridx-- )
                env = ((OsObject *) env[1]) ;
            data = stackp - (rjdx = ridx = ((int) *pc++)) ;
            if ( OS_STACKP (env) ) {
                /* env is in the stack, no hack needed, we use it to reset the
                 * stack pointer, before stack bumping:
                 * -1     get rid of env description vector
                 * -*pc++ get on top of the ext frame
                 * env is set to the normal previous frame */
                stackp = env - 1 - ((int) *pc++) ;
                env    = (OsObject *) env[2] ;
            } else {
                /* This case happen just after a block has been closed: its
                 * extframe has been copied through CLOSE on heap. We use the
                 * last env[2] set at the end of CLOSE instruction to get back
                 * the stack pointer:
                 * - This *has* to be a stack pointer, because env is a heap
                 * frame, so it has been created through CLOSE.
                 * - Once we get the dead frame, we retry the withdrawal of
                 * all freezed frame (this brings us before the first non stack
                 * frame).
                 * We than have the "correct" (hopefully, as I am not really
                 * sure of 2)
                 */
                env    = (OsObject *) env[2] ;
                stackp = env - (((OsVector) *env)->csize + 1) ;
                env    = (OsObject *) env[2] ;
            }
            /* bump the stack, and jump */
            while ( ridx-- )
                *++stackp = *++data ;
            ridx = rjdx ;
            if ( (! OsIsCodeBlock (acc))
                && ( ! OsIsClosedBlock (acc))
                && ( ! OsIsContBlock (acc))
                && ( ! OsIsWindBlock (acc)) ) {
                OS_VM_OUTGOING() ;
                acc = OsErrorTrigger (1, OsE_BadFunction, acc) ;
            }
            pc   = ((OsCodeBlock) acc)->code->string ;
            data = ((OsCodeBlock) acc)->data->cells ;
            continue ;
          case OS_B_RETURN:
            pc         = (unsigned char *) (*stackp--) ;
            data       = (OsObject *) (*stackp--) ;
            /* acc is now set with th result of the call */
            continue ;
          case OS_B_EXIT:
            goto done ;
          case OS_B_BRIFFALSE:
            ridx = OS_GET_SHORT (pc) ;
            if ( acc == OsA_False ) 
                pc += (ridx - 2) ;
            continue ;
          case OS_B_BRIFTRUE:
            ridx = OS_GET_SHORT (pc) ;
            if ( acc != OsA_False ) 
                pc += (ridx - 2) ;
            continue ;
          case OS_B_RGOTO:
            ridx = OS_GET_SHORT (pc) ;
            pc += (ridx - 2) ;
            continue ;

          case OS_B_SAVECONT:
            /*
             * Because we do take the adress of 'cont' we need not declare
             * it as volatile. It would be the case otherwise
             */
            OS_VM_OUTGOING() ;
            stackp -= ridx ; 
            if ( ! OsContBlockMake (interp, acc, env, ridx
                                    , interp->stack, stackp 
                                    , data, pc + sizeof(short)
                                    , (OsContBlock *) &cont) ) {
                acc = data[OS_GET_SHORT(pc)] ;
                OS_VM_SUBR_CHECK (acc) ;
                acc = (((OsSubr) acc)->cfun) (cont, ridx, stackp+1) ;
            } else {		/* restore continuation */
                /* interp was saved on the C stack, and restore through C stack
                 * copying. However, the global current interp variable may be
                 * out of sync:
                 */
                if ( interp != ((OsContBlock) cont)->interp ) {
                    interp = ((OsContBlock) cont)->interp ;
                    stacke = interp->stacke ;
                }
                if ( interp != OsCurrentInterp() )
                    (void) OsInterpSwitch (interp) ;
                acc        = ((OsContBlock) cont)->acc ;
                env        = ((OsContBlock) cont)->env ;
                ridx       = ((OsContBlock) cont)->ridx ;
                bcopy (((OsContBlock) cont)->stack
                       , interp->stack
                       , sizeof(OsObject)*(((OsContBlock) cont)->stacklen+1));
                stackp     = interp->stack + ((OsContBlock) cont)->stacklen ;
                data       = ((OsContBlock) cont)->cells ;
                pc         = ((OsContBlock) cont)->pc ;
                OS_VM_OUTGOING() ;
                os_restore_protection (((OsContBlock) cont)->protection) ;
            }
            continue ;
          case OS_B_RETCONT:
            if ( ridx != 1 ) {
                OS_VM_OUTGOING() ;
                (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                       , OsIntegerMake (1)
                                       , OsIntegerMake (ridx)) ;
            }
            cont = (OsCodeBlock) data[OS_GET_SHORT(pc)] ;
            os_cstack_install ((OsContBlock) cont, *stackp--) ;
            /* NOT REACHED */
            continue ;

          case OS_B_SAVEWIND:
            if ( ridx != 3 ) {
                OS_VM_OUTGOING() ;
                (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                       , OsIntegerMake (3)
                                       , OsIntegerMake (ridx)) ;
            }
            OS_VM_OUTGOING () ;
            (void) OsErrorTrigger (1, OsE_Implementation
                                   , OsStringMake ("OsCodeBlockExecute")
                                   , OsStringMake ("SAVEWIND not impl.")) ;
            continue ;
          case OS_B_RETWIND:
            if ( ridx != 1) {
                  OS_VM_OUTGOING() ;
                  (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                         , OsIntegerMake (1)
                                         , OsIntegerMake (ridx)) ;
            }
            cont = (OsCodeBlock) data[OS_GET_SHORT(pc)] ;
            os_throw_wind ((OsWindBlock) cont, *stackp--) ;
            /* Reaching code below, means an error occured */
            OS_VM_OUTGOING() ;
            (void) OsErrorTrigger (1, ose_bad_wind, cont) ;
            /* NOT REACHED */
            continue ;
            
          case OS_B_PUSH:
            *++stackp = acc ;
            continue ;
          case OS_B_POP:
            acc = *stackp-- ;
            continue ;
          case OS_B_LOAD:
            ridx = ((int) *pc++) ;
            continue ;

          case OS_B_ASUBR0:
            if ( ridx ) {
                OS_VM_OUTGOING() ;
                (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                       , OsIntegerMake (0)
                                       , OsIntegerMake (ridx)) ;
            }
            /* no break intentional */
          case OS_B_SUBR0:
            acc = data[OS_GET_SHORT(pc)] ;
            OS_VM_SUBR_CHECK (acc) ;
            OS_VM_CAPPLY0 (((OsSubr) acc)->cfun) ;
            continue ;
          case OS_B_ASUBR1:
            if ( ridx != 1 ) {
                OS_VM_OUTGOING() ;
                (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                       , OsIntegerMake (1)
                                       , OsIntegerMake (ridx)) ;
            }
            /* no break intentional */
          case OS_B_SUBR1:
            acc = data[OS_GET_SHORT(pc)] ;
            OS_VM_SUBR_CHECK(acc) ;
            OS_VM_CAPPLY1(((OsSubr) acc)->cfun) ;
            continue ;
          case OS_B_ASUBR2:
            if ( ridx != 2 ) {
                OS_VM_OUTGOING() ;
                (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                       , OsIntegerMake (2)
                                       , OsIntegerMake (ridx)) ;
            }
            /* no break intentional */
          case OS_B_SUBR2:
            acc = data[OS_GET_SHORT(pc)] ;
            OS_VM_SUBR_CHECK(acc) ;
            OS_VM_CAPPLY2 (((OsSubr) acc)->cfun) ;
            continue ;
          case OS_B_ASUBR3:
            if ( ridx != 3 ) {
                OS_VM_OUTGOING() ;
                (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                       , OsIntegerMake (3)
                                       , OsIntegerMake (ridx)) ;
            }
            /* no break intentional */
          case OS_B_SUBR3:
            acc = data[OS_GET_SHORT(pc)] ;
            OS_VM_SUBR_CHECK (acc) ;
            OS_VM_CAPPLY3 (((OsSubr) acc)->cfun) ;
            continue ;
          case OS_B_ASUBR4:
            if ( ridx != 4 ) {
                OS_VM_OUTGOING() ;
                (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                       , OsIntegerMake (4)
                                       , OsIntegerMake (ridx)) ;
            }
            /* no break intentional */
          case OS_B_SUBR4:
            acc = data[OS_GET_SHORT(pc)] ;
            OS_VM_SUBR_CHECK (acc) ;
            OS_VM_CAPPLY4 (((OsSubr) acc)->cfun) ;
            continue ;
          case OS_B_ASUBR5:
            if ( ridx != 5 ) {
                OS_VM_OUTGOING() ;
                (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                       , OsIntegerMake (5)
                                       , OsIntegerMake (ridx)) ;
            }
            /* no break intentional */
          case OS_B_SUBR5:
            acc = data[OS_GET_SHORT(pc)] ;
            OS_VM_SUBR_CHECK (acc) ;
            OS_VM_CAPPLY5 (((OsSubr) acc)->cfun) ;
            continue ;
          case OS_B_ASUBR6:
            if ( ridx != 6 ) {
                OS_VM_OUTGOING() ;
                (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                       , OsIntegerMake (6)
                                       , OsIntegerMake (ridx)) ;
            }
            /* no break intentional */
          case OS_B_SUBR6:
            acc = data[OS_GET_SHORT(pc)] ;
            OS_VM_SUBR_CHECK (acc) ;
            OS_VM_CAPPLY6 (((OsSubr) acc)->cfun) ;
            continue ;
          case OS_B_ASUBR7:
            if ( ridx != 7 ) {
                OS_VM_OUTGOING() ;
                (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                       , OsIntegerMake (7)
                                       , OsIntegerMake (ridx)) ;
            }
            /* no break intentional */
          case OS_B_SUBR7:
            acc = data[OS_GET_SHORT(pc)] ;
            OS_VM_SUBR_CHECK(acc) ;
            OS_VM_CAPPLY7 (((OsSubr) acc)->cfun) ;
            continue ;
          case OS_B_ASUBRN:
            acc = data[OS_GET_SHORT(pc)] ;
            OS_VM_SUBR_CHECK (acc) ;
            OS_VM_CAPPLYN (((OsSubr) acc)->cfun) ;
            continue ;

          case OS_B_CLOSE:
            /*
             * Close copy (freeze in fact) the current env. Because we need
             * only one copy (be it on stack or on heap) of any env, closing
             * an env changes the env register to the heap allocated version
             * of env.
             */
            acc = (OsObject) OsCodeBlockClose ((OsCodeBlock) acc) ;
            if ( ! OS_STACKP (env) ) {
                ((OsClosedBlock) acc)->env = env ;
                continue ;
            }
            *++stackp = (OsObject) data ;
            *++stackp = robj = (OsObject) env ;
            for (ridx = 0
                 ; OS_STACKP (robj) 
                 ; robj = ((OsObject *) robj)[1])
                ridx += (((OsVector) *((OsObject *) robj))->csize + 3) ;
            robj = (OsObject) OsGCMalloc (sizeof (OsObject) * ridx) ;
            data = (OsObject *) robj ;
            ridx = ((OsVector) *env)->csize ; 
            ((OsClosedBlock) acc)->env = (OsObject *) (data + ridx) ;

            while ( 1 ) {
                /* copy env frame to the heap */
                env -= ridx ;
                while ( ridx-- >= 0 ) {
                    *data++ = *env++ ;
                }
                robj = (OsObject) (env-1) ;
                /* set up env[1] to previous frame, and env[2] to stackp */
                if ( (env = (OsObject *) *env, ! OS_STACKP (env)) ) {
                    *data++ = (OsObject) env ;
                    *data++ = robj ;
                    break ;
                } else {
                    ridx  = ((OsVector) *env)->csize ;
                    *data = (OsObject) (data + ridx + 2) ; data++ ;
                    *data++ = robj ;
                }
            }
            /* The last saved-env (env[2]) value is set to the old stack
             * frames: this allows UNLINKEXTJUMP to retreive it */
            env     = (OsObject *) *stackp-- ;
            data    = (OsObject *) *stackp-- ;
            robj    = ((OsObject) ((OsClosedBlock) acc)->code) ;

            env = ((OsClosedBlock) acc)->env ;
                    
            continue ;

          /* Messge sending are inlined:
           * We *duplicate* here the code for OsEObjSendMsgx, where x is the
           * arity of the call. This makes a great boost in performance, by
           * saving at least two function calls for lots of cases (which ends
           * up by saving some malloc's)
           * The draw back is that we get code duplication.
           * There is some dependency with bytecompile.c:
           * - it is expected to emit a SAVE opcode, before any message
           * sending (i.e: msg sends are never tailed), because we do not know
           * at compile time the selector implementation type.
           * - a slot has been left in (first) stack position, to put the
           * receiver object (this is due to our particular arg push oredering)
           */
          case OS_B_SENDMSG0:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup ((OsEObj) acc, sel->selid) ;
              switch (imp->type) {
                case OS_SELIMP_CFUN:
                  OS_VM_CSEND0 (imp, sel) ;
                  stackp-- ;
                  pc   = (unsigned char *) (*stackp--) ;
                  data = (OsObject *) (*stackp--) ;
                  break ;
                case OS_SELIMP_CLOS:
                  *stackp = acc ;	/* "set" receiver */
                  ridx = 1 ;		/* single arg to sel imp */
                  acc  = (OsObject) SELIMP_BLOCK (imp) ;
                  pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                  data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                  break ;
                case OS_SELIMP_SLOT:
                  ridx = SELIMP_DIDX (imp) ;
                  while ( (! ((OsEObj) acc)->readers)
                          || (ridx >= ((OsEObj) acc)->readers->csize)
                          || (((OsEObj) acc)->readers->cells[ridx] != (OsObject) sel) ) {
                      acc  = (OsObject) ((OsEObj) acc)->prototype ;
                      imp  = EObjSelimpLookup ((OsEObj) acc, sel->selid) ;
                      ridx = SELIMP_DIDX (imp) ;
                  }
                  acc = ((OsEObj) acc)->values->cells[ridx] ;
                  stackp-- ;
                  pc   = (unsigned char *) (*stackp--) ;
                  data = (OsObject *) (*stackp--) ;
                  break ;
              }
          }
          continue ;
          case OS_B_SENDMSG1:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup ((OsEObj) acc, sel->selid) ;
              switch (imp->type) {
                case OS_SELIMP_CFUN:
                  OS_VM_CSEND1 (imp, sel) ;
                  stackp-- ;
                  pc   = (unsigned char *) (*stackp--) ;
                  data = (OsObject *) (*stackp--) ;
                  break ;
                case OS_SELIMP_CLOS:
                  stackp[-1] = acc ;	/* "set" receiver */
                  ridx = 2 ;		/* signle arg to sel imp */ 
                  acc  = (OsObject) SELIMP_BLOCK (imp) ;
                  pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                  data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                  break ;
                case OS_SELIMP_SLOT:
                  ridx = SELIMP_DIDX (imp) ;
                  if ( (! ((OsEObj) acc)->writers)
                       || (ridx >= ((OsEObj) acc)->writers->csize)
                       || (((OsEObj) acc)->writers->cells[ridx] != (OsObject) sel) ) {
                      OsEObjAllocateSlot ((OsEObj) acc
                                          , os_w_to_r_sel (sel)
                                          , sel
                                          , *stackp--) ;
                  } else {
                      acc = ((OsEObj) acc)->values->cells[ridx] = *stackp-- ;
                  }
                  stackp-- ;
                  pc   = (unsigned char *) (*stackp--) ;
                  data = (OsObject *) (*stackp--) ;
                  break ;
              }
          }
          continue ;
          case OS_B_SENDMSG2:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup ((OsEObj) acc, sel->selid) ;
              switch (imp->type) {
                  case OS_SELIMP_CFUN:
                    OS_VM_CSEND2 (imp, sel) ;
                    stackp-- ;
                    pc   = (unsigned char *) (*stackp--) ;
                    data = (OsObject *) (*stackp--) ;
                    break ;
                  case OS_SELIMP_CLOS:
                    stackp[-2] = acc ;
                    ridx = 3 ;
                    acc  = (OsObject) SELIMP_BLOCK (imp) ;
                    pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                    data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                    break ;
              }
          }
          continue ;
          
          case OS_B_SENDMSG3:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup ((OsEObj) acc, sel->selid) ;
              switch (imp->type) {
                  case OS_SELIMP_CFUN:
                    OS_VM_CSEND3 (imp, sel) ;
                    stackp-- ;
                    pc   = (unsigned char *) (*stackp--) ;
                    data = (OsObject *) (*stackp--) ;
                    break ;
                  case OS_SELIMP_CLOS:
                    stackp[-3] = acc ;
                    ridx = 4 ;
                    acc  = (OsObject) SELIMP_BLOCK (imp) ;
                    pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                    data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                    break ;
              }
          }
          continue ;

          case OS_B_SENDMSG4:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup ((OsEObj) acc, sel->selid) ;
              switch (imp->type) {
                  case OS_SELIMP_CFUN:
                    OS_VM_CSEND4 (imp, sel) ;
                    stackp-- ;
                    pc   = (unsigned char *) (*stackp--) ;
                    data = (OsObject *) (*stackp--) ;
                    break ;
                  case OS_SELIMP_CLOS:
                    stackp[-4] = acc ;
                    ridx = 5 ;
                    acc  = (OsObject) SELIMP_BLOCK (imp) ;
                    pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                    data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                    break ;
              }
          }
          continue ;

          case OS_B_SENDMSG5:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup ((OsEObj) acc, sel->selid) ;
              switch (imp->type) {
                  case OS_SELIMP_CFUN:
                    OS_VM_CSEND5 (imp, sel) ;
                    stackp-- ;
                    pc   = (unsigned char *) (*stackp--) ;
                    data = (OsObject *) (*stackp--) ;
                    break ;
                  case OS_SELIMP_CLOS:
                    stackp[-5] = acc ;
                    ridx = 6 ;
                    acc  = (OsObject) SELIMP_BLOCK (imp) ;
                    pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                    data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                    break ;
              }
          }
          continue ;

          case OS_B_SENDMSG6:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup ((OsEObj) acc, sel->selid) ;
              switch (imp->type) {
                  case OS_SELIMP_CFUN:
                    OS_VM_CSEND6 (imp, sel) ;
                    stackp-- ;
                    pc   = (unsigned char *) (*stackp--) ;
                    data = (OsObject *) (*stackp--) ;
                    break ;
                  case OS_SELIMP_CLOS:
                    stackp[-6] = acc ;
                    ridx = 7 ;
                    acc  = (OsObject) SELIMP_BLOCK (imp) ;
                    pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                    data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                    break ;
              }
          }
          continue ;

          case OS_B_SENDMSG7:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup ((OsEObj) acc, sel->selid) ;
              switch (imp->type) {
                  case OS_SELIMP_CFUN:
                    OS_VM_CSEND7 (imp, sel) ;
                    stackp-- ;
                    pc   = (unsigned char *) (*stackp--) ;
                    data = (OsObject *) (*stackp--) ;
                    break ;
                  case OS_SELIMP_CLOS:
                    stackp[-7] = acc ;
                    ridx = 8 ;
                    acc  = (OsObject) SELIMP_BLOCK (imp) ;
                    pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                    data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                    break ;
              }
          }
          continue ;
              
          case OS_B_SENDSUPERMSG0:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup (((OsEObj) acc)->prototype, sel->selid) ;
              switch (imp->type) {
                case OS_SELIMP_CFUN:
                  OS_VM_CSEND0 (imp, sel) ;
                  stackp-- ;
                  pc   = (unsigned char *) (*stackp--) ;
                  data = (OsObject *) (*stackp--) ;
                  break ;
                case OS_SELIMP_CLOS:
                  *stackp = acc ;	/* "set" receiver */
                  ridx = 1 ;		/* single arg to sel imp */ 
                  acc  = (OsObject) SELIMP_BLOCK (imp) ;
                  pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                  data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                  break ;
                case OS_SELIMP_SLOT:
                  ridx = SELIMP_DIDX (imp) ;
                  while ( (! ((OsEObj) acc)->readers)
                          || (ridx >= ((OsEObj) acc)->readers->csize)
                          || (((OsEObj) acc)->readers->cells[ridx] != (OsObject) sel) ) {
                      acc  = (OsObject) ((OsEObj) acc)->prototype ;
                      imp  = EObjSelimpLookup ((OsEObj) acc, sel->selid) ;
                      ridx = SELIMP_DIDX (imp) ;
                  }
                  acc = ((OsEObj) acc)->values->cells[ridx] ;
                  stackp-- ;
                  pc   = (unsigned char *) (*stackp--) ;
                  data = (OsObject *) (*stackp--) ;
                  break ;
              }
          }
          continue ;
          case OS_B_SENDSUPERMSG1:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup (((OsEObj) acc)->prototype, sel->selid) ;
              switch (imp->type) {
                case OS_SELIMP_CFUN:
                  OS_VM_CSEND1 (imp, sel) ;
                  stackp-- ;
                  pc   = (unsigned char *) (*stackp--) ;
                  data = (OsObject *) (*stackp--) ;
                  break ;
                case OS_SELIMP_CLOS:
                  stackp[-1] = acc ;	/* "set" receiver */
                  ridx = 2 ;		/* signle arg to sel imp */ 
                  acc  = (OsObject) SELIMP_BLOCK (imp) ;
                  pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                  data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                  break ;
                case OS_SELIMP_SLOT:
                  ridx = SELIMP_DIDX (imp) ;
                  if ( (! ((OsEObj) acc)->writers)
                       || (ridx >= ((OsEObj) acc)->writers->csize)
                       || (((OsEObj) acc)->writers->cells[ridx] != (OsObject) sel) ) {
                      OsEObjAllocateSlot ((OsEObj) acc
                                          , os_w_to_r_sel (sel)
                                          , sel
                                          , *stackp--) ;
                  } else {
                      acc = ((OsEObj) acc)->values->cells[ridx] = *stackp-- ;
                  }
                  stackp-- ;
                  pc   = (unsigned char *) (*stackp--) ;
                  data = (OsObject *) (*stackp--) ;
                  break ;
              }
          }
          continue ;
          case OS_B_SENDSUPERMSG2:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup (((OsEObj) acc)->prototype, sel->selid) ;
              switch (imp->type) {
                  case OS_SELIMP_CFUN:
                    OS_VM_CSEND2 (imp, sel) ;
                    stackp-- ;
                    pc   = (unsigned char *) (*stackp--) ;
                    data = (OsObject *) (*stackp--) ;
                    break ;
                  case OS_SELIMP_CLOS:
                    stackp[-2] = acc ;
                    ridx = 3 ;
                    acc  = (OsObject) SELIMP_BLOCK (imp) ;
                    pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                    data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                    break ;
              }
          }
          continue ;
          
          case OS_B_SENDSUPERMSG3:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup (((OsEObj) acc)->prototype, sel->selid) ;
              switch (imp->type) {
                  case OS_SELIMP_CFUN:
                    OS_VM_CSEND3 (imp, sel) ;
                    stackp-- ;
                    pc   = (unsigned char *) (*stackp--) ;
                    data = (OsObject *) (*stackp--) ;
                    break ;
                  case OS_SELIMP_CLOS:
                    stackp[-3] = acc ;
                    ridx = 4 ;
                    acc  = (OsObject) SELIMP_BLOCK (imp) ;
                    pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                    data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                    break ;
              }
          }
          continue ;

          case OS_B_SENDSUPERMSG4:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup (((OsEObj) acc)->prototype, sel->selid) ;
              switch (imp->type) {
                  case OS_SELIMP_CFUN:
                    OS_VM_CSEND4 (imp, sel) ;
                    stackp-- ;
                    pc   = (unsigned char *) (*stackp--) ;
                    data = (OsObject *) (*stackp--) ;
                    break ;
                  case OS_SELIMP_CLOS:
                    stackp[-4] = acc ;
                    ridx = 5 ;
                    acc  = (OsObject) SELIMP_BLOCK (imp) ;
                    pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                    data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                    break ;
              }
          }
          continue ;

          case OS_B_SENDSUPERMSG5:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup (((OsEObj) acc)->prototype, sel->selid) ;
              switch (imp->type) {
                  case OS_SELIMP_CFUN:
                    OS_VM_CSEND5 (imp, sel) ;
                    stackp-- ;
                    pc   = (unsigned char *) (*stackp--) ;
                    data = (OsObject *) (*stackp--) ;
                    break ;
                  case OS_SELIMP_CLOS:
                    stackp[-5] = acc ;
                    ridx = 6 ;
                    acc  = (OsObject) SELIMP_BLOCK (imp) ;
                    pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                    data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                    break ;
              }
          }
          continue ;

          case OS_B_SENDSUPERMSG6:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup (((OsEObj) acc)->prototype, sel->selid) ;
              switch (imp->type) {
                  case OS_SELIMP_CFUN:
                    OS_VM_CSEND6 (imp, sel) ;
                    stackp-- ;
                    pc   = (unsigned char *) (*stackp--) ;
                    data = (OsObject *) (*stackp--) ;
                    break ;
                  case OS_SELIMP_CLOS:
                    stackp[-6] = acc ;
                    ridx = 7 ;
                    acc  = (OsObject) SELIMP_BLOCK (imp) ;
                    pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                    data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                    break ;
              }
          }
          continue ;

          case OS_B_SENDSUPERMSG7:
          {
              os_selimp imp ;
              OsEObjSel sel = (OsEObjSel) data[OS_GET_SHORT(pc)] ;
              
              (void) OsMustBeEObj (acc) ;
              imp = EObjSelimpLookup (((OsEObj) acc)->prototype, sel->selid) ;
              switch (imp->type) {
                  case OS_SELIMP_CFUN:
                    OS_VM_CSEND7 (imp, sel) ;
                    stackp-- ;
                    pc   = (unsigned char *) (*stackp--) ;
                    data = (OsObject *) (*stackp--) ;
                    break ;
                  case OS_SELIMP_CLOS:
                    stackp[-7] = acc ;
                    ridx = 8 ;
                    acc  = (OsObject) SELIMP_BLOCK (imp) ;
                    pc   = ((OsCodeBlock) SELIMP_BLOCK(imp))->code->string ;
                    data = ((OsCodeBlock) SELIMP_BLOCK(imp))->data->cells ;
                    break ;
              }
          }
          continue ;
              
#define OS_INVALIDATE_DEADENV() \
    ridx = 0 ; \
    while ( (OsObject) stackp <= interp->dbg_sp[ridx] ) { \
        interp->dbg_sp[ridx-1] = (OsObject) 0 ; \
        ridx -= OS_DBG_FRAMESZ ; \
    }

          case OS_DBG_PUSH:
            OS_INVALIDATE_DEADENV() ;
            *++(interp->dbg_sp) = (OsObject) data[OS_GET_SHORT(pc)] ;
            *++(interp->dbg_sp) = (OsObject) env ;
            *++(interp->dbg_sp) = (OsObject) stackp ;
            continue ;
          case OS_DBG_PUSHBRK:
            OS_INVALIDATE_DEADENV() ;
            *++(interp->dbg_sp) = (OsObject) data[OS_GET_SHORT(pc)] ;
            *++(interp->dbg_sp) = (OsObject) env ;
            *++(interp->dbg_sp) = (OsObject) stackp ;
            OS_VM_OUTGOING() ;
            (void) OsErrorTrigger (1, ose_break) ;
            continue ;
          case OS_DBG_SAFEPUSH:
            OS_INVALIDATE_DEADENV() ;
            *++(interp->dbg_sp) = (OsObject) data[OS_GET_SHORT(pc)] ;
            *++(interp->dbg_sp) = OS_STACKP(env)?(OsObject) 0 : (OsObject) env;
            *++(interp->dbg_sp) = (OsObject) stackp ;
            continue ;
          case OS_DBG_SAFEPUSHBRK:
            OS_INVALIDATE_DEADENV() ;
            *++(interp->dbg_sp) = (OsObject) data[OS_GET_SHORT(pc)] ;
            *++(interp->dbg_sp) = OS_STACKP(env)?(OsObject) 0 : (OsObject) env;
            *++(interp->dbg_sp) = (OsObject) stackp ;
            OS_VM_OUTGOING() ;
            (void) OsErrorTrigger (1, ose_break) ;
            continue ;

          case OS_DBG_POP:
            /* Tail calls make us miss some DBG_POP instructions, that's why
             * we save the stack pointer, and compare against it here */
            (interp->dbg_sp) -= OS_DBG_FRAMESZ ;
            while ( (OsObject) stackp <= *(interp->dbg_sp) ) {
                interp->dbg_sp -= OS_DBG_FRAMESZ ;
            }
            continue ;
            
          default:
            {
                unsigned char *epc = pc - 1 ;
                OsOutputPortWriteCString (OsCurrentOutputPort
                                          , "*** Implementation bug: \n") ;
                os_decode_instr (OsCurrentOutputPort, &epc, data) ;
            }
            exit (0) ;
            
        }
    }

  done:

#ifdef OS_VM_TIMEABLE
    if ( osa_times_machine->vcell != OsA_False ) {
        times (&tend) ;
        fprintf (stdout, ";; timing: user = %ld, sys = %ld\n"
                 , DELTA_UTIME (&tstart, &tend)
                 , DELTA_STIME (&tstart, &tend)) ;
    }
#endif /* OS_VM_TIMEABLE */
    if ( old_interp )
        (void) OsInterpSwitch (old_interp) ;
    
    OS_VM_OUTGOING() ;
    return (acc) ;
}
