/***************************************************************************
 * cons.c
 * cons.c,v 1.4 1995/03/17 14:20:51 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * Scheme pairs handling.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <os_config.h>
#include <os_interp.h>
#include <os_obj.h>
#include <os_string.h>
#include <os_atom.h>
#include <os_gc.h>
#include <os_vector.h>
#include <os_error.h>
#include <os_cons.h>
#include <os_arit.h>
#include <os_cblock.h>
#include <os_subr.h>
#include <os_port.h>

OsClass OsConsClass = (OsClass) 0 ;

/*
 * Constructors:
 */

OsCons OsConsMake (OsObject car, OsObject cdr) {
    OsCons cons = (OsCons) OsObjectMake (OsConsClass, 0) ;
    cons->car = car ;
    cons->cdr = cdr ;
    return (cons) ;
}

OsCons OsListNMake (int argc, OsObject *argv) {
    OsObject cons = (OsObject) OsA_Nil, *list = &cons ;
    while ( argc-- ) {
        *list = (OsObject) OsConsMake (*argv++, (OsObject) OsA_Nil) ;
        list  = &(((OsCons) *list)->cdr) ;
    }
    return ((OsCons) cons) ;
}

OsCons OsList2Make (OsObject o1, OsObject o2) {
    return (OsConsMake (o1, (OsObject) OsConsMake (o2, OsA_Nil))) ;
}

/*
 * C functions
 */

unsigned int OsConsLength (OsCons cons) {
    unsigned int length = 0 ;
    while ( OsIsCons (cons) ) {
        length++ ;
        cons = (OsCons) OsConsCdr (cons) ;
    }
    return (length) ;
}

OsCons OsList2Append (OsObject o1, OsObject o2, OsObject tail) {
    return (OsConsMake (o1, (OsObject) OsConsMake (o2, tail))) ;
}

OsCons OsListCopy (OsCons lst) {
    OsObject start = OsA_Nil, *plist = &start ;

    while ( OsIsCons (lst) ) {
        *plist = (OsObject) OsConsMake (OsConsCar(lst), OsA_Nil) ;
        plist  = &(OsConsCdr(*plist)) ;
        lst    = (OsCons) OsConsCdr (lst) ;
    }
    return ((OsCons) start) ;
}

/*
 * Subrs for cons:
 */

/* AUTODOC
 * (car PAIR)
 * Returns the contents of the car field of PAIR.
 * Note that it is an error to take the car of the empty list.
 *
 *   (car '(a b c))              ==>  a
 *   (car '((a) b c d))          ==>  (a)
 *   (car '(1 . 2))              ==>  1
 *   (car '())                   ==>  error
 */

static OsObject OsSubrCar (OsObject o) {
    OsMustBeCons (o) ;
    return (OsConsCar ((OsCons) o)) ;
}

/* AUTODOC
 * (cdr PAIR)
 * Returns the contents of the cdr field of PAIR.
 * Note that it is an error to take the cdr of the empty list.
 *
 *   (cdr '((a) b c d))          ==>  (b c d)
 *   (cdr '(1 . 2))              ==>  2
 *   (cdr '())                   ==>  error
 */

static OsObject OsSubrCdr (OsObject o) {
    OsMustBeCons (o) ;
    return (OsConsCdr ((OsCons) o)) ;
}

static OsObject OsSubrCaar (OsObject o) {
    OsCons car ;
    OsMustBeCons (o) ;
    car = (OsCons) OsConsCar (o) ;
    OsMustBeCons (car) ;
    return (OsConsCar (car)) ;
}

static OsObject OsSubrCdar (OsObject o) {
    OsCons car ;
    OsMustBeCons (o) ;
    car = (OsCons) OsConsCar (o) ;
    OsMustBeCons (car) ;
    return (OsConsCdr (car)) ;
}

static OsObject OsSubrCadr (OsObject o) {
    OsCons cdr ;
    OsMustBeCons (o) ;
    cdr = (OsCons) OsConsCdr (o) ;
    OsMustBeCons (cdr) ;
    return (OsConsCar (cdr)) ;
}

static OsObject OsSubrCddr (OsObject o) {
    OsCons cdr ;
    OsMustBeCons (o) ;
    cdr = (OsCons) OsConsCdr (o) ;
    OsMustBeCons (cdr) ;
    return (OsConsCdr (cdr)) ;
}

/* AUTODOC
 * (set-car! PAIR OBJECT)
 * Stores OBJECT in the car field of PAIR. 
 * The value returned by set-car! is unspecified.
 *
 *   (define (f) (list 'foo))
 *   (set-car! (f) 3)            ==>  unspecified
 */

static OsObject OsSubrSetCar (OsObject cons, OsObject obj) {
    OsMustBeCons (cons) ;
    OsConsCar ((OsCons) cons) = obj ;
    return (obj) ;
}

/* AUTODOC
 * (set-cdr! AIR OBJECT)
 * Stores OBJECT in the cdr field of PAIR. 
 * The value returned by set-cdr! is unspecified.
 */

static OsObject OsSubrSetCdr (OsObject cons, OsObject obj) {
    OsMustBeCons (cons) ;
    OsConsCdr ((OsCons) cons) = obj ;
    return (obj) ;
}

/* AUTODOC
 * (pair? OBJECT)
 * Returns #t if obj is a pair, and otherwise returns #f.
 */

static OsObject OsSubrPairP (OsObject cons) {
    return ((OsObject) (OsIsCons (cons) ? OsA_True : OsA_False)) ;
}

/* AUTODOC
 * (cons OBJECT1 OBJECT2)
 * Returns a newly allocated pair whose car is OBJECT1 and cdr OBJECT2.
 *  The pair is guaranteed to be different (in the sense of eqv?) from
 *  every existing object.
 *
 *   (cons 'a '())               ==>  (a)
 *   (cons '(a) '(b c d))        ==>  ((a) b c d)
 *   (cons "a" '(b c))           ==>  ("a" b c)
 *   (cons 'a 3)                 ==>  (a . 3)
 *   (cons '(a b) 'c)            ==>  ((a b) . c)
 */

static OsObject OsSubrCons (OsObject car, OsObject cdr) {
    return ((OsObject) OsConsMake (car, cdr)) ;
}

/* AUTODOC
 * (null? OBJECT)
 * Returns #t if OBJECT is the empty list, otherwise returns #f.
 */

static OsObject OsSubrNullP (OsObject cons) {
    return ((cons == (OsObject) OsA_Nil) ? OsA_True : OsA_False) ;
}

/* AUTODOC
 * (list? OBJECT)
 * Returns #t if OBJECT is a list, otherwise returns #f. 
 */

static OsObject OsSubrListP (OsObject lst) {
    OsObject flst = lst ;
    do {
        if ( ! OsIsCons (flst) )
            return ((flst == OsA_Nil) ? OsA_True : OsA_False) ;
        flst = OsConsCdr (flst) ;
        if ( ! OsIsCons (flst) )
            return ((flst == OsA_Nil) ? OsA_True : OsA_False) ;
        flst = OsConsCdr (flst) ;
        lst  = OsConsCdr (lst) ;
    } while (flst != lst) ;
    return (OsA_False) ;
}

/* AUTODOC
 * (list OBJECT*)
 * Returns a newly allocated list of its arguments.
 *
 *   (list 'a (+ 3 4) 'c)        ==>  (a 7 c)
 *   (list)                      ==>  ()
 */

static OsObject OsSubrList (unsigned int argc, OsObject *argv) {
    return ((OsObject) OsListNMake (argc, argv)) ;
}

/* AUTODOC
 * (length LIST)
 * Returns the length of LIST.
 *
 *   (length '(a b c))           ==>  3
 *   (length '(a (b) (c d e)))   ==>  3
 *   (length '())                ==>  0
 */

static OsObject OsSubrListLength (OsObject cons) {
  retry:
    if ( OsIsCons (cons) ) {
        return ((OsObject) OsIntegerMake (OsConsLength ((OsCons) cons))) ;
    } else if ( cons == (OsObject) OsA_Nil ) {
        return ((OsObject) OsIntegerMake (0)) ;
    } else {
        cons = OsErrorTrigger (1, OsE_BadType, cons, OsConsClass) ;
        goto retry ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (append LIST*)
 * Append all LIST into a single list.
 * Returns a list consisting of the elements of the first list
 * followed by the elements of the other lists.
 *
 *   (append '(x) '(y))          ==>  (x y)
 *   (append '(a) '(b c d))      ==>  (a b c d)
 *   (append '(a (b)) '((c)))    ==>  (a (b) (c))
 *
 * The resulting list is always newly allocated, except that it shares
 * structure with the last list argument. The last argument may
 * actually be any object; an improper list results if the last
 * argument is not a proper list.
 *
 *   (append '(a b) '(c . d))    ==>  (a b c . d)
 *   (append '() 'a)             ==>  a
 */

static OsObject OsSubrAppend (unsigned int argc, OsObject *argv) {
    OsObject start = (OsObject) OsA_Nil, *list = &start, pair ;
    if ( argc < 1 ) {
        return (OsA_Nil) ;
    }
    while ( argc-- > 1 ) {
        if ( *argv == (OsObject) OsA_Nil ) {
            argv++ ;
            continue ;
        } else {
            OsMustBeCons (*argv) ;
            for (pair = *argv++ ; OsIsCons (pair) ; pair = OsConsCdr(pair)) {
                *list = (OsObject) OsConsMake (OsConsCar(pair), OsA_Nil) ;
                list  = &(OsConsCdr (*list)) ;
            }
        }
    }
    *list = *argv ;
    return (start) ;
}

/* AUTODOC
 * (list-tail LIST K)
 * Returns the sublist obtained by omitting the first k elements.
 * List-tail could be defined by:
 *
 *    (define list-tail
 *      (lambda (x k)
 *        (if (zero? k)
 *            x
 *            (list-tail (cdr x) (- k 1)))))
 */

static OsObject OsSubrListTail (OsObject list, OsInteger k) {
    int n ;
    OsMustBeInteger (k) ;
    for (n = k->ival ; n-- ; list = OsConsCdr (list))
        OsMustBeCons (list) ;
    return (list) ;
}

/* AUTODOC
 * (list-ref LIST K)
 * Returns the Kth element of LIST.
 * This is the same as the car of (list-tail list k).
 *
 *   (list-ref '(a b c d) 2)     ==>  c
 */

static OsObject OsSubrListRef (OsObject list, OsInteger k) {
    int n ;
    OsMustBeInteger (k) ;
    for (n = k->ival ; n-- ; list = OsConsCdr (list))
        OsMustBeCons (list) ;
    return (OsConsCar (list)) ;
}

/* AUTODOC
 * (reverse LIST)
 * Returns a new list made of elements of LIST in reverse order.
 *
 *    (reverse '(a b c))          ==>  (c b a)
 *    (reverse '(a (b c) d (e (f))))  
 *       ==>  ((e (f)) d (b c) a)
 */

static OsObject OsSubrReverse (OsCons list) {
    if ( list == (OsCons) OsA_Nil ) {
        return (OsA_Nil) ;
    } else {
        OsObject start = OsA_Nil ;
        OsMustBeCons (list) ;
        while ( list != (OsCons) OsA_Nil ) {
            start = (OsObject) OsConsMake (OsConsCar (list), start) ;
            list  = (OsCons) OsConsCdr (list) ;
        }
        return (start) ;
    }
    /* NOT REACHED */
}

static OsObject OsSubrMemq (OsObject obj, OsCons list) {
    if ( list == (OsCons) OsA_Nil )
        return (OsA_False) ;
    list = OsShouldBeCons (list) ;
    do {
        if ( OsConsCar (list) == obj )
            return ((OsObject) list) ;
    } while (list = (OsCons) OsConsCdr (list), OsIsCons (list) ) ;
    return (OsA_False) ;
}

static OsObject OsSubrMemv (OsObject obj, OsCons list) {
    if ( list == (OsCons) OsA_Nil )
        return (OsA_False) ;
    OsShouldBeCons (list) ;
    do {
        if ( OsSendEqv (obj, OsConsCar (list)) == OsA_True )
            return ((OsObject) list) ;
    } while (list = (OsCons) OsConsCdr (list), OsIsCons (list)) ;
    return (OsA_False) ;
}

static OsObject OsSubrMember (OsObject obj, OsCons list) {
    if ( list == (OsCons) OsA_Nil )
        return (OsA_False) ;
    list = OsShouldBeCons (list) ;
    do {
        if ( OsSendEqual (obj, OsConsCar(list)) == OsA_True )
            return ((OsObject) list) ;
    } while ( (list = (OsCons) OsConsCdr (list), OsIsCons (list)) ) ;
    return (OsA_False) ;
}

static OsObject OsSubrAssq (OsObject obj, OsCons list) {
    if ( list == (OsCons) OsA_Nil )
        return (OsA_False) ;
    list = OsShouldBeCons (list) ;
    do {
        OsMustBeCons (OsConsCar (list)) ;
        if ( obj == OsConsCaar (list) )
            return (OsConsCar(list)) ;
    } while ( list = (OsCons) OsConsCdr (list), OsIsCons (list) ) ;
    return (OsA_False) ;
}
    
static OsObject OsSubrAssv (OsObject obj, OsCons list) {
    if ( list == (OsCons) OsA_Nil )
        return (OsA_False) ;
    list = OsShouldBeCons (list) ;
    do {
        OsMustBeCons (OsConsCar (list)) ;
        if ( OsSendEqv (obj, OsConsCaar (list)) == OsA_True )
            return (OsConsCar(list)) ;
    } while ( list = (OsCons) OsConsCdr (list), OsIsCons (list) ) ;
    return (OsA_False) ;
}
    
static OsObject OsSubrAssoc (OsObject obj, OsCons list) {
    if ( list == (OsCons) OsA_Nil )
        return (OsA_False) ;
    list = OsShouldBeCons (list) ;
    do {
        OsMustBeCons (OsConsCar (list)) ;
        if ( OsSendEqual (obj, OsConsCaar (list)) == OsA_True )
            return (OsConsCar(list)) ;
    } while ( list = (OsCons) OsConsCdr (list), OsIsCons (list) ) ;
    return (OsA_False) ;
}

static OsObject OsSubrListToVector (OsObject list) {
    OsVector vector ;
    if ( list == OsA_Nil ) {
        vector = OsVectorMake (0) ;
    } else {
        OsMustBeCons (list) ;
        vector = OsVectorMake (0) ;
        do {
            OsVectorFastAppend (vector, OsConsCar(list)) ;
        } while ( (list = OsConsCdr (list), OsIsCons (list)) ) ;
    }
    return ((OsObject) vector) ;
}

static OsObject OsSubrListToString (OsObject list) {
    unsigned int  len = 0 ;
    unsigned char *p ;
    OsString str ;
    OsObject obj = list ;
    OsMustBeCons (list) ;
    do {
        OsMustBeChar (OsConsCar (obj)) ;
        len++ ;
    } while ( (obj = OsConsCdr(obj), OsIsCons (obj)) ) ;
    str         = (OsString) OsObjectMake (OsStringClass, 0) ;
    str->len    = len + 1 ;
    str->string = (unsigned char *) OsGCMallocAtomic (str->len+1) ;
    for (p = str->string ; len-- ; p++ ) {
        *p   = ((OsChar) OsConsCar(list))->c ;
        list = OsConsCdr (list) ;
    }
    *p++ = '\0' ;
    return ((OsObject) str) ;
}

/* AUTODOC
 * (map PROC LIST*)
 * Iterate PROC through LIST, and return thelist of results.
 * The LIST must be lists, and proc must be a procedure taking as many
 * arguments as there are lists. If more than one list is given, then
 * they must all be the same length. Map applies PROC element-wise to
 * the elements of the lists and returns a list of the results, in
 * order from left to right. The dynamic order in which PROC is
 * applied to the elements of the LIST is unspecified.
 *
 *   (map cadr '((a b) (d e) (g h)))   
 *             ==>  (b e h)
 *   (map (lambda (n) (expt n n)) '(1 2 3 4 5))                
 *             ==>  (1 4 27 256 3125)
 *   (map + '(1 2 3) '(4 5 6))   ==>  (5 7 9)
 */

static OsObject OsSubrListMap (unsigned int argc, OsObject *argv) {
    unsigned int len, i ;
    OsCons       pair ;
    OsObject     start = OsA_Nil, *list = &start, item ;
    OsCodeBlock  block ;
    
    if ( argc < 2 ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (2)
                               , OsIntegerMake (argc)) ;
    }
    OsMustBeBlock (*argv) ;
    block = (OsCodeBlock) *argv++ ; argc-- ;
    if ( argc == 1 ) {		/* can be fast (no alloc) */
        if ( *argv == OsA_Nil )
            return (OsA_Nil) ;
        pair = (OsCons) OsMustBeCons (*argv) ;
        len  = OsConsLength (pair) ;
        while ( len-- ) {
            item  = OsCodeBlockApplyWithArray (OsCurrentInterp()
                                               , block
                                               , 1, &OsConsCar (pair)
                                               , (OsObject *) 0) ;
            *list = (OsObject) OsConsMake (item, OsA_Nil) ;
            list  = &(OsConsCdr(*list)) ;
            pair  = (OsCons) OsConsCdr (pair) ;
        }
        return (start) ;
    } else {
        if ( *argv == OsA_Nil ) {		/* all lists are nil */
            for (i = 1 ; i < argc ; i++) {
                if ( argv[i] != OsA_Nil ) {
                    (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                           , OsIntegerMake (0)
                                           , OsIntegerMake (-1)) ;
                }
            }
            return (OsA_Nil) ;
        } else if ( argc < 8 ) {		/* 90% gets optimized ;-) */
            OsObject array[8], *pa ;
            OsMustBeCons (*argv) ;
            len = OsConsLength ((OsCons) *argv) ;
            for (i = 1 ; i < argc ; i++) {
                unsigned int got ;
                OsMustBeCons (argv[i]) ;
                if ( len != (got = OsConsLength ((OsCons) argv[i])) ) {
                    (void) OsErrorTrigger(1, OsE_ArityShouldBe
                                          , OsIntegerMake (len)
                                          , OsIntegerMake (got)) ;
                }
            }
            while ( len-- ) {
                for (pa = array, i = 0 ; i < argc ; i++ ) {
                    *pa++   = OsConsCar (argv[i]) ;
                    argv[i] = OsConsCdr (argv[i]) ;
                }
                item  = OsCodeBlockApplyWithArray (OsCurrentInterp()
                                                   , block
                                                   , pa - array, array
                                                   , (OsObject *) 0) ;
                *list = (OsObject) OsConsMake (item, OsA_Nil) ;
                list  = &(OsConsCdr (*list)) ;
            }
            return (start) ;
        } else {				 /* nasty case ! */
            OsObject *array, *pa ;
            OsMustBeCons (*argv) ;
            len = OsConsLength ((OsCons) *argv) ;
            for (i = 1 ; i < argc ; i++) {
                unsigned int got ;
                OsMustBeCons (argv[i]) ;
                if ( len != (got = OsConsLength ((OsCons) argv[i])) ) {
                    (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                           , OsIntegerMake (len)
                                           , OsIntegerMake (got)) ;
                }
            }
            array = (OsObject *) OsGCMalloc (len * sizeof (OsObject)) ;
            while ( len-- ) {
                for (pa = array, i = 0 ; i < argc ; i++ ) {
                    *pa++   = OsConsCar(argv[i]) ;
                    argv[i] = OsConsCdr (argv[i]) ;
                }
                item  = OsCodeBlockApplyWithArray (OsCurrentInterp()
                                                   , block
                                                   , pa - array, array
                                                   , (OsObject *) 0) ;
                *list = (OsObject) OsConsMake (item, OsA_Nil) ;
                list  = &(OsConsCdr (*list)) ;
            }
        }
        return (start) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (for-each PROC LIST*)
 * Iterate PROC through LIST, for side effect.
 * The arguments to for-each are like the arguments to map, but
 * for-each calls PROC for its side effects rather than for its
 * values. Unlike map, for-each is guaranteed to call PROC on the
 * elements of the LIST in order from the first element to the last,
 * and the value returned by for-each is unspecified.
 *
 *   (let ((v (make-vector 5)))
 *     (for-each (lambda (i)
 *                 (vector-set! v i (* i i)))
 *               '(0 1 2 3 4))
 *     v)                        ==>  #(0 1 4 9 16)
 */

static OsObject OsSubrListForEach (unsigned int argc, OsObject *argv) {
    unsigned int len, i ;
    OsCons       pair ;
    OsCodeBlock  block ;
    
    if ( argc < 2 ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (2)
                               , OsIntegerMake (argc)) ;
    }
    OsMustBeBlock (*argv) ;
    block = (OsCodeBlock) *argv++ ; argc-- ;
    if ( argc == 1 ) {		/* can be fast (no alloc) */
        if ( *argv == OsA_Nil )
            return (OsA_Nil) ;
        pair = (OsCons) OsMustBeCons (*argv) ;
        len  = OsConsLength (pair) ;
        while ( len-- ) {
            (void) OsCodeBlockApplyWithArray (OsCurrentInterp()
                                              , block
                                              , 1, &OsConsCar (pair)
                                              , (OsObject *) 0) ;
            pair = (OsCons) OsConsCdr (pair) ;
        }
        return (OsA_Nil) ;
    } else {
        if ( *argv == OsA_Nil ) {		/* all lists are nil */
            for (i = 1 ; i < argc ; i++) {
                if ( argv[i] != OsA_Nil ) {
                    (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                           , OsIntegerMake (0)
                                           , OsIntegerMake (-1)) ;
                }
            }
            return (OsA_Nil) ;
        } else if ( argc < 8 ) {		/* 90% gets optimized ;-) */
            OsObject array[8], *pa ;
            OsMustBeCons (*argv) ;
            len = OsConsLength ((OsCons) *argv) ;
            for (i = 1 ; i < argc ; i++) {
                unsigned int got ;
                OsMustBeCons (argv[i]) ;
                if ( len != (got = OsConsLength ((OsCons) argv[i])) ) {
                    (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                           , OsIntegerMake (len)
                                           , OsIntegerMake (got)) ;
                }
            }
            while ( len-- ) {
                for (pa = array, i = 0 ; i < argc ; i++ ) {
                    *pa++   = OsConsCar (argv[i]) ;
                    argv[i] = OsConsCdr (argv[i]) ;
                }
                (void) OsCodeBlockApplyWithArray (OsCurrentInterp()
                                                  , block
                                                  , pa - array, array
                                                  , (OsObject *) 0) ;
            }
            return (OsA_Nil) ;
        } else {				 /* nasty case ! */
            OsObject     *array, *pa ;
            unsigned int got ;
            OsMustBeCons (*argv) ;
            len = OsConsLength ((OsCons) *argv) ;
            for (i = 1 ; i < argc ; i++) {
                OsMustBeCons (argv[i]) ;
                if ( len != (got = OsConsLength ((OsCons) argv[i])) ) {
                    (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                           , OsIntegerMake (len)
                                           , OsIntegerMake (got)) ;
                }
            }
            array = (OsObject *) OsGCMalloc (len * sizeof (OsObject)) ;
            while ( len-- ) {
                for (pa = array, i = 0 ; i < argc ; i++ ) {
                    *pa++   = OsConsCar(argv[i]) ;
                    argv[i] = OsConsCdr (argv[i]) ;
                }
                (void) OsCodeBlockApplyWithArray (OsCurrentInterp()
                                                  , block
                                                  , pa - array, array
                                                  , (OsObject *) 0) ;
            }
        }
        return (OsA_Nil) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (apply PROC ARG* ARGS)
 * Apply PROC to the list concatenation ARGs to the list ARGS
 * PROC must be a procedure and ARGS must be a list. This form applies PROC to
 * the elements of the list resulting of (append (list ARG*) ARGS).
 *   (apply + (list 3 4))        ==>  7
 */

static OsObject OsSubrApply (unsigned int argc, OsObject *argv) {
    OsCodeBlock block ;
    OsObject start, *list = &start, pair ;

    if ( argc < 2 ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (2)
                               , OsIntegerMake (argc)) ;
    }
    OsMustBeBlock (*argv) ;
    block = (OsCodeBlock) *argv++ ; argc-- ;
    start = *argv ;
    while ( argc-- > 1 ) {
        *list = (OsObject) OsConsMake (*argv++, OsA_Nil) ;
        list  = &(OsConsCdr (*list)) ;
    }
    *list = *argv ;
    return (OsCodeBlockApplyWithList (OsCurrentInterp()
                                      , block
                                      , (OsCons) start
                                      , (OsObject *) 0)) ;
}

/* AUTODOC
 * (cons* OBJECTS*)
 * An N-ARY version of 'cons.
 */

static OsObject OsSubrConsStar (unsigned int argc, OsObject *argv) {
    OsObject start, *list = &start ;
    if ( argc < 1 ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (1)
                               , OsIntegerMake (0)) ;
    }
    while ( argc-- > 1 ) {
        *list = (OsObject) OsConsMake (*argv++, OsA_Nil) ;
        list  = &(OsConsCdr (*list)) ;
    }
    *list = *argv++ ;
    return (start) ;
}
                      
/*
 * Messages impl for conses
 * _display is used both for 'display and 'write, hence the SendMsg instead of
 * SendDisplay...
 */

static OsObject os_cons_display (OsCons icons, int sel, OsOutputPort port) {
    OsCons cons = icons ;
    char   del  = '(' ;
    
    do {
        OsOutputPortWriteChar (port, del) ;
        OsSendMsg (OsConsCar(cons), sel, (OsConsCar(cons), sel, port)) ;
        cons = (OsCons) OsConsCdr (cons) ;
        del  = ' ' ;
    } while ( OsIsCons (cons) ) ;
    if ( cons != ((OsCons) OsA_Nil) ) {
        OsOutputPortWriteCString (port, " . ") ;
        OsSendMsg (cons, sel, (cons, sel, port)) ;
    }
    OsOutputPortWriteChar (port, ')') ;
    return ((OsObject) icons) ;
}

static OsObject os_cons_equal (OsCons self, int sel, OsCons other) {
    if ( OsIsCons (other) ) {
        do {
            if (OsSendEqual(OsConsCar(self),OsConsCar(other)) == OsA_False)
                return (OsA_False) ;
            self  = (OsCons) OsConsCdr (self) ;
            other = (OsCons) OsConsCdr (other) ;
        } while ( OsIsCons (self) && OsIsCons (other) ) ;
        return (OsSendEqual (self, other)) ;
    }
    return (OsA_False) ;
}

/*
 * Initialize :
 */

void OsConsInitialize (OsInterp interp, int *pargc, char **argv) {
    OsConsClass = OsClassDeclare ("Cons", sizeof (OsConsRec)) ;
    OsDefineMsg (OsConsClass, OS_SEL_DISPLAY, os_cons_display) ;
    OsDefineMsg (OsConsClass, OS_SEL_WRITE, os_cons_display) ;
    OsDefineMsg (OsConsClass, OS_SEL_EQUAL, os_cons_equal) ;
    return ;
}

void OsConsSubrInitialize (OsInterp interp) {
    OsSubrMake (interp, "pair?", OsSubrPairP, 1) ;
    OsSubrMake (interp, "cons", OsSubrCons, 2) ;
    OsSubrMake (interp, "car", OsSubrCar, 1) ;
    OsSubrMake (interp, "cdr", OsSubrCdr, 1) ;
    OsSubrMake (interp, "caar", OsSubrCaar, 1) ;
    OsSubrMake (interp, "cdar", OsSubrCdar, 1) ;
    OsSubrMake (interp, "cddr", OsSubrCddr, 1) ;
    OsSubrMake (interp, "cadr", OsSubrCadr, 1) ;
    OsSubrMake (interp, "set-car!", OsSubrSetCar, 2) ;
    OsSubrMake (interp, "set-cdr!", OsSubrSetCdr, 2) ;
    OsSubrMake (interp, "null?", OsSubrNullP, 1) ;
    OsSubrMake (interp, "list?", OsSubrListP, 1) ;
    OsSubrMake (interp, "list", OsSubrList, OS_NARY) ;
    OsSubrMake (interp, "length", OsSubrListLength, 1) ;
    OsSubrMake (interp, "append", OsSubrAppend, OS_NARY) ;
    OsSubrMake (interp, "list-tail", OsSubrListTail, 2) ;
    OsSubrMake (interp, "list-ref", OsSubrListRef, 2) ;
    OsSubrMake (interp, "reverse", OsSubrReverse, 1) ;
    OsSubrMake (interp, "memq", OsSubrMemq, 2) ;
    OsSubrMake (interp, "memv", OsSubrMemv, 2) ;
    OsSubrMake (interp, "member", OsSubrMember, 2) ;
    OsSubrMake (interp, "assq", OsSubrAssq, 2) ;
    OsSubrMake (interp, "assv", OsSubrAssv, 2) ;
    OsSubrMake (interp, "assoc", OsSubrAssoc, 2) ;
    OsSubrMake (interp, "list->vector", OsSubrListToVector, 1) ;
    OsSubrMake (interp, "list->string", OsSubrListToString, 1) ;
    OsSubrMake (interp, "map", OsSubrListMap, OS_NARY) ;
    OsSubrMake (interp, "for-each", OsSubrListForEach, OS_NARY) ;
    OsSubrMake (interp, "apply", OsSubrApply, OS_NARY) ;
    OsSubrMake (interp, "cons*", OsSubrConsStar, OS_NARY) ;
    return ;
}

