/***************************************************************************
 * cont.c
 * cont.c,v 1.2 1995/03/12 02:57:20 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * Control or Continuation. I don't know yet !
 * AUTODOC MODULE
 * Control flow operators implementation.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <os_config.h>
#include <os_interp.h>
#include <os_obj.h>
#include <os_string.h>
#include <os_atom.h>
#include <os_gc.h>
#include <os_vector.h>
#include <os_error.h>
#include <os_cons.h>
#include <os_cblock.h>
#include <os_subr.h>
#include <os_cont.h>

static unsigned long stackstart ;

/*
 * Interanlly used functions;
 * gcc complaints about "returning the adress of a local variable"
 * Because gcc, when compiled in -O6 inline this procedure, we complexify it
 * arbitrarily in order to avoid this. 
 */

static char *top_of_stack (int always_false) {
    long          *l ;
    unsigned long v ;

    v = ((unsigned long) &l) ;
    if ( ! always_false ) {
        /* This is the only executed branch at run time */
        return ((char *) v) ;
    } else { 		/* NOT REACHED */
        /* Let's fool gcc: turning top_of_stack into a rec. function should
         * protect us efficiently against inlin'ing (at least, for a good #
         * of gcc releases :-) */
        fprintf (stderr, "top_of_stack: should always be called with 0 !\n");
        return (top_of_stack (always_false - 1)) ;
    }
    /* NOT REACHED */
}

static void regwin_blower (unsigned int count
                           , void (*cfun)(), void *a1, void *a2) {
    if ( ! count ) {
        (cfun) (a1, a2) ;
        return ;
    } 
    regwin_blower (count - 1, cfun, a1, a2) ;
    return ;
}

/*
 * C API
 */

/* Because we use the local variables *after* copying the stack, we need
 * be sure that the copy doesn't effect them. COMFORT should be at least
 * the size of the os_cstack_install frame call.
 */

#define OS_COMFORT	(8*sizeof (OsObject))

static OsContBlock the_block ;

void cstack_install (OsContBlock cblock, OsObject retvalue) {
    char *instack ;
    unsigned long sp = ((unsigned long) &instack) ;

    the_block = cblock ;
#if ( OS_STACK_GROWS == OS_STACK_DOWN )
    if ( sp + OS_COMFORT >= ((unsigned long) the_block->cstackp) ) 
#else
    if ( ((unsigned long) the_block->cstackp) > sp - OS_COMFORT)
#endif
    {
        VOLATILE char stack_space[64] ;
        bzero (stack_space, sizeof (stack_space)) ;
        os_cstack_install (the_block, retvalue) ;
    }
    the_block->acc = retvalue ;
#ifdef OS_FLUSH_REGWIN_CODE
    OS_FLUSH_REGWIN_CODE ;
#endif /* OS_FLUSH_REGWIN_CODE */
    bcopy (the_block->cstack, the_block->cstackp, the_block->cstack_size) ;
    longjmp (the_block->jmpbuf, 2) ;
    /* NOT REACHED */
}

void os_cstack_install (OsContBlock cblock, OsObject retvalue) {
#ifdef OS_REGWIN_SIZE
    regwin_blower (OS_REGWIN_SIZE, cstack_install, cblock, retvalue) ;
#else
    cstack_install (cblock, retvalue) ;
#endif /* OS_REGWIN_SIZE */
    /* NOT REACHED */
}

int os_cstack_freeze (OsContBlock cblock) {
    VOLATILE int always_false = 0 ;
    int  retvalue ;
    char *stack_top ;

    stack_top = top_of_stack(always_false) ;
    if ( ! (retvalue = setjmp (cblock->jmpbuf)) ) {
#if ( OS_STACK_GROWS == OS_STACK_DOWN )
#ifdef OS_FLUSH_REGWIN_CODE
        OS_FLUSH_REGWIN_CODE ;
#endif /* OS_FLUSH_REGWIN_CODE */
        cblock->cstack_size = stackstart - ((unsigned long) stack_top) ;
        cblock->cstackp     = (char *) stack_top ;
        cblock->cstack      = (char *) OsGCMalloc (cblock->cstack_size) ;
        bcopy (stack_top, cblock->cstack, cblock->cstack_size) ;
#else
#ifdef OS_FLUSH_REGWIN_CODE
        OS_FLUSH_REGWIN_CODE ;
#endif /* OS_FLUSH_REGWIN_CODE */
        cblock->cstack_size = ((unsigned long) stack_top) - stackstart ;
        cblock->cstackp     = (char *) stackstart ;
        cblock->cstack      = (char *) OsGCMalloc (cblock->cstack_size) ;
        bcopy ((char *) stackstart, cblock->cstack, cblock->cstack_size) ;
#endif /* OS_STACK_GROWS */
    }
    return (retvalue) ;
}

/*
 * Subrs:
 */

/* AUTODOC
 * (call/cc THUNK)
 * Capture current continuation, and provide it as arg to THUNK.
 * This function reify the current continuation, and provides it as a
 * parameter to THUNK.
 */

static OsObject OsContSubrCallCC (OsContBlock cont
                                  , unsigned int argc
                                  , OsObject *argv) {
    OsCodeBlock block ;
    if ( argc != 1 ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (1)
                               , OsIntegerMake (argc)) ;
    }
    block = (OsCodeBlock) argv[0] ;
    OsMustBeBlock (block) ;
    return (OsCodeBlockApplyWithArray (OsCurrentInterp()
                                       , block
                                       , 1, (OsObject *) &cont
                                       , (OsObject *) 0)) ;
}

/* AUTODOC
 * (dynamic-wind BEFORE DURING AFTER)
 * Invoke BEFORE, DURING and AFTER, whatever exceptions occurs.
 * BEFORE is always executed before DURING, and when DURING is re-entered
 * through a continuation. AFTER is always executed at the end of DURING.
 * BEFORE and AFTER have no arguments. DURING has a local escape procedure
 * as argument.
 */

static OsObject OsSubrDynamicWind (before, during, after)
    OsCodeBlock before ;
    OsCodeBlock during ;
    OsCodeBlock after ;
{
    OsObject value = (OsObject) 0 ;
    OsObject error = (OsObject) 0 ;
    OsMustBeClosedBlock (before) ;
    OsMustBeClosedBlock (during) ;
    OsMustBeClosedBlock (after) ;
    (void) OsCodeBlockApplyWithArray (OsCurrentInterp()
                                      , before
                                      , 0, (OsObject *) 0
                                      , (OsObject *) 0) ;
    value = OsWindProtectExecute (OsCurrentInterp(), before, during, &error) ;
    (void) OsCodeBlockApplyWithArray (OsCurrentInterp()
                                      , after
                                      , 0 , (OsObject *) 0
                                      , (OsObject *) 0) ;
    if ( error ) {
        os_uncatched_error (error) ;
        /* NOT REACHED */
    } else {
        return (value) ;
    }
}

/*
 * Initialize :
 */

void OsContInitialize (OsInterp interp, int *pargc, char **argv) {
    stackstart = ((unsigned long) pargc) ;
    return ;
}

void OsContSubrInitialize (OsInterp interp) {
    OsSecContSubrMake (interp, "call/cc", OsContSubrCallCC
                       , 1, OS_SECURITY_LEVEL_E) ;
    OsSecSubrMake (interp, "dynamic-wind", OsSubrDynamicWind
                   , 3, OS_SECURITY_LEVEL_E) ;
    return ;
}
