/***************************************************************************
 * os_debug.c
 * debug.c,v 1.1.1.1 1995/02/27 15:15:18 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * Various functions callable from debugger
 * AUTODOC MODULE
 * Internal debugging functions.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <oscheme.h>

/*
 * Displays the given object to (current-output-port). Specially there to be
 * called from your debugger.
 */

void OsDP (OsObject object) {
    OsSendDisplay (object, OsCurrentOutputPort) ;
    OsOutputPortWriteChar (OsCurrentOutputPort, '\n') ;
    OsOutputPortFlush (OsCurrentOutputPort) ;
    return ;
}

#ifdef OS_VM_TRACEABLE

void OsDTraceVM (unsigned int onoff) {
    static OsAtom osa_trace = (OsAtom) 0 ;
    if ( ! osa_trace )
        osa_trace = OsAtomMake (OsCurrentInterp()
                                , "sys:trace-machine", OsA_Unbound) ;
    osa_trace->vcell = (onoff ? OsA_True : OsA_False) ;
    printf ("*** VM %s\n", onoff ? "TRACED" : "UNTRACED") ;
    fflush (stdout) ;
    return ;
}

#endif /* OS_VM_TRACEABLE */

#ifdef OS_COMPILER_TRACEABLE

void OsDTraceCompiler (unsigned int onoff) {
    static OsAtom osa_trace = (OsAtom) 0 ;
    if ( ! osa_trace )
        osa_trace = OsAtomMake (OsCurrentInterp()
                                , "sys:trace-compiler", OsA_Unbound) ;
    osa_trace->vcell = (onoff ? OsA_True : OsA_False) ;
    printf ("*** COMPILER %s\n", onoff ? "TRACED" : "UNTRACED") ;
    fflush (stdout) ;
    return ;
}

#endif /* OS_COMPILER_TRACEABLE */

/* Set breakpoint here when debugging under C */

void os_db_break (char *msg) {
    printf (msg) ;
    return ;
}

/*
 * Subrs:
 */

/* AUTODOC
 * (db:break MESSAGE)
 * Enforce a call to os_db_break, with the MESSAGE string as argument.
 * This function is to be used when debugging oscheme (and sometimes oscheme
 * code). To use it, compile oscheme in debugging mode, run it under your
 * debugger. Set a break point in the C function os_db_break. When db:break
 * is executed, it will stop the oscheme process at the place where the call
 * was evaluated.
 * If you have to enter the C world, you might also want to have a look at
 * the od_debug.c module, which provides other facilities to be used under the
 * debugger.
 */

static OsObject OsSubrDbBreak (msg)
    OsString msg ;
{
    os_db_break (msg->string) ;
    return ((OsObject) msg) ;
}

void OsDebugSubrInitialize (OsInterp interp) {
    OsSubrMake (interp, "db:break", OsSubrDbBreak, 1) ;
    return ;
}

        
