/***************************************************************************
 * eobj.c
 * eobj.c,v 1.2 1995/03/17 14:21:18 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * External objects. This is taken from a package I have written for klone
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <oscheme.h>
#include <os_eobjP.h>

#ifdef OS_EOBJ_STATS
unsigned int stat_nsarrays = 0 ;
unsigned int stat_nduckets = 0 ;
unsigned int stat_ncuckets = 0 ;
unsigned int stat_nbuckets = 0 ;
unsigned int stat_nclasses = 0 ;
#endif /* OS_EOBJ_STATS */

OsClass OsEObjClass       = (OsClass) 0 ;
OsClass OsEObjSelClass    = (OsClass) 0 ;
OsClass OsEObjSelImpClass = (OsClass) 0 ;

OsObject OsEObjRoot      = (OsObject) 0 ;

OsEObjSel OsEObjSelClone = (OsEObjSelId) 0 ;

static OsEObjSel os_sel_define_slot_init   = (OsEObjSel) 0 ;
static OsEObjSel os_sel_responds_to        = (OsEObjSel) 0 ;
static OsEObjSel os_sel_does_not_recognize = (OsEObjSel) 0 ;
static OsEObjSel os_sel_forward            = (OsEObjSel) 0 ;
static OsEObjSel os_sel_error              = (OsEObjSel) 0 ;
static OsEObjSel os_sel_super              = (OsEObjSel) 0 ;
static OsEObjSel os_sel_display            = (OsEObjSel) 0 ;
static OsEObjSel os_sel_display_to         = (OsEObjSel) 0 ;

static OsError ose_reader_redef            = (OsError) 0 ;
static OsError ose_writer_redef            = (OsError) 0 ;
static OsError ose_bad_selector            = (OsError) 0 ;
static OsError ose_bad_imp_for             = (OsError) 0 ;

static OsHashtable os_selimp_hashtable = (OsHashtable) 0 ;

/***************************************************************************
 * Selector implementations:
 * - need to be shared (to optimize bucket sharing)
 * - need to be typed (slot, scheme closure, or c function)
 ***************************************************************************/

/*
 * Default selector implementation, triggers the do_forward c function
 */

static OsObject do_forward () ;
    
static os_selimp_cfun_rec os_defimp_rec = {
    OS_SELIMP_CFUN,
    do_forward,
    (void *) 0
} ;

#define os_defimp	((os_selimp) &os_defimp_rec)

static unsigned long os_selimp_hash (selimp)
    os_selimp selimp ;
{
    unsigned long h = 0 ;
    switch (selimp->type) {
      case OS_SELIMP_CFUN:
        {
            os_selimp_cfun cfun = (os_selimp_cfun) selimp ;
            h = ((unsigned long) cfun->cfun) ^ ((unsigned long) cfun->closure);
        }
        break ;
      case OS_SELIMP_CLOS:
        {
            os_selimp_clos clos = (os_selimp_clos) selimp ;
            h = ((unsigned long) clos->block) ;
        }
        break ;
      case OS_SELIMP_SLOT:
        {
            os_selimp_slot slot = (os_selimp_slot) selimp ;
            h = ((unsigned long) slot->didx) ;
        }
        break ;
      default:
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("os_selimp_hash")
                               , OsStringMake ("invalid selimp type.")) ;
        /* NOT REACHED */
    }
    return ((h << 2) | ((unsigned long) selimp->type)) ;
}

static int os_selimp_cmp (selimp1, selimp2)
    os_selimp selimp1, selimp2 ;
{
    if ( selimp1->type == selimp2->type ) {
        switch (selimp1->type) {
          case OS_SELIMP_CFUN:
            return ((SELIMP_CFUN(selimp1) != SELIMP_CFUN(selimp2))
                    || (SELIMP_CLOSURE(selimp1) != SELIMP_CLOSURE(selimp2))) ;
          case OS_SELIMP_CLOS:
            return (SELIMP_BLOCK(selimp1) != SELIMP_BLOCK(selimp2)) ;
          case OS_SELIMP_SLOT:
            return (SELIMP_DIDX(selimp1) != SELIMP_DIDX(selimp2)) ;
          default:
            (void) OsErrorTrigger (1, OsE_Implementation
                                   , OsStringMake ("os_selimp_cmp")
                                   , OsStringMake ("invalid selimp type")) ;
        }
        /* NOT REACHED */
    }
    return (1) ;
}

static os_selimp os_selimp_make (type, a1, a2)
    unsigned char type ;
    void          *a1, *a2 ;
{
    switch (type) {
      case OS_SELIMP_CFUN:
        {
            os_selimp_cfun_rec cfun ;
            os_selimp_cfun     imp ;
            cfun.type    = OS_SELIMP_CFUN ;
            cfun.cfun    = (OsObject (*)()) a1 ;
            cfun.closure = a2 ;
            imp = (os_selimp_cfun) OsHashtableGet (os_selimp_hashtable, &cfun);
            if ( imp )
                return ((os_selimp) imp) ;
            imp = (os_selimp_cfun) OsGCMalloc (sizeof (os_selimp_cfun_rec)) ;
            imp->type    = OS_SELIMP_CFUN ;
            imp->cfun    = (OsObject (*)()) a1 ;
            imp->closure = a2 ;
            OsHashtablePut (os_selimp_hashtable, (void *) imp, (void *) imp) ;
            return ((os_selimp) imp) ;
        }
        break ;
      case OS_SELIMP_CLOS:
        {
            os_selimp_clos_rec clos ;
            os_selimp_clos     imp ;
            clos.type  = OS_SELIMP_CLOS ;
            clos.block = (OsObject) a1 ;
            imp = (os_selimp_clos) OsHashtableGet (os_selimp_hashtable, &clos);
            if ( imp )
                return ((os_selimp) imp) ;
            imp = (os_selimp_clos) OsGCMalloc (sizeof (os_selimp_clos_rec)) ;
            imp->type  = OS_SELIMP_CLOS ;
            imp->block = (OsObject) a1 ;
            OsHashtablePut (os_selimp_hashtable, (void *) imp, (void *) imp) ;
            return ((os_selimp) imp) ;
        }
        break ;
      case OS_SELIMP_SLOT:
        {
            os_selimp_slot_rec slot ;
            os_selimp_slot     imp ;
            slot.type = OS_SELIMP_SLOT ;
            slot.didx = (unsigned int) a1 ;
            imp = (os_selimp_slot) OsHashtableGet (os_selimp_hashtable, &slot);
            if ( imp )
                return ((os_selimp) imp) ;
            imp = (os_selimp_slot) OsGCMalloc (sizeof (os_selimp_slot_rec)) ;
            imp->type  = OS_SELIMP_SLOT ;
            imp->didx  = (unsigned int) a1 ;
            OsHashtablePut (os_selimp_hashtable, (void *) imp, (void *) imp) ;
            return ((os_selimp) imp) ;
        }
        break ;
      default:
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("os_selimp_make")
                               , OsStringMake ("invliad selimp type")) ;
    }
    /* NOT REACHED */
}

/***************************************************************************
 * Selector
 * Selectors have a uniq (unsigned short) id, a name, and an arity, they
 * are to be retreieved either from name (hashtable), or from id (dyn alloc
 * array).
 ***************************************************************************/

/*
 * We maintain two arrays:
 * - the global array of selectors (selid->sel)
 * - a hash table to go from name to selectors (name->sel), through the
 *   os_sel_bucket type
 */

static OsVector     selectors = (OsVector) 0 ;
unsigned int        free_sel  = 0 ;
static OsHashtable  hselector = (OsHashtable) 0 ;


static unsigned long os_hash_sel (OsEObjSel selector) {
    OsAtom        *argv = selector->selv ;
    unsigned int  argc  = selector->selc ;
    unsigned long ret   = 0 ;
    while ( argc-- ) 
	ret += (((unsigned long) *argv++) >> 2) ;
    return (ret) ;
}

static int os_cmp_sel (OsEObjSel sel1, OsEObjSel sel2) {
    if ( sel1->selc == sel2->selc ) {
        unsigned int selc ;
        for (selc = 0 ; selc < sel1->selc ; selc++ ) {
            if ( sel1->selv[selc] != sel2->selv[selc] )
                return (1) ;
        }
        return (0) ;
    }
    return (1) ;
}

static OsEObjSel os_sel_declare (unsigned int selc
                                      , OsAtom *selv
                                      , unsigned int arity
                                      , int definep) {
    OsEObjSelRec key, *rsel ;
    
/*
 * Create the relevant selector, if needed:
 */
    
    key.selc = selc ;
    key.selv = selv ;
    if ( ! (rsel = (OsEObjSel) OsHashtableGet (hselector, &key)) ) {
        rsel = (OsEObjSel) OsObjectMake (OsEObjSelClass, 0) ;
        rsel->selc  = selc ;
        rsel->selv  = selv ;
        rsel->arity = arity ;
        rsel->selid = (OsEObjSelId) 0 ;
        OsHashtablePut (hselector, rsel, rsel) ;
    }
/*
 * If it is a definition, define it:
 */
    if ( ( ! rsel->selid ) && definep ) {
        if ( free_sel ) {
            rsel->selid = free_sel ;
            free_sel = (unsigned int) selectors->cells[free_sel] ;
            selectors->cells[rsel->selid] = (OsObject) rsel ;
        } else {
            rsel->selid = (OsEObjSelId) OsVectorFastAppend (selectors
                                                            , (OsObject) rsel);
        }
    } 
    return (rsel) ;
}

static INLINE OsEObjSel os_sel_get (selid)
    OsEObjSelId selid ;
{
    OsEObjSel ret = (OsEObjSel) selectors->cells[selid] ;
    if ( ! ret ) {
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("os_sel_get")
                               , OsStringMake ("selector not found.")) ;
    }
    return (ret) ;
}

static OsEObjSel os_sel_lookup (name, createp)
    unsigned char *name ;
    unsigned int  createp ;
{
    unsigned int  arity = 0 ;
    OsAtom        selv[16], *psel = selv, *msel ;
    OsString      pname ;
    unsigned char *start = name, *end = name ;
    
    do {
	if ( *end == ':' ) {
	    arity++ ;
	    if ( *++end ) {
		pname   = OsStringMakeFromBytes (end-start, start) ;
		*psel++ = OsAtomMakeWithString (OsCurrentInterp()
                                                , pname, OsA_Unbound) ;
		start   = end ;
	    } else {
		*psel++ = OsAtomMake (OsCurrentInterp(), start, OsA_Unbound) ;
		break ;
	    }
	} else if ( ! *end ) {
	    *psel++ = OsAtomMake (OsCurrentInterp(), start, OsA_Unbound) ;
	    break ;
	}
	end++ ;
    } while ( 1 ) ;

    msel = (OsAtom *) OsGCMalloc (sizeof (OsObject) * (psel-selv)) ;
    bcopy (selv, msel, sizeof (OsObject) * (psel-selv)) ;
    return (os_sel_declare (psel-selv, msel, arity, createp)) ;
}

OsEObjSel OsEObjSelMake (unsigned int selc
                         , OsAtom *selv
                         , unsigned int arity) {
    return (os_sel_declare (selc, selv, arity, 0)) ;
}

OsEObjSel OsEObjSelDefine (name)
    char *name ;
{
    return (os_sel_lookup (name, 1)) ;
}

OsEObjSel OsEObjSelLookup (name)
    char *name ;
{
    return (os_sel_lookup (name, 0)) ;
}

OsEObjSel os_w_to_r_sel (wsel)
    OsEObjSel wsel ;
{
    OsString wname ;
    char     *rname ;

    if ( wsel->arity != 1 ) {
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("os_w_to_r_sel")
                               , OsStringMake ("invalid write sel arity.")) ;
    }
    wname = wsel->selv[0]->pname ;
    rname = (char *) OsGCMallocAtomic (wname->len + 1) ;
    bcopy (wname->string, rname, wname->len) ;
    rname[wname->len-1] = '\0' ;
    return (OsEObjSelDefine (rname)) ;
}

static OsEObjSel r_to_w_sel (rsel)
    OsEObjSel rsel ;
{
    OsString rname ;
    char     *wname ;

    if ( rsel->arity != 0 ) {
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("r_to_w_sel")
                               , OsStringMake ("invalid read sel arity")) ;
    }
    rname = rsel->selv[0]->pname ;
    wname = (char *) OsGCMallocAtomic (rname->len + 3) ;
    bcopy (rname->string, wname, rname->len) ;
    wname[rname->len]   = ':' ;
    wname[rname->len+1] = '\0' ;
    return (OsEObjSelDefine (wname)) ;
}

static OsEObjSelImp os_selimp_export (object, sel)
    OsEObj object ;
    OsEObjSel sel ;
{
    os_selimp    selimp = EObjSelimpLookup (object, sel->selid) ;
    OsEObjSelImp imp    = (OsEObjSelImp) OsObjectMake (OsEObjSelImpClass,0);
    imp->from   = object ;
    imp->sel    = sel ;
    imp->os_imp = selimp ;
    return (imp) ;
}


/***************************************************************************
 * Sparse arrays:
 * --------------
 * Used for selectors (we use here a four level sparse array, using the
 * assumption that there can not be more than 0xFFFF selectors.
 * Yea ! happy I do not have 6 levels ;-)
 * Four levels are:
 *      ducket       cucket        bucket        selector
 * <-- 4 bits --><-- 4 bits --><-- 4 bits --><-- 4 bits -->
 ***************************************************************************/

static os_sarray os_sarray_make (version, cdef_fun, ldef_fun)
    unsigned long version ;
    OsObject (*cdef_fun)() ;
    OsObject ldef_fun ;
{
    os_sarrayS es = (os_sarrayS) OsGCMalloc (sizeof (os_sarrayS_rec)) ;
    os_sarrayB eb = (os_sarrayB) OsGCMalloc (sizeof (os_sarrayB_rec)) ;
    os_sarrayC ec = (os_sarrayC) OsGCMalloc (sizeof (os_sarrayC_rec)) ;
    os_sarray  sa = (os_sarray)  OsGCMalloc (sizeof (os_sarray_rec)) ;
    int i ;

#ifdef OS_EOBJ_STATS
    stat_nsarrays++ ;
    stat_nbuckets++ ;
    stat_ncuckets++ ;
    stat_nduckets++ ;
#endif /* OS_EOBJ_STATS */
    
    es->version  = version ;
    eb->version  = version ;
    ec->version  = version ;
    sa->version  = version ;

    sa->empty_selector = es ; 
    sa->empty_bucket   = eb ; 
    sa->empty_cucket   = ec ; 

    for ( i = 0 ; i < SA_SELECTOR_SIZE ; i++ ) {
	es->selectors[i] = os_defimp ;
    }
    for ( i = 0 ; i < SA_BUCKET_SIZE ; i++ ) 
	eb->buckets[i] = es ;
    for ( i = 0 ; i < SA_CUCKET_SIZE ; i++ ) 
	ec->cuckets[i] = eb ;
    for ( i = 0 ; i < SA_DUCKET_SIZE ; i++ ) 
	sa->duckets[i] = ec ;
    return (sa) ;
}

static os_sarray os_sarray_lazy_copy (src, version)
    os_sarray     src ;
    unsigned long version ;
{
    os_sarray dst = (os_sarray) OsGCMalloc (sizeof (os_sarray_rec)) ;

#ifdef OS_EOBJ_STATS
    stat_nsarrays++ ;
#endif /* OS_EOBJ_STATS */
    
    bcopy (src, dst, sizeof (os_sarray_rec)) ;
    dst->version  = version ;
    return (dst) ;
}
    
static int os_sarray_put (sa, sel_id, selimp)
    os_sarray    sa ;
    OsEObjSelId  sel_id ;
    os_selimp    selimp ;
{
    unsigned short d, c, b, s ;
    os_sarrayC *pducket ;
    os_sarrayB *pcucket ;
    os_sarrayS *pbucket ;
    int i ;
    
    SplitSelector (sel_id, d, c, b, s) ;
    pducket = &(sa->duckets[d]) ;
    pcucket = &(sa->duckets[d]->cuckets[c]) ;
    pbucket = &(sa->duckets[d]->cuckets[c]->buckets[b]) ;

    if ( (*pbucket)->selectors[s] == selimp )
        return (0) ;
    
/* Ducket level check */
    if ( (*pducket) == sa->empty_cucket ) {
	*pducket = (os_sarrayC) OsGCMalloc (sizeof (os_sarrayC_rec)) ;
#ifdef OS_EOBJ_STATS
	stat_nduckets++ ;
#endif /* OS_EOBJ_STATS */
	for (i = 0 ; i < SA_CUCKET_SIZE ; i++ ) 
	    (*pducket)->cuckets[i] = sa->empty_bucket ;
	(*pducket)->version  = sa->version ;
	pcucket = &((*pducket)->cuckets[c]) ;
	pbucket = &((*pcucket)->buckets[b]) ;
    } else if ( (*pducket)->version != sa->version ) {
	os_sarrayC oducket = *pducket ;
	*pducket = (os_sarrayC) OsGCMalloc (sizeof (os_sarrayC_rec)) ;
#ifdef OS_EOBJ_STATS
	stat_nduckets++ ;
#endif /* OS_EOBJ_STATS */
	for (i = 0 ; i < SA_CUCKET_SIZE ; i++ ) 
	    (*pducket)->cuckets[i] = oducket->cuckets[i] ;
	(*pducket)->version  = sa->version ;
	pcucket = &((*pducket)->cuckets[c]) ;
	pbucket = &((*pcucket)->buckets[b]) ;
    }

/* Cucket level check */    
    if ( (*pcucket) == sa->empty_bucket ) {
	*pcucket = (os_sarrayB) OsGCMalloc (sizeof (os_sarrayB_rec)) ;
#ifdef OS_EOBJ_STATS
	stat_ncuckets++ ;
#endif /* OS_EOBJ_STATS */
	for (i = 0 ; i < SA_BUCKET_SIZE ; i++ ) 
	    (*pcucket)->buckets[i] = sa->empty_selector ;
	(*pcucket)->version  = sa->version ;
	pbucket = &((*pcucket)->buckets[b]) ;
    } else if ( (*pcucket)->version != sa->version ) {
	os_sarrayB ocucket = *pcucket ;
	*pcucket = (os_sarrayB) OsGCMalloc (sizeof (os_sarrayB_rec)) ;
#ifdef OS_EOBJ_STATS
	stat_ncuckets++ ;
#endif /* OS_EOBJ_STATS */
	for (i = 0 ; i < SA_BUCKET_SIZE ; i++ ) 
	    (*pcucket)->buckets[i] = ocucket->buckets[i] ;
	(*pcucket)->version  = sa->version ;
	pbucket = &((*pcucket)->buckets[b]) ;
    }

/* Bucket level */    
    if ( (*pbucket) == sa->empty_selector ) {
	*pbucket = (os_sarrayS) OsGCMalloc (sizeof (os_sarrayS_rec)) ;
#ifdef OS_EOBJ_STATS
	stat_nbuckets++ ;
#endif /* OS_EOBJ_STATS */
	bcopy (sa->empty_selector, *pbucket, sizeof (os_sarrayS_rec)) ;
	(*pbucket)->version  = sa->version ;
    } else if ( (*pbucket)->version != sa->version ) {
	os_sarrayS obucket = *pbucket ;
#ifdef OS_EOBJ_STATS
	stat_nbuckets++ ;
#endif /* OS_EOBJ_STATS */
	*pbucket = (os_sarrayS) OsGCMalloc (sizeof (os_sarrayS_rec)) ;
	bcopy (obucket, *pbucket, sizeof (os_sarrayS_rec)) ;
	(*pbucket)->version  = sa->version ;
    }
    (*pbucket)->selectors[s] = selimp ;
    return (1) ;
}

static INLINE unsigned long os_sarray_selector_version (sa, sel)
    os_sarray sa ;
    OsEObjSelId  sel ;
{
    unsigned int d, c, b, s ;

    SplitSelector (sel, d, c, b, s) ;
    return (sa->duckets[d]->cuckets[c]->buckets[b]->version) ;
}

/***************************************************************************
 * Miscleaneous functions to get/set methods
 ***************************************************************************/

/*
 * Adding/Removing children from subclasses lists
 */

static INLINE void iclass_add_child (parent, child)
    os_iclass parent, child ;

{
    if ( ! parent->subclasses )
        parent->subclasses = OsWeakVectorMake (4) ;
    OsVectorFastAppend (parent->subclasses, (OsObject) child) ;
    return ;
}

static INLINE void iclass_remove_child (parent, child)
    os_iclass parent, child ;
{
    if ( parent->subclasses ) {
        if ( OsVectorFastRemoveq (parent->subclasses, (OsObject) child) >= 0 )
            return ;
    }
    (void) OsErrorTrigger (1, OsE_Implementation
                           , OsStringMake ("iclass_remove_child")
                           , OsStringMake ("child not found.")) ;
    /* NOT REACHED */
}

static INLINE int iclass_passthrough (os_iclass cls, OsEObj obj) {
    OsEObj proto = cls->canonical ;
    while ( proto ) {
	if ( proto == obj )
	    return (1) ;
	else
	    proto = proto->prototype ;
    }
    return (0) ;
}

/*
 * Perform lazy copy of the given class.
 * This is tricky, but how could it be otherwise...the main idea, is that
 * the given object should be canonic by the end of the function (i.e, it
 * should own its class).
 * The major problem this functions deals with is the following: suppose
 * we have a proto A, cloned to B, itself cloned to C. We define a msg on C
 * which make C lazy copy B's class, ie, A's class. Than we come back definigng
 * a message to B...
 * We need a mean to ensure that when creating the class of B, we are able to
 * rattach to it C class.
 *
 * This function returns:
 * - 0 if no subclass was attached to the new lazy copied class
 * - 1 otherwise.
 */

static void iclass_finalizer (objclass)
    os_iclass objclass ;
{
    if ( OsVectorFastRemoveq (objclass->parent->subclasses
                              , (OsObject) objclass) < 0 )
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("iclass_finalizer")
                               , OsStringMake ("not in parent subclasses")) ;
    return ;
}

static int class_lazy_copy (OsEObj obj) {
    static unsigned long class_id = 2 ;	/* root_class is one */
    os_iclass newclass  = (os_iclass) OsGCMalloc (sizeof(os_iclass_rec));
    os_iclass oldclass  = EObjIclass (obj) ;
    int retvalue = 0 ;
    
#ifdef OS_EOBJ_STATS
    stat_nclasses++ ;
#endif /* OS_EOBJ_STATS */

/*
 * - Copy informations from parent class
 * - Trigger the class msg sparse array lazy-copy
 * - Finalize newclass with a function to remove it from its parent subclasses
 */
    
    newclass->version    = class_id++ ;
    newclass->shl        = 0 ;
    newclass->parent     = oldclass ;
    newclass->canonical  = obj ;
    newclass->subclasses = (OsVector) 0 ;
    newclass->msg        = os_sarray_lazy_copy (oldclass->msg
                                                , newclass->version) ;
    OsGCRegisterFinalizer (newclass, iclass_finalizer) ;
    
/*
 * Now, we need take away our children from our master-copy
 * Surely, I'm doing this to often...this is needed only if the parent
 * link is indirect, which *should* not always be the case.
 */
    if ( oldclass->subclasses ) {
        OsVector subclasses = oldclass->subclasses ;
        OsObject *pclass    = subclasses->cells ;
	/* do not precompute eclass, as it *does* change in remove_... */
	while ( pclass < subclasses->cells + subclasses->csize ) {
	    if ( iclass_passthrough ((os_iclass) *pclass, obj) ) {
		retvalue = 1 ;
		iclass_add_child (newclass, *pclass) ;
		iclass_remove_child (oldclass, *pclass) ;
	    } else {
		pclass++ ;
	    }
	}
    }

/* Dirt oldclass, to make prototype reconsider their structure, and update
 * right now our canonical prototype */
    (oldclass->shl)++ ;
    oldclass->canonical->objclass_shl = oldclass->shl ;
    
/* Add new to old parents */
    iclass_add_child (oldclass, newclass) ;
    obj->objclass     = newclass ;
    obj->objclass_shl = newclass->shl ;
    GC_register_disappearing_link ((void **) &(obj->objclass)) ;  
    return (retvalue) ;
}

/*
 * Update subclasses for the given selector (if needed)
 */

static int iclass_implements (os_iclass class, OsEObjSelId selid) {
    if ( os_sarray_selector_version(class->msg, selid) == class->version ) {
	os_selimp sel_imp = os_sarray_get (class->msg, selid) ;
	return (os_defimp != sel_imp) ;
    }
    return (0) ;
}

static void update_subclasses (superclass, sel, imp)
    os_iclass   superclass ;
    OsEObjSelId sel ;
    os_selimp   imp ;
{
    if ( superclass->subclasses ) {
        OsObject *pclass = superclass->subclasses->cells ;
        OsObject *eclass = pclass + superclass->subclasses->csize ;
        while ( pclass < eclass ) {
            if ( (superclass->version != ((os_iclass) (*pclass))->version)
                && ( ! iclass_implements ((os_iclass) *pclass, sel) )
                && os_sarray_put (((os_iclass) (*pclass))->msg, sel, imp) ) {
                update_subclasses ((os_iclass) *pclass, sel, imp) ;
            }
            pclass++ ;
        }
    }
    return ;
}

void OsEObjDefineFullMsg (obj, sel, newimp_ptr)
    OsEObj    obj ;
    OsEObjSel sel ;
    void      *newimp_ptr ;
{
    int needs_update = 0 ;
    int needs_decl   = 1 ;
    os_selimp oldimp, newimp = (os_selimp) newimp_ptr ;
/*
 * If we are to define a method which is the same as our current father, leave
 * it (this is *very* good for slots, so we share their parents selimp)
 */
    oldimp = EObjSelimpLookup (obj, sel->selid) ;
    if ( oldimp == newimp )
        return ;
/*
 * Ok, too bad !
 */
    if ( obj->objclass->canonical != obj ) 
	needs_update = class_lazy_copy (obj) ;
    else 
        needs_decl = ( ! iclass_implements (obj->objclass, sel->selid)) ;
    needs_update |= os_sarray_put (obj->objclass->msg, sel->selid, newimp) ;
    if ( needs_update )
	update_subclasses (obj->objclass, sel->selid, newimp) ;
    if ( needs_decl ) {
        if ( obj->implements )
            (void) OsVectorFastAppend (obj->implements, (OsObject) sel) ;
        else
            obj->implements = OsVectorMakeFromArray (1, (OsObject *) &sel) ;
    }
    return ;
}

void OsEObjDefineMsg (obj, sel, cfun, closure)
    OsEObj    obj ;
    OsEObjSel sel ;
    OsObject  (*cfun)() ;
    void      *closure ;
{
    os_selimp_cfun imp = make_selimp_cfun (cfun, closure) ;
    OsEObjDefineFullMsg (obj, sel, (os_selimp) imp) ;
    return ;
}

void OsEObjDefineOsMsg (obj, sel, block)
    OsEObj    obj ;
    OsEObjSel sel ;
    OsObject  block ;
{
    os_selimp_clos imp = make_selimp_clos (block) ;
    OsEObjDefineFullMsg (obj, sel, (os_selimp) imp) ;
    return ;
}

void OsEObjDefineSlot (obj, sel, didx)
    OsEObj       obj ;
    OsEObjSel    sel ;
    unsigned int didx ;
{
    os_selimp_slot imp = make_selimp_slot (didx) ;
    OsEObjDefineFullMsg (obj, sel, (os_selimp) imp) ;
    return ;
}

OsObject OsEObjAllocateSlot (self, rsel, wsel, value)
    OsEObj    self ;
    OsEObjSel rsel, wsel ;
    OsObject  value ;
{
    unsigned int didx ;
    
    if ( ! rsel )
        rsel = os_sel_does_not_recognize ;
    if ( ! wsel )
        wsel = os_sel_does_not_recognize ;
    if ( self->readers ) {
        if ( (rsel != os_sel_does_not_recognize)
            && (OsVectorFastMemq (self->readers, (OsObject) rsel) >= 0) )
            (void) OsErrorTrigger (1, ose_reader_redef, self, rsel) ;
        if ( (wsel != os_sel_does_not_recognize)
            && (OsVectorFastMemq (self->writers, (OsObject) wsel) >= 0) )
            (void) OsErrorTrigger (1, ose_writer_redef, self, wsel) ;
        didx = OsVectorFastAppend (self->readers, (OsObject) rsel) ;
        (void) OsVectorFastAppend (self->writers, (OsObject) wsel) ;
        (void) OsVectorFastAppend (self->values, OsA_Nil) ;
    } else {
        self->readers = OsVector1Make ((OsObject) rsel) ;
        self->writers = OsVector1Make ((OsObject) wsel) ;
        self->values  = OsVector1Make (OsA_Nil) ;
        didx          = 0 ;
    }
    self->values->cells[didx] = value ;
    if ( rsel != os_sel_does_not_recognize )
        OsEObjDefineSlot (self, rsel, didx) ;
    if ( wsel != os_sel_does_not_recognize )
        OsEObjDefineSlot (self, wsel, didx) ;
    return (value) ;
}

/***************************************************************************
 * Find the corresponding super for the couple (obj,selid)
 * I think this could be done better if I use the classes instead of the
 * object themselves (in some case)...
 ***************************************************************************/

static OsEObj os_eobj_super (obj, selid)
    OsEObj      obj ;
    OsEObjSelId selid ;
{
    os_selimp imp   = EObjSelimpLookup (obj, selid), super_imp ;
    OsEObj    super = obj->prototype ;
    while ( super ) {
	super_imp = EObjSelimpLookup (super, selid) ;
        if ( super_imp != imp )
	    return (super) ;
	else
	    super = super->prototype ;
    }
    return ((OsEObj) 0) ;
}

OsObject _OsEObjSend0 (lookup, rcv, sel, imp)
    OsEObj    lookup, rcv ;
    OsEObjSel sel ;
    os_selimp imp ;
{
    if ( ! imp )
        imp = EObjSelimpLookup (lookup, sel->selid) ;
    switch (imp->type) {
      case OS_SELIMP_CFUN:
        return ((SELIMP_CFUN(imp))(SELIMP_CLOSURE(imp), rcv, sel)) ;
      case OS_SELIMP_CLOS:
        return (OsCodeBlockApplyWithArray (OsCurrentInterp()
                                           , SELIMP_BLOCK(imp)
                                           , 1, (OsObject *) &rcv
                                           , (OsObject *) 0)) ;
      case OS_SELIMP_SLOT:
        {
            unsigned int didx = SELIMP_DIDX (imp) ;
            if ( (! rcv->readers) || (didx >= rcv->readers->csize)
                || (rcv->readers->cells[didx] != (OsObject) sel) ) {
                /* Share parent's value, until slot is written */
                return (OsEObjSendMsg0 (rcv->prototype, rcv->prototype, sel)) ;
            } else {
                /* This slot's now belong to us */
                return (rcv->values->cells[didx]) ;
            }
        }
        /* NOT REACHED */
      default:
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("OsEObjSendMsg0")
                               , OsStringMake ("invalid selimp type.")) ;
    }
    /* NOT REACHED */
}

OsObject _OsEObjSend1 (lookup, rcv, sel, imp, a1)
    OsEObj    lookup, rcv ;
    OsEObjSel sel ;
    os_selimp imp ;
    OsObject  a1 ;
{
    if ( ! imp )
        imp = EObjSelimpLookup (lookup, sel->selid) ;
    switch (imp->type) {
      case OS_SELIMP_CFUN:
        return ((SELIMP_CFUN(imp))(SELIMP_CLOSURE(imp), rcv, sel, a1)) ;
      case OS_SELIMP_CLOS:
        {
            OsObject argv[2] ;
            argv[0] = (OsObject) rcv ;
            argv[1] = a1 ;
            return (OsCodeBlockApplyWithArray (OsCurrentInterp()
                                               , SELIMP_BLOCK(imp)
                                               , 2, argv
                                               , (OsObject *) 0)) ;
        }
      case OS_SELIMP_SLOT:
        {
            unsigned int didx = SELIMP_DIDX (imp) ;

            if ( (! rcv->writers) || (didx >= rcv->writers->csize)
                || (rcv->writers->cells[didx] != (OsObject) sel) ) {
                /* This slot is shared. Copy on write, now */
                return (OsEObjAllocateSlot (rcv
                                            , os_w_to_r_sel (sel)
                                            , sel
                                            , a1)) ;
            } else {
                /* Ok, slot's value belong to us */
                return (rcv->values->cells[didx] = a1) ;
            }
        }
        /* NOT REACHED */
      default:
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("OsEObjSendMsg1")
                               , OsStringMake ("invalid selimp type.")) ;
    }
    /* NOT REACHED */
}

OsObject _OsEObjSend2 (lookup,  rcv, sel, imp, a1, a2)
    OsEObj    lookup, rcv ;
    OsEObjSel sel ;
    os_selimp imp ;
    OsObject  a1, a2 ;
{
    if ( ! imp )
        imp = EObjSelimpLookup (lookup, sel->selid) ;
    switch (imp->type) {
      case OS_SELIMP_CFUN:
        return ((SELIMP_CFUN(imp))(SELIMP_CLOSURE(imp), rcv, sel, a1, a2)) ;
      case OS_SELIMP_CLOS:
        {
            OsObject argv[3] ;
            argv[0] = (OsObject) rcv ;
            argv[1] = a1 ;
            argv[2] = a2 ;
            return (OsCodeBlockApplyWithArray (OsCurrentInterp()
                                               , SELIMP_BLOCK(imp)
                                               , 3, argv
                                               , (OsObject *) 0));
        }
      case OS_SELIMP_SLOT:
        
      default:
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("OsEObjSendMsg2")
                               , OsStringMake ("invalid selimp type.")) ;
    }
    /* NOT REACHED */
}


OsObject _OsEObjSend3 (lookup, rcv, sel, imp, a1, a2, a3)
    OsEObj    lookup, rcv ;
    OsEObjSel sel ;
    os_selimp imp ;
    OsObject  a1, a2, a3 ;
{
    if ( ! imp )
        imp = EObjSelimpLookup (lookup, sel->selid) ;
    switch (imp->type) {
      case OS_SELIMP_CFUN:
        return ((SELIMP_CFUN(imp))(SELIMP_CLOSURE(imp), rcv, sel, a1, a2, a3));
      case OS_SELIMP_CLOS:
        {
            OsObject argv[4] ;
            argv[0] = (OsObject) rcv ;
            argv[1] = a1 ;
            argv[2] = a2 ;
            argv[3] = a3 ;
            return (OsCodeBlockApplyWithArray (OsCurrentInterp()
                                               , SELIMP_BLOCK(imp)
                                               , 4, argv
                                               , (OsObject *) 0)) ;
        }
      case OS_SELIMP_SLOT:

      default:
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("OsEObjSendMsg3")
                               , OsStringMake ("invalid selimp type.")) ;
    }
    /* NOT REACHED */
}

OsObject OsEObjPerformWithArray (rcv, sel, argc, argv)
    OsEObj       rcv ;
    OsEObjSel    sel ;
    unsigned int argc ;
    OsObject     *argv ;
{
    if ( sel->arity != argc ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (sel->arity)
                               , OsIntegerMake (argc)) ;
        /* NOT REACHED */
    }
    switch (argc) {
      case 0:
        return (OsEObjSendMsg0 (rcv, rcv, sel)) ;
      case 1:
        return (OsEObjSendMsg1 (rcv, rcv, sel, argv[0])) ;
      case 2:
        return (OsEObjSendMsg2 (rcv, rcv, sel, argv[0], argv[1])) ;
      case 3:
        return (OsEObjSendMsg3 (rcv, rcv, sel, argv[0], argv[1], argv[2])) ;
    default:
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("OsEObjPerformWithArray")
                               , OsStringMake ("arity not implemented.")) ;
    }
    /* NOT REACHED */
}

/***************************************************************************
 * Classes
 * - Maintains the full inheritance tree
 * - Maintains the associated method array
 ***************************************************************************/

static OsObject do_forward (closure, self, sel, a1, a2, a3, a4, a5, a6, a7, a8)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
    OsObject  a1, a2, a3, a4, a5, a6, a7, a8 ;
{
    OsObject  pargv[8] ;	/* better use the stdarg stuff here */
    
    pargv[0] = a1 ;
    pargv[1] = a2 ;
    pargv[2] = a3 ;
    pargv[3] = a4 ;
    pargv[4] = a5 ;
    pargv[5] = a6 ;
    pargv[6] = a7 ;
    pargv[7] = a8 ;
    
    return (OsEObjSendMsg2 (self, self
			    , os_sel_forward
			    , (OsObject) sel
			    , (OsObject) OsVectorMakeFromArray (sel->arity
                                                                , pargv))) ;
}

static OsObject default_clone (closure, self, sel)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
{
    OsEObj clone = (OsEObj) OsObjectMake (OsEObjClass, 0) ;

    clone->prototype    = self ;
    clone->objclass     = self->objclass ;
    clone->objclass_shl = clone->objclass->shl ;
    clone->implements   = (OsVector) 0 ;
    clone->readers      = (OsVector) 0 ;
    clone->writers      = (OsVector) 0 ;
    clone->values       = (OsVector) 0 ;
    return ((OsObject) clone) ;
}

static OsObject default_prototype (closure, self, sel)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
{
    return (self->prototype ? (OsObject) self->prototype : OsA_False) ;
}

static OsObject default_forward (closure, self, sel, os_sel, args)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel, os_sel ;
    OsObject  args ;
{
    return (OsEObjSendMsg1 (self, self
                            , os_sel_does_not_recognize
                            , (OsObject) os_sel)) ;
}

static OsObject default_is_cloned_from (closure, self, sel, from)
    void      *closure ;
    OsEObj    self, from ;
    OsEObjSel sel ;
{
    while ( self ) {
	if ( self == from )
	    return (OsA_True) ;
	self = self->prototype ;
    }
    return (OsA_False) ;
}

static OsObject default_define_message_to (closure, self, sel, os_sel, os_fun)
    void           *closure ;
    OsEObj         self ;
    OsEObjSel      sel, os_sel ;
    OsClosedBlock  os_fun ;
{
    OsShouldBeEObjSel (os_sel) ;
    OsShouldBeClosedBlock (os_fun) ;		/* too restrictive ? */
    OsEObjDefineOsMsg (self, os_sel, (OsObject) os_fun) ;
    return (OsA_True) ;
}

static OsObject default_define_slot (closure, self, sel, os_slot)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
    OsString  os_slot ;
{
    OsEObjSel rsel, wsel ;
    
    OsShouldBeString (os_slot) ;
    rsel = OsEObjSelDefine (os_slot->string) ;
    wsel = r_to_w_sel (rsel) ;
    return (OsEObjAllocateSlot (self, rsel, wsel, OsA_Nil)) ;
}

static OsObject default_define_slot_init (closure, self, sel, os_slot, os_init)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
    OsString  os_slot ;
    OsObject  os_init ;
{
    OsEObjSel rsel, wsel ;

    OsShouldBeString (os_slot) ;
    rsel = OsEObjSelDefine (os_slot->string) ;
    wsel = r_to_w_sel (rsel) ;
    return (OsEObjAllocateSlot (self, rsel, wsel, os_init)) ;
}

static OsObject default_get_messages (closure, self, sel)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
{
    return ((self->implements)
            ? (OsObject) self->implements
            : (OsObject) OsVectorMake (0)) ;
}

static OsObject default_responds_to (closure, self, sel, os_sel)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel, os_sel ;
{
    OsShouldBeEObjSel (os_sel) ;
    if ( EObjSelimpLookup (self, os_sel->selid) != os_defimp )
        return (OsA_True) ;
    else
        return (OsA_False) ;
}

static OsObject default_does_not_recognize (closure, self, sel, os_sel)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel, os_sel ;
{

    (void) OsErrorTrigger (1, ose_bad_selector, self, os_sel) ;
    /* NOT REACHED */
}

static OsObject default_error (closure, self, sel, os_msg)
    void      *closure ;
    OsObject  self ;
    OsEObjSel sel ;
    OsString  os_msg ;
{
    OsShouldBeString (os_msg) ;
    (void) OsErrorTrigger (1, ose_bad_selector, self, sel) ;
    /* NOT REACHED */
}

static OsObject default_super (closure, self, sel)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
{
    (void) OsErrorTrigger (1, ose_bad_selector, self, sel) ;
    /* NOT REACHED */
}

static OsObject default_display_to (closure, self, sel, port)
    void         *closure ;
    OsEObj       self ;
    OsEObj       sel ;
    OsOutputPort port ;
{
    char tmp[OS_DISPLAY_BUFSIZ] ;
    OsShouldBeOutputPort (port) ;
    sprintf (tmp, "#[ExtObj %p]", self) ;
    OsOutputPortWriteCString (port, tmp) ;
    return ((OsObject) self) ;
}

static OsObject default_display (closure, self, sel)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
{
    return (OsEObjSendMsg1 (self, self
                            , os_sel_display_to
                            , (OsObject) OsCurrentOutputPort)) ;
}

static OsObject default_perform_with (closure, self, sel, the_sel, argv)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
    OsEObjSel the_sel ;
    OsVector  argv ;
{
    OsShouldBeEObjSel (the_sel) ;
    OsShouldBeVector (argv) ;
    return (OsEObjPerformWithArray (self, the_sel, argv->csize, argv->cells));
}

static OsObject default_get_implementation_for (closure, self, sel, the_sel)
    void *closure ;
    OsEObj self ;
    OsEObjSel sel, the_sel ;
{
    OsShouldBeEObjSel (the_sel) ;
    return ((OsObject) os_selimp_export (self, the_sel)) ;
}

static OsObject default_run_implementation_with (closure, self, sel, imp, argv)
    void *closure ;
    OsEObj self ;
    OsEObjSel sel ;
    OsEObjSelImp imp ;
    OsVector argv ;
{
    OsShouldBeEObjSelImp (imp) ;
    OsShouldBeVector (argv) ;
    if ( argv->csize != imp->sel->arity ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (imp->sel->arity)
                               , OsIntegerMake (argv->csize)) ;
        /* NOT REACHED */
    }
    if ( self != imp->from ) {
        (void) OsErrorTrigger (1, ose_bad_imp_for
                               , imp
                               , self) ;
        /* NOT REACHED */
    }
    switch (argv->csize) {
      case 0:
        return (OsEObjSendImp0 (self, self, imp->sel, imp->os_imp)) ;
      case 1:
        return (OsEObjSendImp1 (self, self, imp->sel, imp->os_imp
                                , argv->cells[0])) ;
      case 2:
        return (OsEObjSendImp2 (self, self, imp->sel, imp->os_imp
                                , argv->cells[0]
                                , argv->cells[1])) ;
      case 3:
        return (OsEObjSendImp3 (self, self, imp->sel, imp->os_imp
                                , argv->cells[0]
                                , argv->cells[1]
                                , argv->cells[2])) ;
      default:
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake("default_run_implementation_for")
                               , OsStringMake("arity not implemented.")) ;
        /* NOT REACHED */
    }
    /* NOT REACHED */
}

static void os_define_default_behavior (obj)
    OsEObj obj ;
{
    OsEObjSel sel ;

    OsEObjSelClone = OsEObjSelDefine ("clone") ;
    OsEObjDefineMsg (obj, OsEObjSelClone, default_clone, (void *) 0) ;
    sel = OsEObjSelDefine ("prototype") ;
    OsEObjDefineMsg (obj, sel, default_prototype, (void *) 0) ;
    sel = OsEObjSelDefine ("isClonedFrom:") ;
    OsEObjDefineMsg (obj, sel, default_is_cloned_from, (void *) 0) ;
    sel = OsEObjSelDefine ("defineMessage:to:") ;
    OsEObjDefineMsg (obj, sel, default_define_message_to, (void *) 0) ;
    sel = OsEObjSelDefine ("defineSlot:") ;
    OsEObjDefineMsg (obj, sel, default_define_slot, (void *) 0) ;
    sel = OsEObjSelDefine ("getMessages") ;
    OsEObjDefineMsg (obj, sel, default_get_messages, (void *) 0) ;
    os_sel_define_slot_init = OsEObjSelDefine ("defineSlot:init:") ;
    OsEObjDefineMsg (obj
                     , os_sel_define_slot_init, default_define_slot_init
                     , (void *) 0) ;
    os_sel_forward = OsEObjSelDefine ("forward::") ;
    OsEObjDefineMsg (obj, os_sel_forward, default_forward, (void *) 0) ;
    os_sel_responds_to = OsEObjSelDefine ("respondsTo:") ;
    OsEObjDefineMsg (obj
                     , os_sel_responds_to, default_responds_to
                     , (void *) 0) ;
    os_sel_does_not_recognize = OsEObjSelDefine ("doesNotRecognize:") ;
    OsEObjDefineMsg (obj
                     , os_sel_does_not_recognize, default_does_not_recognize
                     , (void *) 0);
    os_sel_error = OsEObjSelDefine ("error:") ;
    OsEObjDefineMsg (obj, os_sel_error, default_error, (void *) 0) ;
    os_sel_super = OsEObjSelDefine ("super") ;
    OsEObjDefineMsg (obj, os_sel_super, default_super, (void *) 0) ;
    os_sel_display    = OsEObjSelDefine ("display") ;
    OsEObjDefineMsg (obj, os_sel_display, default_display, (void *) 0) ;
    os_sel_display_to = OsEObjSelDefine ("display:") ;
    OsEObjDefineMsg (obj, os_sel_display_to, default_display_to, (void *) 0) ;
    OsEObjDefineMsg (obj
                     , OsEObjSelDefine ("perform:with:")
                     , default_perform_with
                     , (void *) 0) ;
    OsEObjDefineMsg (obj
                     , OsEObjSelDefine ("getImplementationFor:")
                     , default_get_implementation_for
                     , (void *) 0) ;
    OsEObjDefineMsg (obj
                     , OsEObjSelDefine ("runImplementation:with:")
                     , default_run_implementation_with
                     , (void *) 0) ;
    return ;
}

static OsObject os_eobj_bootstrap ()
{
    os_iclass root_class = (os_iclass) OsGCMalloc (sizeof(os_iclass_rec));
    OsEObj root ;
    
    root_class->version    = 1 ;
    root_class->shl        = 0 ;
    root_class->parent     = (os_iclass) 0 ;
    root_class->subclasses = (OsVector) 0 ;
    root_class->msg = os_sarray_make (root_class->version
                                      , (OsObject (*)()) do_forward
                                      , (OsObject) 0) ;

    root = (OsEObj) OsObjectMake (OsEObjClass, 0) ;
    root->prototype    = (OsEObj) 0 ;
    root->objclass     = root_class ;	/* OsEObjRoot *is* canonical */
    root->objclass_shl = root_class->shl ;
    root->implements   = (OsVector) 0 ;
    root->readers      = (OsVector) 0 ;
    root->writers      = (OsVector) 0 ;
    root->values       = (OsVector) 0 ;

    root_class->canonical = (OsEObj) root ;

    os_define_default_behavior (root) ;
    
    return (OsEObjRoot = (OsObject) root) ;
}

static OsObject OsSubrObjSelector (os_selname)
    OsString os_selname ;
{
    OsEObjSel sel ;
    OsShouldBeString (os_selname) ;
    if ( (sel = OsEObjSelLookup (os_selname->string)) )
        return ((OsObject) sel) ;
    else
	return (OsA_False) ;
}

static OsObject OsSubrObjDefineSelector (os_selname)
    OsString os_selname ;
{
    OsShouldBeString (os_selname) ;
    return ((OsObject) OsEObjSelDefine (os_selname->string)) ;
}

#ifdef OS_EOBJ_STATS
static OsObject OsSubrObjStat ()
{
    printf ("----- external object model statistics -----\n") ;
    printf ("# of duckets: %4d (%4d bytes)\n"
	    , stat_nduckets, stat_nduckets * sizeof (os_sarrayC_rec)) ;
    printf ("# of cuckets: %4d (%4d bytes)\n"
	    , stat_ncuckets, stat_ncuckets * sizeof (os_sarrayB_rec)) ;
    printf ("# of buckets: %4d (%4d bytes)\n"
	    , stat_nbuckets, stat_nbuckets * sizeof (os_sarrayS_rec)) ;
    printf ("# of sarrays: %4d (%4d bytes)\n"
	    , stat_nsarrays, stat_nsarrays * sizeof (os_sarray_rec)) ;
    printf ("Total # of bytes used by sarrays: %d\n"
	    , (stat_nduckets * sizeof (os_sarrayC_rec)
	       + stat_ncuckets * sizeof (os_sarrayB_rec)
	       + stat_nbuckets * sizeof (os_sarrayS_rec)
	       + stat_nsarrays * sizeof (os_sarray_rec))) ;
    printf ("Total # of created classes: %d\n", stat_nclasses) ;
    printf ("--------------------------------------------\n") ;
    fflush (stdout) ;
    return (OsA_True) ;
}
#endif /* OS_EOBJ_STATS */

/*
 * Message implementations:
 */

static OsObject os_eobj_display (self, sel, port)
    OsEObj   self ;
    int      sel ;
    OsObject port ;
{
    if ( port == (OsObject) OsCurrentOutputPort ) {
        return (OsEObjSendMsg0 (self, self, os_sel_display)) ;
    } else {
        return (OsEObjSendMsg1 (self, self, os_sel_display_to, port)) ;
    }
    /* NOT REACHED */
}

static OsObject os_eobjsel_display (self, sel, port)
    OsEObjSel    self ;
    int          sel ;
    OsOutputPort port ;
{
    char tmp[OS_DISPLAY_BUFSIZ] ;
    unsigned int selc ;
    sprintf (tmp, "#[ExtSel@%p ", self->selv) ;
    OsOutputPortWriteCString (port, tmp) ;
    for (selc = 0 ; selc < self->selc ; selc++ ) {
        OsSendDisplay (self->selv[selc], port) ;
    }
    sprintf (tmp, " (%d)]", self->selid) ;
    OsOutputPortWriteCString (port, tmp) ;
    return ((OsObject) self) ;
}

static OsObject os_eobjsel_finalize (self, sel)
    OsEObjSel self ;
    int       sel ;
{
    printf ("Deleting selector (%d)\n", self->selid) ;
    OsHashtableRemove (hselector, (OsObject) self) ;
    if ( free_sel ) 
        selectors->cells[self->selid] = (OsObject) ((unsigned long) free_sel) ;
    else
        selectors->cells[self->selid] = (OsObject) 0 ;
    free_sel = self->selid ;
    return ((OsObject) self) ;
}

static OsObject os_eobjselimp_display (self, sel, port)
    OsEObjSelImp self ;
    int sel ;
    OsOutputPort port ;
{
    char tmp[OS_DISPLAY_BUFSIZ] ;
    OsOutputPortWriteCString (port, "#[ExtSelImp ") ;
    OsSendDisplay (self->sel, port) ;
    OsOutputPortWriteCString (port, " for ") ;
    OsSendDisplay (self->from, port) ;
    OsOutputPortWriteChar (port, ']') ;
    return ((OsObject) self) ;
}

void OsEObjInitialize (OsInterp interp, int *pargc, char **argv) {
    OsEObjClass     = OsClassDeclare ("ExtObj", sizeof (OsEObjRec)) ;
    OsDefineMsg (OsEObjClass, OS_SEL_DISPLAY, os_eobj_display) ;
    OsEObjSelClass  = OsClassDeclare ("ExtSel", sizeof (OsEObjSelRec)) ;
    OsDefineMsg (OsEObjSelClass, OS_SEL_DISPLAY, os_eobjsel_display) ;
    OsDefineMsg (OsEObjSelClass, OS_SEL_FINALIZE, os_eobjsel_finalize) ;
    OsEObjSelImpClass = OsClassDeclare ("ExtSelImp", sizeof (OsEObjSelImpRec));
    OsDefineMsg (OsEObjSelImpClass, OS_SEL_DISPLAY, os_eobjselimp_display) ;
/*
 * Create selimp hashtable:
 */
    os_selimp_hashtable = OsHashtableMake (0
                                           , (OsHashFct) os_selimp_hash
                                           , (OsCmpFct) os_selimp_cmp) ;
/*
 * Create the selector array, and the selector hash-table.
 * The selector array is implemented using vectors, it starts with nil to
 *    preserve the special semantic of selector 0.
 */
    selectors = OsWeakVectorMake (0) ;
    (void) OsVectorFastAppend (selectors, OsA_Nil) ;
    hselector = OsWeakHashtableMake (0
                                     , (OsHashFct) os_hash_sel
                                     , (OsCmpFct) os_cmp_sel) ;
    os_eobj_bootstrap () ;
/*
 * Export the root, cloning it to preserve internal stuff (like struct API)
 * from broken users
 */
    OsAtomMake (interp
                , "obj:root"
		, OsEObjSendMsg0 ((OsEObj) OsEObjRoot, (OsEObj) OsEObjRoot
                                  , OsEObjSelClone)) ;
    return ;
}

void OsEObjSubrInitialize (OsInterp interp) {
    ose_reader_redef = OsErrorMake (interp
                                    , "error:obj-reader-redef"
                                    , "reader %1 redefined on %0"
                                    , 2, OSE_CATCHABLE) ;
    ose_writer_redef = OsErrorMake (interp
                                    , "error:obj-writer-redef"
                                    , "writer %1 redefined on %0"
                                    , 2, OSE_CATCHABLE) ;
    ose_bad_selector = OsErrorMake (interp
                                    , "error:obj-bad-selector"
                                    , "invalid selector %1 for %0"
                                    , 2, OSE_CATCHABLE) ;
    ose_bad_imp_for  =  OsErrorMake (interp
                                     , "error:obj-bad-implementation-for"
                                     , "Bad receiver %0 of implementation %1"
                                     , 2, OSE_CATCHABLE) ;
    OsSubrMake (interp, "obj:selector", OsSubrObjSelector, 1) ;
    OsSubrMake (interp, "obj:define-selector", OsSubrObjDefineSelector, 1) ;
#ifdef OS_EOBJ_STATS
    OsSubrMake (interp, "obj:stat", OsSubrObjStat, 0) ;
#endif /* OS_EOBJ_STATS */
    return ;
}
