/***************************************************************************
 * error.c
 * error.c,v 1.2 1995/03/12 02:57:34 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * OScheme error handling module.
 ***************************************************************************/

#include <varargs.h>
#include <stdio.h>
#include <setjmp.h>
#include <sys/types.h>

#include <oscheme.h>

OsObject OsE_Unbound        = (OsObject) 0 ;
OsObject OsE_ArityShouldBe  = (OsObject) 0 ;
OsObject OsE_BadFunction    = (OsObject) 0 ;
OsObject OsE_BadType        = (OsObject) 0 ;
OsObject OsE_Implementation = (OsObject) 0 ;
OsObject OsE_BadSyscall     = (OsObject) 0 ;

OsClass  OsErrorClass = (OsClass) 0 ;

static OsError ose_user   = (OsError) 0 ;
static OsError ose_vmctxt = (OsError) 0 ;

/*
 * C Api :
 */

OsString OsErrorFormat (errv)
    OsVector errv ;
{
    OsString     str  = OsStringMakeFromBytes (0, "") ;
    OsOutputPort port = OsOutputStringMake (str) ;
    unsigned int idx ;

    for (idx = 0 ; idx < errv->csize ; idx += 2) {
        if ( errv->cells[idx] == (OsObject) ose_vmctxt )
            continue ;
        (void) OsFormatWithList (port
                                 , ((OsError) errv->cells[idx])->fmt
                                 , (OsCons) errv->cells[idx+1]) ;
        OsOutputPortWriteCString (port, "\n\t") ;
    }
    OsOutputPortClose (port) ;
    return (str) ;
}

OsObject OsErrorAugment (va_alist)
    va_dcl 
{
    va_list      ap ;
    OsVector     errv ;
    OsError      error ;
    unsigned int arity, icnt ;
    OsCons       start, *list ;
    
    va_start (ap) ;
    errv = va_arg (ap, OsVector) ;
    icnt = va_arg (ap, int) ;
    while ( icnt-- ) {
        error = va_arg (ap, OsError) ;
        (void) OsVectorFastAppend (errv, (OsObject) error) ;
        start = (OsCons) OsA_Nil ;
        list  = &start ;
        arity = error->arity ;
        while ( arity-- ) {
            *list = OsConsMake (va_arg (ap, OsObject), OsA_Nil) ;
            list  = (OsCons *) &(OS_CONS_CDR(*list)) ;
        }
        (void) OsVectorFastAppend (errv, (OsObject) start) ;
    }
    va_end (pa) ;
    return ((OsObject) errv) ;
}
    
OsObject OsErrorTrigger (va_alist)
    va_dcl
{
    va_list      ap ;
    OsVector     fcall, fvm ;
    OsCons       start, *list ;
    OsError      error ;
    unsigned int arity, icnt ;
    OsInterp     interp = OsCurrentInterp() ;
    
    va_start (ap) ;
    icnt  = va_arg (ap, int) ;
    fcall = OsVectorMake (icnt * 2) ;
    while ( icnt-- ) {
        error = va_arg (ap, OsError) ;
        (void) OsVectorFastAppend (fcall, (OsObject) error) ;
        start = (OsCons) OsA_Nil ;
        list  = &start ;
        arity = error->arity ;
        while ( arity-- ) {
            *list = OsConsMake (va_arg (ap, OsObject), OsA_Nil) ;
            list  = (OsCons *) &(OS_CONS_CDR(*list)) ;
        }
        (void) OsVectorFastAppend (fcall, (OsObject) start) ;
    }
    va_end (pa) ;
    if ( (fvm = OsCurrentVMStateToError (interp)) ) {
        OsObject evm = (OsObject) OsConsMake ((OsObject) fvm, OsA_Nil) ;
        (void) OsVectorFastAppend (fcall, (OsObject) ose_vmctxt) ;
        (void) OsVectorFastAppend (fcall, evm) ;
    }
    if ( (error->severity & OSE_CATCHABLE)
#ifdef OS_SECURE
         /* When compiled in secure mode, this test avoids looping on no more
          * clicks error */
         && ( interp->clicks < interp->lclicks )
         && ( interp->read_bytes < interp->lread_bytes )
         && ( interp->written_bytes < interp->lwritten_bytes )
         && ( interp->security < OS_SECURITY_LEVEL_A )
#endif /* OS_SECURE */
         && ( interp->error_handler != OsA_False ) ) {
        OsObject rep ;
        rep = OsCodeBlockApplyWithArray (OsCurrentInterp()
                                         , (OsCodeBlock) interp->error_handler
                                         , fcall->csize
                                         , fcall->cells
                                         , (OsObject *) 0) ;
        if ((error->severity & OSE_REPAIRABLE) && (rep != OsA_Unbound))
            return (rep) ;
    } 
    os_uncatched_error ((OsObject) fcall) ;
    /* NOT REACHED */
    return ((OsObject) 0) ;
}

void OsFatalError (msg)
    char *msg ;
{
    fprintf (stderr, msg) ;
    exit (1) ;
}

OsError OsErrorMake (OsInterp interp
                     , char *cname
                     , char *fmt
                     , unsigned int arity, unsigned int severity) {
    OsError err = (OsError) OsObjectMake (OsErrorClass, 0) ;
    err->severity = severity ;
    err->arity    = arity ;
    err->fmt      = OsStringMake (fmt) ;
    (void) OsAtomMake (interp, cname, (OsObject) err) ;
    return (err) ;
}

/*
 * Subrs:
 */

/* AUTODOC
 * (error? OBJECT)
 * Is the given OBJECT of type error ?
 */

static OsObject OsSubrErrorP (OsObject error) {
    return (OsIsError (error) ? OsA_True : OsA_False) ;
}
  
static OsObject OsSubrErrorToFormat (OsError error) {
    OsShouldBeError (error) ;
    return ((OsObject) error->fmt) ;
}

static OsObject OsSubrErrorToSeverity (OsError error) {
    OsShouldBeError (error) ;
    return ((OsObject) OsIntegerMake (error->severity)) ;
}

static OsObject OsSubrErrorToArity (OsError error) {
    OsShouldBeError (error) ;
    return ((OsObject) OsIntegerMake (error->arity)) ;
}

/* AUTODOC
 * (error FORMAT ARGS*)
 * Trigger as user error.
 * This primitive triggers a user error 'error:user. The associated message is
 * made of the string reulsting from formatting ARGS* according to FORMAT (see
 * the 'format primitive).
 */

static OsObject OsSubrError (unsigned int argc, OsObject *argv) {
    OsString     msg  = OsStringMakeFromBytes (0, "") ;
    OsOutputPort port = OsOutputStringMake (msg) ;
    OsString     fmt ;

    if ( argc < 1 )
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (0)
                               , OsIntegerMake (1)) ;
    fmt = (OsString) *argv++ ; argc-- ;
    fmt = OsShouldBeString (fmt) ;
    (void) OsFormatWithArray (port, fmt, argc, argv) ;
    OsOutputPortClose (port) ;
    (void) OsErrorTrigger (1, ose_user, msg) ;
    /* NOT REACHED */
}

/* AUTODOC
 * (error-handler)
 * Retreive the error handler associated with the current interpreter context.
 * This primitive returns the error handler associated with the current
 * interpreter context, which is #f if none was set.
 */

static OsObject OsSubrErrorHandler (void) {
    return ((OsObject) (OsCurrentInterp()->error_handler)) ;
}
    
/*
 * AUTODOC
 * (set-error-handler! THUNK)
 * Define error handler to THUNK.
 * This primitive defines an error handler for the current interpreter context.
 * When an error occurs, this handler is invoked with the errors and their
 * parameters as argument.
 * To reset the error handler, you can set it to #f (meaning you are no more
 * interested in the error.
 */

static OsObject OsSubrSetErrorHandler (OsCodeBlock thunk) {
    if ( ((OsObject) thunk) != OsA_False )
        OsMustBeClosedBlock (thunk) ;
    OsCurrentInterp()->error_handler = (OsObject) thunk ;
    return (OsA_True) ;
}

/*
 * Initialize :
 */

void OsErrorInitialize (OsInterp interp, int *pargc, char **argv) {
    OsErrorClass = OsClassDeclare ("Error", sizeof (OsErrorRec)) ;
    return ;
}

void OsErrorSubrInitialize (OsInterp interp) {
    OsE_Unbound
        = (OsObject) OsErrorMake (interp
                                  , "error:unbound"
                                  , "unbound symbol: %0"
                                  , 1, OSE_NORMAL);
    OsE_ArityShouldBe
        = (OsObject) OsErrorMake (interp
                                  , "error:arity"
                                  , "invalid arity: should be %0 instead of %1"
                                  , 2, OSE_CATCHABLE) ;
    OsE_BadFunction
        = (OsObject) OsErrorMake (interp
                                  , "error:function"
                                  , "invalid function: %0"
                                  , 1, OSE_NORMAL) ;
    OsE_BadType
        = (OsObject) OsErrorMake (interp
                                  , "error:type"
                                  , "invalid type: %0 should be a %1"
                                  , 2, OSE_NORMAL) ;
    OsE_Implementation =
        (OsObject) OsErrorMake (interp
                                , "error:implementation"
                                , "implementation error %1 in %0"
                                , 2, OSE_CATCHABLE) ;
    OsE_BadSyscall =
        (OsObject) OsErrorMake (interp
                                , "error:bad-syscall"
                                , "system call %0 %2 failed (%1)"
                                , 3, OSE_CATCHABLE) ;
    
    ose_user   = OsErrorMake(interp
                             , "error:user"
                             , "*** error: %0"
                             , 1, OSE_NORMAL) ;
    ose_vmctxt = OsErrorMake(interp
                             , "error:vm-context"
                             , "vm state %0"
                             , 1, OSE_NORMAL) ;
    OsSubrMake (interp, "set-error-handler!", OsSubrSetErrorHandler, 1) ;
    OsSubrMake (interp, "error-handler", OsSubrErrorHandler, 0) ;
    OsSubrMake (interp, "error?", OsSubrErrorP, 1) ;
    OsSubrMake (interp, "error->format", OsSubrErrorToFormat, 1) ;
    OsSubrMake (interp, "error->severity", OsSubrErrorToSeverity, 1) ;
    OsSubrMake (interp, "error->arity", OsSubrErrorToArity, 1) ;
    OsSubrMake (interp, "error", OsSubrError, OS_NARY) ;
    return ;
}
