/***************************************************************************
 * gdld.c
 * gdld.c,v 1.2 1995/03/12 02:57:41 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * Interface to gnu dld dynamic load. To be compatible with the other files
 * defining such an interface, this should define:
 * OsDlInitialize
 * OsDlSubrInitialize
 * And the following set of subrs:
 * - dl:load : to open/initialize a module
 ***************************************************************************/


#include <stdio.h>
#include <setjmp.h>

#include <oscheme.h>

#ifdef OS_HAS_DLD

#include <dld.h>

static OsError ose_bad_dldlink = (OsError) 0 ;

/* This vector should contains the list of libraries oscheme kernel is linked
 * to, in order to allow symbols defined in those lib, but not used by the
 * kernel to be solved for dynamically loadable modules.
 * See the initialize function.
 */

static OsVector extra_libs = (OsVector) 0 ;


/*
 * Subrs:
 */

OsObject OsSubrDlLoad (OsString module) {
    char         initer[512] ;
    void         (*func)() ;
    OsString     basename, full_path ;
    unsigned int i ;
    int          err = 0 ;
    OsInterp     interp = OsCurrentInterp() ;
    
    /* Load given object module */
    OsShouldBeString (module) ;
    if ( ! (full_path = OsPathFindFile (interp, module)) )
        return (OsA_False) ;
    if ( err = dld_link (full_path->string) ) {
        (void) OsErrorTrigger (1, ose_bad_dldlink
                               , full_path
                               , OsStringMake (dld_strerror (err))) ;
    }
    /* (re) link extra libraries, to solve undefined symbols */
    for ( i = 0 ; i < extra_libs->csize ; i++ ) {
        if ( dld_link (((OsString) extra_libs->cells[i])->string) ) {
            (void) OsErrorTrigger (1, ose_bad_dldlink
                                   , extra_libs->cells[i]
                                   , OsStringMake (dld_strerror (err))) ;
        }
    }
    basename = OsPathBasename (module) ;
    sprintf (initer, "Os_%s_initialize", basename->string) ;
    if ( ( ! (func = (void(*)()) dld_get_func (initer)) )
         || ( ! dld_function_executable_p (initer)) ) {
        if ( dld_errno ) {
            (void) OsErrorTrigger (1, ose_bad_dldlink
                                   , module
                                   , OsStringMake (dld_strerror (err))) ;
        } else {
            (void) OsErrorTrigger (1, ose_bad_dldlink
                                   , module
                                   , OsStringMake ("no init function")) ;
        }
    }
    (*func)(interp) ;
    return (OsA_True) ;
}

void dld_debug () {
    int i     = dld_undefined_sym_count ;
    char **us = dld_list_undefined_sym () ;
    while ( i-- ) {
        fprintf (stderr, "%s\n", us[i]) ;
    }
    return ;
}

    
/*
 * Initialize
 */

void OsDlInitialize (OsInterp interp, int *pargc, char **argv) {
    (void) dld_init (dld_find_executable(argv[0])) ;
    extra_libs = OsVectorMake (0) ;
    (void) OsVectorFastAppend (extra_libs
                               , (OsObject) OsStringMake ("/usr/lib/libc.a"));
}

void OsDlSubrInitialize (OsInterp interp) {
    ose_bad_dldlink = OsErrorMake (interp
                                   , "error:bad-dynamic-load"
                                   , "Cannot load %0 (%1)"
                                   , 2, OSE_CATCHABLE) ;
    OsSubrMake (interp, "dl:load", OsSubrDlLoad, 1) ;
    OsPathProvides (interp, OsStringMake ("dld")) ;
}

#endif /* OS_HAS_DLD */
