/***************************************************************************
 * htable.c
 * htable.c,v 1.2 1995/03/12 02:57:43 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * Hashtable management functions.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#define I_HIDE_POINTERS
#include <gc.h>

#include <os_config.h>
#include <os_interp.h>
#include <os_obj.h>
#include <os_string.h>
#include <os_atom.h>
#include <os_gc.h>
#include <os_vector.h>
#include <os_error.h>
#include <os_cons.h>
#include <os_cblock.h>
#include <os_subr.h>
#include <os_htable.h>
#include <os_port.h>

OsClass OsHashtableClass = (OsClass) 0 ;

typedef struct _os_hbucket {
    os_hbucket 	next ;
    void        *key, *item ;
} os_hbucket_rec ;

/*
 * Internals:
 */

static void resize_hashtable (ht) 
    OsHashtable ht ;
{
    os_hbucket *ps = ht->buckets, *pe = ps + ht->bsize ;
    
    ht->bsize <<= 1 ;
    ht->buckets = (os_hbucket *) OsGCMalloc (ht->bsize*sizeof(os_hbucket));
    ht->bused   = 0 ;
    if ( ht->weakp ) {
        while ( ps < pe ) {
            os_hbucket b = *ps ;
            while ( b ) {
                OsHashtablePut (ht
                                , REVEAL_POINTER(b->key)
                                , REVEAL_POINTER(b->item)) ;
                b = b->next ;
            }
            ps++ ;
        }
    } else {
        while ( ps < pe ) {
            os_hbucket b = *ps ;
            while ( b ) {
                OsHashtablePut (ht, b->key, b->item) ;
                b = b->next ;
            }
            ps++ ;
        }
    }
    return ;
}

/*
 * Create an empty hash table:
 */

OsHashtable OsHashtableMake (size, h, c)
    unsigned int size ;
    OsHashFct    h ;
    OsCmpFct     c ;
{
    OsHashtable ht = (OsHashtable) OsObjectMake (OsHashtableClass, 0) ;
    ht->weakp   = 0 ;
    ht->bused   = 0 ;
    ht->bsize   = ((size < OS_HASHTABLE_DEFAULT_SIZE)
                 ? OS_HASHTABLE_DEFAULT_SIZE
                 : size) ;
    ht->hash    = h ;
    ht->cmp     = c ;
    ht->buckets = (os_hbucket *) OsGCMalloc (ht->bsize * sizeof (os_hbucket)) ;
    return (ht) ;
}

OsHashtable OsWeakHashtableMake (size, f, c)
    unsigned int size ;
    OsHashFct    f ;
    OsCmpFct     c ;
    
{
    OsHashtable ht = OsHashtableMake (size, f, c) ;
    ht->weakp = 1 ;
    return (ht) ;
}

void OsDHt (htable)
    OsHashtable htable ;
{
    os_hbucket *pb = htable->buckets, *pe = pb + htable->bsize ;
    if ( htable->weakp ) {
        while (pb < pe) {
            os_hbucket b = *pb ;
            while ( b ) {
                fprintf (stdout, "bucket %p", b) ;
                OsSendDisplay ((OsObject) REVEAL_POINTER(b->key)
                               , OsCurrentOutputPort) ;
                OsOutputPortFlush (OsCurrentOutputPort) ;
                OsOutputPortWriteCString (OsCurrentOutputPort, "\n\t") ;
                OsSendDisplay ((OsObject) REVEAL_POINTER (b->item)
                               , OsCurrentOutputPort) ;
                OsOutputPortFlush (OsCurrentOutputPort) ;
                OsOutputPortWriteChar (OsCurrentOutputPort, '\n') ;
                OsOutputPortFlush (OsCurrentOutputPort) ;
                b = b->next ;
            }
            pb++ ;
        }
    } else {
        while (pb < pe) {
            os_hbucket b = *pb ;
            while ( b ) {
                printf ("bucket %p", b) ;
                OsSendDisplay ((OsObject) b->key, OsCurrentOutputPort) ;
                OsOutputPortFlush (OsCurrentOutputPort) ;
                OsOutputPortWriteCString (OsCurrentOutputPort, "\n\t") ;
                OsSendDisplay ((OsObject) b->item, OsCurrentOutputPort) ;
                OsOutputPortFlush (OsCurrentOutputPort) ;
                OsOutputPortWriteChar (OsCurrentOutputPort, '\n') ;
                OsOutputPortFlush (OsCurrentOutputPort) ;
                b = b->next ;
            }
            pb++ ;
        }
    }
    OsOutputPortFlush (OsCurrentOutputPort) ;
    return ;
}

/*
 * Put/Get/Remove on hashtables:
 */

void *OsHashtableFindKey (htable, item)
    OsHashtable htable ;
    void        *item ;
{
    os_hbucket *ps = htable->buckets, *pe = ps + htable->bsize, b ;

    if ( htable->weakp ) {
        while ( ps < pe ) {
            for (b = *ps ; b ; b = b->next)
                if ( ! (htable->cmp)(item, (void *) REVEAL_POINTER (b->item)) )
                    return ((void *) REVEAL_POINTER (b->key)) ;
            ps++ ;
        }
    } else {
        while ( ps < pe ) {
            for (b = *ps ; b ; b = b->next)
                if ( ! (htable->cmp) (item, b->item) )
                    return (b->key) ;
            ps++ ;
        }
    }
    return ((void *) 0) ;
}
    
void *OsHashtablePut (htable, key, item)
    OsHashtable htable ;
    void        *key, *item ;
{
    unsigned long hval ;
    os_hbucket    b ;
    
    if ( htable->bused > (htable->bsize << 1) ) 
        resize_hashtable (htable) ;
    hval = (htable->hash)(key) % htable->bsize ;
    b    = htable->buckets[hval] ;
    if ( htable->weakp ) {
        while ( b ) {
            if ( ! (htable->cmp)(key, REVEAL_POINTER (b->key)) ) {
                b->item =(void *) HIDE_POINTER (item) ;
                return (item) ;
            }
            b = b->next ;
        }
    } else {
        while ( b ) {
            if ( ! (htable->cmp)(key, b->key) )
                return (b->item = item) ;
            b = b->next ;
        }
    }
    if ( htable->weakp ) {
        b = (os_hbucket) OsGCMalloc (sizeof (os_hbucket_rec)) ;
        b->key  = (void *) HIDE_POINTER (key) ;
        b->item = (void *) HIDE_POINTER (item) ;
    } else {
        b = (os_hbucket) OsGCMalloc (sizeof (os_hbucket_rec)) ;
        b->key  = key ;
        b->item = item ;
    }
    b->next = htable->buckets[hval] ;
    htable->buckets[hval] = b ;
    htable->bused++ ;
    return (item) ;
}

void *OsHashtableGet (htable, key)
    OsHashtable htable ;
    void        *key ;
{
    os_hbucket b = htable->buckets[(htable->hash)(key) % htable->bsize] ;
    if ( htable->weakp ) {
        while ( b ) {
            if ( ! (htable->cmp)(key, REVEAL_POINTER (b->key)) )
                return (REVEAL_POINTER(b->item)) ;
            b = b->next ;
        }
    } else {
        while ( b ) {
            if ( ! (htable->cmp)(key, b->key) )
                return (b->item) ;
            b = b->next ;
        }
    }
    return ((void *) 0) ;
}

void OsHashtableRemove (htable, key)
    OsHashtable htable ;
    void        *key ;
{
    os_hbucket *pb = htable->buckets + ((htable->hash)(key) % htable->bsize) ;
    os_hbucket b   = *pb ;
    
    if ( htable->weakp ) {
        while ( b ) {
            if ( ! (htable->cmp)(key, REVEAL_POINTER (b->key)) ) {
                *pb = b->next ;
                htable->bused-- ;
                return ;
            }
            pb = &(b->next) ;
            b  = b->next ;
        }
    } else {
        while ( b ) {
            if ( ! (htable->cmp)(key, b->key) ) {
                *pb = b->next ;
                htable->bused-- ;
                return ;
            }
            pb = &(b->next) ;
            b  = b->next ;
        }
    }
    return ;
}

void OsHashtableMap (h, func, closure)
    OsHashtable h ;
    void        (*func)() ;
    void        *closure ;
{
    os_hbucket *pb, *pe, b ;
    
    if ( h->weakp ) {
        for (pb = h->buckets, pe = pb + h->bsize ; pb < pe ; pb++ ) {
            for (b = *pb ; b ; b = b->next) {
                (*func) ((void *) REVEAL_POINTER (b->key)
                         , (void *) REVEAL_POINTER (b->item)
                         , closure) ;
            }
        }
    } else {
        for (pb = h->buckets, pe = pb + h->bsize ; pb < pe ; pb++ ) {
            for (b = *pb ; b ; b = b->next) {
                (*func) (b->key, b->item, closure) ;
            }
        }
    }
    return ;
}

static void add_entry (void *key, void *item, OsHashtable dest) {
    OsHashtablePut (dest, key, item) ;
    return ;
}

OsHashtable OsHashtableCopy (OsHashtable source) {
    OsHashtable cp ;

    /* Weak hashtable needs special support wr to the GC (in general, a
     * finalizer is registered on keys), so we do not allow them to be
     * copied */
    if ( source->weakp )
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("OsHashtableCopy")
                               , OsStringMake ("Try to copy a weakp table")) ;
    cp = OsHashtableMake (source->bsize, source->hash, source->cmp) ;
    (void) OsHashtableMap (source, add_entry, cp) ;
    return (cp) ;
}
    
/*
 * A few std cmp/hash function sets
 */

unsigned long OsHashCString (cstr)
    char *cstr ;
{
    unsigned long ret = 0;
    unsigned long ctr = 0;

    while (*(char*)cstr) {
        ret ^= *(char*)cstr++ << ctr;
        ctr = (ctr + 1) % sizeof (void *);
    }
    return (ret) ;
}

int OsCmpCString (s1, s2)
    char *s1, *s2 ;
{
    return (strcmp (s1, s2)) ;
}

unsigned long OsHashPointer (ptr)
    void *ptr ;
{
    return (((unsigned long) ptr) >> 1) ;
}

int OsCmpPointer (p1, p2)
    void *p1, *p2 ;
{
    return (((unsigned long) p1) - ((unsigned long) p2)) ;
}

unsigned long OsHashInteger (i)
    int i ;
{
    return ((unsigned long) i) ;
}

int OsCmpInteger (i, j)
    int i, j ;
{
    return (i - j) ;
}

/*
 * Subrs to hashtable:
 */

/* AUTODOC
 * (hashtable? OBJECT)
 * Is OBJECT a hashtable ?
 * Returns #t if the given OBJECT is a hashtable, #f otherwise.
 */

static OsObject OsSubrHashtableP (OsObject hash) {
    return ((OsObject) (OsIsHashtable(hash) ? OsA_True : OsA_Nil)) ;
}

/* AUTODOC
 * (make-hashtable)
 * Create an empty hashtable.
 * This function returns an empty hashtable object. 
 */

static unsigned long object_hashe (obj)
    OsObject obj ;
{
    return ((unsigned long) OsSendHashe (obj)) ;
}

static int object_cmpe (obj1, obj2)
    OsObject obj1, obj2 ;
{
    if ( OsSendEqual (obj1, obj2) == OsA_True )
        return (0) ;
    return (1) ;	/* no order on objects */
}

static OsObject OsSubrMakeHashtable () {
    return ((OsObject) OsHashtableMake (0, object_hashe, object_cmpe)) ;
/*    return ((OsObject) OsHashtableMake (0, OsHashPointer, OsCmpPointer)) ; */
}

/* AUTODOC
 * (hashtable-ref HASHTABLE KEY DEFAULT)
 * Retreive the item associated to KEY in HASHTABLE.
 * This function retreives the item associated to KEY by the given HASHTABLE.
 * If non is found, than return DEFAULT.
 */

static OsObject OsSubrHashtableRef (OsHashtable h, OsObject k, OsObject d) {
    OsObject item ;
    OsMustBeHashtable (h) ;
    if ( (item = (OsObject) OsHashtableGet (h, k)) )
        return (item) ;
    return (d) ;
}

/* AUTODOC
 * (hashtable-set! HASHTABLE KEY ITEM)
 * Associate KEY to ITEM in HASHTABLE.
 * Creates an new association from KEY to ITEM in the given HASHTABLE.
 */
   
static OsObject OsSubrHashtableSet (OsHashtable h, OsObject k, OsObject i) {
    OsMustBeHashtable (h) ;
    OsHashtablePut (h, k, i) ;
    return (i) ;
}

/* AUTODOC
 * (hashtable-map FUNCTION HASHTABLE)
 * Apply FUNCTION to every key, item couple found in HASHTABLE.
 * The given FUNCTION should take to paraneters, the first one is the key
 * and the second its corresponding association in the HASHTABLE.
 * This functions always returns #t
 */
  
static OsObject OsSubrHashtableMap (OsCodeBlock block, OsHashtable h) {
    os_hbucket *pb, *pe, b ;
    OsObject   args[2] ;
    
    OsMustBeHashtable (h) ;
    OsMustBeBlock (block) ;
    if ( h->weakp ) {
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("OsSubrHashtableMap")
                               , OsStringMake ("on weakp hashtable !")) ;
    }
    for (pb = h->buckets, pe = pb + h->bsize ; pb < pe ; pb++ ) {
        for (b = *pb ; b ; b = b->next) {
            args[0] = b->key ;
            args[1] = b->item ;
            (void) OsCodeBlockApplyWithArray (OsCurrentInterp()
                                              , block
                                              , 2, args
                                              , (OsObject *) 0) ;
        }
    }
    return (OsA_True) ;
}

/*
 * Initialize:
 */

void OsHashtableInitialize (OsInterp interp, int *pargc, char **argv) {
    /* WARNING: interp not initialized yet ! */
    OsHashtableClass = OsClassDeclare ("Hashtable", sizeof (OsHashtableRec)) ;
    return ;
}

void OsHashtableSubrInitialize (OsInterp interp) {
    OsSubrMake (interp, "hashtable?", OsSubrHashtableP, 1) ;
    OsSubrMake (interp, "make-hashtable", OsSubrMakeHashtable, 0) ;
    OsSubrMake (interp, "hashtable-ref", OsSubrHashtableRef, 3) ;
    OsSubrMake (interp, "hashtable-set!", OsSubrHashtableSet, 3) ;
    OsSubrMake (interp, "hashtable-map", OsSubrHashtableMap, 2) ;
    return ;
}
    
