/***************************************************************************
 * init.c
 * init.c,v 1.2 1995/03/17 14:21:47 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * Just the initialize stuff
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <oscheme.h>

static void scheme_initialize (OsInterp interp
                               , int *pargc, char **argv
                               , void (*before_profile_hook)()) {
    unsigned int argc    = 0 ;
    OsVector     vcmd    = OsVectorMake (0) ;
    OsVector     scripts = OsVectorMake (0) ;
    extern       OsSubrSysLoad () ;
    OsObject     error ;
    
    OsString profile = OsGetenv (OsStringMake ("OS_PROFILE")
                                 , OsStringMake (".oschemerc")) ;
/*
 * Get our own options:
 */
    while ( argc < *pargc ) {
        if ( ! strcmp ("-q", argv[argc]) ) {
            profile = (OsString) 0 ;
            OsArgvUsed (pargc, argc, argv) ;
        } else if ( ! strcmp ("-profile", argv[argc]) ) {
            OsArgvUsed (pargc, argc, argv) ;
            if ( profile )
                profile = OsStringMake (argv[argc]) ;
            OsArgvUsed (pargc, argc, argv) ;
        } else if ( ! strcmp ("-f", argv[argc]) ) {
            OsArgvUsed (pargc, argc, argv) ;
            (void) OsVectorFastAppend (scripts
                                       , (OsObject) OsStringMake (argv[argc]));
            OsArgvUsed (pargc, argc, argv) ;
        } else {
            (void) OsVectorFastAppend (vcmd
                                       , (OsObject) OsStringMake (argv[argc]));
            argc++ ;
        }
    }
/*
 * Make command line vector, and launch profile:
 */
    (void) OsAtomMake (interp, "*command-line*", (OsObject) vcmd) ;
    if ( before_profile_hook )
        (*before_profile_hook)(interp) ;
    if ( profile ) {
        OsCProtectExecute (interp
                           , (OsObject (*)()) OsPathLoad 
                           , (OsObject) profile
                           , &error) ;
        if ( error ) {
            OsSendDisplay(OsErrorFormat((OsVector) error),OsCurrentOutputPort);
            exit (1) ;
        }
    }
/*
 * Execute scripts:
 * The loader should now be defined, so we use it to solve pathnames
 */
    for (argc = 0 ; argc < scripts->csize ; argc++ ) {
        OsCProtectExecute (interp
                           , (OsObject (*)()) OsPathLoad
                           , scripts->cells[argc]
                           , &error) ;
        if ( error ) {
            OsSendDisplay(OsErrorFormat((OsVector) error),OsCurrentOutputPort);
            exit (1) ;
        }
    }
    return ;
}

OsInterp OsInitialize (pargc, argv, before_profile_hook)
    int  *pargc ;
    char **argv ;
    void (*before_profile_hook)() ;
{
    OsInterp interp ;
/*
 * First stage: initialize classes:
 * Because interp uses OsObject and OsHashtable objects, we must wait until
 * these classes get initialized before making the OsInterp object.
 * In a clean world, the right place to inititalize OsInterp would be between
 * first and second stage...
 */
    OsGCInitialize (interp, pargc, argv) ;
    OsObjectInitialize (interp, pargc, argv) ;
    OsHashtableInitialize (interp, pargc, argv) ;
    OsInterpInitialize (interp, pargc, argv) ;
    interp = OsInterpMake () ;
    OsSubrInitialize (interp, pargc, argv) ;
    OsVectorInitialize (interp, pargc, argv) ;
    OsCharInitialize (interp, pargc, argv) ;
    OsStringInitialize (interp, pargc, argv) ;
    OsIntegerInitialize (interp, pargc, argv) ;
    OsAtomInitialize (interp, pargc, argv) ;
    OsErrorInitialize (interp, pargc, argv) ;
    OsConsInitialize (interp, pargc, argv) ;
    OsInputPortInitialize (interp, pargc, argv) ;
    OsOutputPortInitialize (interp, pargc, argv) ;
    OsCodeBlockInitialize (interp, pargc, argv) ;
    OsCompInitialize (interp, pargc, argv) ;
    OsExpanderInitialize (interp, pargc, argv) ;
    OsUnixInitialize (interp, pargc, argv) ;
    OsContInitialize (interp, pargc, argv) ;
    OsRegexpInitialize (interp, pargc, argv) ;
    OsPointerInitialize (interp, pargc, argv) ;
    OsEObjInitialize (interp, pargc, argv) ;
    OsEObjStructInitialize (interp, pargc, argv) ;
    OsPathInitialize (interp, pargc, argv) ;
    OsFloatInitialize (interp, pargc, argv) ;
#ifdef OS_HAS_DLD
    OsDlInitialize (interp, pargc, argv) ;
#endif /* OS_HAS_DLD */
#ifdef OS_HAS_SIGNALS
    OsSignalsInitialize (interp, pargc, argv) ;
#endif /* OS_HAS_SIGNALS */

/*
 * Second stage: create predefined subrs:
 */
    OsObjectSubrInitialize (interp) ;
    OsCharSubrInitialize (interp) ;
    OsStringSubrInitialize (interp) ;
    OsConsSubrInitialize (interp) ;
    OsIntegerSubrInitialize (interp) ;
    OsVectorSubrInitialize (interp) ;
    OsAtomSubrInitialize (interp) ;
    OsHashtableSubrInitialize (interp) ;
    OsInputPortSubrInitialize (interp) ;
    OsCodeBlockSubrInitialize (interp) ;
    OsOutputPortSubrInitialize (interp) ;
    OsExpanderSubrInitialize (interp) ;
    OsSubrGCInitialize (interp) ;
    OsContSubrInitialize (interp) ;
    OsCompSubrInitialize (interp) ;
    OsErrorSubrInitialize (interp) ;
    OsRegexpSubrInitialize (interp) ;
    OsPointerSubrInitialize (interp) ;
    OsEObjSubrInitialize (interp) ;
    OsUnixSubrInitialize (interp) ;
#ifdef OS_HAS_DLD
    OsDlSubrInitialize (interp) ;
#endif
    OsDebugSubrInitialize (interp) ;
#ifdef OS_HAS_SIGNALS
    OsSignalsSubrInitialize (interp) ;
#endif /* OS_HAS_SIGNALS */
    OsPathSubrInitialize (interp) ;
    OsInterpSubrInitialize (interp) ;
    OsFloatSubrInitialize (interp) ;
#include "genini.c"
/*
 * Scheme-time initialization:
 */
    scheme_initialize (interp, pargc, argv, before_profile_hook) ;
    return (interp) ;
}
    
