;;***************************************************************************
;; autoload.scm
;; autoload.scm,v 1.1.1.1 1995/02/27 15:16:04 baird Exp
;; See the file COPYRIGHT foe licensing conditions.
;; AUTODOC MODULE
;; A simple autoload system.
;***************************************************************************

(require "system")

;; AUTODOC 
;; (autoload SYMBOL FILE)
;; Declare SYMBOL as being autolodable through FILE.
;; This function declares the atom SYMBOL to be solved in FILE: when an 
;; 'error:unbound is triggered by the system, it is caught. The given FILE is
;; loaded, and the error should be corrected.

(define (autoload sym file) (sys:autoloader 'autoload sym file))

(define (autoload:unbound-corrector sym) (sys:autoloader 'solve sym))

(define sys:autoloader
  (let ((autos (make-hashtable)))
    (define (autoload sym file) 
      (hashtable-set! autos sym file))
    (define (solve sym)
      (let ((file (hashtable-ref autos sym #f)))
        (and file (load file) sym)))
    (define (dispatch msg . args)
      (cond ((eq? msg 'autoload) (apply autoload args))
            ((eq? msg 'solve) (apply solve args))
            (else (display "Invalid command"))))
    dispatch))

;; AUTODOC
;; (autoload:init)
;; Initialize autoload, by loading available autoload files.
;; This function load  "initdir.scm" files of each load path directory.

(define (autoload:init)
  (for-each 
   (lambda (dir) 
     (let ((initdir (path-append (expand-path-for-tilde dir) "initdir.scm")))
       (and (file-exists? initdir) (load initdir))))
   (load-path)))

(autoload:init)

(provide "autoload") 
