;; lstutls.scm
;; lstutls.scm,v 1.1.1.1 1995/02/27 15:16:09 baird Exp
;; See the file COPYRIGHT foe licensing conditions.
;; AUTODOC MODULE
;; Various list utilities
;; Some of these utilitites are borrowed from Meroon code, by C. Queinnec.

;; AUTODOC AUTOLOAD
;; (list:remove-some! LIST PREDICATE)
;; Physically remove objects in LIST that match PREDICATE. 

(define (list:remove-some! lst pred)
  (if (null? lst)
      lst
      (let loop ((prev #f) (cur lst) (nxt (cdr lst)))
        (and (pred (car cur))
             (if prev (set-cdr! prev nxt) (set! lst nxt)))
        (if (pair? nxt) (loop cur nxt (cdr nxt)) lst))))

;; AUTODOC AUTOLOAD
;; (removeq! OBJ LIST)
;; Physically remove all object in LIST which are 'eq? to OBJ.

(define (removeq! lst obj)
  (list:remove-some! lst (lambda (i) (eq? obj i))))

;; AUTODOC AUTOLOAD
;; (reverse! LIST)
;; Physically reverse the given LIST.

(define (reverse! l)
  (define (nreverse l r)
    (if (pair? l)
        (let ((cdrl (cdr l)))
          (set-cdr! l r)
          (nreverse cdrl l) )
        r ) )
  (nreverse l '()) )

;; AUTODOC AUTOLOAD
;; (append! LIST1 LIST2)
;; Physically append LIST1 to LIST2.

(define (append! l1 l2)
  (cond ((pair? l1) 
         (if (pair? (cdr l1))
             (append! (cdr l1) l2)
             (set-cdr! l1 l2) )
         l1 )
        (else l2) ) )

;; AUTODOC AUTOLOAD
;; (remove ITEM LIST)
;; Builds a new list containing all elements of LIST not equal? to ITEM.

(define (remove item list)
  (if (pair? list)
      (if (equal? item (car list))
          (cdr list)
          (cons (car list) (remove item (cdr list))) )
      list ) )

;; AUTODOC AUTOLOAD
;; (last-pair LIST)
;; Returns last pair of LIST.
;; This function triggers an error if LIST is not a pair.

(define (last-pair l)
  (if (pair? l)
      (let scan ((cdrl (cdr l)) (result l))
        (if (pair? cdrl)
            (scan (cdr cdrl) cdrl)
            result))
      (error "last-pair: %0 not a pair." l)) )

(define (caddr l) (car (cddr l)))
(define (cdddr l) (cdr (cddr l)))
(define (cadddr l) (cadr (cddr l)))
(define (cddddr l) (cddr (cddr l)))

(provide "lstutls")
