;; r4rs.scm
;; r4rs.scm,v 1.1 1995/03/17 14:25:08 baird Exp
;; AUTODOC MODULE
;; Scheme R4RS procedures that can easily be defined in scheme.

;; AUTODOC AUTOLOAD
;; (call-with-current-continuation THUNK)
;; Grab current continuation and apply THUNK on it.
;; See also 'call/cc.

(define call-with-current-continuation call/cc)

;; AUTODOC AUTOLOAD
;; (call-with-input-file FILE THUNK)
;; Apply THUNK to the input port obtained by opening FILE for input.
;; This procedure opens the file FILE, and apply THUNK on the resulting input
;; port. If the file can not be opened, an error is triggered, otherwise, if
;; THUNK returns, the underliyng input port is closed.

(define (call-with-input-file file thunk)
  (let* ((port  (open-input-file file))
         (value (thunk port)))
    (close-input-port port)
    value))

;; AUTODOC AUTOLOAD
;; (call-with-output-file FILE thunk)
;; Apply THUNK to the input port obtained by opening FILE for output.
;; This procedure opens the file FILE, and apply THUNK on the resulting output
;; port. If the file can not be opened, an error is triggered, otherwise, if
;; THUNK returns, the underliyng input port is closed.

(define (call-with-output-file file thunk)
  (let* ((port  (open-output-file file))
         (value (thunk port)))
    (close-output-port port)
    value))

(provide "r4rs")
