;;***************************************************************************
;; system.scm
;; system.scm,v 1.1.1.1 1995/02/27 15:16:08 baird Exp
;; See the file COPYRIGHT foe licensing conditions.
;; AUTODOC MODULE
;; Various file system utilities.
;***************************************************************************

;; AUTODOC AUTOLOAD
;; (mktemp)
;; Generate a temporary file name.
;; This function is actually poorly coded. Sometime, I'll come back to it.

(define (mktemp)
  (let loop ((idx 0))
    (let ((filename (string-append "/tmp/oscheme-tmp." (integer->string idx))))
      (if (file-exists? filename)
          (loop (+ idx 1))
          filename))) )

;; AUTODOC AUTOLOAD
;; (directory? PATH)
;; Is the given string PATH a directory.

(define (directory? path)
  (and (= (logand [(file-stat path) mode] 16384) 16384) #t))

;; AUTODOC AUTOLOAD
;; (link? PATH)
;; Is the given string PATH a link file.

(define (link? path)
  (and (= (logand [(file-stat path) mode] 40960) 32768) #t))

;; AUTODOC AUTOLOAD
;; (mkdir DIRECTORY)
;; Create given DIRECTORY.

(define (mkdir dir)
  (system (string-append "mkdir " dir)))

;; AUTOLOAD AUTODOC
;; (mv FILE DEST)
;; Move FILE to DEST.

(define (mv file dest)
  (system (string-append "mv " file " " dest)))

;; AUTODOC AUTOLOAD
;; (path-to-list PATH)
;; Splits string PATH into its list of components (last being the file name)
;; This function breaks a full path into its list of components. Once splitted
;; absolute pathes still have a leading / in their first component.

(define path-to-list
  (let ((split-re (regexp:make "^([^/]*)[/](.*)$")))
    (lambda (path)
      (define (to-list str-path)
        (let loop ((p str-path) (m (regexp:match split-re str-path 1 2)))
          (if m
              (or (and (string=? (car m) "") 
                       (loop (cadr m) (regexp:match split-re (cadr m) 1 2)))
                  (cons (car m) 
                        (loop (cadr m) (regexp:match split-re (cadr m) 1 2))))
              (list p))))
      (let ((path-list (to-list path)))
        (if (char=? #\/ (string-ref path 0))
            (cons (string-append "/" (car path-list)) (cdr path-list))
            path-list)))) )


;; AUTODOC AUTOLOAD
;; (path-append COMPONENTS*)
;; Build a path from the given list of string COMPONENTS.
;; This function creates a path by concatenating all the given components, 
;; making sure there is a / in proper places.

(define path-append
  (let ((needs-slash-re (regexp:make ".*/$")))
    (lambda comps
      (let loop ((c comps) (n (cdr comps)))
        (if (null? n)
            (car c)
            (let ((m (regexp:match needs-slash-re (car c))))
              (if m
                  (string-append (car c) (loop (cdr c) (cdr n)))
                  (string-append (car c) "/" (loop (cdr c) (cdr n))))))))))

;; AUTODOC AUTOLOAD
;; (dirname PATH)
;; Returns the directory for the given string PATH.

(define dirname ())

;; AUTODOC AUTOLOAD
;; (filename PATH)
;; Returns the file name of the given string PATH

(define filename ())

(let ((dir-re (regexp:make "(.*)/([^/]*)")))
  (set! dirname
        (lambda (filename)
          (let ((match (regexp:match dir-re filename 1)))
            (or (and match (car match)) ""))))
  (set! filename
        (lambda (filename)
          (let ((match (regexp:match dir-re filename 2)))
            (or (and match (car match)) filename)))))

(provide "system")
