/***************************************************************************
 * misc.c
 * misc.c,v 1.1.1.1 1995/02/27 15:15:40 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <oscheme.h>

/*
 * C API :
 */

void OsExprEval (expr, silent, perror)
    OsObject expr ;
    int      silent ;
    OsObject *perror ;

{
    OsObject    expobj, value, error ;
    OsCodeBlock block ;
    
    if ( perror )
        *perror = (OsObject) 0 ;
    expobj = OsCProtectExecute (OsCurrentInterp(), OsExpandExpr, expr, &error);
    if ( error ) {
        if ( perror )
            *perror = error ;
        if ( silent )
            return ;
        OsSendDisplay (OsErrorFormat ((OsVector) error), OsCurrentOutputPort) ;
        OsOutputPortWriteChar (OsCurrentOutputPort, '\n') ;
        OsOutputPortFlush (OsCurrentOutputPort) ;
        return ;
    }
    /* Compile */
    block  = (OsCodeBlock) OsCProtectExecute (OsCurrentInterp()
                                              , (OsObject (*)()) OsCompileExpr
                                              , expobj, &error) ;
    if ( error ) {
        if ( perror )
            *perror = error ;
        if ( silent )
            return ;
        OsSendDisplay (OsErrorFormat ((OsVector) error), OsCurrentOutputPort) ;
        OsOutputPortWriteChar (OsCurrentOutputPort, '\n') ;
        OsOutputPortFlush (OsCurrentOutputPort) ;
        return ;
    }
    /* Execute */
    value  = OsCodeBlockProtectExecute (OsCurrentInterp(), block, &error) ;
    if ( error ) {
        if ( perror )
            *perror = error ;
        if ( silent )
            return ;
        OsSendDisplay (OsErrorFormat ((OsVector) error), OsCurrentOutputPort) ;
        OsOutputPortWriteChar (OsCurrentOutputPort, '\n') ;
        OsOutputPortFlush (OsCurrentOutputPort) ;
        return ;
    }
    /* Print */
    if ( ! silent ) {
        OsOutputPortWriteCString (OsCurrentOutputPort, " = ") ;
        OsSendDisplay (value, OsCurrentOutputPort) ;
        OsOutputPortWriteChar (OsCurrentOutputPort, '\n') ;
        OsOutputPortFlush (OsCurrentOutputPort) ;
    }
    return ;
}

/*
 * Perform a single REP interaction:
 */

void OsRepl (OsInterp interp) {
    OsObject    obj, error ;
    
    OsOutputPortWriteCString (OsCurrentOutputPort, "=> ") ;
    OsOutputPortFlush (OsCurrentOutputPort) ;
    obj = OsCProtectExecute (interp
                             , (OsObject (*)()) OsRead
                             , (OsObject) OsCurrentInputPort
                             , &error) ;
    if ( obj == OsA_Eof ) {
        exit (0) ;
    }
    if ( error ) {
        OsSendDisplay (OsErrorFormat ((OsVector) error), OsCurrentOutputPort) ;
        OsOutputPortWriteChar (OsCurrentOutputPort, '\n') ;
        OsOutputPortFlush (OsCurrentOutputPort) ;
    } else {
        OsExprEval (obj, 0, (OsObject *) 0) ;
    }
    return ;
}

/*
 * Usefull to parse command lines inside modules:
 */

void OsArgvUsed (pargc, argc, argv)
    int  *pargc, argc ;
    char **argv ;
{
    if ( argc < (*pargc) - 1 ) {
        bcopy (argv+argc+1, argv+argc, sizeof(char *) * ((*pargc)-argc-1)) ;
    }
    (*pargc)-- ;
    return ;
}
