/***************************************************************************
 * obj.c
 * obj.c,v 1.4 1995/03/17 14:23:24 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * Manage types (as classes).
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <os_config.h>
#include <os_interp.h>
#include <os_obj.h>
#include <os_gc.h>
#include <os_string.h>
#include <os_atom.h>
#include <os_vector.h>
#include <os_error.h>
#include <os_cons.h>
#include <os_cblock.h>
#include <os_subr.h>
#include <os_port.h>
#include <os_arit.h>

OsClass OsMetaClass = (OsClass) 0 ;

static OsError ose_bad_impsel = (OsError) 0 ;

/*
 * Class management functions (C-Api):
 */

OsObject OsSelInvalidSelector (rcv, sel)
    OsObject rcv ;
    int      sel ;
{
    (void) OsErrorTrigger (1, ose_bad_impsel, OsIntegerMake (sel)) ;
    /* NOT REACHED */
}

OsObject OsSelEvalToSelf (self, sel)
    OsObject self ;
    int      sel ;
{
    return (self) ;
}

OsObject OsSelCompileConstant (self, sel, cenv)
    OsObject  self ;
    int       sel ;
    OsCompEnv cenv ;
{
    extern OsObject os_comp_constant() ;
    return (os_comp_constant (self, sel, cenv)) ;
}

static OsObject os_send_finalize (OsObject rcv, void *unused) {
    return (OsSendMsg (rcv, OS_SEL_FINALIZE, (rcv, OS_SEL_FINALIZE))) ;
}

static OsObject os_default_display (OsObject self, int s, OsOutputPort port) {
    char tmp[OS_DISPLAY_BUFSIZ] ;
    sprintf (tmp, "#[%s %p]", self->os_class->name, self) ;
    (void) OsOutputPortWriteCString (port, tmp) ;
    return (self) ;
}

static OsObject os_default_write (OsObject self, int s, OsOutputPort port) {
    char tmp[OS_DISPLAY_BUFSIZ] ;
    sprintf (tmp, "#[%s %p]", self->os_class->name, self) ;
    (void) OsOutputPortWriteCString (port, tmp) ;
    return (self) ;
}

static unsigned long os_default_hashe (self, sel)
    OsObject self ;
    int      sel ;
{
    return ((unsigned long) self) ;
}

OsClass OsClassDeclare (name,  isize)
    char         *name ;
    unsigned int isize ;
{
    OsClass  os_class = (OsClass) OsGCMallocAtomic (sizeof (OsClassRec)) ;
    int midx ;
    
    os_class->os_class = OsMetaClass ;
    os_class->name     = name ;
    os_class->isize    = isize ;
    for (midx = 0 ; midx < OS_SEL_LAST ; midx++ )
        os_class->methods[midx] = OsSelInvalidSelector ;
    os_class->methods[OS_SEL_HASHE]   = (OsMethod) os_default_hashe ;
    os_class->methods[OS_SEL_COMPILE] = OsSelCompileConstant ;
    os_class->methods[OS_SEL_DISPLAY] = os_default_display ;
    os_class->methods[OS_SEL_WRITE]   = os_default_write ;
    return (os_class) ;
}

OsObject OsObjectMake (os_class, collectable)
    OsClass os_class ;
    int     collectable ;
{
    OsObject obj ;
    if ( collectable ) {
        obj = (OsObject) OsGCMallocAtomic (os_class->isize) ;
        bzero (obj, os_class->isize) ;
    } else {
        obj = (OsObject) OsGCMalloc (os_class->isize) ;
        if ( os_class->methods[OS_SEL_FINALIZE] != OsSelInvalidSelector )
            OsGCRegisterFinalizer (obj, (void (*)()) os_send_finalize) ;
    }
    obj->os_class = os_class ;
    return (obj) ;
}

/*
 * Subr available on all objects
 */

static OsObject OsSubrEqP (OsObject o1, OsObject o2) {
    return ((o1 == o2) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrEqvP (OsObject o1, OsObject o2) {
    return (OsSendEqv (o1, o2)) ;
}

static OsObject OsSubrNot (OsObject o) {
    return ((o == OsA_False) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrEqualP (OsObject o1, OsObject o2) {
    return (OsSendEqual (o1, o2)) ;
}

static OsObject OsSubrTypeOf (OsObject obj) {
    return ((OsObject) obj->os_class) ;
}

/*
 * Msgs implementations:
 */

static OsObject os_class_display (OsClass cls, int sel, OsOutputPort port) {
    char tmp[OS_DISPLAY_BUFSIZ] ;
    sprintf (tmp, "#[Class %s]", cls->name) ;
    OsOutputPortWriteCString (port, tmp) ;
    return ((OsObject) cls) ;
}

/*
 * Initialize:
 */

void OsObjectInitialize (OsInterp interp, int *pargc, char **argv) {
    /* WARNING: interp not initialized yet ! */
    OsMetaClass = OsClassDeclare ("MetaClass", sizeof (OsClassRec)) ;
    OsDefineMsg (OsMetaClass, OS_SEL_DISPLAY, os_class_display) ;
    OsMetaClass->os_class = OsMetaClass ;
    return ;
}

void OsObjectSubrInitialize (OsInterp interp) {
    ose_bad_impsel = OsErrorMake (interp
                                  , "error:bad-imp-sel"
                                  , "invalid internal selector %0"
                                  , 1, OSE_SEVERE) ;
    OsSubrMake (interp, "eq?", OsSubrEqP, 2) ;
    OsSubrMake (interp, "eqv?", OsSubrEqvP, 2) ;
    OsSubrMake (interp, "not", OsSubrNot, 1) ;
    OsSubrMake (interp, "equal?", OsSubrEqualP, 2) ;
    OsSubrMake (interp, "type-of", OsSubrTypeOf, 1) ;
    return ;
}
