/***************************************************************************
 * os_cblock.h
 * os_cblock.h,v 1.3 1995/03/17 14:19:50 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * Code nlock stuff
 ***************************************************************************/

#ifndef _OS_CODE_BLOCK_H_
#define _OS_CODE_BLOCK_H_

/*
 * Various kind of codeblocks:
 * - simple code block
 * - closed code block (with env)
 * - cont code block (continuation)
 */

extern OsClass OsCodeBlockClass ;

#define OsIsCodeBlock(b)     (((OsObject)(b))->os_class == OsCodeBlockClass)
#define OsMustBeCodeBlock(b) (OsMustBeOfClass ((b), OsCodeBlockClass))

typedef struct _OsCodeBlock {
    OsClass       os_class ;
    OsString	  code ;
    OsVector      data ;
} OsCodeBlockRec, *OsCodeBlock ;

extern OsClass OsClosedBlockClass ;

#define OsIsClosedBlock(b)     (((OsObject)(b))->os_class==OsClosedBlockClass)
#define OsMustBeClosedBlock(b) (OsMustBeOfClass ((b), OsClosedBlockClass))
#define OsShouldBeClosedBlock(b) \
    ((b) = (OsClosedBlock) (OsMustBeOfClass((b), OsClosedBlockClass)))

typedef struct _OsClosedBlock { /* Can *only* add fields to codeblock */
    OsClass	os_class ;
    OsString    code ;
    OsVector    data ;
    OsObject    *env ;
} OsClosedBlockRec, *OsClosedBlock ;

extern OsClass OsContBlockClass ;

typedef struct _OsContBlock { /* Can *only* add fields to codeblock */
    OsClass       os_class ;
    OsString      code ;
    OsVector      data ;
    OsObject      acc ;
    OsObject      *env ;
    OsInterp      interp ;
    int           ridx ;
    OsObject      *stack ;
    unsigned int  stacklen ;
    OsObject      *cells ;
    unsigned char *pc ;
    char          *cstack ;
    char	  *cstackp ;
    jmp_buf	  jmpbuf ;
    unsigned int  cstack_size ;
    void          *protection ;
} OsContBlockRec, *OsContBlock ;

#define OsIsContBlock(c)     (((OsObject) (c))->os_class == OsContBlockClass)
#define OsMustBeContBlock(c) (OsMustBeOfClass ((c), OsContBlockClass))
#define OsShouldBeContBlock(c) \
    ((c) = (OsContBlock) OsMustBeOfClass ((c), OsContBlockClass))

extern OsClass OsWindBlockClass ;

typedef struct _OsWindBlock {
    OsClass 	  os_class ;
    OsString 	  code ;
    OsVector	  data ;
    unsigned long stamp ;
} OsWindBlockRec, *OsWindBlock ;

#define OsIsWindBlock(c)     (((OsObject) (c))->os_class == OsWindBlockClass)
#define OsMustBeWindBlock(c) (OsMustBeOfClass((c), OsWindBlockClass))
#define OsShouldBeWindBlock(c) \
    ((c) = (OsWindBlock) OsMustBeOfClass ((c), OsWindBlockClass))

#define OsMustBeBlock(o) \
    ( ((((OsObject) (o))->os_class == OsCodeBlockClass) \
       || (((OsObject) (o))->os_class == OsClosedBlockClass) \
       || (((OsObject) (o))->os_class == OsContBlockClass) \
       || (((OsObject) (o))->os_class == OsWindBlockClass)) \
     ? ((OsObject) o) \
     : OsErrorTrigger (1, OsE_BadFunction, o) )

#define OsIsBlock(o) \
    ( (((OsObject) (o))->os_class == OsCodeBlockClass) \
       || (((OsObject) (o))->os_class == OsClosedBlockClass) \
       || (((OsObject) (o))->os_class == OsContBlockClass) \
       || (((OsObject) (o))->os_class == OsWindBlockClass))

/* 
 * Break points:
 */

extern OsClass OsBreakPointClass ;

typedef struct _OsBreakPoint {
    OsClass	  os_class ;
    OsCodeBlock   block ;
    OsObject      expr ;
    OsObject      condition ;
    unsigned char *iloc ;
} OsBreakPointRec, *OsBreakPoint ;

#define OsIsBreakPoint(b)    (((OsObject) (b))->os_class == OsBreakPoitnClass)
#define OsMustBeBreakPoint(b) (OsMustBeOfClass((b), OsBreakPointClass))
#define OsShouldBeBreakPoint(b) \
    ((b) = (OsBreakPoint) OsMustBeOfClass ((b), OsBreakPointClass))

/*
 * Externals:
 */

OsCodeBlock       OsCodeBlockMake (unsigned char *c
                                   , unsigned long l
                                   , OsVector d) ;
OsClosedBlock     OsClosedBlockMake (unsigned char *c
                                     , unsigned long l
                                     , OsVector d
                                     , OsObject *env) ;
    
OsCodeBlock       OsCompileExpr OS_P ((OsObject expr)) ;
OsObject          OsCodeBlockExecute OS_P ((OsInterp interp
                                            , OsCodeBlock blok)) ;
    
OsObject OsCodeBlockApplyWithList OS_P ((OsInterp interp
                                         , OsCodeBlock block
                                         , OsCons list
                                         , OsObject *perror)) ;

OsObject OsCodeBlockApplyWithArray OS_P ((OsInterp interp
                                          , OsCodeBlock block
                                          , unsigned int argc
                                          , OsObject *argv
                                          , OsObject *perror)) ;

void     os_uncatched_error OS_P ((OsObject error)) ;

OsObject OsCodeBlockProtectExecute OS_P ((OsInterp interp
                                          , OsCodeBlock blok
                                          , OsObject *perror));
OsObject OsCProtectExecute OS_P ((OsInterp interp
                                  , OsObject(*c_function)()
                                  , OsObject argument
                                  , OsObject *perror));
OsObject OsWindProtectExecute OS_P ((OsInterp interp
                                     , OsCodeBlock enter
                                     , OsCodeBlock block
                                     , OsObject *perror)) ; 
OsVector OsCurrentVMStateToError OS_P((OsInterp interp)) ;
    
void OsCodeBlockInitialize OS_P ((OsInterp interp, int *pargc, char **argv)) ;
void OsCodeBlockSubrInitialize OS_P ((OsInterp interp)) ;

void OsCompInitialize OS_P ((OsInterp interp, int *pargc, char **argv)) ;    
void OsCompSubrInitialize OS_P ((OsInterp interp)) ;

void OsExpanderInitialize OS_P ((OsInterp interp, int *pargc, char **argv)) ;
void OsExpanderSubrInitialize OS_P ((OsInterp interp)) ;
    
/***************************************************************************
 * Now, the *really* interesting stuff:
 * This describes the byte-code used for the virtual machine. I wanted
 * to find a good encoding scheme for instructions (which would make
 * the decompile op easier), I'm afraid there is not enough room in a single
 * byte for fancy stuff.
 ***************************************************************************/

#define OS_B_NOOP		0
    
/*
 * Accessing local variables:
 */

#define OS_B_LREF0		1	/* (char vidx) */
#define OS_B_LREF1		2       /* (char vidx) */
#define OS_B_LREF2		3       /* (char vidx) */
#define OS_B_LREF3		4       /* (char vidx) */
#define OS_B_LREFN		5       /* (char fidx, char vidx) */
#define OS_B_LREF0PUSH		6       
#define OS_B_LREF1PUSH		7
#define OS_B_LREF2PUSH		8
#define OS_B_LREF3PUSH		9
#define OS_B_LREFNPUSH		10
    
#define OS_B_LSET0		30
#define OS_B_LSET1		31
#define OS_B_LSET2		32
#define OS_B_LSET3		33
#define OS_B_LSETN		34

/*
 * Accessing global variables:
 */
    
#define OS_B_GREF		50	/* (short data_idx) */
#define OS_B_GREFPUSH		51	
#define OS_B_GSET		52
#define OS_B_GDEF		53
    
/*
 * Accessing constants (always through the 'data segment'
 */
    
#define OS_B_CST		60      /* (char data_idx) */
#define OS_B_CSTPUSH		61      /* (char data_idx) */

/*
 * Manipulating the environment register/frames
 */
       
#define OS_B_LINK		70	/* (char arity) */
#define OS_B_LINKN		71	/* (char arity) */
#define OS_B_UNLINK		72	/* (char how-much) */
#define OS_B_LINKEXT		73	/* link closed env (void) */
#define OS_B_LINKNEXT		74
#define OS_B_UNLINKEXT		75	/* unlink closed env (void) */
    
/*
 * Control flow stuff
 */

#define OS_B_SAVE		80	/* (short skip) */
#define OS_B_JUMP		81	/* (char arity) */
#define OS_B_UNLINKJUMP		82	/* (char arity, char envsize) */
#define OS_B_UNLINKEXTJUMP	83    
#define OS_B_RETURN		84
#define OS_B_EXIT		85
#define OS_B_BRIFFALSE		86	/* (short relidx) */
#define OS_B_BRIFTRUE		87	/* (short relidx) */
#define OS_B_RGOTO		88	/* (short relidx) */
#define OS_B_SAVECONT		89	/* (void) */
#define OS_B_RETCONT		90 	/* (short dataidx) */
#define OS_B_SAVEWIND		91
#define OS_B_RETWIND		92
    
/*
 * Stack and 'acc' manipulations:
 */

#define OS_B_PUSH		100
#define OS_B_POP		101
#define OS_B_LOAD		102 	/* (unsigned char arity) */

/*
 * cfunc applications
 * The one with the leading A(rity) do check arity with against the ridx
 * register (default). When the compiler is able to predict correct arity
 * it uses the non A version (which allows to bypass this test).
 * There is of course no non-A version of SUBRN
 */

#define OS_B_ASUBR0		110	/* (short data_idx) */
#define OS_B_ASUBR1		111	
#define OS_B_ASUBR2		112
#define OS_B_ASUBR3		113
#define OS_B_ASUBR4		114
#define OS_B_ASUBR5		115
#define OS_B_ASUBR6		116
#define OS_B_ASUBR7		117
#define OS_B_ASUBRN		118
    
#define OS_B_SUBR0		120	/* (short data_idx) */
#define OS_B_SUBR1		121	
#define OS_B_SUBR2		122
#define OS_B_SUBR3		123
#define OS_B_SUBR4		124
#define OS_B_SUBR5		125
#define OS_B_SUBR6		126
#define OS_B_SUBR7		127
    
/*
 * Closing functions in current env.
 */

#define OS_B_CLOSE		130	/* (void) */

/*
 * Messaging instructions:
 */

#define OS_B_SENDMSG0		140
#define OS_B_SENDMSG1		141
#define OS_B_SENDMSG2		142
#define OS_B_SENDMSG3		143
#define OS_B_SENDMSG4		144
#define OS_B_SENDMSG5		145
#define OS_B_SENDMSG6		146
#define OS_B_SENDMSG7		147
#define OS_B_SENDNMSG		148
    
#define OS_B_SENDSUPERMSG0      150
#define OS_B_SENDSUPERMSG1	151	
#define OS_B_SENDSUPERMSG2	152
#define OS_B_SENDSUPERMSG3	153
#define OS_B_SENDSUPERMSG4	154
#define OS_B_SENDSUPERMSG5	155
#define OS_B_SENDSUPERMSG6	156
#define OS_B_SENDSUPERMSG7	157
#define OS_B_SENDSUPERNMSG	158

/*
 * Debugging (those instructions are emitted only if scheme code is compiled
 * with the -g option
 */

#define OS_DBG_PUSH		160
#define OS_DBG_PUSHBRK		161
#define OS_DBG_SAFEPUSH		162
#define OS_DBG_SAFEPUSHBRK	163
#define OS_DBG_POP		164
    
/*
 * This should be defined to the last available opcode
 */

#define OS_B_OPLAST		165

#endif /* _OS_CODE_BLOCK_H_ */
