/***************************************************************************
 * os_config.h
 * os_config.h,v 1.2 1995/03/07 02:21:10 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 ***************************************************************************/

#ifndef _OS_CONFIG_H_
#define _OS_CONFIG_H_

/*
 * Compiler options (config/template.make, part two, A):
 */

#ifndef REGISTER
#    define REGISTER	   register
#endif
#ifndef VOLATILE
#    define VOLATILE       volatile
#endif
#ifndef INLINE
#    define INLINE    	   inline
#endif
#ifndef OS_STACK_GROWS
#    define OS_STACK_GROWS OS_STACK_DOWN
#endif
#ifdef OS_HAS_PROTOTYPES
#    define OS_P(_argdecl) _argdecl
#else
#    define OS_P(x) ()
#endif

/*
 * System options (config/template.make, part two, B):
 */

#ifndef OS_MACHINE_TYPE
#    define OS_MACHINE_TYPE "unknown"
#endif
#ifndef OS_SYSTEM_TYPE
#    define OS_SYSTEM_TYPE  "unknown"
#endif
#ifndef OS_PATH_SEPARATOR
#    define OS_PATH_SEPARATOR	'/'
#endif
#ifndef OS_PATH_HOME
#    define OS_PATH_HOME	'~'
#endif
#include "genpath.h"
#ifndef OS_DEFAULT_EXTENSION
#    define OS_DEFAULT_EXTENSION ":.scm"
#endif

#ifdef OS_NEEDS_BCOPY
#define bcopy(_s,_d,_sz)	memcpy((_d), (_s), (_sz))
#define bzero(_s,_sz)		memset((_s), (char) 0, (_sz))
#endif /* OS_NEEDS_BCOPY */

#ifdef OS_HAS_SIGNALS
#    ifndef OS_SIGNALS_MAX_PENDING
#        define OS_SIGNALS_MAX_PENDING	32
#    endif
#endif

/*
 * GC options:
 * - GC_DEBUG: is the original Boehm's GC flag to get help when memory
 *   management may get crazy. (DO NOT TURN IT ON, due to some bug...)
 * - GC_MACRO_GC_CALL: All call to Boehm's GC will be turned to macro instead
 *   of function calls.
 */

#undef  GC_DEBUG
#define OS_MACRO_GC_CALL

#define OS_DEFAULT_HEAPSIZE		(1024*256)


/*
 * Vector configuration:
 * - OS_VECTOR_INCREMENT: how many new cells should be allocated at a time
 *     when a vector need to be grown.
 */

#define OS_VECTOR_INCREMENT 		8

#define OS_HASHTABLE_DEFAULT_SIZE	4

/*
 * Integer configuration:
 */

#define OS_LOWER_INTEGER_PREDEFINED	-20
#define OS_HIGHEST_INTEGER_PREDEFINED	100

#define OS_PARSER_MXUNGET		64
#define OS_PARSER_BUFSIZE		4096
#define OS_PARSER_STKSIZE		512

/*
 * Compiler options:
 */

#define OS_COMPILER_DEFAULT_CODELEN	4096
#define OS_COMPILER_DEFAULT_DATALEN	256
#define OS_COMPILER_CODE_INCREMENT	2048
#define OS_COMPILER_TRACEABLE

/*
 * Machine configuration:
 * - OS_VM_TRACEABLE: trace each instruction before execution. This can
 *    be controled through the 'sys:trace-machine value (#t means trace).
 * - OS_VM_TIMEABLE: times the time spend in executing everyblock. This can
 *    be controled through the 'sys:times-machine value (#t means times)
 * - OS_VM_STACKWSIZE: default stack size for vm (in Words)
 * - OS_VM_CHECKSTACK: check stack (debug only), check between each vm
 *   instruction that stack pointer is still correct.
 */

#undef  OS_VM_TRACEABLE
#define OS_VM_TIMEABLE
#define OS_VM_STACKWSIZE		(8192)
#undef  OS_VM_CHECKSTACK

/*
 * Object Model defines
 * - OS_EOBJ_STAT: if you want memory usage stats through '(obj:stat)
 */

#define OS_EOBJ_STATS

/*
 * Check that given (unsigned long) can safely be converted to (unsigned short)
 */

#define OS_CHECK_SHORT(u) ((u) & 0xFFFF0000)

/*
 * Put a short at given adress (changinbg something here means you have to
 * change someting in os_emit_short
 */

#define OS_PUT_SHORT(l, s) \
    *l++ = ((unsigned char) (((s) & 0xFF00) >> 8)) ; \
    *l++ = ((unsigned char) ((s) & 0xFF)) 

/*
 * Get a short:
 */

#define OS_GET_SHORT(l) \
    (l+=2, ((((unsigned short) *(l-2)) << 8) | ((unsigned short) *(l-1))))

#endif /* _OS_CONFIG_H_ */
