/***************************************************************************
 * os_eobjP.h
 * os_eobjP.h,v 1.1.1.1 1995/02/27 15:15:53 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * Private part of external object implementations. Hackers only.
 ***************************************************************************/

#ifndef _OS_EOBJ_P_H_
#define _OS_EOBJ_P_H_


/*
 * Forward declarations
 */

typedef struct _os_sarrayS *os_sarrayS ;
typedef struct _os_sarrayB *os_sarrayB ;
typedef struct _os_sarrayC *os_sarrayC ;
typedef struct _os_sarray  *os_sarray ;

/*
 * Selector implementation are the leaf of hash-tries (see below).
 * They describe how sending a message to an object should be interpreted.
 */

typedef struct _os_selimp {
    unsigned char	type ;
} os_selimp_rec ;

typedef struct _os_selimp_cfun {
    unsigned char 	type ;
    OsObject 		(*cfun)() ;
    void		*closure ;
} os_selimp_cfun_rec, *os_selimp_cfun ;

#define SELIMP_CFUN(s)	  (((os_selimp_cfun) (s))->cfun)
#define SELIMP_CLOSURE(s) (((os_selimp_cfun) (s))->closure)

typedef struct _os_selimp_clos {
    unsigned char	type ;
    OsObject		block ;
} os_selimp_clos_rec, *os_selimp_clos ;

#define SELIMP_BLOCK(s)	  ((OsCodeBlock) ((os_selimp_clos) (s))->block)

typedef struct _os_selimp_slot {
    unsigned char	type ;
    unsigned int	didx ;
} os_selimp_slot_rec, *os_selimp_slot ;

#define SELIMP_DIDX(s)	  (((os_selimp_slot) (s))->didx)

#define OS_SELIMP_CFUN	1
#define OS_SELIMP_CLOS	2
#define OS_SELIMP_SLOT	3

#define make_selimp_cfun(f,c) \
    ((os_selimp_cfun) os_selimp_make (OS_SELIMP_CFUN, (void *) f, (void *) c))
#define make_selimp_clos(c) \
    ((os_selimp_clos) os_selimp_make (OS_SELIMP_CLOS, (void *) c, (void *) 0))
#define make_selimp_slot(n) \
    ((os_selimp_slot) os_selimp_make (OS_SELIMP_SLOT, (void *) n, (void *) 0))

/*
 * Hash trie layout
 * Hash trie maps selector to selector implementation for a hidden class. Both
 * hash-trie themselves and their different bucket levels can be shared.
 */
    
#define SA_DUCKET_BITS		0xF000
#define SA_CUCKET_BITS		0x0F00
#define SA_BUCKET_BITS		0x00F0
#define SA_SELECTOR_BITS	0x000F

#define SA_DUCKET_SIZE		16
#define SA_CUCKET_SIZE		16
#define SA_BUCKET_SIZE		16
#define SA_SELECTOR_SIZE	16

typedef struct _os_sarrayS {
    unsigned long version ;
    os_selimp     selectors[SA_SELECTOR_SIZE] ;
} os_sarrayS_rec ;

typedef struct _os_sarrayB {
    unsigned long version ;
    os_sarrayS    buckets[SA_BUCKET_SIZE] ;
} os_sarrayB_rec ;

typedef struct _os_sarrayC {
    unsigned long version ;
    os_sarrayB    cuckets[SA_CUCKET_SIZE] ;
} os_sarrayC_rec ;

typedef struct _os_sarray {
    unsigned long version ;
    os_sarrayC    empty_cucket ;
    os_sarrayB    empty_bucket ;
    os_sarrayS    empty_selector ;
    os_sarrayC    duckets[SA_DUCKET_SIZE] ;
} os_sarray_rec ;

typedef struct _os_iclass {
    unsigned long version ;
    unsigned long shl ;
    OsEObj        canonical ;		/* back to canonical prototype */
    os_iclass     parent ;
    OsVector      subclasses ;
    os_sarray     msg ;
} os_iclass_rec ;

/*
 * Split a selector OsObject into its various parts.
 */

typedef struct _selector_ {
    union {
	unsigned short id ;
	struct {
	    unsigned s: 4 ;
	    unsigned b: 4 ;
	    unsigned c: 4 ;
	    unsigned d: 4 ;
	} dcbs ;
    } split ;
} _selector ;
	
#define SplitSelector(_id,_d,_c,_b,_s) \
{ \
      _selector the_sel ; \
      the_sel.split.id = (unsigned short) (_id) ; \
      _c = (unsigned int) the_sel.split.dcbs.c ; \
      _d = (unsigned int) the_sel.split.dcbs.d ; \
      _b = (unsigned int) the_sel.split.dcbs.b ; \
      _s = (unsigned int) the_sel.split.dcbs.s ; \
 }

static INLINE os_iclass os_iclass_update (OsEObj obj)
{
    os_iclass    newclass ;
    OsVector     stack = OsVectorMake (0) ;
    OsObject     *pstack, *estack ;
    /* Our objclass might need to be updated, go up to the first canonical
     * prototype, and propagate down the new class shl. */
    while ( obj->objclass->shl != obj->objclass_shl ) {
        (void) OsVectorFastAppend (stack, (OsObject) obj) ;
        obj = obj->prototype ;
    }
    newclass = obj->objclass ;
    pstack = stack->cells ;
    estack  = pstack + stack->csize ;
    while ( pstack < estack ) {
        ((OsEObj) *pstack)->objclass_shl = newclass->shl ;
        ((OsEObj) *pstack++)->objclass   = newclass ;
    }
    return (newclass) ;
}

#define EObjIclass(o)					\
    (((o)->objclass_shl != ((o)->objclass->shl))	\
     ? os_iclass_update (o) 				\
     : (o)->objclass) 

static INLINE os_selimp os_sarray_get (sa, sel_id)
    os_sarray sa ;
    OsEObjSelId  sel_id ;
{
    unsigned int d, c, b, s ;

    SplitSelector (sel_id, d, c, b, s) ;
    return (sa->duckets[d]->cuckets[c]->buckets[b]->selectors[s]) ;
}

#define EObjSelimpLookup(o,sid) os_sarray_get(EObjIclass(o)->msg, (sid))
    
OsEObjSel os_w_to_r_sel OS_P ((OsEObjSel sel)) ;

#endif /* _OS_EOBJ_P_H_ */
