/***************************************************************************
 * os_error.h
 * os_error.h,v 1.1.1.1 1995/02/27 15:15:19 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 ***************************************************************************/

#ifndef _OS_ERROR_H_
#define _OS_ERROR_H_

extern OsClass OsErrorClass ;

#define OsIsError(e)	 (((OsObject) (e))->os_class == OsErrorClass)
#define OsMustBeError(e) (OsMustBeOfClass((e), OsErrorClass))
#define OsShouldBeError(e) \
    ((e) = (OsError) OsMustBeOfClass ((e), OsErrorClass))

typedef struct _OsError {
    OsClass      os_class ;
    unsigned int severity ;
    unsigned int arity ;
    OsString     fmt ;
} OsErrorRec, *OsError ;

/*
 * Different level for errors:
 * OSE_FATAL      : error cannot be treated in any reasonable way,
 * OSE_SEVERE     : just display error, cannot do better
 * OSE_CATCHABLE  : error can be caught at the scheme level,
 * OSE_REPAIRABLE : error can be corrected at the scheme level,
 */

#define OSE_FATAL	0
#define OSE_SEVERE	1
#define OSE_CATCHABLE   2
#define OSE_REPAIRABLE  4
#define OSE_NORMAL      (OSE_CATCHABLE | OSE_REPAIRABLE)

OsError  OsErrorMake OS_P ((OsInterp interp
                            , char *error_name
                            , char *error_format
                            , unsigned int arity
                            , unsigned int severity));

OsObject OsErrorAugment () ;	/* Warn: varargs here */
OsObject OsErrorTrigger () ;	/* Warn: varargs here */

OsString OsErrorFormat OS_P ((OsVector errv)) ;
    
void     OsFatalError OS_P ((char *msg)) ;

void     OsErrorInitialize OS_P ((OsInterp interp, int *pargc, char **argv)) ;
void     OsErrorSubrInitialize OS_P ((OsInterp interp)) ;
    
/*
 * List of pre-defined errors:
 */

extern OsObject OsE_Unbound ;
extern OsObject OsE_ArityShouldBe ;
extern OsObject OsE_BadFunction ;
extern OsObject OsE_BadType ;
extern OsObject OsE_Implementation ;
extern OsObject OsE_BadSyscall ;

#endif /* _OS_ERROR_H_ */
