/***************************************************************************
 * os_htable.h
 * os_htable.h,v 1.2 1995/03/12 02:58:12 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 ***************************************************************************/

#ifndef _OS_HASHTABLE_H_
#define _OS_HASHTABLE_H_

extern OsClass OsHashtableClass ;

#define OsIsHashtable(h)     (((OsObject) (h))->os_class == OsHashtableClass)
#define OsMustBeHashtable(h) (OsMustBeOfClass ((h), OsHashtableClass))

typedef unsigned long (*OsHashFct) OS_P((void *key)) ;
typedef int (*OsCmpFct) OS_P ((void *key1, void *key2)) ;

typedef struct _os_hbucket *os_hbucket ;

typedef struct _OsHashtable {
    OsClass		os_class ;
    int                 weakp ;		/* TRUE if hash is weak */
    unsigned int	bused ;		/* # of used buckets */
    unsigned int	bsize ;		/* size of bucket array */
    OsHashFct           hash ;		/* hash function */
    OsCmpFct		cmp ;		/* compare function */
    os_hbucket		*buckets ;
} OsHashtableRec, *OsHashtable ;

/*
 * Externals:
 */

void *OsHashtablePut OS_P ((OsHashtable htable, void *key, void *item)) ;
void *OsHashtableGet OS_P ((OsHashtable htable, void *key)) ;
void OsHashtableRemove OS_P ((OsHashtable htable, void *key)) ;
void *OsHashtableFindKey OS_P ((OsHashtable htable, void *item)) ;

void OsHashtableMap OS_P ((OsHashtable htable, void (*cfun)(), void *closure));

unsigned long OsHashCString OS_P ((char *cstr)) ;
int           OsCmpCString OS_P ((char *s1, char *s2)) ;
unsigned long OsHashPointer OS_P ((void *p)) ;
int           OsCmpPointer OS_P ((void *p1, void *p2)) ;
unsigned long OsHashInteger OS_P ((int i)) ;
int           OsCmpInteger OS_P ((int i, int j)) ;
    
OsHashtable OsHashtableMake OS_P ((unsigned int s, OsHashFct h, OsCmpFct c));
OsHashtable OsWeakHashtableMake OS_P ((unsigned int size
                                       , OsHashFct hash
                                       , OsCmpFct cmp));
OsHashtable OsHashtableCopy OS_P ((OsHashtable source)) ;

void OsHashtableInitialize OS_P ((OsInterp interp, int *pargc, char **argv)) ;
void OsHashtableSubrInitialize OS_P ((OsInterp interp)) ;
    
#endif /* _OS_HASHTABLE_H_ */
