/***************************************************************************
 * os_interp.h
 * os_interp.h,v 1.2 1995/03/12 02:58:14 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 ***************************************************************************/

#ifndef _OS_INTERP_H_
#define _OS_INTERP_H_

typedef struct _OsObject *OsObject ;
typedef struct _OsClass  *OsClass ;

extern OsClass OsInterpClass ;

#define OsIsInterp(i)      (((OsObject) (i))->os_class == OsInterpClass)
#define OsMustBeInterp(i)  (OsMustBeOfClass((i), OsInterpClass))
#define OsShouldBeInterp(i) \
    ((i) = (OsInterpClass) OsMustBeOfClass ((i), OsInterpClass))
    
typedef struct _OsInterp {
    OsClass             os_class ;
    OsObject 		*stack, *sp, *stacke ;	/* OsObject, in fact */
    OsObject 		*dbg_stack, *dbg_sp ;	/* OsObject, in fact */
    int      		ridx ;
    void     		*name_space ;		/* OsHashtable, in fact */
    void     		*macro_space ;		/* OsHashtable, in fact */
    void                *path_list ;		/* OsCons, load-path */
    void                *ext_list ;		/* OsCons, extension list */
    void                *prov_list ;		/* OsCons provided list */
    void                *hooked_ext ;		/* OsHashtable, ext. hooks */
    unsigned long	read_bytes ;		/* # of readen bytes */
    unsigned long       lread_bytes ;
    unsigned long       written_bytes ;		/* # of written bytes */
    unsigned long       lwritten_bytes ;
    unsigned long	clicks ;		/* # of computing clicks */
    unsigned long       lclicks ;
    int                 security ;		/* security level */
    OsObject 		error_handler ;		/* error handler */
} OsInterpRec, *OsInterp ;

/*
 * Security levels:
 * This is really a try, it might change both in the way subrs are catgorize
 * and in the semantics of levels:
 * LEVEL_F:
 *    All dangerous operations are disabled. 
 * LEVEL_E:
 *    Operations dangerous at the scheme level are allowed, this means for 
 *    example, that you can perform eval, call/cc...
 * LEVEL_D:
 *    Scheme IO input operations allowed
 * LEVEL_C:
 *    Full IO operations are allowed (input/output)
 * LEVEL_B: 
 *    System calls that are dangerous are allowed: chdir, directory, sockets...
 * LEVEL_A:
 *    System calls allowing to create processes are allowed
 * Orthogonaly to the security level that controls the set of primitives, 
 * all oscheme scripts falls under the control of:
 * - # of read bytes,
 * - # of written bytes,
 * - # of clicks
 */

#define OS_SECURITY_HIGHEST	0
#define OS_SECURITY_LEVEL_A	25
#define OS_SECURITY_LEVEL_B	50
#define OS_SECURITY_LEVEL_C	75
#define OS_SECURITY_LEVEL_D	100
#define OS_SECURITY_LEVEL_E	150
#define OS_SECURITY_LEVEL_F	200
#define OS_SECURITY_LOWEST	1024

/*
 * Externals:
 */

OsInterp OsInterpMake OS_P ((void)) ;
OsInterp OsInterpClone OS_P ((OsInterp interp
                              , int stack_size
                              , int security
                              , unsigned long lread_bytes
                              , unsigned long lwritten_bytes
                              , unsigned long lclicks)) ;

OsInterp OsCurrentInterp OS_P ((void)) ;
OsInterp OsInterpSwitch OS_P ((OsInterp source)) ;

unsigned long OsInterpReadBytesExceeded OS_P ((OsInterp interp)) ;
unsigned long OsInterpWrittenBytesExceeded OS_P ((OsInterp interp)) ;
void          OsInterpStackSizeExceeded OS_P ((OsInterp interp)) ;

/*
 * Security check macros:
 */

#ifdef OS_SECURE
#   define OsInterpCheckReadBytes(_i, _cnt) \
    ( (((_i)->read_bytes += (_cnt)) >= (_i)->lread_bytes) \
      ? OsInterpReadBytesExceeded (_i) \
      : ((unsigned long) (_cnt)))
#   define OsInterpCheckWrittenBytes(_i, _cnt) \
    ( (((_i)->written_bytes += (_cnt)) >= (_i)->lwritten_bytes) \
      ? OsInterpWrittenBytesExceeded (_i) \
      : (unsigned long) (_cnt))
#   define OsInterpCheckClicks(_i,_cost) \
    ( (((_i)->clicks += (_cost)) >= (_i)->lclicks) \
      ? OsInterpClicksExceeded (_i) \
      : (unsigned long) (_cost))
#   define OsInterpCheckSecurity(_i,_r) \
    ( ((_i)->security > (_r)) \
      ? OsInterpSecurityFailed (_i) \
      : (_r))
#else /* OS_SECURE */
#   define OsInterpCheckReadBytes(_i,_c)	(_c)
#   define OsInterpCheckWrittenBytes(_i,_c)	(_c)
#   define OsInterpCheckClicks(_i,_c)		(_c)
#   define OsInterpCheckSecurity(_i,_r)		(_r)
#endif /* OS_SECURE */

#endif /* _OS_INTERP_H_ */
