/***************************************************************************
 * os_obj.h
 * os_obj.h,v 1.3 1995/03/17 14:32:14 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * Okay, this is not really an object model, but looks much more like
 * an array of cfuncs. I have intentionnaly fixed the number of 'selectors'
 * because this object model (hu !) should not be extended. Use the external
 * object model for extended object. This will cost you one function call
 * but will delimit the kernel cleanly.
 ***************************************************************************/

#ifndef _OS_OBJECT_H_
#define _OS_OBJECT_H_

/*
 * The list of messages an OsObject should respond to:
 */

#define OS_SEL_EQUAL		0
#define OS_SEL_HASHE		1
#define OS_SEL_DISPLAY		2
#define OS_SEL_WRITE		3
#define OS_SEL_FINALIZE		4
#define OS_SEL_COMPILE		5

#define OS_SEL_LAST    		6

/*
 * Object & Classes :
 * Classes have a MetaClass, which allows us to use classes in the lang
 */

typedef struct _OsObject {
    OsClass 	os_class ;
} OsObjectRec ;

typedef OsObject (*OsMethod)() ;
    
typedef struct _OsClass {
    OsClass      os_class ;
    char         *name ;
    unsigned int isize ;		/* instance size */
    OsMethod	 methods[OS_SEL_LAST] ;	/* method array */
    void         *extension ;
} OsClassRec ;

extern OsClass OsMetaClass ;

#define OsIsClass(c)		(((OsObject) (c))->os_class == OsMetaClass)
#define OsMustBeClass(c)	(OsMustBeOfClass ((c), OsMetaClass))
#define OsShouldBeClass(c) \
    ((c) = (OsClass) OsMustBeOfClass((c), OsMetaClass))

/*
 * Externals:
 */

OsClass  OsClassDeclare OS_P ((char *name, unsigned int isize)) ;
OsObject OsObjectMake OS_P ((OsClass os_class, int collectable)) ;

typedef struct _OsCompEnv *OsCompEnv ;

OsObject OsSelEvalToSelf OS_P ((OsObject self, int sel)) ;
OsObject OsSelInvalidSelector OS_P ((OsObject self, int sel)) ;
OsObject OsSelCompileConstant OS_P ((OsObject self, int sel, OsCompEnv cenv));

void OsObjectInitialize OS_P ((OsInterp interp, int *pargc, char **argv)) ;
void OsObjectSubrInitialize OS_P ((OsInterp interp)) ;
    
/*
 * Macros:
 */

#define OsSendMsg(_obj,_sel,_args)   (((_obj)->os_class)->methods[(_sel)])_args
#define OsDefineMsg(_cls,_sel,_cfun) ((_cls)->methods[(_sel)] = (_cfun))

#define OsSendDisplay(_obj,_port) \
    OsSendMsg (_obj, OS_SEL_DISPLAY, (_obj, OS_SEL_DISPLAY, _port))

#define OsSendWrite(_obj,_port) \
    OsSendMsg (_obj, OS_SEL_WRITE, (_obj, OS_SEL_WRITE, _port))
    
#define OsSendCompile(_obj,_env) \
    OsSendMsg (_obj, OS_SEL_COMPILE, (_obj, OS_SEL_COMPILE, _env))

#define OsSendEqual(_o1,_o2) \
    ((((OsObject) (_o1)) == ((OsObject) (_o2))) \
     ? OsA_True \
     : OsSendMsg (_o1, OS_SEL_EQUAL, ((_o1), OS_SEL_EQUAL, (_o2))))

#define OsSendHashe(_obj) \
    OsSendMsg (_obj, OS_SEL_HASHE, (_obj, OS_SEL_HASHE))

#define OsSendEqv(_o1,_o2) \
    ((((OsObject) (_o1)) == ((OsObject) (_o2)) \
      || ((((OsObject) (_o1))->os_class == ((OsObject) (_o2))->os_class) \
          && ((OsIsInteger (_o1) \
               && (OsIntegerValue (_o1) == OsIntegerValue (_o2))) \
              || (OsIsFloat (_o1) \
                  && (OsFloatValue (_o1) == OsFloatValue (_o2)))))) \
     ? OsA_True \
     : OsA_False)
    
#define OsMustBeOfClass(o,c) \
    ( (((OsObject) (o))->os_class == (c)) \
     ? ((OsObject) o) \
     : OsErrorTrigger (1, OsE_BadType, (o), (c)) )

#endif /* _OS_OBJECT_H_ */
