/***************************************************************************
 * os_port.h
 * os_port.h,v 1.2 1995/03/12 02:58:18 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 ***************************************************************************/

#ifndef _OS_PORT_H_
#define _OS_PORT_H_

#define OS_DISPLAY_BUFSIZ	512

/*
 * Input ports:
 */

extern OsClass OsInputPortClass ;

#define OsIsInputPort(p)     (((OsObject) (p))->os_class == OsInputPortClass)
#define OsMustBeInputPort(p) (OsMustBeOfClass((p), OsInputPortClass))
#define OsShouldBeInputPort(p) \
    ((p) = (OsInputPort) (OsMustBeOfClass((p), OsInputPortClass)))

typedef struct _OsInputPort {
    OsClass      os_class ;
    unsigned int lineno ;
    unsigned int buflen ;
    char 	 *buf, *pbuf, *ebuf ;
    unsigned int stklen ;
    char 	 *stk, *pstk, *estk ;
    int          (*os_read)() ;
    void         (*os_close)() ;
    int          (*os_ready)() ;
    union {
        int      fd ;
        struct {
            int      pos ;
            OsString string ;
        } istring ;
    } medium ;
} OsInputPortRec, *OsInputPort ;

/*
 * Output ports:
 * We have three kind of output ports:
 * - standard file (open-output-port)
 * - async file (open-output-stream)
 * - string file (open-output-string)
 * Dispatching is done according to the type.
 */

extern OsClass OsOutputPortClass ;

#define OsIsOutputPort(p)     (((OsObject) (p))->os_class == OsOutputPortClass)
#define OsMustBeOutputPort(p) (OsMustBeOfClass((p), OsOutputPortClass))
#define OsShouldBeOutputPort(p) \
    ((p) = (OsOutputPort) OsMustBeOfClass ((p), OsOutputPortClass))

typedef struct _OsOutputPort {
    OsClass      os_class ;
    void 	 (*os_putc)() ;
    void         (*os_puts)() ;
    void         (*os_write)() ;
    void         (*os_flush)() ;
    void         (*os_close)() ;
    int          (*os_ready)() ;
    union {
        int      fd ;
        FILE	 *fp ;
        struct {
            char *strbuf, *pbuf ;
            OsString string ;
        } ostring ;
    } medium ;
} OsOutputPortRec, *OsOutputPort ;
    
/*
 * Externals:
 */

extern OsObject     OsA_Eof ;
extern OsInputPort  OsCurrentInputPort ;
extern OsOutputPort OsCurrentOutputPort ;
extern OsOutputPort OsCurrentErrorPort ;

OsInputPort  OsInputPortMakeFromFileno OS_P ((int fileno)) ;
OsInputPort  OsInputPortMakeFromName OS_P ((OsString filename)) ;
OsInputPort  OsInputStringMake OS_P ((OsString string)) ;
OsObject     OsRead OS_P ((OsInputPort port)) ;
OsObject     OsLoadFile OS_P ((OsInterp interp, OsString filename)) ;

void OsInputPortInitialize OS_P ((OsInterp interp, int *pargc, char **argv)) ;
void OsInputPortSubrInitialize OS_P ((OsInterp interp)) ;

OsOutputPort OsOutputPortMakeFromFile OS_P ((FILE *fp)) ;
OsOutputPort OsOutputPortMakeFromName OS_P ((OsString filename)) ;
OsOutputPort OsOutputStringMake OS_P ((OsString string)) ;
OsOutputPort OsFormatWithList OS_P ((OsOutputPort p, OsString s, OsCons c));
OsOutputPort OsFormatWithArray OS_P ((OsOutputPort p
                                      , OsString s
                                      , unsigned int argc, OsObject *argv)) ;

void OsOutputPortInitialize OS_P ((OsInterp interp, int *pargc, char **argv)) ;
void OsOutputPortSubrInitialize OS_P ((OsInterp interp)) ;

/*
 * Macros:
 */
    
#define OsOutputPortWrite(p,l,b) \
    (OsInterpCheckWrittenBytes (OsCurrentInterp(), (l)) \
     , (p)->os_write ((p),(l),(b)))
#define OsOutputPortWriteCString(p,s) \
    (OsInterpCheckWrittenBytes (OsCurrentInterp(), strlen(s)) \
     , (p)->os_puts((p),(s)))
#define OsOutputPortWriteChar(p,c) \
    (OsInterpCheckWrittenBytes (OsCurrentInterp(), 1) \
     , (p)->os_putc((p),(c)))
    
#define OsOutputPortFlush(p)          ((p)->os_flush(p))
#define OsOutputPortClose(p)	      ((p)->os_close(p))

#define OsInputPortClose(p)	      ((p)->os_close (p))

#endif /* _OS_PORT_H_ */
