/***************************************************************************
 * path.c
 * path.c,v 1.3 1995/03/12 02:58:22 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * Path, load and dynamic loading management functions.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>
#include <sys/param.h>		/* for MAXPATHLEN */

#include <oscheme.h>

static OsError ose_bad_filename          = (OsError) 0 ;
static OsError ose_required_not_provided = (OsError) 0 ;

/*
 * Internal API:
 */

static OsCons shell_list_to_list (lst)
    char *lst ;
{
    OsCons start = (OsCons) OsA_Nil, *plist = &start ;
    char   *s = lst, *w = s ;
    while ( *w ) {
        if ( *w == ':' ) {
            *plist = OsConsMake ((OsObject) OsStringMakeFromBytes (w - s, s)
                                 , OsA_Nil) ;
            plist  = (OsCons *) &(OS_CONS_CDR(*plist)) ;
            s = w + 1 ;
        } 
        w++ ;
    }
    if ( w != s )
        *plist = OsConsMake ((OsObject) OsStringMakeFromBytes (w - s, s)
                             , OsA_Nil) ;
    return (start) ;
}

/*
 * C API:
 */

void OsPathProvides (OsInterp interp, OsString file) {
    interp->prov_list = OsConsMake ((OsObject) file
                                    , (OsObject) interp->prov_list) ;
    return ;
}

OsObject OsPathProvided (OsInterp interp, OsString file) {
    OsCons prov = interp->prov_list ;
    while ( OsIsCons (prov) ) {
        if ( OsSendEqual (OS_CONS_CAR (prov), file) == OsA_True )
            return (OsA_True) ;
        prov = (OsCons) OS_CONS_CDR (prov) ;
    }
    return (OsA_False) ;
}

OsString OsPathBasename (file)
    OsString file ;
{
    char *s = file->string, *e = s + file->len-1, *ext = e+1 ;
    
    if ( file->len ) {
        while ( e > s ) {
            if ( *e == OS_PATH_SEPARATOR ) {
                return (OsStringMakeFromBytes (ext - e - 1, e + 1)) ;
            } else if ( *e == '.' ) {
                ext = e ;
            }
            e-- ;
        }
    }
    return (OsStringMakeFromBytes (ext - e, e)) ;
}

OsString OsPathExtension (file)
    OsString file ;
{
    char *s = file->string, *e = s + file->len ;

    if ( file->len ) {
        while ( e > s ) {
            if ( *e == '.' )
                return (OsStringMake (e)) ;
            e-- ;
        }
    }
    return ((OsString) 0) ;
}

OsString OsPathExpandForTilde (file)
    OsString file ;
{
    char name[MAXPATHLEN+1] ;
    char *rest ;
    char *filename = file->string ;
    
    if ( filename[0] == OS_PATH_HOME ) {
        OsString user, home ;
        for (rest = filename + 1 ; *rest != OS_PATH_SEPARATOR ; rest++ ) {
            if ( ! *rest )
                break ;
        }
        user = OsStringMakeFromBytes (rest-filename-1, filename+1) ;
        if ( ! (home = OsGetHomeDirectory (user)) )
            return (file) ;
        if ( home->string[home->len-1] == OS_PATH_SEPARATOR ) {
            sprintf (name, "%s%s"
                     , home->string
                     , (*rest == OS_PATH_SEPARATOR) ? rest + 1 : rest) ;
        } else {
            sprintf (name, "%s/%s"
                     , home->string
                     , (*rest == OS_PATH_SEPARATOR) ? rest + 1 : rest) ;
        }
        return (OsStringMake (name)) ;
    }
    return (file) ;
}

OsString OsPathFindFile (OsInterp interp, OsString file) {
    char     *filename = file->string ;
    char     name[MAXPATHLEN+1] ;
    OsCons   dir = interp->path_list ;
    OsCons   ext = interp->ext_list ;
    char     *dirend ;
    OsString fext = OsPathExtension (file) ;
    
    switch (filename[0]) {
      case OS_PATH_SEPARATOR:
        /* absolute path */
        strcpy (name, filename) ;
        break ;
      case OS_PATH_HOME:
        /* home relative path */
        {
            OsString exppath = OsPathExpandForTilde (file) ;
            bcopy (exppath->string, name, exppath->len+1) ;
        }
        break ;
      default:
        /* relative path */
        while ( OsIsCons (dir) ) {
            OsString dirnam = (OsString) OS_CONS_CAR (dir) ;
            
            OsShouldBeString (dirnam) ;
            dirnam = OsPathExpandForTilde (dirnam) ;
            if ( dirnam->string[dirnam->len-1] == OS_PATH_SEPARATOR ) {
                sprintf (name, "%s%s", dirnam->string, filename) ;
            } else {
                sprintf (name, "%s/%s", dirnam->string, filename) ;
            }
            dirend = name + strlen (name) ;
            if ( fext ) {
                if ( OsFileExists (name) == OsA_True )
                    return (OsStringMake (name)) ;
            } else {
                ext = interp->ext_list ;
                while ( OsIsCons (ext) ) {
                    OsString extnam = (OsString) OS_CONS_CAR(ext) ;
                    
                    OsShouldBeString (extnam) ;
                    bcopy (extnam->string, dirend, extnam->len+1) ;
                    if ( OsFileExists (name) == OsA_True )
                        return (OsStringMake (name)) ;
                    ext = (OsCons) OS_CONS_CDR (ext) ;
                }
            }
            dir = (OsCons) OS_CONS_CDR (dir) ;
        }
        return ((OsString) 0) ;
    }
    /* Look for an extension if needed */
    dirend = name + strlen (name) ;
    if ( fext ) {
        if ( OsFileExists (name) == OsA_True )
            return ((OsString) OsStringMake (name)) ;
        else
            return ((OsString) 0) ;
    } else {
        while ( OsIsCons (ext) ) {
            OsString extnam = (OsString) OS_CONS_CAR (ext) ;

            OsShouldBeString (extnam) ;
            bcopy (extnam->string, dirend, extnam->len+1) ;
            if ( OsFileExists (name) == OsA_True )
                return (OsStringMake (name)) ;
            ext = (OsCons) OS_CONS_CDR (ext) ;
        }
    }
    return ((OsString) 0) ;
}

OsObject OsPathLoad (OsString filename) {
    OsString full_filename, ext ;
    OsInterp interp = OsCurrentInterp() ;
    
    while ( ! (full_filename = OsPathFindFile (interp, filename)) ) {
        filename = (OsString) OsErrorTrigger (1, ose_bad_filename, filename) ;
        OsShouldBeString (filename) ;
    }
    if ( (ext = OsPathExtension (full_filename)) ) {
        OsCodeBlock hook ;
        if ( hook = (OsCodeBlock) OsHashtableGet (interp->hooked_ext, ext) ) {
            (void) OsCodeBlockApplyWithArray (interp
                                              , hook
                                              , 1, (OsObject *) &full_filename
                                              , (OsObject *) 0) ;
            return (OsA_True) ;
        }
    }
    OsLoadFile (interp, full_filename) ;
    return (OsA_True) ;
}

/*
 * Subrs:
 */

/* AUTODOC
 * (set-load-path! PATH*)
 * Append each string PATH to current load path list.
 * The load path list is a list of directory to be searched when loading a
 * file through 'require or 'load.
 */

static OsObject OsSubrSetLoadPath (pathl)
    OsCons pathl ;
{
    OsShouldBeCons (pathl) ;
    OsCurrentInterp()->path_list = pathl ;
    return ((OsObject) pathl) ;
}

/* AUTODOC
 * (load-path)
 * Get the current load path list
 * Returns the current list of directory to be searched for files on 'require
 * or 'load requests.
 */

static OsObject OsSubrLoadPath ()
{
    return ((OsObject) (OsCurrentInterp()->path_list)) ;
}

/* AUTODOC
 * (set-load-extension! EXTENSIONS*)
 * Append each string EXTENSION to the current extension list.
 * The extension list is the list of extension the 'load or 'require operations
 * support. When a file is requested through 'load or 'request, it is searched
 * in the current load path list in the following fashion:
 * For each directory in load path list, look for a file having as extension
 * any of the extension in the load extension list. When such file is found,
 * it is either passed as an argument to a extension load hook (if any), or
 * loaded as a scheme source file.
 */

static OsObject OsSubrSetLoadExtension (extl)
    OsCons extl ;
{
    OsShouldBeCons (extl) ;
    OsCurrentInterp()->ext_list = extl ;
    return ((OsObject) extl) ;
}

/* AUTODOC
 * (load-extension)
 * Returns the current list of load extensions.
 */

static OsObject OsSubrLoadExtension ()
{
    return ((OsObject) (OsCurrentInterp()->ext_list)) ;
}

/* AUTODOC
 * (find-file FILE)
 * This function looks for a file named "FILE.ext" where ext is any extension
 * of the current extension list. The search is perform in all the directories
 * contained in the current load path list.
 */

static OsObject OsSubrFindFile (OsString filename) {
    OsString path ;
    OsShouldBeString (filename) ;
    if ( (path = OsPathFindFile (OsCurrentInterp(), filename)) )
        return ((OsObject) path) ;
    return (OsA_False) ;
}

/* AUTODOC
 * (require FILE)
 * Make sure that string FILE has been loaded.
 * First looks if the file has been 'provided. If not tries to 'load it.
 */

static OsObject OsSubrRequire (filename)
    OsString filename ;
{
    OsCons   prov = OsCurrentInterp()->prov_list ;
    
    while ( OsIsCons (prov) ) {
        if ( OsSendEqual (OS_CONS_CAR(prov), filename) == OsA_True )
            return ((OsObject) filename) ;
        prov = (OsCons) OS_CONS_CDR (prov) ;
    }
    OsPathLoad (filename) ;
    if ( OsPathProvided (OsCurrentInterp(), filename) != OsA_True ) {
        (void) OsErrorTrigger (1, ose_required_not_provided, filename) ;
    }
    return (OsA_True) ;
}

/* AUTODOC
 * (set-extension-hook! EXTENSION HOOK)
 * Set HOOK to be invoked when a file with EXTENSION is being loaded.
 * This function define an extension load hook. Such a hook is invoked
 * when a file with the given extension is to be loaded. The HOOK function
 * is invoked with the full path of the file to be loaded, and has the
 * resposibility of loading it.
 */

static OsObject OsSubrSetExtensionHook (ext, hook)
    OsString ext ;
    OsCodeBlock hook ;
{
    OsShouldBeString (ext) ;
    OsMustBeBlock (hook) ;
    (void) OsHashtablePut (OsCurrentInterp()->hooked_ext, ext, hook) ;
    return (OsA_True) ;
}

/* AUTODOC
 * (extension-hook EXTENSION)
 * Retreives the load hook for the given EXTENSION
 * Returns the hook for this EXTENSION, or #f if none was find. This is
 * usefull if one wants to chain extension hooks.
 */
 
static OsObject OsSubrExtensionHook (ext)
    OsString ext ;
{
    OsObject hook ;
    
    OsShouldBeString (ext) ;
    if ( (hook = (OsObject) OsHashtableGet (OsCurrentInterp()->hooked_ext
                                            , (void *) ext)) ) {
        return ((OsObject) hook) ;
    } else {
        return (OsA_False) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (remove-extension-hook! EXTENSION)
 * Removes any load hook for the given EXTENSION.
 * This function unregister any load hook for the given EXTENSION. It always
 * returns #t
 */

static OsObject OsSubrRemoveExtensionHook (ext)
    OsString ext ;
{
    OsShouldBeString (ext) ;
    (void) OsHashtableRemove (OsCurrentInterp()->hooked_ext, (void *) ext) ;
    return (OsA_True) ;
}

/* AUTODOC
 * (provide FEATURE)
 * Provide the string FEATURE.
 * Add the given FEATURE to the list of loaded features. A 'require ed feature
 * must be provided by the corresponding file (otherwise an error occurs).
 */

static OsObject OsSubrProvide (OsString filename) {
    OsInterp interp = OsCurrentInterp() ;
    
    OsShouldBeString (filename) ;
    interp->prov_list = OsConsMake ((OsObject) filename
                                    , (OsObject) interp->prov_list) ;
    return ((OsObject) filename) ;
}

/* AUTODOC
 * (load FILE)
 * Load the given FILE.
 * The FILE is first seached for using 'find-file, and depending on its
 * extension (see 'set-extension-hook!), loaded into the scheme process.
 */

static OsObject OsSubrLoad (OsString filename) {
    OsShouldBeString (filename) ;
    return (OsPathLoad (filename)) ;
}

/* AUTODOC
 * (expand-path-for-tilde PATH)
 * Expand the tilde of PATH, if any.
 * This functions look for a leading ~ in the given PATH. If found, it is
 * expanded either to the home directory of the user running the oscheme
 * process, or to the home directory of the user mentionned just after the
 * ~ character.
 */

static OsObject OsSubrExpandPathForTilde (filename)
    OsString filename ;
{
    OsString exp ;
    OsShouldBeString (filename) ;
    if ( (exp = OsPathExpandForTilde (filename)) )
        return ((OsObject) exp) ;
    return ((OsObject) filename) ;
}

/* AUTODOC
 * (path-extension FILENAME)
 * Get the extension of FILENAME
 * This function retruns the (string) extension of the given FILENAME. The
 * extension of a file is defined to be the smallest trailinbg part of the
 * name beginning with a '.'.
 */

static OsObject OsSubrPathExtension (filename)
    OsString filename ;
{
    OsString ext ;
    OsShouldBeString (filename) ;
    if ( (ext = OsPathExtension (filename)) )
        return ((OsObject) ext) ;
    return (OsA_False) ;
}

/* AUTODOC
 * (provided? FEATURE)
 * Is given (string) FEATURE provided.
 * This function checks the list of provided features, and returns #t if the
 * FEATURE has been provided, or #f otherwise.
 */

static OsObject OsSubrProvidedP (filename)
    OsString filename ;
{
    OsShouldBeString (filename) ;
    return (OsPathProvided (OsCurrentInterp(), filename)) ;
}

/* AUTODOC
 * (basename PATH)
 * Removes directories and suffix of given PATH.
 * This function returns the basneame of PATH, i.e, the PATH without leading
 * directories, and whitout its suffix.
 */

static OsObject OsSubrBasename (filename)
    OsString filename ;
{
    OsShouldBeString (filename) ;
    return ((OsObject) OsPathBasename (filename)) ;
}
    
/*
 * Initialize :
 */

void OsPathInitialize (OsInterp interp, int *pargc, char **argv) {
    unsigned int argc = 1 ;
    OsString     foo ;
    OsCons       path_list, ext_list ;
    OsHashtable  hooked_ext ;
    
    foo = OsGetenv (OsStringMake ("OS_PATH")
                    , OsStringMake (OS_DEFAULT_PATH));
    path_list = shell_list_to_list (foo->string) ;

    foo = OsGetenv (OsStringMake ("OS_EXTENSION")
                    , OsStringMake (OS_DEFAULT_EXTENSION)) ;
    ext_list = shell_list_to_list (foo->string) ;

    hooked_ext = OsHashtableMake (0
                                  , (OsHashFct) OsStringHash
                                  , (OsCmpFct) OsStringCompare) ;
    while ( argc < *pargc ) {
        if ( ! strcmp ("-path", argv[argc]) ) {
            OsArgvUsed (pargc, argc, argv) ;
            path_list = OsConsMake ((OsObject) OsStringMake (argv[argc])
                                    , (OsObject) path_list) ;
            OsArgvUsed (pargc, argc, argv) ;
        } else if ( ! strcmp ("-ext", argv[argc]) ) {
            OsArgvUsed (pargc, argc, argv) ;
            ext_list = OsConsMake ((OsObject) OsStringMake (argv[argc])
                                   , (OsObject) ext_list) ;
            OsArgvUsed (pargc, argc, argv) ;
        } else {
            argc++ ;
        }
    }
    interp->path_list  = path_list ;
    interp->ext_list   = ext_list ;
    interp->prov_list  = OsA_Nil ;
    interp->hooked_ext = hooked_ext ;
    return ;
}

void OsPathSubrInitialize (OsInterp interp) {
    ose_bad_filename
        = OsErrorMake (interp
                       , "error:invalid-file-name"
                       , "invalid file name %0"
                       , 1, OSE_NORMAL) ;
    ose_required_not_provided
        = OsErrorMake (interp
                       , "error:required-not-provided"
                       , "required feature %0 not provided"
                       , 1, OSE_CATCHABLE) ;
    OsSecSubrMake (interp, "set-load-path!", OsSubrSetLoadPath
                   , 1, OS_SECURITY_LEVEL_E) ;
    OsSecSubrMake (interp, "load-path", OsSubrLoadPath
                   , 0, OS_SECURITY_LEVEL_E) ;
    OsSecSubrMake (interp, "set-load-extension!", OsSubrSetLoadExtension
                   , 1, OS_SECURITY_LEVEL_E);
    OsSecSubrMake (interp, "load-extension", OsSubrLoadExtension
                   , 0, OS_SECURITY_LEVEL_E) ;
    OsSubrMake (interp, "find-file", OsSubrFindFile, 1) ;
    OsSubrMake (interp, "set-extension-hook!", OsSubrSetExtensionHook, 2) ;
    OsSubrMake (interp, "extension-hook", OsSubrExtensionHook, 1) ;
    OsSubrMake (interp, "remove-extension-hook!", OsSubrRemoveExtensionHook,1);
    OsSubrMake (interp, "require", OsSubrRequire, 1) ;
    OsSubrMake (interp, "provide", OsSubrProvide, 1) ;
    OsSubrMake (interp, "load", OsSubrLoad, 1) ;
    OsSubrMake (interp, "expand-path-for-tilde", OsSubrExpandPathForTilde, 1) ;
    OsSubrMake (interp, "path-extension", OsSubrPathExtension, 1) ;
    OsSubrMake (interp, "provided?", OsSubrProvidedP, 1) ;
    OsSubrMake (interp, "basename", OsSubrBasename, 1) ;
    return ;
}
