/***************************************************************************
 * port.c
 * port.c,v 1.4 1995/03/17 14:33:18 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * Port handling. OScheme allows for files and strings port.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>
#include <sys/types.h>
#include <fcntl.h>
#ifdef OS_HAS_UNISTD_H
#include <unistd.h>
#endif
#include <ctype.h>
#include <errno.h>

extern int errno ;

#include <oscheme.h>

OsClass OsInputPortClass  = (OsClass) 0 ;
OsClass OsOutputPortClass = (OsClass) 0 ;

OsObject     OsA_Eof             = (OsObject) 0 ;
OsInputPort  OsCurrentInputPort  = (OsInputPort) 0 ;
OsOutputPort OsCurrentOutputPort = (OsOutputPort) 0 ;
OsOutputPort OsCurrentErrorPort  = (OsOutputPort) 0 ;

static OsObject osa_send_msg = (OsObject) 0 ;
static OsError  ose_filectxt = (OsError) 0 ;
static OsError  ose_syntax   = (OsError) 0 ;
static OsError  ose_wclose   = (OsError) 0 ;
static OsError  ose_rclose   = (OsError) 0 ;

/*
 * Char classes macros (to begin with...), this *will* be optimized some day
 */

#define OS_SYMCHAR(c) \
    (isalpha(c) || isdigit(c) \
     || (c=='!') || (c=='?') || (c=='-') || (c=='>') || (c=='<') \
     || (c=='+') || (c=='-') || (c=='*') || (c=='\\') || (c==':') \
     || (c=='=') || (c == '/') || (c == '_') || (c == '.'))

#define OS_SEP(c) \
    ((c == ' ') || (c == '\t') || (c == '\n') || (c == ')') || (c == '(') \
     || (c == ';') || (c == ';'))

/*
 * Parsers buffer their internal state:
 */

static char os_fill_buffer (OsInputPort port) ;
    
#define OS_NEEDSFILL(p) ((p)->pbuf >= (p)->ebuf)

#define OS_GETC(p) \
    (((p)->pbuf < (p)->ebuf) ? *((p)->pbuf++) : os_fill_buffer (p))

#define OS_INCLINE(p) ((p)->lineno++)
#define OS_DECLINE(p) ((p)->lineno--)

#define OS_UNGETC(c,p) \
    ((c == '\n') \
      ? (OS_DECLINE(p), (*--((p)->pbuf) = c)) \
      : (*--((p)->pbuf) = c))

#define OS_PUSHC(c,p) (*((p)->pstk++) = c)

#define OS_POP(p) ((p)->pstk = (p)->stk)

#define OS_SLEN(p) ((p)->pstk - (p)->stk)

/*
 * This function always leaves room for at least OL_PARSER_MXUNGET chars
 * This is really poor, I will code soon the two buffers scheme describe in
 * Aho et all
 */

static INLINE char os_fill_buffer (p)
    OsInputPort p ;
{
    int          got ;
    
    if ( (p->ebuf - p->buf) < OS_PARSER_MXUNGET ) {
	p->pbuf = p->ebuf ;
    } else {
	bcopy (p->ebuf - OS_PARSER_MXUNGET, p->buf, OS_PARSER_MXUNGET) ;
	p->pbuf = p->buf + OS_PARSER_MXUNGET ;
    }
    if ( (got = (p->os_read) (p, p->buflen-(p->pbuf-p->buf), p->pbuf)) < 0 ) {
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("os_fill_buffer")
                               , OsStringMake ("read method failed on port"));
        /* NOT REACHED */
    }
    OsInterpCheckReadBytes (OsCurrentInterp(), got) ;
    p->ebuf = p->pbuf + got ;
    return ((p->ebuf == p->pbuf) ? ((char) 0) : OS_GETC(p)) ;
}

/*
 * Reading chars:
 */

static OsObject os_read_char (parser)
    OsInputPort parser ;
{
    unsigned char c = OS_GETC (parser) ;
    unsigned char n = OS_GETC (parser) ;
    if ( OS_SEP (n) ) {
        OS_INCLINE (parser) ;
        OS_UNGETC (n, parser) ;
        return ((OsObject) OsCharMake (c)) ;
    } else {
        char *symchar ;
        
        OS_PUSHC (c, parser) ;
        OS_PUSHC (n, parser) ;
        while ( 1 ) {
            c = OS_GETC(parser) ;
            if ( OS_SEP (c) ) {
                if ( c == '\n' ) {
                    OS_INCLINE (parser) ;
                }
                break ;
            } else {
                OS_PUSHC (c, parser) ;
            }
        }
        OS_UNGETC (c, parser) ;
        OS_PUSHC ('\0', parser) ;
        symchar = OS_POP (parser) ;
        if ( (! strcmp (symchar, "space")) || (! strcmp (symchar, "Space")) ) {
            return ((OsObject) OsCharMake (' ')) ;
        } else if ( ! strcmp ("newline", symchar) ) {
            return ((OsObject) OsCharMake ('\n')) ;
        } else {
            (void) OsErrorTrigger (1, ose_syntax
                                   , OsStringMake ("invalid char syntax")) ;
        }
    }
    /* NOT REACHED */
}

/*
 * Read strings:
 */

static OsObject os_read_string (parser)
    OsInputPort parser ;
{
    char c ;
    unsigned int len ;
    while ( (c = OS_GETC (parser)) ) {
        switch (c) {
          case '"':
            len = OS_SLEN(parser) ;
            return ((OsObject) OsStringMakeFromBytes (len, OS_POP(parser))) ;
          case '\\':
            switch (c = OS_GETC (parser)) {
              case 'n':
                OS_PUSHC ('\n', parser) ;
                break ;
              case 'r':
                OS_PUSHC ('\r', parser) ;
                break ;
              case 't':
                OS_PUSHC ('\t', parser) ;
                break ;
              default:
                OS_PUSHC (c, parser) ;
            }
            break ;
          default:
            OS_PUSHC(c, parser) ;
        }
    }
    (void) OsErrorTrigger (1, ose_syntax, OsStringMake ("non closed string")) ;
}

/*
 * Read atoms:
 */

static OsObject os_read_atom (parser)
    OsInputPort parser ;
{
    char c ;
    while ( (c = OS_GETC(parser)) ) {
        if ( c == '\n' )		/* next UNGETC might decr it */
            OS_INCLINE (parser) ;
	if ( OS_SYMCHAR (c) ) {
	    OS_PUSHC (c, parser) ;
	} else {
	    OS_UNGETC (c, parser) ;
	    OS_PUSHC ('\0', parser) ;
	    return ((OsObject) OsAtomMake (OsCurrentInterp()
                                           , OS_POP(parser), OsA_Unbound)) ;
	}
    }
    OS_PUSHC ('\0', parser) ;
    return ((OsObject) OsAtomMake (OsCurrentInterp()
                                   , OS_POP(parser), OsA_Unbound)) ;
}

/*
 * Read numbers:
 */

static OsObject os_read_exp (OsInputPort parser) { /* [+-][0-9]* */
    double fval ;
    char   c = OS_GETC (parser) ;
    if ( (c == '+') || (c == '-')) {
        OS_PUSHC (c, parser) ;
        c = OS_GETC(parser) ;
    }
    do {
        if ( isdigit (c) ) {
            OS_PUSHC (c, parser) ;
        } else {
            OS_UNGETC (c, parser) ;
            OS_PUSHC ('\0', parser) ;
            return ((OsObject) OsFloatMakeFromCString (OS_POP (parser))) ;
        }
    } while ( c = OS_GETC (parser) ) ;
    OS_PUSHC ('\0', parser) ;
    return ((OsObject) OsFloatMakeFromCString (OS_POP (parser))) ;
}

static OsObject os_read_float (OsInputPort parser) { /* [0-9]*{read_exp} */
    char c = OS_GETC (parser) ;
    do {
        if ( isdigit (c) ) {
            OS_PUSHC (c, parser) ;
        } else if ( (c == 'e') || (c == 'E') ) {
            OS_PUSHC (c, parser) ;
            return (os_read_exp (parser)) ;
        } else {
            OS_UNGETC (c, parser) ;
            OS_PUSHC ('\0', parser) ;
            return ((OsObject) OsFloatMakeFromCString (OS_POP(parser))) ;
        }
        c = OS_GETC (parser) ;
    } while ( c ) ;
    OS_PUSHC ('\0', parser) ;
    return ((OsObject) OsFloatMakeFromCString (OS_POP(parser))) ;
}

static OsObject os_read_number (OsInputPort parser) {
    char c ;
    
    while ( (c = OS_GETC (parser)) ) {
        if ( isdigit (c) ) {
            OS_PUSHC (c, parser) ;
        } else if ( (c == 'e') || (c == 'E') ) {
            OS_PUSHC (c, parser) ;
            return (os_read_exp (parser)) ;
        } else if ( c == '.' ) {
            OS_PUSHC (c, parser) ;
            return (os_read_float (parser)) ;
        } else {
            OS_UNGETC (c, parser) ;
            OS_PUSHC ('\0', parser) ;
            return ((OsObject) OsIntegerMake (atoi (OS_POP(parser)))) ;
        }
    }
    OS_PUSHC ('\0', parser) ;
    return ((OsObject) OsIntegerMake (atoi (OS_POP(parser)))) ;
}

/*
 * Read vectors:
 */

static OsObject os_read_vector (parser)
    OsInputPort parser ;
{

    OsVector v = OsVectorMake (0) ;
    char c ;
    while ( (c = OS_GETC (parser)) ) {
        switch (c) {
          case ')':
	    return ((OsObject) v) ;
          case '\n': 
            OS_INCLINE (parser) ; 
            /* no break intentional */ 
          case ';': 
            while ( (c = OS_GETC (parser)) && ( c != '\n') ) 
                ; 
            OS_INCLINE (parser) ; 
            break ;
            /* NOT REACHED */
          case ' ': 
          case '\t': 
            break ;
          default:
            OS_UNGETC (c, parser) ;
            (void) OsVectorFastAppend (v, OsRead (parser)) ;
        }
    }
    /* NOT REACHED */
}

/*
 * Read lists:
 */

static OsObject os_read_list (parser, end)
    OsInputPort   parser ;
    unsigned char end ;
{
    OsObject start = OsA_Nil, *cont = &start ;
    unsigned char c ;

    while ( (c = OS_GETC (parser)) ) {
        if ( c == end )
            return (start) ;
        switch (c) {
        case '\n': 
          OS_INCLINE (parser) ; 
          /* no break intentional */ 
        case ' ': 
        case '\t': 
          break ;
        case ';': 
          while ( (c = OS_GETC (parser)) && ( c != '\n') ) 
              ; 
          OS_INCLINE (parser) ; 
          break ;
          /* NOT REACHED */
        case '.':
          c = OS_GETC (parser) ;
          if ( OS_SEP (c) ) {
              OsObject cdr = OsRead (parser) ;
              while ( (c = OS_GETC (parser)) ) {
                  switch (c) {
                    case '\n':
                      OS_INCLINE (parser) ;
                      /* no break intentional */
                    case ' ':
                    case '\t':
                      continue ;
                    case ')':
                      *cont = cdr ;
                      return (start) ;
                    default:
                      (void)OsErrorTrigger(1, ose_syntax
                                           ,OsStringMake("bad dotted list"));
                      /* NOT REACHED */
                  }
              }
              (void) OsErrorTrigger (1, ose_syntax
                                     , OsStringMake ("unclosed list")) ;
              /* NOT REACHED */
          } else {
              if ( c == '\n' )
                  OS_INCLINE (parser) ;
              OS_UNGETC ('.', parser) ;
              OS_UNGETC (c, parser) ;
              *cont = (OsObject) OsConsMake (OsRead (parser), OsA_Nil) ;
              cont  = &(OS_CONS_CDR (*cont)) ;
          }
          break ;
          default:
            OS_UNGETC (c, parser) ;
            *cont = (OsObject) OsConsMake (OsRead (parser), OsA_Nil) ;
            cont  = &(OS_CONS_CDR (*cont)) ;
        }
    }
    (void) OsErrorTrigger (1, ose_syntax, OsStringMake ("unclosed list")) ;
}

OsObject OsRead (parser) 
    OsInputPort parser ;
{
    OsObject start ;
    char c ;

    while ( (c = OS_GETC (parser)) ) {
        switch (c) {
          case '(':
            return (os_read_list (parser, ')')) ;
          case '"':
            return (os_read_string (parser)) ;
          case '\n': 
            OS_INCLINE (parser) ; 
            /* no break intentional */ 
          case ' ': 
          case '\t': 
            break ;
          case '#':
            switch (c = OS_GETC (parser)) {
              case '(':
                return (os_read_vector (parser)) ;
              case '\\':
                return (os_read_char (parser)) ;
              case '!':
                while ( (c = OS_GETC (parser)) ) {
                    switch (c) {
                      case '\n':
                        OS_INCLINE (parser) ;
                        break ;
                      case '!':
                        if ( (c = OS_GETC(parser)) == '#' ) {
                            return (OsRead (parser)) ;
                        } else if ( c == '\n' ) {
                            OS_INCLINE (parser) ;
                        } else if ( ! c ) {
                            OsErrorTrigger(1, ose_syntax
                                           , OsStringMake("unclosed comment"));
                        }
                        break ;
                      default:
                        break ;
                    }
                }
                (void) OsErrorTrigger (1, ose_syntax
                                       , OsStringMake ("unclosed comment")) ;
                break ;
              default:
                OS_PUSHC ('#', parser) ;
                OS_PUSHC (c, parser) ;
                return (os_read_atom (parser)) ;
            }
            /* NOT REACHED */
            break ;
          case ';': 
            while ( (c = OS_GETC (parser)) && ( c != '\n') ) 
                ; 
            OS_INCLINE (parser) ; 
            break ;
            /* NOT REACHED */
          case '-':
          case '+':
            {
                char n = OS_GETC (parser) ;
                OS_PUSHC (c, parser) ;
                if ( isdigit (n) ) {
                    OS_PUSHC (n, parser) ;
                    return (os_read_number (parser)) ;
                } else {
                    OS_UNGETC (n, parser) ;
                    return (os_read_atom (parser)) ;
                }
            }
            /* NOT REACHED */
          case '0': case '1': case '2': case '3': case '4':
          case '5': case '6': case '7': case '8': case '9':
            OS_PUSHC (c, parser) ;
            return (os_read_number(parser)) ;
          case '[':
            start = os_read_list (parser, ']') ;
            return ((OsObject) OsConsMake (osa_send_msg, start)) ;
          case '\'':
            start = OsRead (parser) ;
            return ((OsObject) OsList2Make (OsA_Quote, start)) ;
          case '`':
            start = OsRead (parser) ;
            return ((OsObject) OsList2Make (OsA_Quasiquote, start)) ;
          case ',':
            c = OS_GETC (parser) ;
            if ( c == '@' ) {
                start = OsRead (parser) ;
                return ((OsObject) OsList2Make (OsA_UnquoteSplice, start)) ;
            } else {
                OS_UNGETC (c, parser) ;
                start = OsRead (parser) ;
                return ((OsObject) OsList2Make (OsA_Unquote, start)) ;
            }
            /* NOT REACHED */
          default:
            OS_PUSHC (c, parser) ;
            return (os_read_atom (parser)) ;
        }
    } 
    return (OsA_Eof) ;
}

/***************************************************************************
 * Inputports
 ***************************************************************************/

/*
 * Input port method definitions:
 */

static int iport_read (OsInputPort port, int len, char *buffer) {
    int cnt ;
    if ( port->medium.fd < 0 ) {
        (void) OsErrorTrigger (1, ose_rclose, port) ;
    }
  retry:
    if ( (cnt = read (port->medium.fd, buffer, len)) < 0 ) {
        int error = errno ;
#ifdef EINTR
        if ( error == EINTR ) {
#ifdef OS_HAS_SIGNALS
            {
                if ( os_signals_pending )
                    OsDoSignals () ;
            }
#endif /* OS_HAS_SIGNALS */
            goto retry ;
        }
#endif /* EINTR */
#ifdef EAGAIN
        if ( error == EAGAIN ) {
            return (0) ;
        }
#endif /* EAGAIN */
#ifdef EWOULDBLOCK
        if ( error == EWOULDBLOCK ) {
            return (0) ;
        }
#endif /* EWOULDBLOCK */
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("read")
                               , OsIntegerMake (error)
                               , OsConsMake ((OsObject) port, OsA_Nil)) ;
    }
    return (cnt) ;
}

static void iport_close (OsInputPort port) {
    if ( port->medium.fd >= 0 ) {
        close (port->medium.fd) ;
        port->medium.fd = -1 ;
    }
    return ;
}

static int iport_ready (OsInputPort port) {
    return (port->medium.fd) ;
}

/*
 * Input string method definitions:
 */

#define ISTRING_STRING(p)	((p)->medium.istring.string)
#define ISTRING_POS(p)		((p)->medium.istring.pos)

static int istring_read (OsInputPort port, int len, char *buffer) {
    if ( ! ISTRING_STRING (port) ) {
        (void) OsErrorTrigger (1, ose_rclose, port) ;
    }
    if ( ISTRING_POS (port) < ISTRING_STRING(port)->len ) {
        unsigned int cnt = ISTRING_STRING(port)->len - ISTRING_POS(port) ;
        cnt = ((cnt < len) ? cnt : len) ;
        bcopy (ISTRING_STRING(port)->string + ISTRING_POS(port), buffer, cnt) ;
        ISTRING_POS (port) += cnt ;
        return (cnt) ;
    }
    return (0) ;
}

static void istring_close (OsInputPort port) {
    if ( ISTRING_STRING(port) ) {
        ISTRING_STRING (port) = (OsString) 0 ;
    }
    return ;
}

static int istring_ready (OsInputPort port) {
    return (-1) ;
}

/*
 * Constructors:
 */

OsInputPort OsInputPortMakeFromFileno (int fd) {
    OsInputPort p = (OsInputPort) OsObjectMake (OsInputPortClass, 0) ;
    p->lineno = 1 ;
    p->buflen = OS_PARSER_BUFSIZE ;
    p->ebuf = p->pbuf = p->buf = (char *) OsGCMallocAtomic (p->buflen) ;
    p->stklen = OS_PARSER_STKSIZE ;
    p->estk = p->pstk = p->stk = (char *) OsGCMallocAtomic (p->stklen) ;
    p->os_read   = iport_read ;
    p->os_close  = iport_close ;
    p->os_ready  = iport_ready ;
    p->medium.fd = fd ;
    return (p) ;
}

OsInputPort OsInputPortMakeFromName (OsString filename) {
    int fd = open (filename->string, O_RDONLY) ;
    if ( fd >= 0 ) {
        return (OsInputPortMakeFromFileno (fd)) ;
    } else {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("open")
                               , OsIntegerMake (error)
                               , OsConsMake ((OsObject) filename, OsA_Nil)) ;
    }
    /* NOT REACHED */
}

OsInputPort OsInputStringMake (OsString buffer) {
    OsInputPort p = (OsInputPort) OsObjectMake (OsInputPortClass, 0) ;
    p->lineno = 1 ;
    p->buflen = OS_PARSER_BUFSIZE ;
    p->ebuf = p->pbuf = p->buf = (char *) OsGCMallocAtomic (p->buflen) ;
    p->stklen = OS_PARSER_STKSIZE ;
    p->estk = p->pstk = p->stk = (char *) OsGCMallocAtomic (p->stklen) ;
    p->os_read  = istring_read ;
    p->os_close = istring_close ;
    p->os_ready = istring_ready ;
    p->medium.istring.pos    = 0 ;
    p->medium.istring.string = buffer ;
    return (p) ;
}

OsObject OsLoadFile (OsInterp interp, OsString filename) {
    OsInputPort port ;
    OsObject    obj, error ;

    port = OsInputPortMakeFromName (filename) ;
    while ( 1 ) {
        obj = OsCProtectExecute (interp
                                 , (OsObject (*)()) OsRead
                                 , (OsObject) port
                                 , &error) ;
        if ( obj == OsA_Eof )
            return (OsA_True) ;
        if ( ! error ) {
            OsExprEval (obj, 1, &error) ;
        }
        if ( error ) {
            os_uncatched_error (OsErrorAugment(error
                                               , 1, ose_filectxt
                                               , filename
                                               , OsIntegerMake(port->lineno)));
            /* NOT REACHED */
        }
    }
    /* NOT REACHED */
}

/*
 * input port subrs:
 */

/* AUTODOC
 * (sys:load FILENAME)
 * Load the given scheme FILENAME into the oscheme process.
 * This functions does not perform any expansion of FILENAME, it just read
 * eval the FILENAME expression by expression.
 * Hackers only.
 */

OsObject OsSubrSysLoad (OsString filename) {
    OsShouldBeString (filename) ;
    return (OsLoadFile (OsCurrentInterp(), filename)) ;
}

/* AUTODOC
 * (input-port? OBJECT)
 * Is OBJECT an input port ?
 * Returns #t if OBJECT is an input port, #f atherwise.
 */

static OsObject OsSubrInputPortP (OsObject port) {
    return (OsIsInputPort (port) ? OsA_True : OsA_False) ;
}

/* AUTODOC
 * (open-input-file FILENAME)
 * Open the given FILENAME, and returns an input port to read it.
 * This function tries to open the given FILENAME (as is), and returns an
 * input port object suitable to get data out of it.
 * If the file cannot be opened, an 'error:bad-syscall is triggered.
 */

static OsObject OsSubrOpenInputFile (OsString filename) {
    OsShouldBeString (filename) ;
    return ((OsObject) OsInputPortMakeFromName (filename)) ;
}

/* AUTODOC
 * (open-input-string STRING)
 * Open given STRING as an input port.
 * This function creates an input port whose output will be the given STRING.
 * The resulting port can be used as any other input port: it supports exactly
 * the same set of operations.
 */

static OsObject OsSubrOpenInputString (OsString buffer) {
    OsShouldBeString (buffer) ;
    return ((OsObject) OsInputStringMake (buffer)) ;
}

/* AUTODOC
 * (read [INPUT-PORT])
 * Read an expression from the (optional) INPUT-PORT
 * This functions read a full EXPRESSION from the given INPUT-PORT. It will
 * block until a full expression is readable from the port, or an end of file
 * is encountered.
 * This function may trigger 'error:bad-syntax errors.
 */

static OsObject OsSubrRead (unsigned int argc, OsInputPort *argv) {
    OsInputPort port = OsCurrentInputPort ;
    switch (argc) {
      case 0:
        break ;
      case 1: 
        OsShouldBeInputPort (*argv) ;
        port = *argv ;
        break ;
      default:
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (1)
                               , OsIntegerMake (argc)) ;
    }
    return (OsRead(port)) ;
}

/* AUTODOC
 * (read-char [INPUT-PORT])
 * Read a single character from the optional INPUT-PORT.
 * This function blocks until a character is available on the given INPUT-PORT
 * and returns it when ready (or returns an end of file object if appropriate).
 * If no port is provided, it is defaulted to current input port.
 */

static OsObject OsSubrReadChar (unsigned int argc, OsInputPort *argv) {
    OsInputPort   port = OsCurrentInputPort ;
    unsigned char r ;
    switch (argc) {
      case 0:
        break ;
      case 1: 
        OsShouldBeInputPort (*argv) ;
        port = *argv ;
        break ;
      default:
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (1)
                               , OsIntegerMake (argc)) ;
    }
    if ( (r = OS_GETC (port)) )
        return ((OsObject) OsCharMake (r)) ;
    else
        return (OsA_Eof) ;
    /* NOT REACHED */
}


/* AUTODOC
 * (read-line [INPUT-PORT])
 * Read a line from the optional INPUT-PORT
 * Blocks until a line is available on the given INPUT-PORT (or current input
 * port if none is provided). A line is delimited by a '\n' character, the
 * returned string doesn't contain it.
 * If end of file is reached, an end of file object is returned.
 */

static OsObject OsSubrReadLine (unsigned int argc, OsInputPort *argv) {
    OsInputPort   port = OsCurrentInputPort ;
    unsigned char r ;
    unsigned int  len = 0 ;
    
    switch (argc) {
      case 0:
        break ;
      case 1:
        OsShouldBeInputPort (*argv) ;
        port = *argv ;
        break ;
      default:
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (1)
                               , OsIntegerMake (argc)) ;
    }
    while ( (r = OS_GETC (port)) ) {
        if ( r == '\n' ) {
            len = OS_SLEN (port) ;
            return ((OsObject) OsStringMakeFromBytes (len, OS_POP(port))) ;
        } else {
            OS_PUSHC (r, port) ;
        }
    }
    if ( (len = OS_SLEN(port)) )
        return ((OsObject) OsStringMakeFromBytes (len, OS_POP(port))) ;
    else
        return (OsA_Eof) ;
    /* NOT REACHED */
}

/* AUTODOC
 * (read-bytes INPUT-PORT COUNT)
 * Read as mush as COUNT bytes from the INPUT-PORT
 * This function reads as much as COUNT bytes on the given INPUT-PORT and
 * returns a string. If it reaches the end of file, it retrusn an eof object.
 * If no bytes are available on the port, it returns #f. This last situation
 * occurs on socket stream, for example.
 * This function may trigger the 'error:bad-syscall error.
 */

static OsObject OsSubrReadBytes (OsInputPort port, OsInteger count) {
    char buffer[BUFSIZ] ;
    int  got ;
    OsShouldBeInputPort (port) ;
    OsShouldBeInteger (count) ;
    if ( port->pbuf < port->ebuf ) {	/* check for buffered bytes first */
        OsString result = OsStringMakeFromBytes (port->ebuf - port->pbuf
                                                 , port->pbuf) ;
        port->pbuf = port->ebuf ;
        return ((OsObject) result) ;
    } else {				/* read from file descr */
        got = (count->ival > BUFSIZ) ? BUFSIZ : count->ival ;
        if ( (got = (port->os_read) (port, got, buffer)) < 0 ) {
            int error = errno ;
            (void) OsErrorTrigger (1, OsE_BadSyscall
                                   , OsStringMake ("read")
                                   , OsIntegerMake (error)
                                   , OsConsMake ((OsObject) port, OsA_Nil)) ;
        } else if ( ! got ) {
            return (OsA_False) ;
        }
        OsInterpCheckReadBytes (OsCurrentInterp(), got) ;
        return ((OsObject) OsStringMakeFromBytes (got, buffer)) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (peek-char [INPUT-PORT])
 * Peek a char on given optional INPUT-PORT.
 * This function get next available character on the given INPUT-PORT (or
 * current input port of none is provided) and returns it.
 * An eof object might be returned if the end of file is reached.
 */

static OsObject OsSubrPeekChar (unsigned int argc, OsInputPort *argv) {
    OsInputPort   port = OsCurrentInputPort ;
    unsigned char r ;
    switch (argc) {
      case 0:
        break ;
      case 1: 
        OsShouldBeInputPort (*argv) ;
        port = *argv ;
        break ;
      default:
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (1)
                               , OsIntegerMake (argc)) ;
        /* NOT REACHED */
    }
    if ( (r = OS_GETC (port)) ) {
        OS_UNGETC (r, port) ;
        return ((OsObject) OsCharMake (r)) ;
    } else {
        return (OsA_Eof) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (eobj-object? OBJECT)
 * Is OBJECT an end of file.
 * This function returns #t if OBJECT if an end of file, #f otherwise.
 */

static OsObject OsSubrEofObjectP (OsObject foo) {
    return ((foo == OsA_Eof) ? OsA_True : OsA_False) ;
}

/* AUTODOC
 * (close-input-port INPUT-PORT)
 * Closes the given input port.
 * This function closes the given INPUT-PORT. Trying to get data from a closed
 * input port will trigger a 'error:read-on-closed-port error.
 */

static OsObject OsSubrCloseInputPort (OsInputPort port) {
    OsShouldBeInputPort (port) ;
    OsInputPortClose (port) ;
    return ((OsObject) port) ;
}

/* AUTODOC
 * (with-input-from-file FILENAME THUNK)
 * Run THUNK with current input port set to FILENAME's input port.
 * This function first opens FILENAME as an input port, and than sets the
 * current input port to it. Finally it runs THUNK wit no argument, and close
 * the underlying input port.
 */
   
static OsObject OsSubrWithInputFromFile (OsString file, OsCodeBlock block) {
    OsInputPort  port ;
    OsInputPort  oldport = OsCurrentInputPort ;
    OsObject     error = (OsObject) 0, value = (OsObject) 0 ;
    
    OsShouldBeString (file) ;
    OsMustBeBlock (block) ;
    port = OsInputPortMakeFromName (file) ;
    OsCurrentInputPort = port ;
    value = OsCodeBlockApplyWithArray (OsCurrentInterp()
                                       , block
                                       , 0, (OsObject *) 0
                                       , &error) ;
    OsCurrentInputPort = oldport ;
    if ( error ) 
        os_uncatched_error (error) ;
    else
        return (value) ;
    /* NOT REACHED */
}

/* AUTODOC
 * (with-input-from-string STRING THUNK)
 * Run THUNK with current input port ser to STRING's input port.
 * This function first open STRING as an input port, than set current input
 * port to it, and finally runs THUNK is this environment.
 * Once the THUNK run is completed, the underlying input port is closed.
 */
 
static OsObject OsSubrWithInputFromString (OsString str, OsCodeBlock block) {
    OsInputPort  port ;
    OsInputPort  oldport = OsCurrentInputPort ;
    OsObject     error = (OsObject) 0, value = (OsObject) 0 ;
    
    OsShouldBeString (str) ;
    OsMustBeBlock (block) ;
    port = OsInputStringMake (str) ;
    OsCurrentInputPort = port ;
    value = OsCodeBlockApplyWithArray (OsCurrentInterp()
                                       , block
                                       , 0, (OsObject *) 0
                                       , &error) ;
    OsInputPortClose (port) ;
    OsCurrentInputPort = oldport ;
    if ( error ) 
        os_uncatched_error (error) ;
    else
        return (value) ;
    /* NOT REACHED */
}

/* AUTODOC
 * (current-input-port)
 * Returns current input port.
 * This function returns the current input port.
 */

static OsObject OsSubrCurrentInputPort () {
    return ((OsObject) OsCurrentInputPort) ;
}

/*
 * Message implementations:
 */

static OsObject os_inputport_finalize (OsInputPort iport) {
    OsInputPortClose (iport) ;
    return (OsA_True) ;
}

/*
 * Initialize :
 */

void OsInputPortInitialize (OsInterp interp, int *pargc, char **argv) {
    OsInputPortClass = OsClassDeclare ("InputPort", sizeof (OsInputPortRec)) ;
    OsDefineMsg (OsInputPortClass, OS_SEL_FINALIZE, os_inputport_finalize) ;
    OsA_Eof = (OsObject) OsAtomMake (interp, "<eof>", OsA_Unbound) ;
    osa_send_msg = (OsObject) OsAtomMake (interp, "sys:send-msg", OsA_Unbound);
    OsCurrentInputPort = OsInputPortMakeFromFileno (fileno(stdin)) ;
    return ;
}

void OsInputPortSubrInitialize (OsInterp interp) {
    ose_filectxt = OsErrorMake (interp
                                , "error:filectxt"
                                , "file %0 line %1"
                                , 2, OSE_NORMAL) ;
    ose_syntax = OsErrorMake (interp
                              , "error:bad-syntax"
                              , "invalid syntax: %0"
                              , 1, OSE_CATCHABLE) ;
    ose_rclose = OsErrorMake (interp
                              , "error:read-on-closed-port"
                              , "read on closed port %0"
                              , 1, OSE_CATCHABLE) ;
    ose_wclose = OsErrorMake (interp
                              , "error:write-on-closed-port"
                              , "write on closed port %0"
                              , 1, OSE_CATCHABLE) ;
    OsSubrMake (interp, "input-port?", OsSubrInputPortP, 1) ;
    OsSubrMake (interp, "close-input-port", OsSubrCloseInputPort, 1) ;
    OsSubrMake (interp, "read", OsSubrRead, OS_NARY) ;
    OsSubrMake (interp, "read-line", OsSubrReadLine, OS_NARY) ;
    OsSubrMake (interp, "read-bytes", OsSubrReadBytes, 2) ;
    OsSubrMake (interp, "read-char", OsSubrReadChar, OS_NARY) ;
    OsSubrMake (interp, "peek-char", OsSubrPeekChar, OS_NARY) ;
    OsSubrMake (interp, "eof-object?", OsSubrEofObjectP, 1) ;
    OsSubrMake (interp, "sys:load", OsSubrSysLoad, 1) ;
    OsSubrMake (interp, "current-input-port", OsSubrCurrentInputPort, 0) ;
    OsSecSubrMake (interp, "open-input-file", OsSubrOpenInputFile
                   , 1, OS_SECURITY_LEVEL_D) ;
    OsSecSubrMake (interp, "open-input-string", OsSubrOpenInputString
                   , 1, OS_SECURITY_LEVEL_E) ;
    OsSecSubrMake (interp, "with-input-from-file", OsSubrWithInputFromFile
                   , 2, OS_SECURITY_LEVEL_D) ;
    OsSecSubrMake (interp, "with-input-from-string", OsSubrWithInputFromString
                   , 2, OS_SECURITY_LEVEL_E);
    return ;
}

/***************************************************************************
 * Output ports
 ***************************************************************************/

/*
 * Internals method definitions for output ports
 */

static void oport_putc (port, c)
    OsOutputPort  port ;
    unsigned char c ;
{
    if ( ! port->medium.fp ) {
        (void) OsErrorTrigger (1, ose_wclose, port) ;
    }
    fputc (c, port->medium.fp) ;
    return ;
}

static void oport_puts (OsOutputPort port, char *cstr) {
    if ( ! port->medium.fp ) {
        (void) OsErrorTrigger (1, ose_wclose, port) ;
    }
    fputs (cstr, port->medium.fp) ;
    return ;
}

static void oport_write (OsOutputPort port, unsigned int len, char *bytes) {
    if ( ! port->medium.fp ) {
        (void) OsErrorTrigger (1, ose_wclose, port) ;
    }
    while ( len-- ) {
        fputc (*bytes, port->medium.fp) ;
        bytes++ ;
    }
    return ;
}

static void oport_flush (OsOutputPort port) {
    if ( port->medium.fp )
        fflush (port->medium.fp) ;
    return ;
}

static void oport_close (OsOutputPort port) {
    if ( port->medium.fp ) {
        fclose (port->medium.fp) ;
        port->medium.fp = (FILE *) 0 ;
    }
    return ;
}

static int oport_ready (OsOutputPort port) {
    return (fileno(port->medium.fp)) ;
}


/*
 * Internal method defintions for output strings:
 * We use here CBytesAppend, which side-effect the string, instead of the more
 * conventional CAppend, which returns a new string.
 */

#define OSTRING_BUFLEN	256
#define OSTRING_STRING(p) ((p)->medium.ostring.string)
#define OSTRING_STRBUF(p) ((p)->medium.ostring.strbuf)
#define OSTRING_PBUF(p)   ((p)->medium.ostring.pbuf)
#define OSTRING_BUFSIZ(p) (OSTRING_PBUF(p) - OSTRING_STRBUF(p))

static void ostring_putc (port, c)
    OsOutputPort  port ;
    unsigned char c ;
{
    if ( ! OSTRING_STRING (port) ) {
        (void) OsErrorTrigger (1, ose_wclose, port) ;
    }
    if ( OSTRING_STRBUF(port) && ( OSTRING_BUFSIZ(port) >= OSTRING_BUFLEN ) ) {
        port->os_flush (port) ;
    } else if ( ! OSTRING_STRBUF (port) ) {
        OSTRING_STRBUF(port) = OsGCMallocAtomic(OSTRING_BUFLEN) ;
        OSTRING_PBUF(port)   = OSTRING_STRBUF(port) ;
    }
    *(OSTRING_PBUF(port)++) = c ;
    return ;
}

static void ostring_puts (OsOutputPort port, char *cstr) {
    unsigned int len ;
    if ( ! OSTRING_STRING (port) ) {
        (void) OsErrorTrigger (1, ose_wclose, port) ;
    }
    len = strlen (cstr) ;
    if (OSTRING_STRBUF(port) && (OSTRING_BUFSIZ(port)+len < OSTRING_BUFLEN)) {
        bcopy (cstr, OSTRING_PBUF(port), len) ;
        OSTRING_PBUF (port) += len ;
    } else {
        port->os_flush (port) ;
        OsStringCBytesAppend (OSTRING_STRING (port), len, cstr) ;
    }
    return ;
}

static void ostring_write (OsOutputPort port, unsigned int len, char *bytes) {
    if ( ! OSTRING_STRING (port) ) {
        (void) OsErrorTrigger (1, ose_wclose, port) ;
    }
    if (OSTRING_STRBUF(port) && (OSTRING_BUFSIZ(port)+len < OSTRING_BUFLEN) ) {
        bcopy (bytes, OSTRING_PBUF (port), len) ;
        OSTRING_PBUF (port) += len ;
    } else {
        port->os_flush (port) ;
        OsStringCBytesAppend (OSTRING_STRING(port), len, bytes) ;
    }
    return ;
}

static void ostring_flush (OsOutputPort port) {
    if ( OSTRING_STRBUF (port) ) {
        OsStringCBytesAppend (OSTRING_STRING(port)
                              , OSTRING_BUFSIZ(port)
                              , OSTRING_STRBUF(port)) ;
        OSTRING_PBUF (port) = OSTRING_STRBUF (port) ;
    }
    return ;
}

static void ostring_close (OsOutputPort port) {
    if ( OSTRING_STRING (port) ) {
        port->os_flush (port) ;
        OSTRING_STRING (port) = (OsString) 0 ;
    }
    return ;
}

static int ostring_ready (OsOutputPort port) {
    return (-1) ;
}

/*
 * C-API
 */

OsOutputPort OsOutputPortMakeFromFile (FILE *fp) {
    OsOutputPort port = (OsOutputPort) OsObjectMake (OsOutputPortClass, 0) ;
    port->os_putc   = oport_putc ;
    port->os_puts   = oport_puts ;
    port->os_write  = oport_write ;
    port->os_flush  = oport_flush ;
    port->os_close  = oport_close ;
    port->os_ready  = oport_ready ;
    port->medium.fp = fp ;
    return (port) ;
}

OsOutputPort OsOutputPortMakeFromName (OsString filename) {
    FILE *fp = fopen (filename->string, "w") ;
    if ( fp ) {
        return (OsOutputPortMakeFromFile (fp)) ;
    } else {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("fopen")
                               , OsIntegerMake (error)
                               , OsConsMake ((OsObject) filename, OsA_Nil)) ;
    }
    /* NOT REACHED */
}

OsOutputPort OsOutputStringMake (OsString buffer) {
    OsOutputPort port = (OsOutputPort) OsObjectMake (OsOutputPortClass, 0) ;
    port->os_putc  = ostring_putc ;
    port->os_puts  = ostring_puts ;
    port->os_write = ostring_write ;
    port->os_flush = ostring_flush ;
    port->os_close = ostring_close ;
    port->os_ready = ostring_ready ;
    port->medium.ostring.string = buffer ;
    return (port) ;
}

OsOutputPort OsFormatWithList (OsOutputPort port, OsString fmt, OsCons lst) {
    unsigned char *pf, *ps ;
    unsigned int  len = OsConsLength (lst) ;
    OsCons        plst ;
    
    for (ps = pf = fmt->string ; *pf ; ) {
        switch (*pf++) {
          case '\\':
            if (*pf) pf++ ;
            break ;
          case '%':
            if ( isdigit (*pf) ) {
                int idx = (*pf - '0') ;
                if ( idx >= len ) {
                    (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                           , OsIntegerMake (len)
                                           , OsIntegerMake (idx)) ;
                } else {
                    for (plst = lst ; idx ; idx--)
                        plst = (OsCons) OS_CONS_CDR (plst) ;
                    OsOutputPortWrite (port, pf - ps - 1, ps) ;
                    OsSendDisplay (OS_CONS_CAR(plst), port) ;
                    ps = ++pf ;
                }
            }
            break ;
          default:
            continue ;
        }
    }
    if ( ps != pf )
        OsOutputPortWrite (port, pf - ps, ps) ;
    return (port) ;
}

OsOutputPort OsFormatWithArray (port, fmt, argc, argv)
    OsOutputPort port ;
    OsString     fmt ;
    unsigned int argc ;
    OsObject     *argv ;
{
    unsigned char *pf, *ps ;
    
    for (ps = pf = fmt->string ; *pf ; ) {
        switch (*pf++) {
          case '\\':
            if (*pf) pf++ ;
            break ;
          case '%':
            if ( isdigit (*pf) ) {
                int idx = (*pf - '0') ;
                if ( idx >= argc ) {
                    (void) OsErrorTrigger (1, OsE_ArityShouldBe
                                           , OsIntegerMake (argc)
                                           , OsIntegerMake (idx)) ;
                } else {
                    OsOutputPortWrite (port, pf - ps - 1, ps) ;
                    OsSendDisplay (argv[idx], port) ;
                    ps = ++pf ;
                }
            }
            break ;
          default:
            continue ;
        }
    }
    if ( ps != pf )
        OsOutputPortWrite (port, pf - ps, ps) ;
    return (port) ;
}

/*
 * Output port subrs:
 */

/* AUTODOC
 * (output-port? OBJECT)
 * Is OBJECT an output port ?
 * Returns #t if object is an output port, #f otherwise.
 */

static OsObject OsSubrOutputPortP (OsObject port) {
    return (OsIsOutputPort (port) ? OsA_True : OsA_False) ;
}
            
/* AUTODOC
 * (display OBJECT [OUTPUT-PORT])
 * Display OBJECT to the optional OUTPUT-PORT.
 * Display an external representation of OBJECT on the provided OUTPUT-PORT. If
 * no port is provided, than the current output port is used.
 */

static OsObject OsSubrDisplay (unsigned int argc, OsObject *argv) {
    OsOutputPort port = OsCurrentOutputPort ;

    if ( argc == 2 ) {
        port = (OsOutputPort) argv[1] ;
        OsShouldBeOutputPort (port) ;
    } else if ( argc != 1 ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (1)
                               , OsIntegerMake (argc)) ;
        /* NOT REACHED */
    }
    OsSendDisplay (*argv, port) ;
    return (*argv) ;
}

/* AUTODOC
 * (write OBJECT [OUTPUT-PORT])
 * Write OBJECT to the optional OUTPUT-PORT.
 * Write an external representation of OBJECT, that can be read back through
 * the 'read primitive. This means, in particular that strings get displayed
 * with escaped characters, etc. If no OUTPUT-PORT is provided, default is to
 * write to the current output port.
 */

static OsObject OsSubrWrite (unsigned int argc, OsObject *argv) {
    OsOutputPort port = OsCurrentOutputPort ;

    if ( argc == 2 ) {
        port = (OsOutputPort) argv[1] ;
        OsShouldBeOutputPort (port) ;
    } else if ( argc != 1 ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (1)
                               , OsIntegerMake (argc)) ;
        /* NOT REACHED */
    }
    OsSendWrite (*argv, port) ;
    return (*argv) ;
}

/* AUTODOC
 * (open-output-file FILENAME)
 * Make an output port from the given FILENAME.
 * This primitive creates an output port from the given FILENAME. If the 
 * FILENAME already exists, it truncated to zero, otherwise the corresponding 
 * FILENAME is created.
 */

static OsObject OsSubrOpenOutputFile (OsString filename) {
    OsShouldBeString (filename) ;
    return ((OsObject) OsOutputPortMakeFromName (filename)) ;
}

/* AUTODOC
 * (output-string-make STRING)
 * Make an output port from the given STRING.
 * This primitive creates an output port whose output will be appended to the 
 * given STRING. The string is side-effected on each write operation to the
 * port, take care !
 */

static OsObject OsSubrOpenOutputString (OsString buffer) {
    OsShouldBeString (buffer) ;
    return ((OsObject) OsOutputStringMake (buffer)) ;
}

/* AUTODOC
 * (newline [OUTPUT-PORT])
 * Display a newline character to optional OUTPUT-PORT.
 * Newline display a suitable line break to the provided OUTPUT-PORT, or to
 * the current output port if none is provided. The same effect can be reached 
 * be display the special character "\n" in a port.
 */

static OsObject OsSubrNewline (unsigned int argc, OsObject *argv) {
    OsOutputPort port = OsCurrentOutputPort ;
    if ( argc == 1 ) {
        port = (OsOutputPort) argv[0] ;
        OsShouldBeOutputPort (port) ;
    } else if ( argc ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (0)
                               , OsIntegerMake (argc)) ;
    }
    OsOutputPortWriteChar (port, '\n') ;
    OsOutputPortFlush (port) ;
    return ((OsObject) port) ;
}

/* AUTODOC
 * (write-char CHAR [OUTPUT-PORT])
 * Display the given CHAR to the optional OUTPUT-PORT
 * The CHAR is displayed to the optional OUTPUT-PORT or to the current output
 * port if none is provided.
 */

static OsObject OsSubrWriteChar (unsigned int argc, OsObject *argv) {
    OsOutputPort port = OsCurrentOutputPort ;
    OsChar       chr ;
    if ( argc == 2 ) {
        port = (OsOutputPort) argv[1] ;
        OsShouldBeOutputPort (port) ;
    } else if ( argc != 1 ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (1)
                               , OsIntegerMake (argc)) ;
        /* NOT REACHED */
    }
    chr = (OsChar) argv[0] ;
    OsShouldBeChar (chr) ;
    OsOutputPortWriteChar (port, chr->c) ; 
    return (argv[0]) ;
}

/* AUTODOC
 * (current-output-port)
 * Returns the current output port.
 * Returns the current output port.
 */

static OsObject OsSubrCurrentOutputPort () {
    return ((OsObject) OsCurrentOutputPort) ;
}

/* AUTODOC
 * (close-output-port OUTPUT-PORT)
 * Close the given OUTPUT-PORT.
 * This function flushes any pending output to the given port, and then closes
 * it. 
 * Writing to a closed output port will trigger the 'error:write-on-closed-port
 * error.
 */

static OsObject OsSubrCloseOutputPort (OsOutputPort port) {
    OsShouldBeOutputPort (port) ;
    OsOutputPortClose (port) ;
    return ((OsObject) port) ;
}

/* AUTODOC
 * (with-output-to-file FILENAME THUNK)
 * Run THUNK with current output port set to FILENAME
 * This function creates an output port from the given FILENAME, set current
 * output port to it, and runs THUNK in this enviornement.
 * Once THUNK is completed, it closes the underlying port, and returns THUNK
 * returned value.
 */

static OsObject OsSubrWithOutputToFile (OsString file, OsCodeBlock block) {
    OsOutputPort port ;
    OsOutputPort oldport = OsCurrentOutputPort ;
    OsObject     error = (OsObject) 0, value = (OsObject) 0 ;
    
    OsShouldBeString (file) ;
    OsMustBeBlock (block) ;
    port = OsOutputPortMakeFromName (file) ;
    OsCurrentOutputPort = port ;
    value = OsCodeBlockApplyWithArray (OsCurrentInterp()
                                       , block
                                       , 0, (OsObject *) 0
                                       , &error) ;
    OsCurrentOutputPort = oldport ;
    OsOutputPortClose (port) ;
    if ( error ) 
        os_uncatched_error (error) ;
    else
        return (value) ;
    /* NOT REACHED */
}

/* AUTODOC
 * (with-output-to-string FILENAME THUNK)
 * Run THUNK with current output port set to STRING
 * This function creates an output port from the given STRING, set current
 * output port to it, and runs THUNK in this enviornement.
 * Once THUNK is completed, it closes the underlying port, and returns THUNK
 * returned value.
 */

static OsObject OsSubrWithOutputToString (OsString string, OsCodeBlock block) {
    OsOutputPort port ;
    OsOutputPort oldport = OsCurrentOutputPort ;
    OsObject     error = (OsObject) 0, value = (OsObject) 0 ;
    
    OsShouldBeString (string) ;
    OsMustBeBlock (block) ;
    port = OsOutputStringMake (string) ;
    OsCurrentOutputPort = port ;
    value = OsCodeBlockApplyWithArray (OsCurrentInterp()
                                       , block
                                       , 0, (OsObject *) 0
                                       , &error) ;
    OsCurrentOutputPort = oldport ;
    OsOutputPortClose (port) ;
    if ( error ) 
        os_uncatched_error (error) ;
    else
        return (value) ;
    /* NOT REACHED */
}

/* AUTODOC
 * (output-port-flush [OUTPUT-PORT])
 * Flush current output port.
 * This function writes all pending character of the optional OUTPUT-PORT to
 * its corresponding physical device, and returns the prot. If OUTPUT-PORT is
 * omitted, it defaults to current output port.
 */

static OsObject OsSubrOutputPortFlush (argc, argv)
    unsigned int argc ;
    OsOutputPort *argv ;
{
    OsOutputPort port = OsCurrentOutputPort ;

    switch (argc) {
      case 0:
        break ;
      case 1:
        port = argv[0] ;
        OsShouldBeOutputPort (port) ;
        break ;
      default:
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (1)
                               , OsIntegerMake (argc)) ;
        /* NOT REACHED */
    }
    OsOutputPortFlush (port) ;
    return ((OsObject) port) ;
}

/* AUTODOC
 * (format OUTPUT-PORT FORMAT-STRING [OBJECTS*])
 * A simple formattinbg utility.
 * The format primitive output the resulting of formatting OBJECTS according
 * to FORMAT-STRING to OUTPUT-PORT. FORMAT-STRING is any string where the 
 * special characters '%n' where n is a digit will be replaced by the nth 
 * OBJECTS.
 */


static OsObject OsSubrFormat (unsigned int argc, OsObject *argv) {
    OsOutputPort  port = OsCurrentOutputPort ;
    OsString      fmt ;
    
    if ( argc < 1 )
        return (OsA_True) ;
    port = (OsOutputPort) argv[0] ;
    OsShouldBeOutputPort (port) ;
    if ( argc < 2 )
        return (OsA_True) ;
    fmt = (OsString) argv[1] ;
    OsShouldBeString (fmt) ;
    argv += 2 ; argc -= 2 ;
    (void) OsFormatWithArray (port, fmt, argc, argv) ;
    return (OsA_True) ;
}

/* AUTODOC
 * (current-error-port)
 * Returns current error port.
 */

static OsObject OsSubrCurrentErrorPort () {
    return ((OsObject) OsCurrentErrorPort) ;
}

/*
 * Message implementations:
 */

static OsObject os_outputport_finalize (OsOutputPort port) {
    OsOutputPortClose (port) ;
    return ((OsObject) port) ;
}

/*
 * Initialize :
 */

void OsOutputPortInitialize (OsInterp interp, int *pargc, char **argv) {
    OsOutputPortClass = OsClassDeclare ("OutputPort", sizeof(OsOutputPortRec));
    OsDefineMsg (OsOutputPortClass, OS_SEL_FINALIZE, os_outputport_finalize) ;
    OsCurrentOutputPort = OsOutputPortMakeFromFile (stdout) ;
    OsCurrentErrorPort  = OsOutputPortMakeFromFile (stderr) ;
    return ;
}

void OsOutputPortSubrInitialize (OsInterp interp) {
    OsSubrMake (interp, "output-port?", OsSubrOutputPortP, 1) ;
    OsSubrMake (interp, "display", OsSubrDisplay, OS_NARY) ;
    OsSubrMake (interp, "write", OsSubrWrite, OS_NARY) ;
    OsSubrMake (interp, "newline", OsSubrNewline, OS_NARY) ;
    OsSubrMake (interp, "write-char", OsSubrWriteChar, OS_NARY) ;
    OsSubrMake (interp, "current-output-port", OsSubrCurrentOutputPort, 0) ;
    OsSubrMake (interp, "close-output-port", OsSubrCloseOutputPort, 1) ;
    OsSubrMake (interp, "output-port-flush", OsSubrOutputPortFlush, OS_NARY) ;
    OsSubrMake (interp, "format", OsSubrFormat, OS_NARY) ;
    OsSecSubrMake (interp, "current-error-port", OsSubrCurrentErrorPort
                   , 0, OS_SECURITY_LEVEL_C) ;
    OsSecSubrMake (interp, "open-output-file", OsSubrOpenOutputFile
                   , 1, OS_SECURITY_LEVEL_C) ;
    OsSecSubrMake (interp, "open-output-string", OsSubrOpenOutputString
                   , 1, OS_SECURITY_LEVEL_C) ;
    OsSecSubrMake (interp, "with-output-to-file", OsSubrWithOutputToFile
                   , 2, OS_SECURITY_LEVEL_C) ;
    OsSecSubrMake (interp, "with-output-to-string", OsSubrWithOutputToString
                   , 2, OS_SECURITY_LEVEL_C) ;
    return ;
}
