/***************************************************************************
 * ptr.c
 * ptr.c,v 1.1.1.1 1995/02/27 15:15:31 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * External pointer handling. Created objects are WEAK in the sense of
 * Bohem's GC. 
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <oscheme.h>

OsClass OsPointerClass = (OsClass) 0 ;

static OsHashtable pointer_hashtable = (OsHashtable) 0 ;

OsError OsE_InvalidPointer = (OsError) 0 ;

/*
 * C API:
 */

OsPointer OsSharedPointerMake (void *pointer) {
    OsPointer ptr = (OsPointer) OsHashtableGet (pointer_hashtable, pointer) ;
    if ( ! ptr ) {
        ptr          = (OsPointer) OsObjectMake (OsPointerClass, 1) ;
        ptr->flags  |= OS_POINTER_SHARED ;
        ptr->pointer = pointer ;
        OsHashtablePut (pointer_hashtable, pointer, (void *) ptr) ;
    }
    return (ptr) ;
}

OsPointer OsPointerMake (void *pointer) {
    OsPointer ptr = (OsPointer) OsObjectMake (OsPointerClass, 1) ;
    ptr->pointer = pointer ;
    return (ptr) ;
}

/*
 * Subrs:
 */

static OsObject OsSubrPointerP (OsObject o) {
    return (OsIsPointer(o) ? OsA_True : OsA_False) ;
}

/*
 * Message implementations:
 */

static OsObject os_pointer_equal (OsPointer p1, OsPointer p2) {
    return (((p1 == p2) && (p1->pointer != (void *) 0))
            ? OsA_True
            : OsA_False) ;
}

static OsObject os_pointer_display (OsPointer p, int sel, OsOutputPort port) {
    char tmp[OS_DISPLAY_BUFSIZ] ;
    sprintf (tmp, "#[pointer(%s) -> %p]"
             , OsPointerIsValid (p) ? "alife" : "dead"
             , p->pointer) ;
    OsOutputPortWriteCString (port, tmp) ;
    return ((OsObject) p) ;
}

/*
 * Initialize :
 */

void OsPointerInitialize (OsInterp interp, int *pargc, char **argv) {
    OsPointerClass = OsClassDeclare ("Pointer", sizeof (OsPointerRec)) ;
    OsDefineMsg (OsPointerClass, OS_SEL_EQUAL, os_pointer_equal) ;
    OsDefineMsg (OsPointerClass, OS_SEL_DISPLAY, os_pointer_display) ;
    pointer_hashtable = OsHashtableMake (0
                                         , (OsHashFct) OsHashPointer
                                         , (OsCmpFct) OsCmpPointer) ;
    return ;
}

void OsPointerSubrInitialize (OsInterp interp) {
    OsE_InvalidPointer = OsErrorMake (interp
                                      , "error:invalid-pointer"
                                      , "freed pointer %0"
                                      , 1, OSE_CATCHABLE) ;
    OsSubrMake (interp, "pointer?", OsSubrPointerP, 1) ;
}
