/***************************************************************************
 * string.c
 * string.c,v 1.3 1995/03/17 14:34:09 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * OScheme string implementation.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>
#include <ctype.h>

#include <os_config.h>
#include <os_interp.h>
#include <os_obj.h>
#include <os_string.h>
#include <os_atom.h>
#include <os_gc.h>
#include <os_vector.h>
#include <os_error.h>
#include <os_cons.h>
#include <os_arit.h>
#include <os_cblock.h>
#include <os_subr.h>
#include <os_port.h>

/***************************************************************************
 * Chars
 ***************************************************************************/

OsClass OsCharClass = (OsClass) 0 ;

static OsError ose_bad_index = (OsError) 0 ;

static OsChar os_chars[256] = { (OsChar) 0, } ;

/*
 * Constructor:
 */

OsChar OsCharMake (unsigned char c) {
    return (os_chars[c & 0xFF]) ;
}

/*
 * Subrs:
 */

static OsObject OsSubrCharP (OsObject chr) {
    return (OsIsChar (chr) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharEgalP (OsChar c1, OsChar c2) {
    OsMustBeChar (c1) ;
    OsMustBeChar (c2) ;
    return ((c1 == c2) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharLessP (OsChar c1, OsChar c2) {
    OsMustBeChar (c1) ;
    OsMustBeChar (c2) ;
    return ((c1->c < c2->c) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharGreaterP (OsChar c1, OsChar c2) {
    OsMustBeChar (c1) ;
    OsMustBeChar (c2) ;
    return ((c1->c > c2->c) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharLessEP (OsChar c1, OsChar c2) {
    OsMustBeChar (c1) ;
    OsMustBeChar (c2) ;
    return ((c1->c <= c2->c) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharGreaterEP (OsChar c1, OsChar c2) {
    OsMustBeChar (c1) ;
    OsMustBeChar (c2) ;
    return ((c1->c >= c2->c) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharEgalCiP (OsChar c1, OsChar c2) {
    OsMustBeChar (c1) ;
    OsMustBeChar (c2) ;
    return ((tolower(c1->c) == tolower(c2->c)) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharLessCiP (OsChar c1, OsChar c2) {
    OsMustBeChar (c1) ;
    OsMustBeChar (c2) ;
    return ((tolower(c1->c) < tolower(c2->c)) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharGreaterCiP (OsChar c1, OsChar c2) {
    OsMustBeChar (c1) ;
    OsMustBeChar (c2) ;
    return ((tolower(c1->c) > tolower(c2->c)) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharLessCiEP (OsChar c1, OsChar c2) {
    OsMustBeChar (c1) ;
    OsMustBeChar (c2) ;
    return ((tolower(c1->c) <= tolower(c2->c)) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharGreaterCiEP (OsChar c1, OsChar c2) {
    OsMustBeChar (c1) ;
    OsMustBeChar (c2) ;
    return ((tolower(c1->c) >= tolower(c2->c)) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharAlphabeticP (OsChar chr) {
    OsMustBeChar (chr) ;
    return (isalpha (chr->c) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharNumericP (OsChar chr) {
    OsMustBeChar (chr) ;
    return (isdigit (chr->c) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharWhitespaceP (OsChar chr) {
    OsMustBeChar (chr) ;
    return ((chr->c == ' ') ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharUpperCaseP (OsChar chr) {
    OsMustBeChar (chr) ;
    return (isupper(chr->c) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharLowerCaseP (OsChar chr) {
    OsMustBeChar (chr) ;
    return (islower(chr->c) ? OsA_True : OsA_False) ;
}

static OsObject OsSubrCharToInteger (OsChar chr) {
    OsMustBeChar (chr) ;
    return ((OsObject) OsIntegerMake ((unsigned int) chr->c)) ;
}

static OsObject OsSubrIntegerToChar (OsInteger i) {
    OsMustBeInteger (i) ;
    return ((OsObject) OsCharMake ((unsigned char) (i->ival & 0xFF))) ;
}

static OsObject OsSubrCharUpcase (OsChar chr) {
    OsMustBeChar (chr) ;
    return ((OsObject) OsCharMake ((unsigned char) toupper(chr->c))) ;
}

static OsObject OsSubrCharDowncase (OsChar chr) {
    OsMustBeChar (chr) ;
    return ((OsObject) OsCharMake ((unsigned char) tolower(chr->c))) ;
}

/*
 * Msg implementations:
 */

static OsObject os_char_display (OsChar chr, int sel, OsOutputPort port) {
    OsOutputPortWriteChar (port, chr->c) ;
    return ((OsObject) chr) ;
}

static OsObject os_char_write (OsChar chr, int sel, OsOutputPort port) {
    OsOutputPortWriteChar (port, '#') ;
    OsOutputPortWriteChar (port, '\\') ;
    OsOutputPortWriteChar (port, chr->c) ;
    return ((OsObject) chr) ;
}

static OsObject os_char_equal (OsObject self, int sel, OsObject other) {
    /* Chars have uniq instances */
    return ((other == self) ? OsA_True : OsA_False) ;
}
        
void OsCharInitialize (OsInterp interp, int *pargc, char **argv) {
    OsChar        chr ;
    unsigned char c = ((unsigned char) 0) ;
    OsCharClass = OsClassDeclare ("Char", sizeof (OsCharRec)) ;
    OsDefineMsg (OsCharClass, OS_SEL_DISPLAY, os_char_display) ;
    OsDefineMsg (OsCharClass, OS_SEL_WRITE, os_char_write) ;
    OsDefineMsg (OsCharClass, OS_SEL_EQUAL, os_char_equal) ;
    do {
        chr = (OsChar) OsGCMallocAtomic (sizeof (OsCharRec)) ;
        chr->os_class = OsCharClass ;
        chr->c        = c ;
        os_chars[c] = chr ;
    } while ( ++c !=  0 ) ;
    return ;
}

void OsCharSubrInitialize (OsInterp interp) {
    OsSubrMake (interp, "char?", OsSubrCharP, 1) ;
    OsSubrMake (interp, "char=?", OsSubrCharEgalP, 2) ;
    OsSubrMake (interp, "char<?", OsSubrCharLessP, 2) ;
    OsSubrMake (interp, "char>?", OsSubrCharGreaterP, 2) ;
    OsSubrMake (interp, "char<=?", OsSubrCharLessEP, 2) ;
    OsSubrMake (interp, "char>=?", OsSubrCharGreaterEP, 2) ;
    OsSubrMake (interp, "char-ci=?", OsSubrCharEgalCiP, 2) ;
    OsSubrMake (interp, "char-ci<?", OsSubrCharLessCiP, 2) ;
    OsSubrMake (interp, "char-ci>?", OsSubrCharGreaterCiP, 2) ;
    OsSubrMake (interp, "char-ci<=?", OsSubrCharLessCiEP, 2) ;
    OsSubrMake (interp, "char-ci>=?", OsSubrCharGreaterCiEP, 2) ;
    OsSubrMake (interp, "char-alphabetic?", OsSubrCharAlphabeticP, 1) ;
    OsSubrMake (interp, "char-numeric?", OsSubrCharNumericP, 1) ;
    OsSubrMake (interp, "char-whitespace?", OsSubrCharWhitespaceP, 1) ;
    OsSubrMake (interp, "char-upper-case?", OsSubrCharUpperCaseP, 1) ;
    OsSubrMake (interp, "char-lower-case?", OsSubrCharLowerCaseP, 1) ;
    OsSubrMake (interp, "char->integer", OsSubrCharToInteger, 1) ;
    OsSubrMake (interp, "integer->char", OsSubrIntegerToChar, 1) ;
    OsSubrMake (interp, "char-upcase", OsSubrCharUpcase, 1) ;
    OsSubrMake (interp, "char-downcase", OsSubrCharDowncase, 1) ;
    return ;
}

/***************************************************************************
 * Strings
 ***************************************************************************/

OsClass OsStringClass = (OsClass) 0 ;

/*
 * Constructors:
 */

OsString OsStringMake (unsigned char *cstring) {
    OsString str = (OsString) OsObjectMake (OsStringClass, 0) ;
    str->len    = strlen (cstring) ;
    str->string = (unsigned char *) OsGCMallocAtomic (str->len + 1) ;
    if ( str->len )
        bcopy (cstring, str->string, str->len) ;
    str->string[str->len] = '\0' ;
    return (str) ;
}

OsString OsStringMakeFromBytes (unsigned int len, unsigned char *bytes) {
    OsString str = (OsString) OsObjectMake (OsStringClass, 0) ;
    str->string = (unsigned char *) OsGCMallocAtomic (len + 1) ;
    if ( (str->len = len) )
        bcopy (bytes, str->string, len) ;
    str->string[len] = '\0' ;
    return (str) ;
}

/*
 * C API:
 */

unsigned long OsStringHash (OsString str) {
    char *s = str->string ;
    unsigned long ret = 0;
    unsigned long ctr = 0;

    while (*(char*)s) {
        ret ^= *(char*)s++ << ctr;
        ctr = (ctr + 1) % sizeof (void *);
    }
    return (ret) ;
}

int OsStringCompare (OsString str1, OsString str2) {
    unsigned char *s1 = str1->string ;
    unsigned char *s2 = str2->string ;
    unsigned int len  = (str1->len > str2->len) ? str2->len : str1->len ;
    while ( len-- ) {
        if ( *s1++ != *s2++ )
            return ( *(s1-1) - *(s2-1) ) ;
    }
    return (((int) str1->len) - ((int) str2->len)) ;
}

int OsStringCiCompare (OsString str1, OsString str2) {
    unsigned char *s1 = str1->string ;
    unsigned char *s2 = str2->string ;
    unsigned int len  = (str1->len > str2->len) ? str2->len : str1->len ;
    while ( len-- ) {
        if ( tolower (*s1) != tolower (*s2) )
            return ( tolower (*s1) - tolower(*s2) ) ;
    }
    return (((int) str1->len) - ((int) str2->len)) ;
}

OsString OsStringCAppend (OsString str, char *cstr) {
    unsigned int  clen = strlen (cstr) ;
    unsigned char *s, *e ;
    OsString app = (OsString) OsObjectMake (OsStringClass, 0) ;

    app->len    = str->len + clen ;
    app->string = (unsigned char *) OsGCMallocAtomic (app->len+1) ;
    if ( str->len )
        bcopy (str->string, app->string, str->len) ;
    s = app->string + str->len ; e = s + clen ;
    while ( s < e )
        *s++ = *cstr++ ;
    *s++ = '\0' ;
    return (app) ;
}

OsString OsStringCBytesAppend (OsString str, unsigned int len, char *bytes) {
    if ( len ) {
        str->string = OsGCRealloc (str->string, str->len + len + 1) ;
        bcopy (bytes, str->string+str->len, len) ;
        str->len += len ;
        str->string[str->len] = '\0' ;
    }
    return (str) ;
}

int OsStringDepth (OsString string) {
    int  level       = 0 ;
    int  in_string   = 0 ;
    char *s          = string->string ;
    unsigned int len = string->len ;
    
    while ( len-- ) {
        switch (*s++) {
          case '\0':
            return ((level < 0) ? 0 : level) ;
          case '\\':
            s++ ;
            break ;
          case '"':
            if ( ( in_string = ( ! in_string ) ) )
                level++ ;
            else
                level-- ;
            break ;
          case '(': case '[':
            if ( ! in_string )
                level++ ;
            break ;
          case ')': case ']':
            if ( ! in_string )
                level-- ;
            break ;
        }
    }
    /* NOT REACHED */
}

/*
 * Subrs:
 */

/* AUTODOC
 * (string? OBJECT)
 * Returns #t if OBJECT is a string, otherwise returns #f.
 */

static OsObject OsSubrStringP (OsObject string) {
    return (OsIsString (string) ? OsA_True : OsA_False) ;
}

/* AUTODOC
 * (make-string K CHAR)
 * Make-string returns a newly allocated string of length K.
 * If CHAR is given, then all elements of the string are initialized to
 *  CHAR, otherwise the contents of the string are unspecified.
 */

static OsObject OsSubrMakeString (unsigned int argc, OsObject *argv) {
    unsigned char c = 0, *p, *e ;
    unsigned int  len ;
    OsString      string ;
    
    switch (argc) {
      case 2:
        OsMustBeChar (argv[1]) ;
        c = ((OsChar) argv[1])->c ;
        /* no break intentional */
      case 1:
        OsMustBeInteger (*argv) ;
        len = ((OsInteger) *argv)->ival ;
        break ;
      default:
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (-1)
                               , OsIntegerMake (argc)) ;
    }
    string         = (OsString) OsObjectMake (OsStringClass, 0) ;
    string->len    = len ;
    string->string = (unsigned char *) OsGCMallocAtomic (len + 1) ;
    for (p = string->string, e=p+string->len ; p < e ; p++ )
        *p = c ;
    *p++ = '\0' ;
    return ((OsObject) string) ;
}

/* AUTODOC
 * (string CHAR*)
 * Returns a newly allocated string composed of the arguments.
 */

static OsObject OsSubrString (unsigned int argc, OsObject *argv) {
    OsString string = (OsString) OsObjectMake (OsStringClass, 0) ;
    unsigned char *p, *e ;
    
    string->len    = argc ;
    string->string = (unsigned char *) OsGCMallocAtomic (string->len+1) ;
    for (p = string->string, e=p+string->len ; p < e ; p++ ) {
        OsMustBeChar (*argv) ;
        *p = ((OsChar) *argv++)->c ;
    }
    *p++ = '\0' ;
    return ((OsObject) string) ;
}
                                                                  
/* AUTODOC
 * (string-length STRING)
 * Returns the number of characters in the given string.
 */

static OsObject OsSubrStringLength (OsString str) {
    OsMustBeString (str) ;
    return ((OsObject) OsIntegerMake (str->len)) ;
}

/* AUTODOC
 * (string-ref STRING K)
 * String-ref returns character k of string using zero-origin indexing.
 * K must be a valid index of STRING.
 */

static OsObject OsSubrStringRef (OsString str, OsInteger idx) {
    OsMustBeString (str) ;
    OsMustBeInteger (idx) ;
    while ( (idx->ival < 0) || (idx->ival >= str->len) ) {
        idx = (OsInteger) OsErrorTrigger (1, ose_bad_index, idx, str) ;
    }
    return ((OsObject) OsCharMake (str->string[idx->ival])) ;
    /* NOT REACHED */
}

/* AUTODOC
 * (string-set! STRING K CHAR)
 * String-set! stores CHAR in element K of string.
 * It returns an unspecified value.
 */

static OsObject OsSubrStringSet (OsString str, OsInteger idx, OsChar chr) {
    OsMustBeString (str) ;
    OsMustBeInteger (idx) ;
    OsMustBeChar (chr) ;
    while ( (idx->ival < 0) || (idx->ival >= str->len) ) {
        idx = (OsInteger) OsErrorTrigger (1, ose_bad_index, idx, str) ;
    }
    str->string[idx->ival] = chr->c ;
    return ((OsObject) chr) ;
}

/* AUTODOC
 * (string-=? STRING1 STRING2)
 * Compares STRING1 and STRING2 for equality.
 * Returns #t if the two strings are the same length and contain the same
 * characters in the same positions, otherwise returns #f.
 */

static OsObject OsSubrStringEgalP (OsString s1, OsString s2) {
    OsMustBeString (s1) ;
    OsMustBeString (s2) ;
    if ( s1->len == s2->len ) {
        unsigned int len = s1->len ;
        unsigned char *s = s1->string, *d = s2->string ;
        while ( len-- ) {
            if ( *s++ != *d++ )
                return (OsA_False) ;
        }
        return (OsA_True) ;
    }
    return (OsA_False) ;
}

/* AUTODOC
 * (string-<? STRING1 STRING2)
 * Compares STRING1 and STRING2 for strict inferority.
 * String<? is the lexicographic ordering on strings induced by the ordering
 * char<? on characters. It returns #t if STRING1 is smaller than STRING2, #f
 * otherwise.
 */

static OsObject OsSubrStringLessP (OsString s1, OsString s2) {
    OsMustBeString (s1) ;
    OsMustBeString (s2) ;
    return ((OsStringCompare (s1, s2) < 0) ? OsA_True : OsA_False) ;
}
        
/* AUTODOC
 * (string->? STRING1 STRING2)
 * Compares STRING1 and STRING2 for strict superority.
 * String>? is the lexicographic ordering on strings induced by the ordering
 * char>? on characters. It returns #t if STRING1 is bigger than STRING2, #f
 * otherwise.
 */

static OsObject OsSubrStringGreaterP (OsString s1, OsString s2) {
    OsMustBeString (s1) ;
    OsMustBeString (s2) ;
    return ((OsStringCompare(s1, s2) > 0) ? OsA_True : OsA_False) ;
}
        
/* AUTODOC
 * (string-<=? STRING1 STRING2)
 * Compares STRING1 and STRING2 for inferority or equality.
 * String<=? is the lexicographic ordering on strings induced by the ordering
 * char<=? on characters. It returns #t if STRING1 is smaller than STRING2, #f
 * otherwise.
 */

static OsObject OsSubrStringLessEP (OsString s1, OsString s2) {
    OsMustBeString (s1) ;
    OsMustBeString (s2) ;
    return ((OsStringCompare (s1, s2) <= 0) ? OsA_True : OsA_False) ;
}
        
/* AUTODOC
 * (string->=? STRING1 STRING2)
 * Compares STRING1 and STRING2 for superiority or equality.
 * String>=? is the lexicographic ordering on strings induced by the ordering
 * char>=? on characters. It returns #t if STRING1 is bigger than STRING2, #f
 * otherwise.
 */

static OsObject OsSubrStringGreaterEP (OsString s1, OsString s2) {
    OsMustBeString (s1) ;
    OsMustBeString (s2) ;
    return ((OsStringCompare (s1, s2) >= 0) ? OsA_True : OsA_False) ;
}
        
/* AUTODOC
 * (string-ci=? STRING1 STRING2)
 * Compares STRING1 and STRING2 for equality, case insensitive.
 * String-ci=? is a case insensitive version of string-=?.
 */

static OsObject OsSubrStringEgalCiP (OsString s1, OsString s2) {
    OsMustBeString (s1) ;
    OsMustBeString (s2) ;
    return (( ! OsStringCiCompare (s1, s2)) ? OsA_True : OsA_False) ;
}
        
/* AUTODOC
 * (string-ci<? STRING1 STRING2)
 * Compares STRING1 and STRING2 for inferiority, case insensitive.
 * String-ci<? is a case insensitive version of string-<?.
 */

static OsObject OsSubrStringLessCiP (OsString s1, OsString s2) {
    OsMustBeString (s1) ;
    OsMustBeString (s2) ;
    return ((OsStringCiCompare (s1, s2) < 0) ? OsA_True : OsA_False) ;
}
        
/* AUTODOC
 * (string-ci>? STRING1 STRING2)
 * Compares STRING1 and STRING2 for superiority, case insensitive.
 * String-ci>? is a case insensitive version of string->?.
 */

static OsObject OsSubrStringGreaterCiP (OsString s1, OsString s2) {
    OsMustBeString (s1) ;
    OsMustBeString (s2) ;
    return ((OsStringCiCompare (s1, s2) > 0) ? OsA_True : OsA_False) ;
}
        
/* AUTODOC
 * (string-ci<=? STRING1 STRING2)
 * Compares STRING1 and STRING2 for inferiority or equality, case insensitive.
 * String-ci<=? is a case insensitive version of string-<=?.
 */

static OsObject OsSubrStringLessCiEP (OsString s1, OsString s2) {
    OsMustBeString (s1) ;
    OsMustBeString (s2) ;
    return ((OsStringCiCompare (s1, s2) <= 0) ? OsA_True : OsA_False) ;
}
        
/* AUTODOC
 * (string-ci>=? STRING1 STRING2)
 * Compares STRING1 and STRING2 for superiority or equality, case insensitive.
 * String-ci>=? is a case insensitive version of string->=?.
 */

static OsObject OsSubrStringGreaterCiEP (OsString s1, OsString s2) {
    OsMustBeString (s1) ;
    OsMustBeString (s2) ;
    return ((OsStringCiCompare (s1, s2) >= 0) ? OsA_True : OsA_False) ;
}

/* AUTODOC
 * (substring STRING START END)
 * Get part of a string.
 * String must be a string, and start and end must be exact integers satisfying
 * 0 <= start <= end <= (string-length STRING)
 * Substring returns a newly allocated string formed from the characters of
 * string beginning with index start (inclusive) and ending with index end
 * (exclusive). 
 */

static OsObject OsSubrSubstring (OsString s, OsInteger i1, OsInteger i2) {
    OsMustBeString (s) ;
    OsMustBeInteger (i1) ;
    OsMustBeInteger (i2) ;
    while ((i1->ival < 0) || (i1->ival > i2->ival) || (i2->ival > s->len)) {
        i1 = (OsInteger) OsErrorTrigger (1, ose_bad_index, i1, s) ;
    }
    return ((OsObject) OsStringMakeFromBytes (i2->ival - i1->ival
                                              , s->string + i1->ival)) ;
}

/* AUTODOC
 * (string-append STRING*)
 * Concatenate STRINGS to a signle string, and return it.
 * Returns a newly allocated string whose characters form the concatenation of
 * the given strings.
 */

static OsObject OsSubrStringAppend (unsigned int argc, OsString *argv) {
    OsString      str ;
    unsigned int  len, i ;
    unsigned char *d ;
    for (i = len = 0 ; i < argc ; i++ ) {
        OsMustBeString (argv[i]) ;
        len += argv[i]->len ;
    }
    str         = (OsString) OsObjectMake (OsStringClass, 0) ;
    str->len    = len ;
    str->string = d = (unsigned char *) OsGCMallocAtomic (len + 1) ;
    while ( argc-- ) {
        bcopy ((*argv)->string, d, (*argv)->len) ;
        d += (*argv++)->len ;
    }
    *d++ = '\0' ;
    return ((OsObject) str) ;
}

/* AUTODOC
 * (string->list STRING)
 * Get the list of characters that STRING is made of.
 * String->list returns a newly allocated list of the characters that make up
 * the given string.
 */

static OsObject OsSubrStringToList (OsString str) {
    OsObject start = OsA_Nil, *list = &start ;
    unsigned char *s, *e ;
    OsMustBeString (str) ;
    s = str->string ; e = s + str->len ;
    while ( s < e ) {
        *list = (OsObject) OsConsMake ((OsObject) OsCharMake (*s), OsA_Nil) ;
        list  = &(OS_CONS_CDR (*list)) ;
        s++ ;
    }
    return ((OsObject) start) ;
}

/* AUTODOC
 * (list->string CHARS*)
 * Create a string from the given CHARS.
 * Liststring returns a newly allocated string formed from the characters in
 * the list chars. Stringlist and liststring are inverses so far as equal? is
 * concerned.
 */

static OsObject OsSubrListToString (OsCons list) {
    if ( list == (OsCons) OsA_Nil ) {
        return ((OsObject) OsStringMakeFromBytes (0, (unsigned char *) 0)) ;
    } else {
        unsigned int  len = 0 ;
        unsigned char *d ;
        OsCons        p = list ;
        OsString      str ;
        
        OsMustBeCons (list) ;
        do {
            OsMustBeChar (OS_CONS_CAR (p)) ;
            len++ ;
            p = (OsCons) OS_CONS_CDR (p) ;
        } while ( OsIsCons (p) ) ;
        str         = (OsString) OsObjectMake (OsStringClass, 0) ;
        str->len    = len ;
        str->string = d =(unsigned char *) OsGCMallocAtomic (len + 1) ;
        while ( len-- ) {
            *d++ = ((OsChar) OS_CONS_CAR(list))->c ;
            list = (OsCons) OS_CONS_CDR(list) ;
        }
        *d++ = '\0' ;
        return ((OsObject) str) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (string-copy STRING)
 * Returns a newly allocated copy of the given string.
 */

static OsObject OsSubrStringCopy (OsString str) {
    OsMustBeString (str) ;
    return ((OsObject) OsStringMakeFromBytes (str->len, str->string)) ;
}

/* AUTODOC
 * (string-fill! STRING CHAR)
 * Fill the STRING with CHAR.
 *  Stores char in every element of the given string and returns an
 * unspecified value.
 */

static OsObject OsSubrStringFill (OsString str, OsChar chr) {
    unsigned char *s, *e ;
    OsMustBeString (str) ;
    OsMustBeChar (chr) ;
    s = str->string ; e = s + str->len ;
    while ( s < e )
        *s++ = chr->c ;
    return ((OsObject) str) ;
}

/*
 * Messages impl:
 */

static OsObject os_string_display (OsString str, int sel, OsOutputPort port) {
    char *s = str->string, *e = s + str->len ;
    while ( s < e ) {
        OsOutputPortWriteChar (port, *s) ;
        s++ ;
    }
    return ((OsObject) str) ;
}

static OsObject os_string_write (OsString str, int sel, OsOutputPort port) {
    char *s = str->string, *e = s + str->len ;
    OsOutputPortWriteChar (port, '"') ;
    while ( s < e ) {
        switch (*s) {
          case '\n':
            OsOutputPortWriteChar (port, '\\') ;
            OsOutputPortWriteChar (port, 'n') ;
            break ;
          case '\r':
            OsOutputPortWriteChar (port, '\\') ;
            OsOutputPortWriteChar (port, 'r') ;
            break ;
          case '\t':
            OsOutputPortWriteChar (port, '\\') ;
            OsOutputPortWriteChar (port, 't') ;
            break ;
          case '\\':
          case '"':
            OsOutputPortWriteChar (port, '\\') ;
            /* NO BREAK INTENTIONAL ! */
          default:
            OsOutputPortWriteChar (port, *s) ;
        }
        s++ ;
    }
    OsOutputPortWriteChar (port, '"') ;
    return ((OsObject) str) ;
}

static OsObject os_string_equal (OsString self, int sel, OsString other) {
    if ( OsIsString (other) )
        return ((! OsStringCompare (self, other)) ? OsA_True : OsA_False) ;
    return (OsA_False) ;
}

static unsigned long os_string_hashe (self, sel)
    OsString self ;
    int      sel ;
{
    return (OsStringHash (self)) ;
}

void OsStringInitialize (OsInterp interp, int *pargc, char **argv) {
    OsStringClass = OsClassDeclare ("String", sizeof (OsStringRec)) ;
    OsDefineMsg (OsStringClass, OS_SEL_DISPLAY, os_string_display) ;
    OsDefineMsg (OsStringClass, OS_SEL_WRITE, os_string_write) ;
    OsDefineMsg (OsStringClass, OS_SEL_EQUAL, os_string_equal) ;
    OsDefineMsg (OsStringClass, OS_SEL_HASHE, (OsMethod) os_string_hashe) ;
    return ;
}

void OsStringSubrInitialize (OsInterp interp) {
    ose_bad_index = OsErrorMake (interp
                                 , "error:bad-string-index"
                                 , "Invalid index %0 in string %1"
                                 , 2, OSE_NORMAL) ;
    OsSubrMake (interp, "string?", OsSubrStringP, 1) ;
    OsSubrMake (interp, "make-string", OsSubrMakeString, OS_NARY) ;
    OsSubrMake (interp, "string", OsSubrString, OS_NARY) ;
    OsSubrMake (interp, "string-length", OsSubrStringLength, 1) ;
    OsSubrMake (interp, "string-ref", OsSubrStringRef, 2) ;
    OsSubrMake (interp, "string-set!", OsSubrStringSet, 3) ;
    OsSubrMake (interp, "string=?", OsSubrStringEgalP, 2) ;
    OsSubrMake (interp, "string<?", OsSubrStringLessP, 2) ;
    OsSubrMake (interp, "string>?", OsSubrStringGreaterP, 2) ;
    OsSubrMake (interp, "string<=?", OsSubrStringLessEP, 2) ;
    OsSubrMake (interp, "string>=?", OsSubrStringGreaterEP, 2) ;
    OsSubrMake (interp, "string-ci=?", OsSubrStringEgalCiP, 2) ;
    OsSubrMake (interp, "string-ci<?", OsSubrStringLessCiP, 2) ;
    OsSubrMake (interp, "string-ci>?", OsSubrStringGreaterCiP, 2) ;
    OsSubrMake (interp, "string-ci<=?", OsSubrStringLessCiEP, 2) ;
    OsSubrMake (interp, "string-ci>=?", OsSubrStringGreaterCiEP, 2) ;
    OsSubrMake (interp, "substring", OsSubrSubstring, 3) ;
    OsSubrMake (interp, "string-append", OsSubrStringAppend, OS_NARY) ;
    OsSubrMake (interp, "string->list", OsSubrStringToList, 1) ;
    OsSubrMake (interp, "list->string", OsSubrListToString, 1) ;
    OsSubrMake (interp, "string-copy", OsSubrStringCopy, 1) ;
    OsSubrMake (interp, "string-fill!", OsSubrStringFill, 2) ;
    return ;
}
