/***************************************************************************
 * struct.c
 * struct.c,v 1.1.1.1 1995/02/27 15:15:22 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <oscheme.h>

typedef struct _struct_closure {
    OsObject     (*cfun)() ;
    unsigned int offset ;
    void         *data ;
} struct_closure_rec, *struct_closure ;

static OsEObjSel sel_pointer = (OsEObjSel) 0 ;

/*
 * Privates:
 */

static OsObject struct_slot_read (closure, self, sel)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
{
    struct_closure clos = (struct_closure) closure ;
    unsigned char *adr ;
    adr = (unsigned char *) OsPointerDeref (self->values->cells[0]) ;
    return ((clos->cfun)((void *) (adr + clos->offset)
			 , (OsObject) 0
			 , clos->data
			 , adr)) ;
}

static OsObject struct_slot_write (closure, self, sel, value)
    void      *closure ;
    OsEObj    self ;
    OsObject  value ;
    OsEObjSel sel ;
{
    struct_closure clos = (struct_closure) closure ;
    unsigned char *adr ;

    adr = (unsigned char *) OsPointerDeref (self->values->cells[0]) ;
    return ((clos->cfun)((unsigned char*) adr + clos->offset
			 , value
			 , clos->data
			 , adr)) ;
}

/*
 * Accessors :
 */

OsObject OsStructAccessorChar (loc, value, unused1, unused2)
    char   *loc ;
    OsChar value ;
    void   *unused1, *unused2 ;
{
    if ( value ) {
        OsShouldBeChar (value) ;
        *loc = (char) value->c ;
        return ((OsObject) value) ;
    } else {
        return ((OsObject) OsCharMake (*loc)) ;
    }
}

OsObject OsStructAccessorShort (loc, value, u1, u2)
    short     *loc ;
    OsInteger value ;
    void      *u1, *u2 ;
{
    if ( value ) {
        OsShouldBeInteger (value) ;
        *loc = (short) value->ival ;
        return ((OsObject) value) ;
    } else {
        return ((OsObject) OsIntegerMake ((int) *loc)) ;
    }
}

OsObject OsStructAccessorUnsignedShort (loc, value, u1, u2)
    unsigned short *loc ;
    OsInteger      value ;
    void           *u1, *u2 ;
{
    if ( value ) {
        OsShouldBeInteger (value) ;
        *loc = (unsigned short) value->ival ;
        return ((OsObject) value) ;
    } else {
        return ((OsObject) OsIntegerMake ((int) *loc)) ;
    }
}

OsObject OsStructAccessorInt (loc, value, u1, u2)
    int       *loc ;
    OsInteger value ;
    void      *u1, *u2 ;
{
    if ( value ) {
        OsShouldBeInteger (value) ;
        *loc = value->ival ;
        return ((OsObject) value) ;
    } else {
        return ((OsObject) OsIntegerMake (*loc)) ;
    }
}

OsObject OsStructAccessorLong (loc, value, u1, u2)
    long      *loc ;
    OsInteger value ;
    void      *u1, *u2 ;
{
    if ( value ) {
        OsShouldBeInteger (value) ;
        *loc = value->ival ;
        return ((OsObject) value) ;
    } else {
        return ((OsObject) OsIntegerMake ((int) *loc)) ;
    }
}

OsObject OsStructAccessorString (loc, value, u1, u2)
    char     **loc ;
    OsString value ;
    void     *u1, *u2 ;
{
    if ( value ) {
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("OsStructAccessorString")
                               , OsStringMake ("bad error handling")) ;
    } else {
	return ((OsObject) OsStringMake (*loc)) ;
    }
}

OsObject OsStructAccessorObject (loc, value, u1, u2)
    OsObject *loc ;
    OsObject value ;
    void     *u1, *u2 ;
{
    if ( value ) {
	return (*loc = value) ;
    } else {
	return (*loc) ;
    }
    /* NOT REACHED */
}

OsObject (*OsStructAccessorFind (unsigned int size, int sign))() {
    if ( size == sizeof (char) ) {
        return (OsStructAccessorChar) ;
    } else if ( size == sizeof (short) ) {
        return (sign
                ? (OsObject (*)()) OsStructAccessorShort
                : (OsObject (*)()) OsStructAccessorUnsignedShort) ;
    } else if ( size == sizeof (int) ) {
        return (OsStructAccessorInt) ;
    } else if (size == sizeof (long) ) {
        return (OsStructAccessorLong) ;
    } else {
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("OsStructFindAccessor")
                               , OsStringMake ("invalid accessor size")) ;
    }
    /* NOT REACHED */
}
        
/*
 * C API:
 */

OsEObj OsStructClassMake (char *name, unsigned int size) {
    OsEObj    sclass ; 
    OsPointer pointer ;
    
    sclass  = (OsEObj) OsEObjSendMsg0 ((OsEObj) OsEObjRoot
                                       , (OsEObj) OsEObjRoot
                                       , OsEObjSelClone);
    pointer = OsPointerMake ((void *) 0) ;
    OsPointerInvalidate (pointer) ;
    OsEObjAllocateSlot (sclass
                        , sel_pointer
                        , (OsEObjSel) 0
                        , (OsObject) pointer) ;
    return (sclass) ;
}

/*
 * This function assumes that the pointer slot of struct is always the first
 * one. This is really unfortunate, because it prevents ADDING the struct
 * functionnality to an existing objet. This is needed to bypass the write
 * protection for the pointer slot.
 * Anyway, we need to AllocateSlot() here, because pointer is copy-on-write
 * from the structclass.
 */
 
OsEObj OsStructMake (OsEObj sclass, OsPointer pointer) {
    OsEObj os_struct = (OsEObj) OsEObjSendMsg0 (sclass,sclass,OsEObjSelClone) ;
    OsEObjAllocateSlot (os_struct
                        , sel_pointer
                        , (OsEObjSel) 0
                        , (OsObject) pointer) ;
    return ((OsEObj) os_struct) ;
}

void *OsStructCPointer (sobj)
    OsEObj sobj ;
{
    OsPointer ptr = (OsPointer) OsEObjSendMsg0 (sobj, sobj, sel_pointer) ;
    return (OsPointerDeref (ptr)) ;
}

void OsStructRWSlotDeclare (OsEObj sclass
                            , char *name
                            , OsObject (*cfun)()
                            , unsigned int offset
                            , void *data) {
    struct_closure reader, writer ;
    char tmp[256] ;
    
    reader = (struct_closure) OsGCMalloc (sizeof (struct_closure_rec)) ;
    writer = (struct_closure) OsGCMalloc (sizeof (struct_closure_rec)) ;
    reader->cfun   = writer->cfun   = cfun ;
    reader->data   = writer->data   = data ;
    reader->offset = writer->offset = offset ;
    OsEObjDefineMsg (sclass
                     , OsEObjSelDefine (name)
                     , struct_slot_read
                     , (void *) reader) ;
    sprintf (tmp, "%s:", name) ;
    OsEObjDefineMsg (sclass
                     , OsEObjSelDefine (tmp)
                     , struct_slot_write
                     , (void *) writer) ;
    return ;
}

void OsStructRSlotDeclare (OsEObj sclass
                           , char *name
                           , OsObject (*cfun)()
                           , unsigned int offset
                           , void *data) {
    struct_closure reader ;

    reader = (struct_closure) OsGCMalloc (sizeof (struct_closure_rec)) ;
    reader->cfun   = cfun ;
    reader->data   = data ;
    reader->offset = offset ;
    OsEObjDefineMsg (sclass
                     , OsEObjSelDefine (name) 
                     , struct_slot_read
                     , (void *) reader) ;
    return ;
}

void OsStructWSlotDeclare (OsEObj sclass
                           , char *name
                           , OsObject (*cfun)()
                           , unsigned int offset
                           , void *data) {
    struct_closure writer ;
    char tmp[256] ;

    writer = (struct_closure) OsGCMalloc (sizeof (struct_closure_rec)) ;
    writer->cfun   = cfun ;
    writer->data   = data ;
    writer->offset = offset ;
    sprintf (tmp, "%s:", name) ;
    OsEObjDefineMsg (sclass
                     , OsEObjSelDefine (tmp) 
                     , struct_slot_write
                     , (void *) writer) ;
    return ;
}

void OsEObjStructInitialize (OsInterp interp, int *pargc, char **argv) {
    sel_pointer = OsEObjSelDefine ("pointer") ;
    return ;
}
    
        
