/***************************************************************************
 * vector.c
 * vector.c,v 1.3 1995/03/17 14:34:26 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * OScheme vector implementation.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <os_config.h>
#include <os_interp.h>
#include <os_obj.h>
#include <os_string.h>
#include <os_atom.h>
#include <os_gc.h>
#include <os_vector.h>
#include <os_error.h>
#include <os_arit.h>
#include <os_cons.h>
#include <os_cblock.h>
#include <os_subr.h>
#include <os_port.h>

OsClass OsVectorClass = (OsClass) 0 ;

static OsError ose_bad_index = (OsError) 0 ;

/*
 * Constructors:
 */

OsVector OsVectorNMakeWithDefault (unsigned int size, OsObject def) {
    OsVector vector = (OsVector) OsObjectMake (OsVectorClass, 0) ;
    if ( (vector->csize = vector->cmax = size) ) {
        OsObject *cells, *ecells ;
        vector->cells = (OsObject *) OsGCMalloc (sizeof (OsObject) * size) ;
        cells  = vector->cells ;
        ecells = vector->cells + vector->cmax ;
        while ( cells < ecells ) 
            *cells++ = def ;
    }
    return (vector) ;
}

OsVector OsVectorMake (unsigned int cmax) {
    OsVector vector = (OsVector) OsObjectMake (OsVectorClass, 0) ;
    vector->csize = 0 ;
    if ( vector->cmax = cmax ) 
        vector->cells = (OsObject *) OsGCMalloc (sizeof(OsObject) * cmax) ;
    return (vector) ;
}

OsVector OsWeakVectorMake (unsigned int cmax) {
    OsVector vector  = (OsVector) OsObjectMake (OsVectorClass, 0) ;
    vector->os_class = OsVectorClass ;
    vector->cmax     = cmax = (cmax ? cmax : OS_VECTOR_INCREMENT) ;
    vector->cells    = (OsObject *) OsGCMallocAtomic (sizeof(OsObject)*cmax) ;
    return (vector) ;
}

OsVector OsVectorMakeFromArray (unsigned int argc, OsObject *argv) {
    OsVector vector = OsVectorMake (argc) ;
    OsObject *cells = vector->cells ;
    vector->csize = argc ;
    while ( argc-- )
        *cells++ = *argv++ ;
    return (vector) ;
}

OsVector OsVector1Make (OsObject o1) {
    OsVector vector = OsVectorMake (1) ;
    vector->cells[0] = o1 ;
    vector->csize    = 1 ;
    return (vector) ;
}

OsVector OsVector2Make (OsObject o1, OsObject o2) {
    OsVector vector = OsVectorMake (2) ;
    vector->cells[0] = o1 ;
    vector->cells[1] = o2 ;
    vector->csize    = 2 ;
    return (vector) ;
}

OsVector OsVector3Make (OsObject o1, OsObject o2, OsObject o3) {
    OsVector vector = OsVectorMake (3) ;
    vector->cells[0] = o1 ;
    vector->cells[1] = o2 ;
    vector->cells[2] = o3 ;
    vector->csize    = 3 ;
    return (vector) ;
}

/*
 * C functions :
 */

unsigned int OsVectorFastAppend (OsVector vector, OsObject object) {
    unsigned int idx = vector->csize ;
    if ( vector->csize >= vector->cmax ) {
        unsigned int sz = sizeof(OsObject)*(vector->cmax+=OS_VECTOR_INCREMENT);
        if ( vector->cells )
            vector->cells = (OsObject *) OsGCRealloc (vector->cells, sz) ;
        else
            vector->cells = (OsObject *) OsGCMalloc (sz) ;
    }
    vector->cells[vector->csize++] = object ;
    return (idx) ;
}

int OsVectorFastRemoveq (OsVector vector, OsObject object) {
    OsObject *cells = vector->cells, *ecells = cells + vector->csize ;
    while ( cells < ecells ) {
        if ( *cells == object ) {
            if ( cells + 1 < ecells )
                bcopy (cells+1, cells, sizeof (OsObject) * (ecells-cells-1)) ;
            vector->csize-- ;
            return (cells - vector->cells) ;
        }
        cells++ ;
    }
    return (-1) ;
}

int OsVectorFastMemq (OsVector vector, OsObject object) {
    OsObject *cells = vector->cells, *ecells = cells + vector->csize ;
    while ( cells < ecells ) {
        if ( *cells++ == object )
            return (cells-vector->cells-1) ;
    }
    return (-1) ;
}

OsVector OsVectorFix (OsVector vector) {
    if ( vector->csize + OS_VECTOR_INCREMENT < vector->cmax ) {
        OsObject *oldcells = vector->cells, *cells, *ecells ;
        vector->cmax  = vector->csize ;
        vector->cells = (OsObject*) OsGCMalloc (sizeof(OsObject)*vector->cmax);
        cells  = vector->cells ;
        ecells = vector->cells + vector->cmax ;
        while ( cells < ecells )
            *cells++ = *oldcells++ ;
    }
    return (vector) ;
}

OsVector OsVectorTail (OsVector vector, unsigned int tail) {
    OsVector vtail = (OsVector) OsObjectMake (OsVectorClass, 0) ;
    if ( tail < vector->csize ) {
        OsObject *cells, *ecells, *vcells ;
        vtail->csize = vtail->cmax = vector->csize - tail ;
        vtail->cells = (OsObject *) OsGCMalloc (sizeof(OsObject)*vtail->cmax);
        vcells = vector->cells + tail ;
        cells  = vtail->cells ;
        ecells = vtail->cells + vtail->cmax ;
        while ( cells < ecells )
            *cells++ = *vcells++ ;
    }
    return (vtail) ;
}

/*
 * Subrs for vectors:
 */

/* AUTODOC
 * (vector OBJECT*)
 * Create a vector containing all OBJECT.
 * Returns a newly allocated vector whose elements contain the
 * given arguments. Analogous to list.
 *
 * (vector 'a 'b 'c)           ==>  #(a b c)
 */

static OsObject OsSubrVector (unsigned int argc, OsObject *argv) {
    return ((OsObject) OsVectorMakeFromArray (argc, argv)) ;
}

/* AUTODOC
 * (vector? OBJECT) 
 * Returns #t if OBJECT is a vector, otherwise returns #f.
 */

static OsObject OsSubrVectorP (OsObject object) {
    return ((OsObject) ((OsIsVector (object)) ? OsA_True : OsA_False)) ;
}

/* AUTODOC
 * (make-vector K [FILL]) 
 * Creates a vector of lnegth K, filled optionally with FILL.
 * Returns a newly allocated vector of K elements. If a second
 * argument is given, then each element is initialized to FILL.
 * Otherwise the initial contents of each element is unspecified.
 */

static OsObject OsSubrMakeVector (unsigned int argc, OsObject *argv) {
    OsObject fill = (OsObject) OsA_Nil ;
    if ( argc == 2 ) {
        fill = argv[1] ;
    } else if ( argc != 1 ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (1)
                               , OsIntegerMake (argc)) ;
    }
    OsMustBeInteger (*argv) ;
    return ((OsObject)OsVectorNMakeWithDefault(((OsInteger)*argv)->ival,fill));
}

/* AUTODOC
 * (vector-length VECTOR) 
 * Returns the number of elements in vector.
 */

static OsObject OsSubrVectorLength (OsVector vector) {
    OsMustBeVector (vector) ;
    return ((OsObject) OsIntegerMake (vector->csize)) ;
}

/* AUTODOC
 * (vector-ref VECTOR K)
 * Get the K element of VECTOR.
 * K must be a valid index of vector. Vector-ref returns the
 * contents of element K of VECTOR.
 *
 * (vector-ref '#(1 1 2 3 5 8 13 21) 5)  
 *              ==>  8
 */

static OsObject OsSubrVectorRef (OsVector vector, OsInteger idx) {
    OsMustBeVector (vector) ;
    OsMustBeInteger (idx) ;
    while ( (idx->ival < 0) || (idx->ival >= vector->csize) ) {
        idx = (OsInteger) OsErrorTrigger (1, ose_bad_index, idx, vector) ;
    }
    return (vector->cells[idx->ival]) ;
}

/* AUTODOC 
 * (vector-set! VECTOR K OBJECT) 
 * Replace the K element of VECTOR by OBJECT.  
 * K must be a valid index of VECTOR. Vector-set!  stores OBJECT in
 * element k of vector. The value returned by vector-set! is
 * unspecified.
 *
 *   (let ((vec (vector 0 '(2 2 2 2) "Anna")))
 *      (vector-set! vec 1 '("Sue" "Sue"))
 *      vec)      
 *              ==>  #(0 ("Sue" "Sue") "Anna")
 */

static OsObject OsSubrVectorSet (OsVector vector, OsInteger idx, OsObject o) {
    OsMustBeVector (vector) ;
    OsMustBeInteger (idx) ;
    while ( (idx->ival < 0) || (idx->ival >= vector->csize) ) {
        idx = (OsInteger) OsErrorTrigger (1, ose_bad_index, idx, vector) ;
    }
    return (vector->cells[idx->ival] = o) ;
}

/* AUTODOC
 * (vector-fill! VECTOR FILL)
 * Stores FILL in every element of VECTOR.
 * The value returned by vector-fill! is unspecified.
 */

static OsObject OsSubrVectorFill (OsVector vector, OsObject fill) {
    OsObject *cells, *ecells ;
    OsMustBeVector (vector) ;
    cells  = vector->cells ;
    ecells = vector->cells + vector->csize ;
    while ( cells < ecells )
        *cells++ = fill ;
    return ((OsObject) vector) ;
}

/* AUTODOC
 * (subvector VECTOR START END)
 * Returns a vector containing elements from START to END of VECTOR.
 */

static OsObject OsSubrSubvector (OsVector vector, OsInteger s, OsInteger e) {
    OsVector vsub ;

    OsShouldBeVector (vector) ;
    OsShouldBeInteger (s) ;
    OsShouldBeInteger (e) ;
    while ((s->ival < 0) || (e->ival < s->ival) || (e->ival >= vector->csize))
        s = (OsInteger) OsErrorTrigger (1, ose_bad_index, s, vector) ;
    vsub   = OsVectorMake (e->ival - s->ival + 1) ;
    vsub->csize = e->ival - s->ival + 1 ;
    bcopy (vector->cells + s->ival, vsub->cells, vsub->csize*sizeof(OsObject));
    return ((OsObject) vsub) ;
}

/* AUTODOC
 * (vector->list VECTOR)
 * Returns a newly allocated list made of the elements of VECTOR.
 *   (vector->list '#(dah dah didah))  
 *   ==>  (dah dah didah)
 */

static OsObject OsSubrVectorToList (OsVector vector) {
    OsObject start = (OsObject) OsA_Nil, *list = &start ;
    OsObject *cells = vector->cells, *ecells = vector->cells + vector->csize ;
    OsShouldBeVector (vector) ;
    while ( cells < ecells ) {
        *list = (OsObject) OsConsMake (*cells++, OsA_Nil) ;
        list  = &(OS_CONS_CDR (*list)) ;
    }
    return (start) ;
}

/* AUTODOC
 * (vector-removeq! VECTOR OBJECT)
 * Physically remove all elements of VECTOR 'eq? to OBJECT.
 */

static OsObject OsSubrVectorRemoveq (OsVector vector, OsObject object) {
    OsShouldBeVector (vector) ;
    if ( OsVectorFastRemoveq (vector, object) >= 0 )
        return (OsA_True) ;
    else
        return (OsA_False) ;
    /* NOT REACHED */
}

/* AUTODOC
 * (vector-append! VECTOR OBJECT)
 * Physically appends OBJECT to VECTOR.
 */

static OsObject OsSubrVectorAppend (vector, object)
    OsVector vector ;
    OsObject object ;
{
    OsShouldBeVector (vector) ;
    (void) OsVectorFastAppend (vector, object) ;
    return ((OsObject) vector) ;
}
        
/*
 * Message implementations
 * _display is used both for display and write, hence the SendMsg instead of
 * SendDisplay.
 */

static OsObject os_vector_display (OsVector vector, int s, OsOutputPort port) {
    OsObject *cells  = vector->cells ;
    
    OsOutputPortWriteCString (port, "#(") ;
    if ( cells ) {
        OsObject *ecells = vector->cells + vector->csize ;
        while ( cells < ecells ) {
            OsSendMsg (*cells, s, (*cells, s, port)) ;
            cells++ ;
            if ( cells < ecells )
                OsOutputPortWriteChar (port, ' ') ;
        }
    }
    OsOutputPortWriteChar (port, ')') ;
    return ((OsObject) vector) ;
}

static OsObject os_vector_equal (OsVector self, int sel, OsVector other) {
    if ( OsIsVector (other) && (other->csize == self->csize) ) {
        OsObject *sc = self->cells, *se = self->cells + self->csize ;
        OsObject *oc = other->cells ;
        while ( sc < se ) {
            if ( OsSendEqual (*sc, *oc) == OsA_False )
                return (OsA_False) ;
            sc++ ; oc++ ;
        }
        return (OsA_True) ;
    }
    return (OsA_False) ;
}

/*
 * Initialize :
 */

void OsVectorInitialize (OsInterp interp, int *pargc, char **argv) {
    OsVectorClass = OsClassDeclare ("Vector", sizeof (OsVectorRec)) ;
    OsDefineMsg (OsVectorClass, OS_SEL_DISPLAY, os_vector_display) ;
    OsDefineMsg (OsVectorClass, OS_SEL_WRITE, os_vector_display) ;
    OsDefineMsg (OsVectorClass, OS_SEL_EQUAL, os_vector_equal) ;
    return ;
}

void OsVectorSubrInitialize (OsInterp interp) {
    ose_bad_index = OsErrorMake (interp
                                 , "error:bad-vector-index"
                                 , "Invalid index %0 to vector %1"
                                 , 2, OSE_NORMAL) ;
    OsSubrMake (interp, "vector?", OsSubrVectorP, 1) ;
    OsSubrMake (interp, "make-vector", OsSubrMakeVector, OS_NARY) ;
    OsSubrMake (interp, "vector-length", OsSubrVectorLength, 1) ;
    OsSubrMake (interp, "vector-ref", OsSubrVectorRef, 2) ;
    OsSubrMake (interp, "vector-set!", OsSubrVectorSet, 3) ;
    OsSubrMake (interp, "vector", OsSubrVector, OS_NARY) ;
    OsSubrMake (interp, "vector-fill!", OsSubrVectorFill, 2) ;
    OsSubrMake (interp, "vector->list", OsSubrVectorToList, 1) ;
    OsSubrMake (interp, "subvector", OsSubrSubvector, 3) ;
    OsSubrMake (interp, "vector-removeq!", OsSubrVectorRemoveq, 2) ;
    OsSubrMake (interp, "vector-append!", OsSubrVectorAppend, 2) ;
    return ;
}
