/***************************************************************************
 * os_motif.c
 * os_motif.c,v 1.2 1995/03/07 02:21:29 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <oscheme.h>

#include <X11/IntrinsicP.h>
#include <X11/ObjectP.h>	/* For XtIs<Classname> macros. */
#include <X11/CoreP.h>		/* for CoreRec */

#include <Xm/Xm.h>
#include <Xm/XmAll.h>

#include <os_motif.h>

static OsEObj factory_eobj     = (OsEObj) 0 ;
static OsEObj widget_eobj      = (OsEObj) 0 ;
static OsEObj widget_list_eobj = (OsEObj) 0 ;
static OsEObj widget_text_eobj = (OsEObj) 0 ;

/*
 * Widget global slot's accessors:
 */

static OsEObjSel rsel_widget = (OsEObjSel) 0 ;

/*
 * Factory global slot's accessor:
 */

static OsEObjSel rsel_display         = (OsEObjSel) 0 ; 
static OsEObjSel rsel_appname         = (OsEObjSel) 0 ; 
static OsEObjSel rsel_classname       = (OsEObjSel) 0 ;
static OsEObjSel rsel_dpyname         = (OsEObjSel) 0 ;
static OsEObjSel rsel_toplevel_widget = (OsEObjSel) 0 ;
static OsEObjSel rsel_hashcb	      = (OsEObjSel) 0 ;
static OsEObjSel rsel_factory         = (OsEObjSel) 0 ;

static OsHashtable os_widgets         = (OsHashtable) 0 ;

/***************************************************************************
 * resource handling
 ***************************************************************************/

Widget OsCvtOsWidgetToWidget (OsEObj os_widget) {
    OsPointer wptr ;

    wptr = (OsPointer) OsEObjSendMsg0 (os_widget, os_widget, rsel_widget) ;
    return ((Widget) OsPointerDeref (wptr)) ;
}

XmStringTable OsCvtOsVectorToStringTable (OsVector os_vector) {
    XmStringTable tbl ;
    unsigned int  i ;
    
    tbl = (XmStringTable) XtMalloc (os_vector->csize * sizeof (XmString)) ;
    for (i = 0 ; i < os_vector->csize ; i++ ) {
        OsString os_string = (OsString) os_vector->cells[i] ;
        OsShouldBeString (os_string) ;
        tbl[i] = XmStringCreateLocalized (os_string->string) ;
    }
    return (tbl) ;
}

OsString OsXmStringToOsString (XmString xmstr) {
    XmStringContext   context;
    XmStringCharSet   charset;
    XmStringDirection direction;
    Boolean           separator;
    char              *text;

    OsString os_string = OsStringMake ("") ;
    if ( ! XmStringInitContext(&context, xmstr) )  
	return (os_string) ;
    while ( XmStringGetNextSegment(context
                                   , &text
                                   , &charset
                                   , &direction, &separator)) {
	XtFree(charset);
        os_string = OsStringCAppend (os_string, text) ;
	if (separator == True) 
	    os_string = OsStringCAppend (os_string, "\n") ;
	XtFree(text);
    }
    XmStringFreeContext(context);
    return (os_string) ;
}

/***************************************************************************
 * widget handling
 ***************************************************************************/

static OsEObj create_widget (OsEObj factory, Widget widget, OsEObj wrapper) ;
static OsEObj make_widget (Widget widget) ;

static Widget widget_of_os_widget (OsEObj self) {
    OsPointer wptr   = (OsPointer) OsEObjSendMsg0 (self,self,rsel_widget);
    Widget    widget = (Widget) OsPointerDeref (wptr) ;
    return (widget) ;
}

static OsObject widget_forward (closure, self, sel, os_sel, args)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel, os_sel ;
    OsVector  args ;
{
    OsPointer wptr ;
    Widget    widget ;
    
    OsShouldBeEObjSel (os_sel) ;
    OsShouldBeVector (args) ;
    widget = widget_of_os_widget (self) ;
    if ( ! os_sel->arity ) {		/* getValue */
        return (OsWidgetGetValue (widget, os_sel)) ;
    } else { 				/* (os_sel->arity > 0) => setValues */
        OsWidgetSetValues (widget, os_sel, args) ;
        return (OsA_True) ;
    }
    /* NOT REACHED */
}

static OsObject widget_realize (closure, self, sel)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
{
    Widget widget = widget_of_os_widget (self) ;
    XtRealizeWidget (widget) ;
    return (OsA_True) ;
}

static void oscheme_xt_callback (widget, code, unused)
    Widget      widget ;
    OsCodeBlock code ;
    XtPointer   unused ;
{
    OsObject os_widget = (OsObject) make_widget (widget) ;
    OsObject error     = (OsObject) 0 ;
    
    (void) OsCodeBlockApplyWithArray (OsCurrentInterp()
                                      , code
                                      , 1 , (OsObject *) &os_widget
                                      , &error) ;
    if ( error ) {
        OsSendDisplay (OsErrorFormat ((OsVector) error), OsCurrentOutputPort) ;
        OsOutputPortWriteChar (OsCurrentOutputPort, '\n') ;
        OsOutputPortFlush (OsCurrentOutputPort) ;
    }
    return ;
}

static OsObject widget_define_callback_to (closure, self, sel, name, to)
    void        *closure ;
    OsEObj      self ;
    OsEObjSel   sel ;
    OsString    name ;
    OsCodeBlock to ;
{
    OsString  callbackName ;
    OsPointer wptr    = (OsPointer) OsEObjSendMsg0(self, self, rsel_widget) ;
    OsHashtable hashcb = (OsHashtable) OsEObjSendMsg0(self, self, rsel_hashcb);
    
    OsShouldBeString (name) ;
    OsMustBeBlock (to) ;
    callbackName = OsStringCAppend (name, "Callback") ;
/*
 * We need this hashtable to make our callbacks visible to GC, because
 * widgets are allocated through Xt malloc, not through Boehm's malloc.
 * Before being inited, we inherit this slot's (and its value) from our
 * parent (which has set it to nil). If needed we allocate it right now.
 */
    if ( hashcb != (OsHashtable) OsA_Nil ) {
        OsHashtableRemove (hashcb, callbackName->string) ;
    } else {
        hashcb = OsHashtableMake (0
                                  , (OsHashFct) OsHashCString
                                  , (OsCmpFct) OsCmpCString) ;
        OsEObjAllocateSlot (self
                            , rsel_hashcb
                            , (OsEObjSel) 0
                            , (OsObject) hashcb) ;
    }
    XtAddCallback ((Widget) OsPointerDeref (wptr)
                   , callbackName->string
                   , oscheme_xt_callback
                   , (XtPointer) to) ;
    OsHashtablePut (hashcb, callbackName->string, (void *) to) ;
    return (OsA_True) ;
}

static OsObject widget_class (closure, self, sel)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
{
    Widget      widget = widget_of_os_widget (self) ;
    WidgetClass wclass = XtClass (widget) ;
    
    return ((OsObject) OsStringMake (wclass->core_class.class_name)) ;
}

static OsObject widget_managed (void *closure, OsEObj self, OsEObjSel sel) {
    Widget widget = widget_of_os_widget (self) ;
    return (XtIsManaged (widget) ? OsA_True : OsA_False) ;
}

static OsObject widget_managed_to (void *closure
                                   , OsEObj self, OsEObjSel sel
                                   , OsObject value) {
    Widget widget = widget_of_os_widget (self) ;
    if ( value == OsA_False )
        XtUnmanageChild (widget) ;
    else
        XtManageChild (widget) ;
    return (value) ;
}

static OsObject widget_parent (void *closure, OsEObj self, OsEObjSel sel) {
    OsPointer wptr   = (OsPointer) OsEObjSendMsg0 (self, self, rsel_widget) ;
    OsObject factory = OsEObjSendMsg0 (self, self, rsel_factory) ;
    Widget   parent  = XtParent ((Widget) OsPointerDeref (wptr)) ;
    
    /* Note that the wrapper for widget can really be widget_eobj, because
     * if this is an already created widget, it already has been created
     * hence, the wrapper parameter has no importance.
     * If the widget was not already created, then widget_eobj is the right
     * wrapper for it :-)
     */
    return ((parent) 
            ? (OsObject) create_widget ((OsEObj) factory, parent, widget_eobj)
            : OsA_False) ;
}

static OsObject widget_call_action_with (void *closure
                                         , OsEObj self, OsEObjSel sel
                                         , OsAtom action
                                         , OsVector with) {
    char         *argv[OS_MOTIF_NUM_PARAMS] ;
    unsigned int argc ;
    XEvent       fake_event ;
    OsShouldBeAtom (action) ;
    OsShouldBeVector (with) ;
    if ( with->csize >= OS_MOTIF_NUM_PARAMS ) {
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("NUM_PARAMS exceeded")
                               , OsStringMake ("widget_call_action_with")) ;
    }
    for ( argc = 0 ; argc < with->csize ; argc++ ) {
        OsString arg = (OsString) with->cells[argc] ;
        OsShouldBeString (arg) ;
        argv[argc] = arg->string ;
    }
    bzero ((char *) &fake_event, sizeof (fake_event)) ;	/* no risks */
    XtCallActionProc (widget_of_os_widget (self)
                      , action->pname->string
                      , &fake_event
                      , argv, argc) ;
    return (OsA_True) ;
}
    
/* Special widget_list stuff */

static OsObject widget_list_addItem_at (void *closure
                                        , OsEObj self, OsEObjSel sel
                                        , OsString item
                                        , OsInteger at) {
    Widget    widget = widget_of_os_widget (self) ;
    XmString  xmstr ;
    OsShouldBeString (item) ;
    OsShouldBeInteger (at) ;
    xmstr = XmStringCreateLocalized (item->string) ;
    XmListAddItemUnselected (widget, xmstr, at->ival) ;
    XtFree ((char *) xmstr) ;
    return (OsA_True) ;
}

static OsObject widget_list_selectPosition (void *closure
                                            , OsEObj self, OsEObjSel sel
                                            , OsInteger position
                                            , OsObject callCallback) {
    Widget widget  = widget_of_os_widget (self) ;
    OsShouldBeInteger (position) ;
    XmListSelectPos (widget
                     , position->ival
                     , (callCallback == OsA_False) ? 0 : 1) ;
    return (OsA_True) ;
}

static OsObject widget_list_selectItem (void *closure
                                        , OsEObj self, OsEObjSel sel
                                        , OsString item
                                        , OsObject callCallback) {
    Widget    widget = widget_of_os_widget (self) ;
    XmString  xmstr ;
    OsShouldBeString (item) ;
    xmstr = XmStringCreateLocalized (item->string) ;
    XmListSelectItem (widget
                      , xmstr
                      , (callCallback == OsA_False) ? 0 : 1) ;
    XtFree ((char *) xmstr) ;
    return (OsA_True) ;
}

static OsObject widget_list_setItem (void *closure
                                     , OsEObj self, OsEObjSel sel
                                     , OsString item) {
    Widget    widget = widget_of_os_widget (self) ;
    XmString  xmstr ;
    OsShouldBeString (item) ;
    xmstr = XmStringCreateLocalized (item->string) ;
    XmListSetItem (widget, xmstr) ;
    XtFree ((char *) xmstr) ;
    return (OsA_True) ;
}

/* Text widget specific stuff here */

static OsObject widget_text_insertionPosition (void *closure
                                               , OsEObj self, OsEObjSel sel) {
    Widget         widget = widget_of_os_widget (self) ;
    XmTextPosition pos ;
    pos = XmTextGetInsertionPosition (widget) ;
    return ((OsObject) OsIntegerMake ((int) pos)) ;
}

static OsObject widget_text_insertionPosition_at (void *closure
                                                  , OsEObj self, OsEObjSel sel
                                                  , OsInteger pos) {
    Widget widget = widget_of_os_widget (self) ;
    OsShouldBeInteger (pos) ;
    XmTextSetInsertionPosition (widget, (XmTextPosition) pos->ival) ;
    return ((OsObject) pos) ;
}

static OsObject widget_text_lastPosition (void *closure
                                          , OsEObj self, OsEObjSel sel) {
    Widget widget = widget_of_os_widget (self) ;
    XmTextPosition pos ;
    pos = XmTextGetLastPosition (widget) ;
    return ((OsObject) OsIntegerMake ((int) pos)) ;
}

static OsObject widget_text_insert_at (void *closure
                                       , OsEObj self, OsEObjSel sel
                                       , OsString text, OsInteger pos) {
    Widget   widget = widget_of_os_widget (self) ;
    OsShouldBeString (text) ;
    OsShouldBeInteger (pos) ;
    XmTextInsert (widget, (XmTextPosition) pos->ival, text->string) ;
    return ((OsObject) text) ;
}

static OsObject widget_text_replaceFrom_to_with (void *closure
                                                 , OsEObj self, OsEObjSel sel
                                                 , OsInteger from, OsInteger to
                                                 , OsString with) {
    Widget widget = widget_of_os_widget (self) ;
    OsShouldBeInteger (from) ;
    OsShouldBeInteger (to) ;
    OsShouldBeString (with) ;
    XmTextReplace (widget
                   , (XmTextPosition) from, (XmTextPosition) to
                   , with->string) ;
    return ((OsObject) with) ;
}

static OsObject widget_text_enableDisplay (void *closure
                                          , OsEObj self, OsEObjSel sel
                                          , OsObject bool) {
    Widget widget = widget_of_os_widget (self) ;
    if ( bool == OsA_False )
        XmTextDisableRedisplay (widget) ;
    else
        XmTextEnableRedisplay (widget) ;
    return (bool) ;
}

static void initialize_widget_eobj (void) {
    /* Standard widget scheme methods */
    widget_eobj = (OsEObj) OsEObjSendMsg0 ((OsEObj) OsEObjRoot
                                           , (OsEObj) OsEObjRoot
                                           , OsEObjSelClone) ;
    rsel_widget  = OsEObjSelDefine ("_widget") ;
    rsel_hashcb  = OsEObjSelDefine ("_hashcb") ;
    rsel_factory = OsEObjSelDefine ("factory") ;
    OsEObjDefineMsg (widget_eobj
                     , OsEObjSelDefine ("forward::")
                     , widget_forward
                     , (void *) 0) ;
    OsEObjDefineMsg (widget_eobj
                     , OsEObjSelDefine ("realize")
                     , widget_realize
                     , (void *) 0) ;
    OsEObjDefineMsg (widget_eobj
                     , OsEObjSelDefine ("defineCallback:to:")
                     , widget_define_callback_to
                     , (void *) 0) ;
    OsEObjDefineMsg (widget_eobj
                     , OsEObjSelDefine ("widgetClass")
                     , widget_class
                     , (void *) 0) ;
    OsEObjDefineMsg (widget_eobj
                     , OsEObjSelDefine ("managed")
                     , widget_managed
                     , (void *) 0) ;
    OsEObjDefineMsg (widget_eobj
                     , OsEObjSelDefine ("managed:")
                     , widget_managed_to
                     , (void *) 0) ;
    OsEObjDefineMsg (widget_eobj
                     , OsEObjSelDefine ("parent")
                     , widget_parent
                     , (void *) 0) ;
    OsEObjDefineMsg (widget_eobj
                     , OsEObjSelDefine ("callAction:with:")
                     , widget_call_action_with
                     , (void *) 0) ;
    OsEObjAllocateSlot (widget_eobj, rsel_widget,  (OsEObjSel) 0, OsA_Nil) ;
    OsEObjAllocateSlot (widget_eobj, rsel_factory, (OsEObjSel) 0, OsA_Nil) ;
    OsEObjAllocateSlot (widget_eobj, rsel_hashcb,  (OsEObjSel) 0, OsA_Nil) ;
    /* List widgets deserve special methods if they are to be used */
    widget_list_eobj = (OsEObj) OsEObjSendMsg0 (widget_eobj
                                                , widget_eobj
                                                , OsEObjSelClone) ;
    OsEObjDefineMsg (widget_list_eobj
                     , OsEObjSelDefine ("addItem:at:")
                     , widget_list_addItem_at
                     , (void *) 0) ;
    OsEObjDefineMsg (widget_list_eobj
                     , OsEObjSelDefine ("selectPosition:callCallback:")
                     , widget_list_selectPosition
                     , (void *) 0) ;
    OsEObjDefineMsg (widget_list_eobj
                     , OsEObjSelDefine ("selectItem:callCallback:")
                     , widget_list_selectItem
                     , (void *) 0) ;
    OsEObjDefineMsg (widget_list_eobj
                     , OsEObjSelDefine ("setItem:")
                     , widget_list_setItem
                     , (void *) 0) ;
    /* Text widgets deserver sepcial methods also */
    widget_text_eobj = (OsEObj) OsEObjSendMsg0 (widget_eobj
                                                , widget_eobj
                                                , OsEObjSelClone) ;
    OsEObjDefineMsg (widget_text_eobj
                     , OsEObjSelDefine ("insertionPosition")
                     , widget_text_insertionPosition
                     , (void *) 0) ;
    OsEObjDefineMsg (widget_text_eobj
                     , OsEObjSelDefine ("insertionPosition:")
                     , widget_text_insertionPosition_at
                     , (void *) 0) ;
    OsEObjDefineMsg (widget_text_eobj
                     , OsEObjSelDefine ("lastPosition")
                     , widget_text_lastPosition
                     , (void *) 0) ;
    OsEObjDefineMsg (widget_text_eobj
                     , OsEObjSelDefine ("insert:at:")
                     , widget_text_insert_at
                     , (void *) 0) ;
    OsEObjDefineMsg (widget_text_eobj
                     , OsEObjSelDefine ("replaceFrom:to:with:")
                     , widget_text_replaceFrom_to_with
                     , (void *) 0) ;
    OsEObjDefineMsg (widget_text_eobj
                     , OsEObjSelDefine ("enableDisplay:")
                     , widget_text_enableDisplay
                     , (void *) 0) ;
    return ;
}

static void destroy_widget (widget, wptr, unused)
    Widget    widget ;
    OsPointer wptr ;
    XtPointer unused ;
{
    OsHashtableRemove (os_widgets, (void *) wptr) ;
    OsPointerInvalidate (wptr) ;
    return ;
}

static OsEObj make_widget (widget)
    Widget widget ;
{
    OsPointer wptr   = OsSharedPointerMake (widget) ;
    OsEObj    os_widget ;
    
    if ( os_widget = (OsEObj) OsHashtableGet (os_widgets, (void *) wptr) ) {
        return (os_widget) ;
    }
    (void) OsErrorTrigger (1, OsE_Implementation
                           , OsStringMake ("make_widget")
                           , OsStringMake ("make_widget: unexsting widget.")) ;
    /* NOT REACHED */
}

static OsEObj create_widget (OsEObj factory, Widget widget, OsEObj wrapper) {
    OsPointer wptr   = OsSharedPointerMake (widget) ;
    OsEObj    os_widget ;

    if ( os_widget = (OsEObj) OsHashtableGet (os_widgets, (void *) wptr) ) {
        return (os_widget) ;
    }
    os_widget = (OsEObj) OsEObjSendMsg0 (wrapper
                                         , wrapper
                                         , OsEObjSelClone) ;
    OsEObjAllocateSlot (os_widget
                        , rsel_widget, (OsEObjSel) 0
                        , (OsObject)wptr) ;
    OsEObjAllocateSlot (os_widget
                        , rsel_factory, (OsEObjSel) 0
                        , (OsObject) factory) ;
                        
    XtAddCallback (widget, XtNdestroyCallback, destroy_widget, (void *) wptr) ;
    OsHashtablePut (os_widgets, (void *) wptr, (void *) os_widget) ;
    return (os_widget) ;
}

/***************************************************************************
 * motif:factory
 * This prototype is the widget factory: iot just creates widgets.
 ***************************************************************************/

static OsObject factory_create_toplevel (closure, self, sel, os_name, os_class)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
    OsString  os_name ;
    OsString  os_class ;
{
    Widget    toplevel ;
    OsPointer dptr = (OsPointer) OsEObjSendMsg0 (self, self, rsel_display) ;
    OsShouldBeString (os_name) ;
    OsShouldBeString (os_class) ;
    
    toplevel = XtAppCreateShell (os_name->string
                                 , os_class->string
                                 , applicationShellWidgetClass
                                 , (Display *) OsPointerDeref (dptr) 
                                 , 0, 0) ;
    return ((OsObject) create_widget (self, toplevel, widget_eobj)) ;
}

static Widget os_CreateOptionMenu (Widget widget, char *name
                                   , ArgList argc, Cardinal argcount) {
    /* Warning: arglist and argcount not used */
    Widget omenu, pmenu ;
    Arg args[8] ;
    pmenu = XmCreatePulldownMenu (widget, "_menu", 0, 0) ;
    XtSetArg (args[0], XmNsubMenuId, pmenu) ;
    omenu = XmCreateOptionMenu (widget, name, args, 1) ;
    XtManageChild (omenu) ;
    return (pmenu) ;
}

static OsObject factory_create_widget (fcreate, self, sel, os_name, os_parent)
    Widget    (*fcreate)() ;
    OsEObj    self ;
    OsEObjSel sel ;
    OsString  os_name ;
    OsEObj    os_parent ;
{
    OsPointer wptr ;
    Widget    widget ;
    
    OsShouldBeString (os_name) ;
    wptr = (OsPointer) OsEObjSendMsg0 (os_parent, os_parent, rsel_widget) ;
    widget = (*fcreate)((Widget) OsPointerDeref (wptr)
                        , os_name->string
                        , 0, 0) ;
    if ( fcreate != os_CreateOptionMenu )
        XtManageChild (widget) ;
    return ((OsObject) create_widget (self, widget, widget_eobj)) ;
}

/* ListWidgets needs a few additional scheme methods to be usefull, so their
 * creation function doesn't use 'create_widget, but a enriched clone of it
 */

static OsObject factory_create_ListWidget (Widget (*fcreate)()
                                           , OsEObj self
                                           , OsEObjSel sel
                                           , OsString os_name
                                           , OsEObj os_parent) {
    Widget parent, widget ;
    OsShouldBeString (os_name) ;
    parent = widget_of_os_widget (os_parent) ;
    widget = (*fcreate)(parent, os_name->string, 0, 0) ;
    XtManageChild (widget) ;
    return ((OsObject) create_widget (self, widget, widget_list_eobj)) ;
}

static OsObject factory_create_TextWidget (Widget (*fcreate)()
                                           , OsEObj self
                                           , OsEObjSel sel
                                           , OsString os_name
                                           , OsEObj os_parent) {
    Widget parent, widget ;
    OsShouldBeString (os_name) ;
    parent = widget_of_os_widget (os_parent) ;
    widget = (*fcreate)(parent, os_name->string, 0, 0) ;
    XtManageChild (widget) ;
    return ((OsObject) create_widget (self, widget, widget_text_eobj)) ;
}

static void initialize_factory_eobj () {
    factory_eobj = (OsEObj) OsEObjSendMsg0 ((OsEObj) OsEObjRoot
                                            , (OsEObj) OsEObjRoot
                                            , OsEObjSelClone) ;
    rsel_display         = OsEObjSelDefine ("_display") ;
    rsel_appname         = OsEObjSelDefine ("appName") ;
    rsel_classname       = OsEObjSelDefine ("appClass") ;
    rsel_dpyname         = OsEObjSelDefine ("appDisplay") ;
    rsel_toplevel_widget = OsEObjSelDefine ("appToplevel") ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("arrowButton:parent:")
                     , factory_create_widget
                     , (void *) XmCreateArrowButton) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("bulletinBoard:parent:")
                     , factory_create_widget
                     , (void *) XmCreateBulletinBoard) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("bulletinBoardDialog:parent:")
                     , factory_create_widget
                     , (void *) XmCreateBulletinBoardDialog) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("cascadeButton:parent:")
                     , factory_create_widget
                     , (void *) XmCreateCascadeButton) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("command:parent:")
                     , factory_create_widget
                     , (void *) XmCreateCommand) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("commandDialog:parent:")
                     , factory_create_widget
                     , (void *) XmCreateCommandDialog) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("dialogShell:parent:")
                     , factory_create_widget
                     , (void *) XmCreateDialogShell) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("dragIcon:parent:")
                     , factory_create_widget
                     , (void *) XmCreateDragIcon) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("drawingArea:parent:")
                     , factory_create_widget
                     , (void *) XmCreateDrawingArea) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("drawnButton:parent:")
                     , factory_create_widget
                     , (void *) XmCreateDrawnButton) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("fileSelectionBox:parent:")
                     , factory_create_widget
                     , (void *) XmCreateFileSelectionBox) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("fileSelectionDialog:parent:")
                     , factory_create_widget
                     , (void *) XmCreateFileSelectionDialog) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("form:parent:")
                     , factory_create_widget
                     , (void *) XmCreateForm) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("formDialog:parent:")
                     , factory_create_widget
                     , (void *) XmCreateFormDialog) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("frame:parent:")
                     , factory_create_widget
                     , (void *) XmCreateFrame) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("label:parent:")
                     , factory_create_widget
                     , (void *) XmCreateLabel) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("list:parent:")
                     , factory_create_ListWidget
                     , (void *) XmCreateList) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("scrolledList:parent:")
                     , factory_create_ListWidget
                     , (void *) XmCreateScrolledList) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("mainWindow:parent:")
                     , factory_create_widget
                     , (void *) XmCreateMainWindow) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("menuShell:parent:")
                     , factory_create_widget
                     , (void *) XmCreateMenuShell) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("messageBox:parent:")
                     , factory_create_widget
                     , (void *) XmCreateMessageBox) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("messageDialog:parent:")
                     , factory_create_widget
                     , (void *) XmCreateMessageDialog) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("errorDialog:parent:")
                     , factory_create_widget
                     , (void *) XmCreateErrorDialog) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("informationDialog:parent:")
                     , factory_create_widget
                     , (void *) XmCreateInformationDialog) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("questionDialog:parent:")
                     , factory_create_widget
                     , (void *) XmCreateQuestionDialog) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("warningDialog:parent:")
                     , factory_create_widget
                     , (void *) XmCreateWarningDialog) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("workingDialog:parent:")
                     , factory_create_widget
                     , (void *) XmCreateWorkingDialog) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("templateDialog:parent:")
                     , factory_create_widget
                     , (void *) XmCreateTemplateDialog) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("panedWindow:parent:")
                     , factory_create_widget
                     , (void *) XmCreatePanedWindow) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("pushButton:parent:")
                     , factory_create_widget
                     , (void *) XmCreatePushButton) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("rowColumn:parent:")
                     , factory_create_widget
                     , (void *) XmCreateRowColumn) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("workArea:parent:")
                     , factory_create_widget
                     , (void *) XmCreateWorkArea) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("radioBox:parent:")
                     , factory_create_widget
                     , (void *) XmCreateRadioBox) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("optionMenu:parent:")
                     , factory_create_widget
                     , (void *) os_CreateOptionMenu) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("menuBar:parent:")
                     , factory_create_widget
                     , (void *) XmCreateMenuBar) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("popupMenu:parent:")
                     , factory_create_widget
                     , (void *) XmCreatePopupMenu) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("pulldownMenu:parent:")
                     , factory_create_widget
                     , (void *) XmCreatePulldownMenu) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("scale:parent:")
                     , factory_create_widget
                     , (void *) XmCreateScale) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("scrollbar:parent:")
                     , factory_create_widget
                     , (void *) XmCreateScrollBar) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("scrolledWindow:parent:")
                     , factory_create_widget
                     , (void *) XmCreateScrolledWindow) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("selectionBox:parent:")
                     , factory_create_widget
                     , (void *) XmCreateSelectionBox) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("selectionDialog:parent:")
                     , factory_create_widget
                     , (void *) XmCreateSelectionDialog) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("promptDialog:parent:")
                     , factory_create_widget
                     , (void *) XmCreatePromptDialog) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("separator:parent:")
                     , factory_create_widget
                     , (void *) XmCreateSeparator) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("scrolledText:parent:")
                     , factory_create_TextWidget
                     , (void *) XmCreateScrolledText) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("text:parent:")
                     , factory_create_TextWidget
                     , (void *) XmCreateText) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("textField:parent:")
                     , factory_create_TextWidget
                     , (void *) XmCreateTextField) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("toggleButton:parent:")
                     , factory_create_widget
                     , (void *) XmCreateToggleButton) ;
    OsEObjDefineMsg (factory_eobj
                     , OsEObjSelDefine ("toplevel:appClass:")
                     , factory_create_toplevel
                     , (void *) 0) ;
    OsEObjAllocateSlot (factory_eobj
                        , rsel_display 
                        , (OsEObjSel) 0
                        , OsA_Nil) ;
    OsEObjAllocateSlot (factory_eobj
                        , rsel_appname
                        , (OsEObjSel) 0
                        , OsA_Nil) ;
    OsEObjAllocateSlot (factory_eobj
                        , rsel_classname
                        , (OsEObjSel) 0
                        , OsA_Nil) ;
    OsEObjAllocateSlot (factory_eobj
                        , rsel_dpyname
                        , (OsEObjSel) 0
                        , OsA_Nil) ;
    OsEObjAllocateSlot (factory_eobj
                        , rsel_toplevel_widget
                        , (OsEObjSel) 0
                        , OsA_Nil) ;
    return ;
}

OsEObj MotifFactoryMake (xtapp, dpyname, app_name, app_class)
    XtAppContext xtapp ;
    OsString     dpyname, app_name, app_class ;
{
    Widget  toplevel ;
    Display *dpy ;
    int     argc = 0 ;
    OsEObj  factory ;
    
    dpy = XtOpenDisplay (xtapp
                         , dpyname->string
                         , app_name->string
                         , app_class->string
                         , 0, 0
                         , &argc, (char **) 0) ;
    toplevel = XtAppCreateShell (app_name->string
                                 , app_class->string
                                 , applicationShellWidgetClass
                                 , dpy
                                 , 0, 0) ;
/*
 * Create the factory for this display/toplevel
 */
    
    factory = (OsEObj) OsEObjSendMsg0 (factory_eobj, factory_eobj
                                       , OsEObjSelClone) ;
    OsEObjAllocateSlot (factory
                        , rsel_display
                        , (OsEObjSel) 0
                        , (OsObject) OsSharedPointerMake (dpy)) ;
    OsEObjAllocateSlot (factory
                        , rsel_appname
                        , (OsEObjSel) 0
                        , (OsObject) app_name) ;
    OsEObjAllocateSlot (factory
                        , rsel_classname
                        , (OsEObjSel) 0
                        , (OsObject) app_class) ;
    OsEObjAllocateSlot (factory
                        , rsel_dpyname
                        , (OsEObjSel) 0
                        , (OsObject) dpyname) ;
    OsEObjAllocateSlot (factory
                        , rsel_toplevel_widget
                        , (OsEObjSel) 0
                        , (OsObject) create_widget (factory
                                                    , toplevel
                                                    , widget_eobj)) ;
    return (factory) ;
}

void Os_osmotif_initialize (OsInterp interp) {
    (void) initialize_factory_eobj () ;
    (void) initialize_widget_eobj () ;
    Os_setget_initialize (interp) ;
    os_widgets = OsHashtableMake (0
                                  , (OsHashFct) OsHashPointer
                                  , (OsCmpFct) OsCmpPointer) ;
    return ;
}
