/***************************************************************************
 * setvalue.c
 * setvalue.c,v 1.2 1995/03/07 02:21:35 baird Exp
 * See the file COPYRIGHT for the licensing conditions.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <oscheme.h>

#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <Xm/Xm.h>
#include <Xm/XmAll.h>

#include <os_motif.h>

/* This is private in the r5 distrib */
#if (XtSpecificationRelease < 6)
typedef struct {
    String      name;   /* resource name */
    String      type;   /* representation type name */
    XtArgVal    value;  /* representation */
    int         size;   /* size of representation */
} XtTypedArg, *XtTypedArgList;
#endif /*  (XtSpecificationRelease < 6) */

/* Resource convertion consumes resources that are to be freed *after* the
 * conversion. This little package maintains a list of free function and
 * objets that they are to be applied to: */

static int os_xt_error = 0 ;

static OsError     ose_xt_badtype     = (OsError) 0 ;
static OsError     ose_xt_badresource = (OsError) 0 ;

#define OS_XT_ERROR_BADRESOURCE	1
#define OS_XT_ERROR_BADTYPE	2

static void xt_error_handler (name, type, class, defaultp, params, num_params)
    String   name, type, class, defaultp, *params ;
    Cardinal *num_params ;
{
    if ( ! strcmp (name, "unknownType") ) {
        /* Xt{Set,Get}Values failed, because of invalid resource name */
        os_xt_error = OS_XT_ERROR_BADRESOURCE ;
    } else if ( (! strcmp (name, "conversionError"))
               || ( ! strcmp (name, "conversionFailed"))
               || ( ! strcmp (name, "typeConversionError")) ) {
        /* Cannot convert to requested type */
        os_xt_error = OS_XT_ERROR_BADTYPE ;
    } else {
        fprintf (stderr, "Unkown error name: %s\n", name) ;
    }
    return ;
}

static char *xt_selpart_to_resource (OsString selname) {
    char buffer[256] ;
    strcpy (buffer, selname->string) ;
    buffer[selname->len-1] = '\0' ;		/* remove trailing ':' */
    return (OsGCStrdup (buffer)) ;
}

void OsWidgetSetValues (Widget widget, OsEObjSel sel, OsVector values) {
    XtTypedArgList arg, nested_list ;
    unsigned int i, argc = sel->arity ;
    char *resource ;
    OsObject os_value ;
    
    nested_list = (XtTypedArgList) XtMalloc (sizeof (XtTypedArg) * (1+argc)) ;
    arg = nested_list ;
    for (i = 0 ; i < argc ; i++, arg++ ) {
        arg->name = xt_selpart_to_resource (sel->selv[i]->pname) ;
        os_value = values->cells[i] ;
        if ( OsIsInteger (os_value) ) {		/* integer are allowed */
            arg->type  = XtRInt ;
            arg->value = (XtArgVal) ((OsInteger) os_value)->ival ;
            arg->size  = sizeof (arg->value) ;  /* NOT sizeof(int) ! alpha */
        } else if ( OsIsVector (os_value) ) {	/* vectors, as list of ... */
            arg->type  = XmRXmStringTable ;
            arg->value = (XtArgVal) OsCvtOsVectorToStringTable((OsVector)os_value) ;
            arg->size  = sizeof (XmRStringTable) ;
        } else if ( OsIsEObj (os_value) ) {	/* should be a widget */
            arg->value = (XtArgVal) OsCvtOsWidgetToWidget ((OsEObj) os_value) ;
            arg->size  = sizeof (Widget) ;
            arg->type  = ((( ! strcmp (arg->name, XmNmenuHelpWidget))
                           || ( ! strcmp (arg->name, XmNsubMenuId)))
                          ? XmRMenuWidget
                          : XmRWidget) ;
        } else if ( OsIsString (os_value) ) {	/* fallback to strings */
            arg->type  = XtRString ;
            arg->value = (XtArgVal) ((OsString) os_value)->string ;
            arg->size  = ((OsString) os_value)->len + 1 ;
        } else {
            (void) OsErrorTrigger (1, OsE_BadType
                                   , os_value
                                   , OsList2Make ((OsObject) os_value->os_class
                                                  , (OsObject) OsStringClass));
        }
    }
    arg->name   = NULL ;
    arg->type   = NULL ;
    arg->value  = (XtArgVal) NULL ;
    arg->size   = 0 ;
    os_xt_error = 0 ;
    XtVaSetValues (widget, XtVaNestedList, nested_list, 0) ;
    XtFree ((char *) nested_list) ;
    switch (os_xt_error) {
      case 0:
        return ;
      case OS_XT_ERROR_BADRESOURCE:
        (void) OsErrorTrigger (1, ose_xt_badresource
                               , sel->selv[0]->pname) ;
        /* NOT REACHED */
      case OS_XT_ERROR_BADTYPE:
        (void) OsErrorTrigger (1, ose_xt_badtype
                               , sel->selv[0]->pname) ;
        /* BIT REACHED */
      default:
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("widget_get_value")
                               , OsStringMake ("bad xt_error")) ;
    }
    /* NOT REACHED */
}

/* Get value stuff is more complicated then set... */

static OsHashtable osconv_hash = (OsHashtable) 0 ;

static OsObject (*osconv_get (char *xt_type))() {
    return ((OsObject (*)()) OsHashtableGet (osconv_hash, xt_type)) ;
}

static void osconv_set (char *xt_type, OsObject (*conv)()) {
    OsHashtablePut (osconv_hash, xt_type, conv) ;
    return ;
}

static OsObject osconv_stringtable (Widget widget, XtResourceList resource) {
    XmStringTable xt_value ;
    OsVector      vector = OsVectorMake (0) ;
    unsigned int  i, item_count ;
    XtVaGetValues (widget, resource->resource_name, &xt_value, 0) ;

    if (resource== NULL) {
	return ((OsObject) vector) ;
    } else if ( ! strcmp (resource->resource_name, XmNitems)) {
        XtVaGetValues (widget, XmNitemCount, &item_count, 0) ;
    } else if ( ! strcmp (resource->resource_name, XmNselectedItems) ) {
        XtVaGetValues (widget, XmNselectedItemCount, &item_count, 0) ;
    } else {
        return ((OsObject) vector) ;
    }

    for (i = 0 ; i < item_count ; i++ ) {
        OsVectorFastAppend (vector
                            , (OsObject) OsXmStringToOsString (xt_value[i])) ;
    }
    return ((OsObject) vector) ;
}

static OsObject osconv_dimension (Widget widget, XtResourceList resource) {
    Dimension dim ;
    XtVaGetValues (widget, resource->resource_name, &dim, 0) ;
    return ((OsObject) OsIntegerMake (dim)) ;
}

static OsObject osconv_short (Widget widget, XtResourceList resource) {
    short sh ;
    XtVaGetValues (widget, resource->resource_name, &sh, 0) ;
    return ((OsObject) OsIntegerMake (sh)) ;
}

static OsObject osconv_position (Widget widget, XtResourceList resource) {
    Position pos ;
    XtVaGetValues (widget, resource->resource_name, &pos, 0) ;
    return ((OsObject) OsIntegerMake (pos)) ;
}

static OsObject osconv_int (Widget widget, XtResourceList resource) {
    int i ;
    XtVaGetValues (widget, resource->resource_name, &i, 0) ;
    return ((OsObject) OsIntegerMake (i)) ;
}

static OsObject osconv_cardinal (Widget widget, XtResourceList resource) {
    Cardinal card ;
    XtVaGetValues (widget, resource->resource_name, &card, 0) ;
    return ((OsObject) OsIntegerMake (card)) ;
}

OsObject OsWidgetGetValue (Widget widget, OsEObjSel sel) {
    /* We cache the current resource list, in order to save seom alloc/free */
    static WidgetClass    widget_class  = (WidgetClass) 0 ;
    static XtResourceList resources     = (XtResourceList) 0 ;
    static Cardinal       num_resources = (Cardinal) 0 ;
    XtResourceList pres ;
    char           *resource = sel->selv[0]->pname->string ;
    char           *value ;
    OsObject       os_value ;
    unsigned int   i ;
    
    if ( XtClass (widget) != widget_class ) {	/* update cache */
        if ( num_resources )
            XtFree ((char *) resources) ;
        widget_class = XtClass (widget) ;
        XtGetResourceList (widget_class, &resources, &num_resources) ;
    }
    for (pres = resources ; pres < resources + num_resources ; pres++ ) {
        if ( ! strcmp (pres->resource_name, resource) ) {
            OsObject (*os_conv)() ;
            if ( os_conv = osconv_get (pres->resource_type) ) {
                return (os_conv (widget, pres)) ;
            } else { /* Fallback to string */
                os_xt_error = 0 ;
                XtVaGetValues (widget
                               , XtVaTypedArg
                               , resource, XtRString
                               , &value, sizeof (char *), 0) ;
                switch (os_xt_error) {
                  case 0:
                    os_value = (OsObject) OsStringMake (value) ;
                    XtFree (value) ;
                    return ((OsObject) os_value) ;
                  case OS_XT_ERROR_BADRESOURCE:
                    (void) OsErrorTrigger (1, ose_xt_badresource
                                           , sel->selv[0]->pname) ;
                    /* NOT REACHED */
                  case OS_XT_ERROR_BADTYPE:
                    (void) OsErrorTrigger (1, ose_xt_badtype
                                           , sel->selv[0]->pname) ;
                    /* BIT REACHED */
                  default:
                    (void) OsErrorTrigger (1, OsE_Implementation
                                           , OsStringMake ("widget_get_value")
                                           , OsStringMake ("bad xt_error")) ;
                }
            }
        }
    }
    /* Resource description not found */
    (void) OsErrorTrigger (1, ose_xt_badresource, sel->selv[0]->pname) ;
    /* NOT REACHED */
}

void Os_setget_initialize (OsInterp interp) {
    XtSetWarningMsgHandler (xt_error_handler) ;
    /* Because osconv_hash maps pre-allocated string to function pointer, it
     * can safely be weak */
    osconv_hash = OsWeakHashtableMake (0
                                       , (OsHashFct) OsHashCString
                                       , (OsCmpFct) OsCmpCString) ;
    ose_xt_badtype     = OsErrorMake (interp
                                      , "error:xt-bad-type"
                                      , "Bad resource type for %0"
                                      , 1, OSE_CATCHABLE) ;
    ose_xt_badresource = OsErrorMake (interp
                                      , "error:xt-bad-resource"
                                      , "Invalid resource %0"
                                      , 1, OSE_CATCHABLE) ;
    osconv_set (XmRDimension, osconv_dimension) ;
    osconv_set (XmRHorizontalDimension, osconv_dimension) ;
    osconv_set (XmRVerticalDimension, osconv_dimension) ;
    osconv_set (XmRInt, osconv_int) ;
    osconv_set (XmRCardinal, osconv_cardinal) ;
    osconv_set (XmRPosition, osconv_position) ;
    osconv_set (XmRHorizontalPosition, osconv_position) ;
    osconv_set (XmRVerticalPosition, osconv_position) ;
    osconv_set (XmRTextPosition, osconv_position) ;
    osconv_set (XmRShort, osconv_short) ;
    osconv_set (XmRStringTable, osconv_stringtable) ;
    osconv_set (XmRXmStringTable, osconv_stringtable) ;
    return ;
}


