/***************************************************************************
 * os_cvt.c
 * $Id*
 * See the file COPYRIGHT for licensing conditions.
 * Various converters stollen from tclMotif
 ***************************************************************************/

#include <X11/IntrinsicP.h>
#include <X11/CoreP.h>		/* for CoreRec */
#include <string.h>
#include <ctype.h>

#include <Xm/Xm.h>

#define done(address, type)                               \
    {                                                   \
        if(toVal->addr)                                 \
        {                                               \
            if(toVal->size < sizeof(type))              \
            {                                           \
                toVal->size = sizeof(type);             \
                return( FALSE) ;                        \
            }                                           \
            else                                        \
            {   *((type *) (toVal->addr)) = *address;   \
            }                                           \
        }                                               \
        else                                            \
        {                                               \
            toVal->addr = (XPointer) address;           \
        }                                               \
        toVal->size = sizeof(type);                     \
    }

/* Because just after the XtVaGetValues, we XtFree the returned value,
 * converters should always duplicate the strings they return */

static char *xt_strdup (char *string) {
    int  len = strlen (string) + 1 ;
    char *p  = XtMalloc (len) ;
    bcopy (string, p, len) ;
    return (p) ;
}

Xt_CvtXmStringToString(dpy, args, nargs, fromVal, toVal, destructor_data)
    Display *dpy;
    XrmValuePtr args;
    Cardinal *nargs;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    XmStringContext context;
    static char *buf = NULL;
    char *p, *text;
    XmStringCharSet charset;
    XmStringDirection direction;
    Boolean separator;
    int size;
    int current_len, segment_len;
    XmString xmstr;

    /* whatever buf has grown to, or started from, 
       make it a reasonable size
     */
    if ( buf ) {
        XtFree(buf);
        size = 128;
        buf = XtMalloc(size);
        p = buf;
    }

    xmstr = *(XmString *) fromVal->addr;
    if (!XmStringInitContext(&context, xmstr)) {
	return False;
    }
    while (XmStringGetNextSegment(context, &text, &charset,
		&direction, &separator)) {
	XtFree(charset);
	segment_len = strlen(text);
	if (p + segment_len >= buf + size) {
	    /* string too large, double space */
	    current_len = p - buf;
	    size *= 2;
	    buf = XtRealloc(buf, size);
	    p = buf + current_len;
	}
	strcpy(p, text);
	p += segment_len;
	if (separator == True) {
	    *p++ = '\n';
	}
	*p = 0;
	XtFree(text);
    }
    XmStringFreeContext(context);

    /* destroy the XmString we started from */
    XmStringFree(xmstr);
 
    done(&buf, char *);

    return True;
}

static Boolean
Xt_CvtIntToString(dpy, args, nargs, fromVal, toVal, destructor_data)
    Display *dpy;
    XrmValuePtr args;
    Cardinal *nargs;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf = NULL;
    int intVal;
 
    intVal = *(int *) (fromVal->addr);

    buf = XtMalloc(16);
    sprintf(buf, "%d", intVal);
 
    done(&buf, char *);

    return True;
}

static Boolean
Xt_CvtUnsignedIntToString(dpy, args, num_args, fromVal, toVal, destructor_data)
    Display *dpy;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf = NULL;
    unsigned int intVal;
 
    intVal = *(int *) (fromVal->addr);

    buf = XtMalloc(16);
    sprintf(buf, "%u", intVal);
 
    done(&buf, char *);

    return True;
}

static Boolean
Xt_CvtBooleanToString(dpy, args, nargs, fromVal, toVal, destructor_data)
    Display *dpy;
    XrmValuePtr args;
    Cardinal *nargs;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf;
    Boolean boolVal;
 
    boolVal = *(Boolean *) (fromVal->addr);

    if (boolVal) {
	buf = xt_strdup ("true") ;
    } else {
	buf = xt_strdup ("false") ;
    } 
 
    done(&buf, char *);

    return True;
}

static Boolean
Xt_CvtDimensionToString(dpy, args, nargs, fromVal, toVal, destructor_data)
    Display *dpy;
    XrmValuePtr args;
    Cardinal *nargs;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf = NULL;
    Dimension dimVal;
 
    dimVal = *(Dimension *) (fromVal->addr);

    buf = XtMalloc(16);
    sprintf(buf, "%hu", dimVal);
 
    done(&buf, char *);

    return True;
}

static Boolean
Xt_CvtShortToString(dpy, args, nargs, fromVal, toVal, destructor_data)
    Display *dpy;
    XrmValuePtr args;
    Cardinal *nargs;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf = NULL;
    Dimension shortVal;
 
    shortVal = *(short *) (fromVal->addr);

    buf = XtMalloc(16);
    sprintf(buf, "%hd", shortVal);
 
    done(&buf, char *);

    return True;
}


static Boolean
Xt_CvtLongToString(dpy, args, nargs, fromVal, toVal, destructor_data)
    Display *dpy;
    XrmValuePtr args;
    Cardinal *nargs;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf = NULL;
    Dimension shortVal;
 
    shortVal = *(short *) (fromVal->addr);
    buf = XtMalloc(16);
    sprintf(buf, "%ld", shortVal);
 
    done(&buf, char *);

    return True;
}

static Boolean
Xt_CvtPixelToString(dpy, args, nargs, fromVal, toVal, destructor_data)
    Display *dpy;
    XrmValuePtr args;
    Cardinal *nargs;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static char *buf = NULL;
    Pixel pixelVal;
    Screen *screen;
    Colormap colormap;
    XColor color;
 
    color.pixel = *(unsigned long *) fromVal->addr;
    screen = *((Screen **) args[0].addr);
    colormap = *((Colormap *) args[1].addr);

    XQueryColor(DisplayOfScreen(screen), colormap, &color);

    buf = XtMalloc(32);
    sprintf(buf, "#%04x%04x%04x", color.red, color.green, color.blue);

    done(&buf, char *);

    return True;
}

static Boolean
Xt_CvtIntToShort (dpy, args, nargs, fromVal, toVal, destructor_data)
    Display *dpy;
    XrmValuePtr args;
    Cardinal *nargs;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static short ival ;
    ival = (short) (*((int *) fromVal->addr)) ;
    done (&ival, short) ;
    return True ;
}

static Boolean
Xt_CvtIntToDimension (dpy, args, nargs, fromVal, toVal, destructor_data)
    Display *dpy;
    XrmValuePtr args;
    Cardinal *nargs;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static Dimension dim ;
    dim = (Dimension) (*((int *) fromVal->addr)) ;
    done (&dim, Dimension) ;
    return True ;
}

static Boolean
Xt_CvtIntToPosition (dpy, args, nargs, fromVal, toVal, destructor_data)
    Display *dpy;
    XrmValuePtr args;
    Cardinal *nargs;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static Position pos ;
    pos = (Position) (*((int *) fromVal->addr)) ;
    done (&pos, Position) ;
    return True ;
}

static Boolean
Xt_CvtStringToPixmap(dpy, args, nargs, fromVal, toVal, destructor_data)
    Display *dpy;
    XrmValuePtr args;
    Cardinal *nargs;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    Widget w = *((Widget *) args[0].addr);

    /* use the Motif registered converter for this! (why reinvent?) */
    if (XtConvertAndStore(w, XmRString, fromVal, 
		XmRPrimForegroundPixmap, toVal)) {
        return True;
    }
    return False;
}

void Xt_converters_initialize ()
{
    XtConvertArgRec convert_arg;
    XtConvertArgRec convert_args[128];

    /* XmString -> String */
    XtSetTypeConverter(XmRXmString, XtRString,
                       Xt_CvtXmStringToString,
                       NULL, 0,
                       XtCacheNone,
                       NULL);
    /* XmInt -> String */
    XtSetTypeConverter(XmRInt, XtRString,
                       Xt_CvtIntToString,
                       NULL, 0,
                       XtCacheNone,
                       NULL);
    /* Cardinal -> String */
    XtSetTypeConverter(XmRCardinal, XtRString,
                       Xt_CvtUnsignedIntToString,
                       NULL, 0,
                       XtCacheNone,
                       NULL);
    /* HorizontalInt -> String */
    XtSetTypeConverter(XmRHorizontalInt, XtRString,
                       Xt_CvtIntToString,
                       NULL, 0,
                       XtCacheNone,
                       NULL);
    /* VerticalInt -> String */
    XtSetTypeConverter(XmRVerticalInt, XtRString,
                       Xt_CvtIntToString,
                       NULL, 0,
                       XtCacheNone,
                       NULL);
    /* Set -> String */
#ifdef MOTIF_2_0
    XtSetTypeConverter(XmRSet, XtRString,
                       Xt_CvtBooleanToString,
                       NULL, 0,
                       XtCacheNone,
                       NULL);
#endif /* MOTIF_2_0 */
    /* Boolean -> String */
    XtSetTypeConverter(XmRBoolean, XtRString,
                       Xt_CvtBooleanToString,
                       NULL, 0,
                       XtCacheNone,
                       NULL);
    /* Dimension -> String */
    XtSetTypeConverter(XmRDimension, XtRString,
                       Xt_CvtDimensionToString,
                       NULL, 0,
                       XtCacheNone,
                       NULL);

    /* BooleanDimension -> String */
    XtSetTypeConverter(XmRBooleanDimension, XtRString,
                       Xt_CvtDimensionToString,
                       NULL, 0,
                       XtCacheNone,
                       NULL);
    /* HorizontalDimension -> String */
    XtSetTypeConverter(XmRHorizontalDimension, XtRString,
                       Xt_CvtDimensionToString,
                       NULL, 0,
                       XtCacheNone,
                       NULL);
    /* VerticalDimension -> String */
    XtSetTypeConverter(XmRVerticalDimension, XtRString,
                       Xt_CvtDimensionToString,
                       NULL, 0,
                       XtCacheNone,
                       NULL);
    /* short -> String */
    XtSetTypeConverter (XmRShort, XtRString,
                        Xt_CvtShortToString,
                        NULL, 0,
                        XtCacheNone,
                        NULL);
    /* Position -> String */
    XtSetTypeConverter(XmRPosition, XtRString,
                       Xt_CvtShortToString,
                       NULL, 0,
                       XtCacheNone,
                       NULL);
    /* HorizontalPosition -> String */
    XtSetTypeConverter(XmRHorizontalPosition, XtRString,
                       Xt_CvtShortToString,
                       NULL, 0,
                       XtCacheNone,
                       NULL);
    /* VerticalPoisition -> String */
    XtSetTypeConverter( XmRVerticalPosition, XtRString,
                       Xt_CvtShortToString,
                       NULL, 0,
                       XtCacheNone,
                       NULL);
    /* TextPosition -> String */
    XtSetTypeConverter( XmRTextPosition, XtRString,
                       Xt_CvtLongToString,
                       NULL, 0,
                       XtCacheNone,
                       NULL);
    /* Int -> Short ! Is this an Xt nightmare ? */
    XtSetTypeConverter (XtRInt, XtRShort
                        , Xt_CvtIntToShort
                        , NULL, 0
                        , XtCacheNone
                        , NULL) ;
    XtSetTypeConverter (XtRInt, XmRDimension
                        , Xt_CvtIntToDimension
                        , NULL, 0
                        , XtCacheNone
                        , NULL) ;
    XtSetTypeConverter (XtRInt, XmRHorizontalDimension
                        , Xt_CvtIntToDimension
                        , NULL, 0
                        , XtCacheNone
                        , NULL) ;
    XtSetTypeConverter (XtRInt, XmRVerticalDimension
                        , Xt_CvtIntToDimension
                        , NULL, 0
                        , XtCacheNone
                        , NULL) ;
    XtSetTypeConverter (XtRInt, XmRPosition
                        , Xt_CvtIntToPosition
                        , NULL, 0
                        , XtCacheNone
                        , NULL) ;
    XtSetTypeConverter (XtRInt, XmRVerticalPosition
                        , Xt_CvtIntToPosition
                        , NULL, 0
                        , XtCacheNone
                        , NULL) ;
    XtSetTypeConverter (XtRInt, XmRHorizontalPosition
                        , Xt_CvtIntToPosition
                        , NULL, 0
                        , XtCacheNone
                        , NULL) ;
    /* Pixel -> String */
    convert_args[0].address_mode = XtWidgetBaseOffset;
    convert_args[0].address_id = (XtPointer) XtOffsetOf(WidgetRec,core.screen);
    convert_args[0].size = sizeof(Screen *);

    convert_args[1].address_mode = XtWidgetBaseOffset;
    convert_args[1].address_id = (XtPointer) XtOffsetOf(CoreRec,core.colormap);
    convert_args[1].size = sizeof(Colormap);

    XtSetTypeConverter(
		XmRPixel, XmRString,
		Xt_CvtPixelToString,
		convert_args,
		2,
		XtCacheNone,
		NULL);
    /* String -> Pixmap */
    convert_args[0].address_mode = XtWidgetBaseOffset;
    convert_args[0].address_id = (XtPointer) XtOffsetOf(CoreRec, core.self);
    convert_args[0].size = sizeof(CoreWidget);

    XtSetTypeConverter(XmRString, XtRPixmap,
                       Xt_CvtStringToPixmap,
                       convert_args,
                       1,
                       XtCacheNone,
                       NULL);
    return ;
}
