/***************************************************************************
 * sock.c
 * sock.c,v 1.2 1995/03/12 02:59:10 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE 
 * Use socket as streams. This module provides socket streams, in addition to
 * kernel provided type of streams.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <netdb.h>
#include <errno.h>
extern int errno ;

#include <oscheme.h>

static OsEObj network_stream = (OsEObj) 0 ;
static OsEObj socket_stream  = (OsEObj) 0 ;

static OsEObjSel rsel_network_stream_input = (OsEObjSel) 0 ;
static OsEObjSel rsel_network_stream_port  = (OsEObjSel) 0 ;
static OsEObjSel sel_accept = (OsEObjSel) 0 ;

static OsEObjSel rsel_socket_stream_input  = (OsEObjSel) 0 ;
static OsEObjSel rsel_socket_stream_output = (OsEObjSel) 0 ;

static OsError ose_bad_hostname = (OsError) 0 ;

/*
 * Internal C API:
 */

static OsEObj make_socket_stream_from_fd (fd)
    int fd ;
{
    OsEObj sock = (OsEObj) OsEObjSendMsg0 (socket_stream, socket_stream
                                           , OsEObjSelClone) ;
    int  flags ;
    FILE *fout ;

    /* hack given file descr, to make scheme ports happy */
/*
    fcntl (fd, F_GETFL, &flags) ;
    flags |= O_NDELAY ;
    fcntl (fd, F_SETFL, flags) ;
 */
    if ( ! (fout = fdopen (fd, "w")) ) {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("fdopen")
                               , OsIntegerMake (error)
                               , OsA_Nil) ;
    }
    OsEObjAllocateSlot (sock
                        , rsel_socket_stream_input
                        , (OsEObjSel) 0
                        , (OsObject) OsInputPortMakeFromFileno (fd)) ;
    OsEObjAllocateSlot (sock
                        , rsel_socket_stream_output
                        , (OsEObjSel) 0
                        , (OsObject) OsOutputPortMakeFromFile (fout)) ;
    return (sock) ;
}

static OsEObj make_socket_stream_from_netid (os_machine, os_port)
    OsString  os_machine ;
    OsInteger os_port ;
{
    OsEObj              sock ;
    char		*hostname = os_machine->string ;
    int			fd, port  = os_port->ival ;
    struct sockaddr_in	inaddr ;
    struct sockaddr     *addr ;
    struct hostent	*h ;
    unsigned long       hostinetaddr ;
    int                 addrlen ;
    
    /* Translate machine to some INET address */
    if ( isascii (*hostname) && isdigit (*hostname) ) {
	hostinetaddr             = inet_addr (hostname) ;
	inaddr.sin_family        = AF_INET ;
	inaddr.sin_addr.s_addr   = hostinetaddr ;
    } else {
	if ( (! (h = gethostbyname (hostname))) 
	    || (! h->h_addr_list) 
	    || (h->h_addrtype != AF_INET) ) {
            (void) OsErrorTrigger (1, ose_bad_hostname, os_machine) ;
	}
	inaddr.sin_family = h->h_addrtype ;
	bcopy (h->h_addr, &(inaddr.sin_addr), sizeof (inaddr.sin_addr)) ;
    }
    addr            = (struct sockaddr *) &inaddr ;
    addrlen         = sizeof (struct sockaddr_in) ;
    inaddr.sin_port = htons (port) ;
    /* Connect to target server */
    if ( (fd = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0 ) {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("socket")
                               , OsIntegerMake (error)
                               , OsA_Nil) ;
    }
    if ( connect (fd, addr, addrlen) < 0 ) {
        int error = errno ;
        close (fd) ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("connect")
                               , OsIntegerMake (error)
                               , OsA_Nil) ;
    }
    /* Create object, and initialize */
    return (make_socket_stream_from_fd (fd)) ;
}

static OsEObj make_network_stream (family, type, protocol)
    int family, type, protocol ;
{
    OsEObj             net ;
    int                fd, port = 2000 ;
    struct sockaddr_in saddr ;


    /* Create the socket */
    if ( (fd = socket (family, type, protocol)) < 0 ) {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("socket")
                               , OsIntegerMake (error)
                               , OsA_Nil) ;
    }
    /* Bind to some adress */
    memset ( &saddr, 0, sizeof (saddr) ) ;
    saddr.sin_port        = htons (port) ;
    saddr.sin_family      = AF_INET;
    saddr.sin_addr.s_addr = htonl (INADDR_ANY) ;
    while (bind (fd, (struct sockaddr *) &saddr, sizeof(saddr)))
	saddr.sin_port = htons (++port) ;
    listen (fd, 5) ;
    /* Initialize object */
    net = (OsEObj) OsEObjSendMsg0 (network_stream, network_stream
                                   , OsEObjSelClone) ;
    OsEObjAllocateSlot (net
                        , rsel_network_stream_port
                        , (OsEObjSel) 0
                        , (OsObject) OsIntegerMake (port)) ;
    OsEObjAllocateSlot (net
                        , rsel_network_stream_input
                        , (OsEObjSel) 0
                        , (OsObject) OsInputPortMakeFromFileno (fd)) ;
    return (net) ;
}

/*
 * Network stream objects implementation:
 */

static OsObject network_stream_accept (closure, self, sel)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
{
    int             afd, iaddr_len ;
    struct sockaddr iaddr ;
    OsInputPort     iport ;

    iport = (OsInputPort) OsEObjSendMsg0 (self, self
                                          , rsel_network_stream_input) ;

    if ( (afd = accept (iport->medium.fd, &iaddr, &iaddr_len)) < 0 ) {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("accept")
                               , OsIntegerMake (error)
                               , OsConsMake ((OsObject) iport, OsA_Nil)) ;
    }
    return ((OsObject) make_socket_stream_from_fd (afd)) ;
}

static void initialize_network_stream () {
    network_stream = (OsEObj) OsEObjSendMsg0 ((OsEObj) OsEObjRoot
                                              , (OsEObj) OsEObjRoot
                                              , OsEObjSelClone) ;
    rsel_network_stream_input = OsEObjSelDefine ("input") ;
    rsel_network_stream_port  = OsEObjSelDefine ("port") ;
    sel_accept                = OsEObjSelDefine ("accept") ;
    OsEObjAllocateSlot (network_stream
                        , rsel_network_stream_port
                        , (OsEObjSel) 0
                        , OsA_Nil) ;
    OsEObjAllocateSlot (network_stream
                        , rsel_network_stream_input
                        , (OsEObjSel) 0
                        , OsA_Nil) ;
    OsEObjDefineMsg (network_stream
                     , sel_accept
                     , network_stream_accept
                     , (void *) 0) ;
    return ;
}

/*
 * Socket streams object implementation:
 */

static void initialize_socket_stream () {
    socket_stream = (OsEObj) OsEObjSendMsg0 ((OsEObj) OsEObjRoot
                                             , (OsEObj) OsEObjRoot
                                             , OsEObjSelClone) ;
    rsel_socket_stream_input  = OsEObjSelDefine ("input") ;
    rsel_socket_stream_output = OsEObjSelDefine ("output") ;
    OsEObjAllocateSlot (socket_stream
                        , rsel_socket_stream_input
                        , (OsEObjSel) 0
                        , (void *) 0) ;
    OsEObjAllocateSlot (socket_stream
                        , rsel_socket_stream_output
                        , (OsEObjSel) 0
                        , (void *) 0) ;
    return ;
}

/*
 * Subrs:
 */

/* AUTODOC (AUTOLOAD "socket")
 * (open-network-stream)
 * Opens a network stream suitable to accept connections.
 * This function creates an accepting socket for the oscheme process. The
 * created socket will bind itself to the first available port after 2000. It
 * returns an object whcih have the following set of slots:
 *    input : the input stream for the accepting socket, which can bve selected
 *            on to know when someone is willing to connect to the socket.
 *    port  : the port number where the socket is bind.
 * It implements:
 *    accept: accepts a pending connection, and returns a socket object.
 */

static OsObject OsSubrOpenNetworkStream ()
{
    return ((OsObject) make_network_stream(AF_INET, SOCK_STREAM, IPPROTO_TCP));
}

/* AUTODOC (AUTOLOAD "socket")
 * (open-socket-stream MACHINE PORT)
 * Open a socket stream to the target MACHINE at PORT.
 * This function opens a socket stream to MACHINE, bound at PORT. It returns
 * a socket object, which has the following set of slots:
 *    input : the socket input port, for reading
 *    output: the socket output port, for writing.
 */

static OsObject OsSubrOpenSocketStream (machine, port)
    OsString  machine ;
    OsInteger port ;
{
    OsShouldBeString (machine) ;
    OsShouldBeInteger (port) ;
    return ((OsObject) make_socket_stream_from_netid (machine, port)) ;
}

/*
 * Initialization:
 */

void Os_socket_initialize (OsInterp interp) {
    initialize_network_stream () ;
    initialize_socket_stream () ;
    ose_bad_hostname = OsErrorMake (interp
                                    , "error:bad-hostname"
                                    , "Invalid hostname %0"
                                    , 1, OSE_CATCHABLE) ;
    OsSecSubrMake (interp
                   , "open-network-stream", OsSubrOpenNetworkStream
                   , 0, OS_SECURITY_LEVEL_B) ;
    OsSecSubrMake (interp
                   , "open-socket-stream", OsSubrOpenSocketStream
                   , 2, OS_SECURITY_LEVEL_B) ;
    OsPathProvides (interp, OsStringMake ("socket")) ;
    return ;
}

