/***************************************************************************
 * anchor.c
 * anchor.c,v 1.1.1.1 1995/02/27 15:14:56 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * C level interface to libWWW anchor objects.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include "HText.h"
#include "HTUtils.h"
#include "HTString.h"
#include "HTStream.h"

#include <oscheme.h>
#include "w3scheme.h"

static OsHashtable anchor_hashtable = (OsHashtable)  0 ;

static OsEObj anchor_class = (OsEObj) 0 ;

static OsEObjSel anchor_sel_ptr     = (OsEObjSel) 0 ;
static OsEObjSel anchor_sel_address = (OsEObjSel) 0 ;

/* Anchor's main link */

static OsObject anchor_main_link (closure, self, sel)
    void *closure ;
    OsEObj self ;
    OsEObjSel sel ;
{
    HTAnchor *anchor ;
    anchor = ((OsPointer) OsEObjSendMsg0 (self,self,anchor_sel_ptr))->pointer;
    if ( anchor->mainLink.dest ) 
        return ((OsObject) OsW3MakeAnchor (anchor->mainLink.dest));
    else
        return (OsA_False) ;
}

/* List of anchor's children */

static OsObject anchor_children (closure, self, sel)
    void *closure ;
    OsEObj self ;
    OsEObjSel sel ;
{
    OsVector ret = OsVectorMake (0) ;
    OsAtom   atom ;
    OsEObj   child ;
    HTList   *links ;
    HTAnchor *anchor ;

    anchor = ((OsPointer) OsEObjSendMsg0 (self,self,anchor_sel_ptr))->pointer;
    for (links = anchor->links ; links ; links = links->next) {
        HTLink *link = (HTLink *) links->object ;
        if ( link->type )
            atom = OsAtomMake (OsCurrentInterp()
                               , link->type->name, OsA_Unbound) ;
        else
            atom = OsAtomMake (OsCurrentInterp()
                               , "www:unknown-link-type", OsA_Unbound) ;
        child  = OsW3MakeAnchor (link->dest) ;
        OsVectorFastAppend (ret,OsConsMake((OsObject) atom, (OsObject) child));
    }
    return ((OsObject) ret) ;
}

/* Anchor's adress */

static OsObject anchor_address (closure, self, sel)
    void *closure ;
    OsEObj self ;
    OsEObjSel sel ;
{
    char *adr ;
    HTAnchor *anchor ;

    anchor = ((OsPointer) OsEObjSendMsg0 (self,self,anchor_sel_ptr))->pointer;
    if ( adr = HTAnchor_address (anchor) ) {
        OsString os_adr = OsStringMake (adr) ;
        free (adr) ;
        return ((OsObject) os_adr) ;
    } else {
        return (OsA_False) ;
    }
    /* NOT REACHED */
}

/* Anchor's protocol */

static OsObject anchor_protocol (closure, self, sel)
    void *closure ;
    OsEObj self ;
    OsEObjSel sel ;
{
    HTParentAnchor *anchor ;
    HTProtocol     *proto ;
    
    anchor = ((OsPointer) OsEObjSendMsg0 (self,self,anchor_sel_ptr))->pointer;
    while ( anchor->parent != anchor )	/* go up to first adult */
        anchor = anchor->parent ;
    proto  = (HTProtocol *) anchor->protocol ;
    if ( proto )
        return ((OsObject) OsAtomMake (OsCurrentInterp()
                                       , proto->name, OsA_Unbound)) ;
    else
        return (OsA_False) ;
}

/* Anchor's title */

static OsObject anchor_title (closure, self, sel)
    void *closure ;
    OsEObj self ;
    OsEObjSel sel ;
{
    HTParentAnchor *anchor ;
    char *title ;
    
    anchor = ((OsPointer) OsEObjSendMsg0 (self,self,anchor_sel_ptr))->pointer;
    while ( anchor->parent != anchor )	/* go up to first adult */
        anchor = anchor->parent ;
    if ( title = anchor->title ) 
        return ((OsObject) OsStringMake (title)) ;
    else
        return (OsA_False) ;
}

/* Index stuff:
 * isIndex,
 * setIndex,
 * clearIndex
 */

static OsObject anchor_is_index (closure, self, sel)
    void *closure ;
    OsEObj self ;
    OsEObjSel sel ;
{
    HTParentAnchor *anchor ;
    anchor = ((OsPointer) OsEObjSendMsg0 (self,self,anchor_sel_ptr))->pointer;
    if ( (anchor->parent == anchor) && HTAnchor_isIndex (anchor) )
        return (OsA_True) ;
    else
        return (OsA_False) ;
}

/* List anchor methods */

static OsObject anchor_methods (closure, self, sel)
    void *closure ;
    OsEObj self ;
    OsEObjSel sel ;
{
    HTParentAnchor *anchor ;
    HTList         *methods ;
    OsVector       vmethods = OsVectorMake (0) ;
    
    anchor = ((OsPointer) OsEObjSendMsg0 (self,self,anchor_sel_ptr))->pointer;
    while ( anchor->parent != anchor )	/* go up to first adult */
        anchor = anchor->parent ;
    methods = HTAnchor_methods (anchor) ;
    while ( methods ) {
        HTAtom *atom = (HTAtom *) methods->object ;
        if ( atom )
            OsVectorFastAppend (vmethods
                                , (OsObject) OsAtomMake (OsCurrentInterp()
                                                         , HTAtom_name (atom)
                                                         , OsA_Unbound)) ;
        methods = methods->next ;
    }
    return ((OsObject) vmethods) ;
}

/* DIsplay anchors */

static OsObject anchor_display_to (closure, self, sel, port)
    void *closure ;
    OsEObj self ;
    OsEObjSel sel ;
    OsOutputPort port ;
{
    char     tmp[OS_DISPLAY_BUFSIZ] ;
    OsObject address = OsEObjSendMsg0 (self, self, anchor_sel_address) ;
    
    sprintf (tmp, "#[WWWAnchor (%p) ", OsW3AnchorOfAnchor (self)) ;
    OsOutputPortWriteCString (port, tmp) ;
    OsSendDisplay (address, port) ;
    OsOutputPortWriteChar (port, ']') ;
    return ((OsObject) self) ;
}

static OsObject anchor_is_child (closure, self, sel)
    void *closure ;
    OsEObj self ;
    OsEObjSel sel ;
{
    HTParentAnchor *anchor ;
    anchor = ((OsPointer) OsEObjSendMsg0 (self,self,anchor_sel_ptr))->pointer;
    return ((anchor->parent == anchor) ? OsA_True : OsA_False) ;
}

static OsObject anchor_parent (closure, self, sel)
    void *closure ;
    OsEObj self ;
    OsEObjSel sel ;
{
    HTParentAnchor *anchor ;
    anchor = ((OsPointer) OsEObjSendMsg0 (self,self,anchor_sel_ptr))->pointer;
    return ((anchor->parent != anchor)
            ? (OsObject) OsW3MakeAnchor ((HTAnchor *) anchor->parent)
            : OsA_False) ;
}


static OsObject anchor_tag (void *closure, OsEObj self, OsEObjSel sel) {
    HTChildAnchor *child ;
    child = ((OsPointer) OsEObjSendMsg0 (self,self,anchor_sel_ptr))->pointer ;
    return (((child->parent != (HTParentAnchor *) child) && child->tag)
            ? (OsObject) OsStringMake (child->tag)
            : OsA_False) ;
}

static void anchor_class_initialize () {
    anchor_class = (OsEObj) OsEObjSendMsg0 ((OsEObj) OsEObjRoot
                                            , (OsEObj) OsEObjRoot
                                            , OsEObjSelClone) ;
    OsEObjAllocateSlot (anchor_class
                        , anchor_sel_ptr = OsEObjSelDefine ("_ptr")
                        , (OsEObjSel) 0
                        , OsA_Unbound) ;
    /* display anchor */
    OsEObjDefineMsg (anchor_class
                     , OsEObjSelDefine ("display:")
                     , anchor_display_to
                     , (void *) 0) ;
    /* returns a vector of children anchors */
    OsEObjDefineMsg (anchor_class
                     , OsEObjSelDefine ("children")
                     , anchor_children
                     , (void *) 0) ;
    /* returns anchor adress */
    anchor_sel_address = OsEObjSelDefine ("address") ;
    OsEObjDefineMsg (anchor_class
                     , anchor_sel_address
                     , anchor_address
                     , (void *) 0) ;
    /* returns anchor protocol */
    OsEObjDefineMsg (anchor_class
                     , OsEObjSelDefine ("protocol")
                     , anchor_protocol
                     , (void *) 0) ;
    /* returns anchor title */
    OsEObjDefineMsg (anchor_class
                     , OsEObjSelDefine ("title")
                     , anchor_title
                     , (void *) 0) ;
    /* anchor's main link */
    OsEObjDefineMsg (anchor_class
                     , OsEObjSelDefine ("mainLink")
                     , anchor_main_link
                     , (void *) 0) ;
    /* Index stuff: */
    OsEObjDefineMsg (anchor_class
                     , OsEObjSelDefine ("isIndex")
                     , anchor_is_index
                     , (void *) 0) ;
    /* List available anchors method */
    OsEObjDefineMsg (anchor_class
                     , OsEObjSelDefine ("methods")
                     , anchor_methods
                     , (void *) 0) ;
    /* Is given anchor a child anchor */
    OsEObjDefineMsg (anchor_class
                     , OsEObjSelDefine ("isChild")
                     , anchor_is_child
                     , (void *) 0) ;
    /* Child anchor tag */
    OsEObjDefineMsg (anchor_class
                     , OsEObjSelDefine ("tag")
                     , anchor_tag
                     , (void *) 0) ;
    /* Is given anchor parent anchor */
    OsEObjDefineMsg (anchor_class
                     , OsEObjSelDefine ("parent")
                     , anchor_parent
                     , (void *) 0) ;
    return ;
}

/* External API */

HTAnchor *OsW3AnchorOfAnchor (os_anchor)
    OsEObj os_anchor ;
{
    OsPointer ptr ;

    ptr = (OsPointer) OsEObjSendMsg0 (os_anchor, os_anchor, anchor_sel_ptr) ;
    OsShouldBePointer (ptr) ;
    return ((HTAnchor *) ptr->pointer) ;
}

OsEObj OsW3MakeAnchor (HTAnchor *anchor) {
    OsEObj obj ;

    if ( (obj = (OsEObj) OsHashtableGet (anchor_hashtable, anchor)) ) 
        return (obj) ;
    obj = (OsEObj) OsEObjSendMsg0 (anchor_class, anchor_class, OsEObjSelClone);
    OsEObjAllocateSlot (obj
                        , anchor_sel_ptr
                        , (OsEObjSel) 0
                        , (OsObject) OsSharedPointerMake (anchor)) ;
    return (obj) ;
}

/*
 * Subr
 */

/* AUTODOC
 * (www:make-anchor HREF)
 * Create an anchor object, given an HREF.
 * This primitive creates an anchor, given its HREF. The provided HREF should
 * be absolute. The returned anchor object provides the following set of
 * messages:
 *    (children): list of anchor's children. Available only if the anchor has
 *        been loaded,
 *    (address): the full HREF (as an absolute path) of the given anchor,
 *    (protocol): the access protocol for this anchor,
 *    (mainLink): the link to which this anchor leads (available only once the
 *        anchor has been loaded),
 *    (title): the anchor title (if any),
 *    (isIndex): is the given anchor an index ?
 *    (methods): what are the HTTP methods supported by this anchor ?
 *    (isChild): is this anchor a child anchor,
 *    (tag): the tag of this anchor (which should be a child), if any,
 *    (parent): the parent of the anchor, or #f if the anchor is a parent.
 */
 
static OsObject OsSubrW3MakeAnchor (address)
    OsString address ;
{
    HTAnchor *anchor ;
    
    OsShouldBeString (address) ;
    if ( anchor = HTAnchor_findAddress (address->string) ) {
        return ((OsObject) OsW3MakeAnchor (anchor)) ;
    } else {
        return (OsA_False) ;
    }
    /* NOT REACHED */
}

void Os_anchor_initialize (OsInterp interp) {
    anchor_class_initialize () ;
    anchor_hashtable = OsHashtableMake (0
                                        , (OsHashFct) OsHashPointer
                                        , (OsCmpFct)  OsCmpPointer) ;
    OsSubrMake (interp, "www:make-anchor", OsSubrW3MakeAnchor, 1) ;
    return ;
}

                                        
    
