;; head.scm
;; head.scm,v 1.1.1.1 1995/02/27 15:14:58 baird Exp
;; See the file COPYRIGHT for licensing conditions.
;; AUTODOC MODULE
;; This module provide utilities to get informations related to some anchor
;; or url. 

(require "date")

(define mime:parse-date
  (let ((date-re (regexp:make 
                  (string-append "[^A-Za-z]*([A-Za-z]*)," 	;; day
                                 "[^0-9]*([0-9]*)-" 		;; day in month
                                 "[^A-Za-z]*([A-Za-z]*)-"	;; month
                                 "[^0-9]*([0-9]*) "		;; year
                                 "([0-9][0-9]):"		;; hour
                                 "([0-9][0-9]):"		;; minute
                                 "([0-9][0-9])"))))		;; seconds
    ;; Convert month name into month number:
    (define (month-name-to-number name)
      (cond ((string-ci=? "Jan" name) 1)
            ((string-ci=? "Feb" name) 2)
            ((string-ci=? "Mar" name) 3)
            ((string-ci=? "Apr" name) 4)
            ((string-ci=? "May" name) 5)
            ((string-ci=? "Jun" name) 6)
            ((string-ci=? "Jul" name) 7)
            ((string-ci=? "Aug" name) 8)
            ((string-ci=? "Sep" name) 9)
            ((string-ci=? "Oct" name) 10)
            ((string-ci=? "Nov" name) 11)
            ((string-ci=? "Dec" name) 12)
            (t (error "Invalid month name (mime:parse-date) \"%0\"" name))))
    (lambda (date)
      (let ((match (regexp:match date-re date 1 2 3 4 5 6 7)))
        (or match
            (error "Invalid MIME date (mime:parse-date) \"%0\"" date))
        (date:make-date
         (+ (string->integer (list-ref match 3)) 1900)	;; year
         (month-name-to-number (list-ref match 2))	;; month
         (string->integer (list-ref match 1))		;; day in month
         (string->integer (list-ref match 4))		;; hour
         (string->integer (list-ref match 5))		;; minute
         (string->integer (list-ref match 6)))))) )	;; seconds
         
(define mime:parse-header
  ;; Parse mime header:
  ;; I'm really not sure I'm getting all info right ...
  (let* ((s13 (string (integer->char 13)))
         (fre (string-append "([^" s13 "\n]*)(" s13 "|\n).*")))
    (let ((server-re 
           (regexp:make (string-append "[sS][eE][rR][vV][eE][rR]: " fre)))
          (date-re   
           (regexp:make (string-append "[dD][aA][tT][eE]: " fre)))
          (type-re   
           (regexp:make 
            (string-append 
             "[cC][oO][nN][tT][eE][nN][tT]-[tT][yY][pP][eE]: " 
             fre)))
          (length-re 
           (regexp:make 
            (string-append 
             "[cC][oO][nN][tT][eE][nN][tT]-[lL][eE][nN][gG][tT][hH]: " 
             fre)))
          (mod-re    
           (regexp:make 
            (string-append 
             "[lL][aA][sS][tT]-[mM][oO][dD][iI][fF][iI][eE][dD]: " fre)))
          (stat [obj:root clone]))
      [stat defineSlot: "server"]
      [stat defineSlot: "date"]
      [stat defineSlot: "content-type"]
      [stat defineSlot: "content-length"]
      [stat defineSlot: "last-modified"]
      (define-message stat (display: port)
        (format port "#[AnchorStat date=%0]" [self date]))
      (lambda (header)
        (let ((new [stat clone]))
          ;; Update each field:
          (let ((match (regexp:match server-re header 1)))
            [new server: (and match (car match))])
          (let ((match (regexp:match date-re header 1)))
            [new date: (and match (mime:parse-date (car match)))])
          (let ((match (regexp:match type-re header 1)))
            [new content-type: (and match (car match))])
          (let ((match (regexp:match length-re header 1)))
            [new content-length: (and match (string->integer (car match)))])
          (let ((match (regexp:match mod-re header 1)))
            [new last-modified: (and match (mime:parse-date (car match)))])
          new)))) )

;; AUTODOC AUTOLOAD
;; (www:stat-anchor ANCHOR)
;; Returns available informations on ANCHOR, using HEAD http method.
;; This function returns an object having the following slots:
;;     server            server version
;;     date              document creation date
;;     content-type      document type MIME content 
;;     content-length    document length (in bytes)
;;     last-modified     document last modifictaion date
;; On failure, it returns #f. Warning: some servers don't (still) implement
;; properly the HEAD method, hence some of the above slot might be set to #f
;; when their value wasn't provided.

(define www:stat-anchor
  (let ((request (www:make-request))
        (stream  (www:make-stream)))
    ;; Customize stream to accumulate into info.
    [stream defineSlot: "info" init: #f]
    (define-message stream (char: c)   (display c [self info]))
    (define-message stream (string: s) (display s [self info]))
    (define-message stream (block: b)  (display b [self info]))
    (define-message stream (new: txt)
      (let ((new-stream [self clone]))
        [new-stream info: (open-output-string txt)]
        new-stream))
    ;; Customize request:
    [request method: 'HEAD]
    [request output-format: 'www/mime]
    ;; Main entry point 
    (lambda (anchor)
      (let* ((mime-header (string-copy ""))
             (my-stream   [stream new: mime-header])
             (my-request  [request clone]))
        [my-request anchor: anchor]
        [my-request output-stream: my-stream]
        (let ((ret (www:load my-request)))
          (close-output-port [my-stream info])
          (and ret (mime:parse-header mime-header)))))) )
        
(provide "head")
