;; http.scm
;; http.scm,v 1.2 1995/03/17 14:29:05 baird Exp
;; See the file COPYRIGHT for licensing conditions.
;; Various stuff test
;; AUTODOC MODULE
;; This module provides some utilities to use the WWW interface to oscheme.

(require "www")

;; AUTODOC AUTOLOAD
;; (www:list URL)
;; Returns the list of anchors found in given URL.

(define www:list 
  (let ((handler (www:make-html-handler)))
    [handler defineSlot: "list" init: ()]
    (define-message handler (anchor: anchor href: unused)
      (and [anchor mainLink] 
           [self list: (cons [anchor mainLink] [self list])]))
    (lambda (url)
      (let ((my-handler [handler clone])
            (my-request (www:make-request)))
        [my-request handler: my-handler]
        [my-request anchor: (www:make-anchor url)]
        (www:load my-request)
        [my-handler list]))) )
        
;; AUTODOC AUTOLOAD
;; (www:display-url URL)
;; Display the source of URL.
;; This function display to standard output the source of the given URL.

(define (www:display-url url)
  (let ((request (www:make-request))
        (stream  (www:make-stream)))
    [request output-format: '*/*]
    [request output-stream: stream]
    [request anchor: (www:make-anchor url)]
    (www:load request)) )

;; Source reader builder:

(define (www:make-source-getter load-func)
  (lambda (object)
    (let ((stream  (www:make-stream))
          (request (www:make-request))
          (text    (string-copy "")))
      [stream defineSlot: "source" init: (open-output-string text)]
      (define-message stream (char: c)
        (display c [self source]))
      (define-message stream (string: str)
        (display str [self source]))
      (define-message stream (block: blk)
        (display blk [self source]))
      [request output-format: '*/*]
      [request output-stream: stream]
      (load-func object request)
      (close-output-port [stream source])
      text)) )

;; AUTODOC AUTOLOAD
;; (www:url-source URL)
;; Retrieve the source text of URL (in ASCII).

(define www:url-source 
  (www:make-source-getter 
   (lambda (url request) 
     [request anchor: (www:make-anchor url)]
     (www:load request))) )

;; AUTODOC AUTOLOAD
;; (www:anchor-source ANCHOR)
;; Retreive the source text of ANCHOR (in ASCII).

(define www:anchor-source 
  (www:make-source-getter 
   (lambda (anchor request)
     [request anchor: anchor]
     (www:load request))) )

;; Dump url/anchors to file:

(define (www:make-object-dumper load-func)
  (lambda (object local)
    (let ((stream  (www:make-stream))
          (ostream (open-output-file local))
          (request (www:make-request)))
    (define-message stream (char: c)
      (display c ostream))
    (define-message stream (string: str)
      (display str ostream))
    (define-message stream (block: blk)
      (display blk ostream))
    [request output-format: '*/*]
    [request output-stream: stream]
    (let ((result (load-func object request)))
      (close-output-port ostream)
      result))) )
    
;; AUTODOC AUTOLOAD
;; (www:url-to-file URL FILE)
;; Dump URL's content into local file FILE.

(define www:url-to-file 
  (www:make-object-dumper 
   (lambda (url request)
     [request anchor: (www:make-anchor url)]
     (www:load request))) )

;; AUTODOC AUTOLOAD
;; (www:anchor-to-file URL FILE)
;; Dump anchor's content into local file FILE.

(define www:anchor-to-file 
  (www:make-object-dumper 
   (lambda (anchor request)
     [request anchor: anchor]
     (www:load request))) )

(provide "http")
